Package: badpackage
Title: Advice on R package building
Date: 2016 February
Version: 1.0.0
Author: Gabor Csardi
Maintainer: Gabor Csardi <gcsardi@mango-solutions.com>
Description: Give advice about good practices when building R packages.
License: Public domain
Suggests:
    testthat
Depends:
    goodpractice
Imports:
    desc
VignetteBuilder:
    knitr
RoxygenNote: 5.0.1

Package: goodpractice
Title: Advice on R Package Building
Version: 1.2.0
Authors@R: 
    c(
      person(given = "Mark",
             family = "Padgham",
             role = c("aut", "cre"),
             email = "mark@ropensci.org",
             comment = c(ORCID = "0000-0003-2172-5265")),
      person(given = "Ascent Digital Services UK Limited",
             role = "cph",
             comment = c(GitHub = "MangoTheCat")),
      person(given = "Karina",
             family = "Marks",
             role = "aut",
             email = "karina.marks@ascent.io",
             comment = c(GitHub = "KarinaMarks")),       
      person(given = "Daniel",
             family = "de Bortoli",
             role = c("aut"),
             comment = c(GitHub = "ddbortoli")),
      person(given = "Gabor",
             family = "Csardi",
             role = "aut",
             email = "csardi.gabor@gmail.com"),
      person(given = "Hannah",
             family = "Frick",
             role = "aut",
             email = "hannah.frick@gmail.com"),
      person(given = "Owen",
             family = "Jones",
             role = "aut",
             email = "owenjonesuob@gmail.com",
             comment = c(GitHub = "owenjonesuob")),
      person(given = "Hannah",
             family = "Alexander",
             role = "aut",
             email = "halexander@mango-solutions.com"),
      person(given = "Ana",
             family = "Simmons",
             role = c("ctb"),
             email = "ana.simmons@ascent.io",
             comment = c(GitHub = "anasimmons")),
      person(given = "Fabian",
             family = "Scheipl",
             role = "ctb",
             comment = c(GitHub = "fabian-s")),
      person(given = "Athanasia Mo",
             family = "Mowinckel",
             email = "a.m.mowinckel@psykologi.uio.no",
             role = c("aut"),
            comment = c(
                GitHub = "drmowinckels", 
                ORCID = "0000-0002-5756-0223")
       ))
Description: Give advice about good practices when building R packages.
    Advice includes functions and syntax to avoid, package structure, code
    complexity, code formatting, etc.
License: MIT + file LICENSE
URL: https://docs.ropensci.org/goodpractice/,
        https://github.com/ropensci-review-tools/goodpractice
BugReports: https://github.com/ropensci-review-tools/goodpractice/issues
Depends: R (>= 4.3.0)
Imports: cli, covr, curl, cyclocomp (>= 1.1.0), desc, jsonlite, lintr
        (>= 3.0.0), praise, rcmdcheck, roxygen2, rstudioapi, spelling,
        tools, treesitter, treesitter.r, urlchecker, utils, whoami,
        withr
Suggests: future, future.apply, kableExtra, knitr, rmarkdown, testthat
        (>= 3.0.0)
VignetteBuilder: knitr
Encoding: UTF-8
Config/testthat/edition: 3
Config/ropensci/maintainer: staff
Collate: 'api.R' 'customization.R' 'lists.R' 'chk_avoided_packages.R'
        'treesitter.R' 'chk_code_structure.R' 'chk_covr.R'
        'chk_cyclocomp.R' 'chk_description.R' 'chk_generic.R'
        'chk_lintr.R' 'chk_namespace.R' 'chk_rcmdcheck.R' 'chk_rd.R'
        'chk_revdep.R' 'chk_roxygen2.R' 'chk_spelling.R'
        'chk_tidyverse.R' 'chk_urlchecker.R' 'chk_vignette.R' 'gp.R'
        'package.R' 'prep_utils.R' 'prep_covr.R' 'prep_cyclocomp.R'
        'prep_description.R' 'prep_lintr.R' 'prep_namespace.R'
        'prep_rcmdcheck.R' 'prep_rd.R' 'prep_revdep.R'
        'prep_roxygen2.R' 'prep_source.R' 'prep_spelling.R'
        'prep_tidyverse.R' 'prep_urlchecker.R' 'prep_vignette.R'
        'print.R' 'rstudio_markers.R' 'utils.R'
Config/roxygen2/version: 8.1.0
NeedsCompilation: no
Packaged: 2026-08-31 11:01:10 UTC; smexus
Author: Mark Padgham [aut, cre] (ORCID:
    <https://orcid.org/0000-0003-2172-5265>),
  Ascent Digital Services UK Limited [cph] (GitHub: MangoTheCat),
  Karina Marks [aut] (GitHub: KarinaMarks),
  Daniel de Bortoli [aut] (GitHub: ddbortoli),
  Gabor Csardi [aut],
  Hannah Frick [aut],
  Owen Jones [aut] (GitHub: owenjonesuob),
  Hannah Alexander [aut],
  Ana Simmons [ctb] (GitHub: anasimmons),
  Fabian Scheipl [ctb] (GitHub: fabian-s),
  Athanasia Mo Mowinckel [aut] (GitHub: drmowinckels, ORCID:
    <https://orcid.org/0000-0002-5756-0223>)
Maintainer: Mark Padgham <mark@ropensci.org>
Repository: CRAN
Date/Publication: 2026-08-31 18:40:08 UTC

Package: goodpackage
Title: Advice on R package building
Version: 1.0.0
Author: Gabor Csardi
Maintainer: Gabor Csardi <gcsardi@mango-solutions.com>
Description: Give advice about good practices when building R packages.
License: GPL-2
URL: https://github.com/mangothecat/goodpractice
BugReports: https://github.com/mangothecat/goodpractice/issues
RoxygenNote: 5.0.1

Package: noroxygenspkg
Title: Test Package Without Roxygen2
Version: 1.0.0
Author: Test
Maintainer: Test <test@test.com>
Description: Fixture for testing that roxygen2 checks skip non-roxygen packages.
License: MIT

Package: badpackage
Title: Advice on R package building
Version: 1.0.0
Author: Gabor Csardi
Maintainer: Gabor Csardi <gcsardi@mango-solutions.com>
Description: Give advice about good practices when building R packages.
License: GPL-2
URL: http://github.com/mangothecat/goodpractice
BugReports: http://github.com/mangothecat/goodpractice/issues
RoxygenNote: 5.0.1

Package: badpackage
Title: Advice on R package building
Date: 2016-05-20
Version: 1.0.0
Author: Gabor Csardi
Maintainer: Gabor Csardi <gcsardi@mango-solutions.com>
Description: Give advice about good practices when building R packages.
License: GPL-2
Suggests:
    testthat
Imports:
    goodpractice
RoxygenNote: 5.0.1

Package: badroxygenspkg
Title: Test Package for Roxygen2 Checks
Version: 1.0.0
Author: Test
Maintainer: Test <test@test.com>
Description: Fixture for testing roxygen2-specific checks.
License: MIT
RoxygenNote: 7.3.3

Package: badspelling
Title: Test Packge for Spellign Check
Version: 1.0.0
Author: Test
Maintainer: Test <test@test.com>
Description: Package with misspeled words in documentation.
License: GPL-2
Language: en-US

Package: badvignettes
Title: Test Package with Bad Vignette Patterns
Version: 1.0.0
Author: Test
Maintainer: Test <test@test.com>
Description: A test package with anti-patterns in vignettes.
License: GPL-2

Package: goodprint
Title: Test Package for print Return Check
Version: 1.0.0
Author: Test
Maintainer: Test <test@test.com>
Description: Package with print methods that correctly return invisibly.
License: GPL-2

# Generated by roxygen2: do not edit by hand

S3method(print,goodPractice)
export(all_check_groups)
export(all_checks)
export(checks)
export(checks_by_group)
export(default_checks)
export(describe_check)
export(describe_check_groups)
export(export_json)
export(failed_checks)
export(failed_positions)
export(goodpractice)
export(gp)
export(learn_skill_gp)
export(make_check)
export(make_prep)
export(results)
export(tidyverse_checks)
export(use_skill_gp)
importFrom(covr,
  package_coverage,
  percent_coverage,
  zero_coverage
)
importFrom(cyclocomp,cyclocomp_package_dir)
importFrom(desc,
  desc_get,
  desc_get_collate,
  description
)
importFrom(jsonlite,toJSON)
importFrom(lintr,lint_package)
importFrom(praise,praise)
importFrom(rcmdcheck,rcmdcheck)
importFrom(roxygen2,parse_package)
importFrom(rstudioapi,
  callFun,
  hasFun
)
importFrom(spelling,spell_check_package)
importFrom(stats,setNames)
importFrom(urlchecker,url_check)
importFrom(utils,
  getParseData,
  head
)
importFrom(whoami,
  fullname,
  username
)
importFrom(withr,with_options)

# goodpractice <img src="man/figures/logo.png" align="right" width="20%" height="20%" />

<!-- badges: start -->
[![R-CMD-check](https://github.com/ropensci-review-tools/goodpractice/workflows/R-CMD-check/badge.svg)](https://github.com/ropensci-review-tools/goodpractice/actions)
[![CRAN
status](https://www.r-pkg.org/badges/version/goodpractice)](https://CRAN.R-project.org/package=goodpractice)
[![CRAN RStudio mirror
downloads](https://cranlogs.r-pkg.org/badges/goodpractice)](https://www.r-pkg.org/pkg/goodpractice)
[![Codecov test
coverage](https://codecov.io/gh/ropensci-review-tools/goodpractice/branch/main/graph/badge.svg)](https://app.codecov.io/gh/ropensci-review-tools/goodpractice?branch=main)
<!-- badges: end -->

## Advice on R Package Building

Give advice about good practices when building R packages. Advice
includes functions and syntax to avoid, package structure, code
complexity, code formatting, etc.

## Installation

You can install the release version from CRAN:

``` r
install.packages("goodpractice")
```

The development version can be installed from R-universe by enabling the
[ropensci-review-tools
`r-universe`](https://ropensci-review-tools.r-universe.dev/)
and installing the usual way:

``` r
options(repos = c(
    ropenscireviewtools = "https://ropensci-review-tools.r-universe.dev",
    CRAN = "https://cloud.r-project.org"
))
install.packages("goodpractice")
```

Or the development version can be installed directly from GitHub with:

``` r
pak::pak("ropensci-review-tools/goodpractice")
```

## Usage

``` r
library(goodpractice)
gp("<my-package>")
```

## Example

``` r
library(goodpractice)
# use example package contained in the goodpractice package
pkg_path <- system.file("bad1", package = "goodpractice")
g <- gp(pkg_path)
g
```

    #> ── GP badpackage ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────
    #> 
    #> It is good practice to
    #> 
    #>   ✖ not use "Depends" in DESCRIPTION, as it can cause name clashes, and poor interaction with other packages. Use "Imports" instead.
    #>   ✖ omit "Date" in DESCRIPTION. It is not required and it gets invalid quite often. A build date will be added to the package when you perform `R CMD build` on it.
    #>   ✖ add a "URL" field to DESCRIPTION. It helps users find information about your package online. If your package does not have a homepage, add an URL to GitHub, or the CRAN package package page.
    #>   ✖ add a "BugReports" field to DESCRIPTION, and point it to a bug tracker. Many online code hosting services provide bug trackers for free, https://github.com, https://gitlab.com, etc.
    #>   ✖ add a README.md (or README.Rmd) file to the top-level directory. A good README describes what the package does, how to install it, and includes a short example.
    #>   ✖ add a NEWS.md file to track user-visible changes between releases. See <https://style.tidyverse.org/news.html> for formatting guidance.
    #>   ✖ use the .R file extension for R scripts, not .r or .q. CRAN requires the uppercase .R extension.
    #> 
    #>     'R/bad_extension.r'
    #> 
    #>   ✖ omit trailing semicolons from code lines. They are not needed and most R coding standards forbid them
    #> 
    #>     'R/semicolons.R:4:30'
    #>     'R/semicolons.R:5:29'
    #>     'R/semicolons.R:9:38'
    #> 
    #>   ✖ not import packages as a whole, as this can cause name clashes between the imported packages, especially over time as packages change. Instead, import only the specific functions you need.
    #>   ✖ fix this R CMD check ERROR: VignetteBuilder package not declared: ‘knitr’ See section ‘The DESCRIPTION file’ in the ‘Writing R Extensions’ manual.
    #> ───────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────

``` r
# show all available checks
# all_checks()

# run only a specific check
g_url <- gp(pkg_path, checks = "description_url")
g_url
```

    #> ── GP badpackage ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────
    #> 
    #> It is good practice to
    #> 
    #>   ✖ add a "URL" field to DESCRIPTION. It helps users find information about your package online. If your package does not have a homepage, add an URL to GitHub, or the CRAN package package page.
    #> ───────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────

``` r
# which checks were carried out?
checks(g_url)
```

    #> [1] "description_url"

``` r
# which checks failed?
failed_checks(g)
```

    #>  [1] "no_description_depends"                
    #>  [2] "no_description_date"                   
    #>  [3] "description_url"                       
    #>  [4] "description_bugreports"                
    #>  [5] "has_readme"                            
    #>  [6] "has_news"                              
    #>  [7] "r_file_extension"                      
    #>  [8] "lintr_semicolon_linter"                
    #>  [9] "no_import_package_as_a_whole"          
    #> [10] "rcmdcheck_package_dependencies_present"

``` r
# show the first 5 checks carried out and their results
results(g)[1:5,]
```

    #>                    check passed
    #> 1                   covr     NA
    #> 2              cyclocomp   TRUE
    #> 3 no_description_depends  FALSE
    #> 4    no_description_date  FALSE
    #> 5        description_url  FALSE

## Contributing

We welcome any and all contributions to this package. See
[CONTRIBUTING.md](https://docs.ropensci.org/goodpractice/CONTRIBUTING.html)
for details.

## License

MIT © 2022 Ascent Digital Services UK Limited

## Contributors


<!-- ALL-CONTRIBUTORS-LIST:START - Do not remove or modify this section -->
<!-- prettier-ignore-start -->
<!-- markdownlint-disable -->

All contributions to this project are gratefully acknowledged using the [`allcontributors` package](https://github.com/ropensci/allcontributors) following the [allcontributors](https://allcontributors.org) specification. Contributions of any kind are welcome!

### Code

<table>

<tr>
<td align="center">
<a href="https://github.com/drmowinckels">
<img src="https://avatars.githubusercontent.com/u/14014329?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=drmowinckels">drmowinckels</a>
</td>
<td align="center">
<a href="https://github.com/mpadge">
<img src="https://avatars.githubusercontent.com/u/6697851?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=mpadge">mpadge</a>
</td>
<td align="center">
<a href="https://github.com/gaborcsardi">
<img src="https://avatars.githubusercontent.com/u/660288?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=gaborcsardi">gaborcsardi</a>
</td>
<td align="center">
<a href="https://github.com/hfrick">
<img src="https://avatars.githubusercontent.com/u/12950918?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=hfrick">hfrick</a>
</td>
<td align="center">
<a href="https://github.com/owenjonesuob">
<img src="https://avatars.githubusercontent.com/u/21007837?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=owenjonesuob">owenjonesuob</a>
</td>
<td align="center">
<a href="https://github.com/ddbortoli">
<img src="https://avatars.githubusercontent.com/u/25244497?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=ddbortoli">ddbortoli</a>
</td>
<td align="center">
<a href="https://github.com/KarinaMarks">
<img src="https://avatars.githubusercontent.com/u/20110563?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=KarinaMarks">KarinaMarks</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/olivroy">
<img src="https://avatars.githubusercontent.com/u/52606734?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=olivroy">olivroy</a>
</td>
<td align="center">
<a href="https://github.com/dougmet">
<img src="https://avatars.githubusercontent.com/u/5878305?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=dougmet">dougmet</a>
</td>
<td align="center">
<a href="https://github.com/maelle">
<img src="https://avatars.githubusercontent.com/u/8360597?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=maelle">maelle</a>
</td>
<td align="center">
<a href="https://github.com/fabian-s">
<img src="https://avatars.githubusercontent.com/u/998541?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=fabian-s">fabian-s</a>
</td>
<td align="center">
<a href="https://github.com/noamross">
<img src="https://avatars.githubusercontent.com/u/571752?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=noamross">noamross</a>
</td>
<td align="center">
<a href="https://github.com/MichaelChirico">
<img src="https://avatars.githubusercontent.com/u/7606389?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=MichaelChirico">MichaelChirico</a>
</td>
<td align="center">
<a href="https://github.com/fkohrt">
<img src="https://avatars.githubusercontent.com/u/12914806?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=fkohrt">fkohrt</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/anasimmons">
<img src="https://avatars.githubusercontent.com/u/95026699?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=anasimmons">anasimmons</a>
</td>
<td align="center">
<a href="https://github.com/andrewl776">
<img src="https://avatars.githubusercontent.com/u/64008720?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=andrewl776">andrewl776</a>
</td>
<td align="center">
<a href="https://github.com/HAlexander23">
<img src="https://avatars.githubusercontent.com/u/70958859?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=HAlexander23">HAlexander23</a>
</td>
<td align="center">
<a href="https://github.com/jsta">
<img src="https://avatars.githubusercontent.com/u/7844578?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=jsta">jsta</a>
</td>
<td align="center">
<a href="https://github.com/LiNk-NY">
<img src="https://avatars.githubusercontent.com/u/4392950?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=LiNk-NY">LiNk-NY</a>
</td>
<td align="center">
<a href="https://github.com/nfultz">
<img src="https://avatars.githubusercontent.com/u/418638?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=nfultz">nfultz</a>
</td>
<td align="center">
<a href="https://github.com/russHyde">
<img src="https://avatars.githubusercontent.com/u/7734886?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=russHyde">russHyde</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/marberts">
<img src="https://avatars.githubusercontent.com/u/62676717?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=marberts">marberts</a>
</td>
</tr>

</table>


### Issue Authors

<table>

<tr>
<td align="center">
<a href="https://github.com/peterhurford">
<img src="https://avatars.githubusercontent.com/u/5100840?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Apeterhurford">peterhurford</a>
</td>
<td align="center">
<a href="https://github.com/stillmatic">
<img src="https://avatars.githubusercontent.com/u/4743676?u=ac3fc940694e8fea8013e09465b70ba9980e1482&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Astillmatic">stillmatic</a>
</td>
<td align="center">
<a href="https://github.com/eribul">
<img src="https://avatars.githubusercontent.com/u/7790927?u=3b760b9b319fe68178c9b19a56196028b311f743&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Aeribul">eribul</a>
</td>
<td align="center">
<a href="https://github.com/richelbilderbeek">
<img src="https://avatars.githubusercontent.com/u/2098230?u=eb0321e52ef07f95057edef9a69350878b4e7a0d&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Arichelbilderbeek">richelbilderbeek</a>
</td>
<td align="center">
<a href="https://github.com/nathaneastwood">
<img src="https://avatars.githubusercontent.com/u/9799530?u=59f2fcbe9ba5ec17471672f351e10014fbd92068&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Anathaneastwood">nathaneastwood</a>
</td>
<td align="center">
<a href="https://github.com/daroczig">
<img src="https://avatars.githubusercontent.com/u/495736?u=0b78c574d39075f2b7e0b8c059cf3be834241a95&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Adaroczig">daroczig</a>
</td>
<td align="center">
<a href="https://github.com/mdozmorov">
<img src="https://avatars.githubusercontent.com/u/864945?u=73720ea73dedcc007bc666956113fc22d3d3cac2&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Amdozmorov">mdozmorov</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/vdicolab">
<img src="https://avatars.githubusercontent.com/u/5879492?u=47750b7543a8a588494ef89aa7db563a3a8c6384&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Avdicolab">vdicolab</a>
</td>
<td align="center">
<a href="https://github.com/barryrowlingson">
<img src="https://avatars.githubusercontent.com/u/888980?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Abarryrowlingson">barryrowlingson</a>
</td>
<td align="center">
<a href="https://github.com/HenrikBengtsson">
<img src="https://avatars.githubusercontent.com/u/1616850?u=3db13be6479d854fd363b262ae8d379dbd982f91&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3AHenrikBengtsson">HenrikBengtsson</a>
</td>
<td align="center">
<a href="https://github.com/kbenoit">
<img src="https://avatars.githubusercontent.com/u/2182246?u=7b5ff3ddf6c621c156bad611654bbeca6cbe128c&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Akbenoit">kbenoit</a>
</td>
<td align="center">
<a href="https://github.com/erleholgersen">
<img src="https://avatars.githubusercontent.com/u/5060086?u=4ab0621f32b784ff78ed98ba2c6c763601db1aa0&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Aerleholgersen">erleholgersen</a>
</td>
<td align="center">
<a href="https://github.com/adfi">
<img src="https://avatars.githubusercontent.com/u/1760262?u=9d618125bc4ade9569e7122611735eb3673c6b2c&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Aadfi">adfi</a>
</td>
<td align="center">
<a href="https://github.com/maurolepore">
<img src="https://avatars.githubusercontent.com/u/5856545?u=640bf30b4798fd06becb5a22d56f38d63cab78d2&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Amaurolepore">maurolepore</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/jasonserviss">
<img src="https://avatars.githubusercontent.com/u/11668021?u=f4e519e3edf5f12658b0e440403afd29b0ac39bf&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Ajasonserviss">jasonserviss</a>
</td>
<td align="center">
<a href="https://github.com/jackwasey">
<img src="https://avatars.githubusercontent.com/u/7832270?u=5d02a5e29f6fc02f17ad8277f8b6caa33bd143ad&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Ajackwasey">jackwasey</a>
</td>
<td align="center">
<a href="https://github.com/dragosmg">
<img src="https://avatars.githubusercontent.com/u/13176361?u=971be341a6706a4dc6e6e68458dc3eddee5efbac&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Adragosmg">dragosmg</a>
</td>
<td align="center">
<a href="https://github.com/Bisaloo">
<img src="https://avatars.githubusercontent.com/u/10783929?u=92fdf68eafbcdcf8da283b460ab65a2cafb8c7fe&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3ABisaloo">Bisaloo</a>
</td>
<td align="center">
<a href="https://github.com/bfgray3">
<img src="https://avatars.githubusercontent.com/u/20310144?u=12e6c4b62ad37dc6ff193cbaec61a7828a11b91a&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Abfgray3">bfgray3</a>
</td>
<td align="center">
<a href="https://github.com/wlandau">
<img src="https://avatars.githubusercontent.com/u/1580860?u=6ed1edc717e0853259312206ae59a3aa81fe3bbc&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Awlandau">wlandau</a>
</td>
<td align="center">
<a href="https://github.com/florianm">
<img src="https://avatars.githubusercontent.com/u/762815?u=4cc24043df2493b6395f3d6c1dbc83491ef23163&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Aflorianm">florianm</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/wibeasley">
<img src="https://avatars.githubusercontent.com/u/1372890?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Awibeasley">wibeasley</a>
</td>
<td align="center">
<a href="https://github.com/dpprdan">
<img src="https://avatars.githubusercontent.com/u/1423562?u=641a09e8d193d9a34951e623a97a8ab67e8bf3e4&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Adpprdan">dpprdan</a>
</td>
<td align="center">
<a href="https://github.com/kwstat">
<img src="https://avatars.githubusercontent.com/u/484260?u=31a8a31b1c2877c2abda463a8332684f69ec6341&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Akwstat">kwstat</a>
</td>
<td align="center">
<a href="https://github.com/HenningLorenzen-ext-bayer">
<img src="https://avatars.githubusercontent.com/u/89191115?u=c3f1c23c102201cfc09c9365b6f0f5e8a608f291&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3AHenningLorenzen-ext-bayer">HenningLorenzen-ext-bayer</a>
</td>
<td align="center">
<a href="https://github.com/mccroweyclinton-EPA">
<img src="https://avatars.githubusercontent.com/u/37708338?u=58f68265725538054769fe2d2cd39a01bacb3161&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Amccroweyclinton-EPA">mccroweyclinton-EPA</a>
</td>
<td align="center">
<a href="https://github.com/njtierney">
<img src="https://avatars.githubusercontent.com/u/6488485?u=3eacd57f61342d1c3cecd5c8ac741b1c4897e1de&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Anjtierney">njtierney</a>
</td>
<td align="center">
<a href="https://github.com/abichat">
<img src="https://avatars.githubusercontent.com/u/30300749?u=42f51df6ad8d29d7d61c6b640bc1588422a1d517&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Aabichat">abichat</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/damianooldoni">
<img src="https://avatars.githubusercontent.com/u/33662631?u=7c81d46771e348bf1e7b0f2b8ec31b7cdbf72ea3&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Adamianooldoni">damianooldoni</a>
</td>
<td align="center">
<a href="https://github.com/JesseAlderliesten">
<img src="https://avatars.githubusercontent.com/u/59017087?u=4a82cd707aadbfb30225ef9641a6649b21feb5a7&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3AJesseAlderliesten">JesseAlderliesten</a>
</td>
<td align="center">
<a href="https://github.com/wkmor1">
<img src="https://avatars.githubusercontent.com/u/1680870?u=1171ebc0a6b1af5eb5f19f62247cb93bc0db8622&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Awkmor1">wkmor1</a>
</td>
</tr>

</table>


### Issue Contributors

<table>

<tr>
<td align="center">
<a href="https://github.com/drisso">
<img src="https://avatars.githubusercontent.com/u/8451432?u=e5a0e3d08f829497bdfc50b368902698ac3ab9d2&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+commenter%3Adrisso">drisso</a>
</td>
<td align="center">
<a href="https://github.com/joelnitta">
<img src="https://avatars.githubusercontent.com/u/13459362?u=205e0d1ae66884c3b361a78dd6802cd10160b041&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+commenter%3Ajoelnitta">joelnitta</a>
</td>
<td align="center">
<a href="https://github.com/annakrystalli">
<img src="https://avatars.githubusercontent.com/u/5583057?u=1e992c8f98d38a959fa064afaeb182bee6e9f666&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+commenter%3Aannakrystalli">annakrystalli</a>
</td>
<td align="center">
<a href="https://github.com/sda030">
<img src="https://avatars.githubusercontent.com/u/13221371?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+commenter%3Asda030">sda030</a>
</td>
<td align="center">
<a href="https://github.com/etiennebacher">
<img src="https://avatars.githubusercontent.com/u/52219252?u=66331618d799d2d4567ecab2812236c9928be368&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+commenter%3Aetiennebacher">etiennebacher</a>
</td>
</tr>

</table>

<!-- markdownlint-enable -->
<!-- prettier-ignore-end -->
<!-- ALL-CONTRIBUTORS-LIST:END -->

# goodpractice <img src="man/figures/logo.png" align="right" width="20%" height="20%" />

<!-- badges: start -->
[![R-CMD-check](https://github.com/ropensci-review-tools/goodpractice/workflows/R-CMD-check/badge.svg)](https://github.com/ropensci-review-tools/goodpractice/actions)
[![CRAN
status](https://www.r-pkg.org/badges/version/goodpractice)](https://CRAN.R-project.org/package=goodpractice)
[![CRAN RStudio mirror
downloads](https://cranlogs.r-pkg.org/badges/goodpractice)](https://www.r-pkg.org/pkg/goodpractice)
[![Codecov test
coverage](https://codecov.io/gh/ropensci-review-tools/goodpractice/branch/main/graph/badge.svg)](https://app.codecov.io/gh/ropensci-review-tools/goodpractice?branch=main)
<!-- badges: end -->

## Advice on R Package Building

Give advice about good practices when building R packages. Advice
includes functions and syntax to avoid, package structure, code
complexity, code formatting, etc.

## Installation

You can install the release version from CRAN:

``` r
install.packages("goodpractice")
```

The development version can be installed from R-universe by enabling the
[ropensci-review-tools
`r-universe`](https://ropensci-review-tools.r-universe.dev/)
and installing the usual way:

``` r
options(repos = c(
    ropenscireviewtools = "https://ropensci-review-tools.r-universe.dev",
    CRAN = "https://cloud.r-project.org"
))
install.packages("goodpractice")
```

Or the development version can be installed directly from GitHub with:

``` r
pak::pak("ropensci-review-tools/goodpractice")
```

## Usage

``` r
library(goodpractice)
gp("<my-package>")
```

## Example

``` r
library(goodpractice)
# use example package contained in the goodpractice package
pkg_path <- system.file("bad1", package = "goodpractice")
g <- gp(pkg_path)
g
```

    #> ── GP badpackage ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────
    #> 
    #> It is good practice to
    #> 
    #>   ✖ not use "Depends" in DESCRIPTION, as it can cause name clashes, and poor interaction with other packages. Use "Imports" instead.
    #>   ✖ omit "Date" in DESCRIPTION. It is not required and it gets invalid quite often. A build date will be added to the package when you perform `R CMD build` on it.
    #>   ✖ add a "URL" field to DESCRIPTION. It helps users find information about your package online. If your package does not have a homepage, add an URL to GitHub, or the CRAN package package page.
    #>   ✖ add a "BugReports" field to DESCRIPTION, and point it to a bug tracker. Many online code hosting services provide bug trackers for free, https://github.com, https://gitlab.com, etc.
    #>   ✖ add a README.md (or README.Rmd) file to the top-level directory. A good README describes what the package does, how to install it, and includes a short example.
    #>   ✖ add a NEWS.md file to track user-visible changes between releases. See <https://style.tidyverse.org/news.html> for formatting guidance.
    #>   ✖ use the .R file extension for R scripts, not .r or .q. CRAN requires the uppercase .R extension.
    #> 
    #>     'R/bad_extension.r'
    #> 
    #>   ✖ omit trailing semicolons from code lines. They are not needed and most R coding standards forbid them
    #> 
    #>     'R/semicolons.R:4:30'
    #>     'R/semicolons.R:5:29'
    #>     'R/semicolons.R:9:38'
    #> 
    #>   ✖ not import packages as a whole, as this can cause name clashes between the imported packages, especially over time as packages change. Instead, import only the specific functions you need.
    #>   ✖ fix this R CMD check ERROR: VignetteBuilder package not declared: ‘knitr’ See section ‘The DESCRIPTION file’ in the ‘Writing R Extensions’ manual.
    #> ───────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────

``` r
# show all available checks
# all_checks()

# run only a specific check
g_url <- gp(pkg_path, checks = "description_url")
g_url
```

    #> ── GP badpackage ──────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────
    #> 
    #> It is good practice to
    #> 
    #>   ✖ add a "URL" field to DESCRIPTION. It helps users find information about your package online. If your package does not have a homepage, add an URL to GitHub, or the CRAN package package page.
    #> ───────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────────

``` r
# which checks were carried out?
checks(g_url)
```

    #> [1] "description_url"

``` r
# which checks failed?
failed_checks(g)
```

    #>  [1] "no_description_depends"                
    #>  [2] "no_description_date"                   
    #>  [3] "description_url"                       
    #>  [4] "description_bugreports"                
    #>  [5] "has_readme"                            
    #>  [6] "has_news"                              
    #>  [7] "r_file_extension"                      
    #>  [8] "lintr_semicolon_linter"                
    #>  [9] "no_import_package_as_a_whole"          
    #> [10] "rcmdcheck_package_dependencies_present"

``` r
# show the first 5 checks carried out and their results
results(g)[1:5,]
```

    #>                    check passed
    #> 1                   covr     NA
    #> 2              cyclocomp   TRUE
    #> 3 no_description_depends  FALSE
    #> 4    no_description_date  FALSE
    #> 5        description_url  FALSE

## Contributing

We welcome any and all contributions to this package. See
[CONTRIBUTING.md](https://docs.ropensci.org/goodpractice/CONTRIBUTING.html)
for details.

## License

MIT © 2022 Ascent Digital Services UK Limited

## Contributors

<!-- ALL-CONTRIBUTORS-LIST:START - Do not remove or modify this section -->

<!-- prettier-ignore-start -->

<!-- markdownlint-disable -->

All contributions to this project are gratefully acknowledged using the [`allcontributors` package](https://github.com/ropensci/allcontributors) following the [allcontributors](https://allcontributors.org) specification. Contributions of any kind are welcome!

### Code

<table>

<tr>
<td align="center">
<a href="https://github.com/mpadge">
<img src="https://avatars.githubusercontent.com/u/6697851?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=mpadge">mpadge</a>
</td>
<td align="center">
<a href="https://github.com/drmowinckels">
<img src="https://avatars.githubusercontent.com/u/14014329?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=drmowinckels">drmowinckels</a>
</td>
<td align="center">
<a href="https://github.com/gaborcsardi">
<img src="https://avatars.githubusercontent.com/u/660288?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=gaborcsardi">gaborcsardi</a>
</td>
<td align="center">
<a href="https://github.com/hfrick">
<img src="https://avatars.githubusercontent.com/u/12950918?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=hfrick">hfrick</a>
</td>
<td align="center">
<a href="https://github.com/owenjonesuob">
<img src="https://avatars.githubusercontent.com/u/21007837?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=owenjonesuob">owenjonesuob</a>
</td>
<td align="center">
<a href="https://github.com/ddbortoli">
<img src="https://avatars.githubusercontent.com/u/25244497?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=ddbortoli">ddbortoli</a>
</td>
<td align="center">
<a href="https://github.com/KarinaMarks">
<img src="https://avatars.githubusercontent.com/u/20110563?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=KarinaMarks">KarinaMarks</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/olivroy">
<img src="https://avatars.githubusercontent.com/u/52606734?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=olivroy">olivroy</a>
</td>
<td align="center">
<a href="https://github.com/JesseAlderliesten">
<img src="https://avatars.githubusercontent.com/u/59017087?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=JesseAlderliesten">JesseAlderliesten</a>
</td>
<td align="center">
<a href="https://github.com/dougmet">
<img src="https://avatars.githubusercontent.com/u/5878305?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=dougmet">dougmet</a>
</td>
<td align="center">
<a href="https://github.com/maelle">
<img src="https://avatars.githubusercontent.com/u/8360597?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=maelle">maelle</a>
</td>
<td align="center">
<a href="https://github.com/fabian-s">
<img src="https://avatars.githubusercontent.com/u/998541?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=fabian-s">fabian-s</a>
</td>
<td align="center">
<a href="https://github.com/MichaelChirico">
<img src="https://avatars.githubusercontent.com/u/7606389?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=MichaelChirico">MichaelChirico</a>
</td>
<td align="center">
<a href="https://github.com/noamross">
<img src="https://avatars.githubusercontent.com/u/571752?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=noamross">noamross</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/fkohrt">
<img src="https://avatars.githubusercontent.com/u/12914806?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=fkohrt">fkohrt</a>
</td>
<td align="center">
<a href="https://github.com/anasimmons">
<img src="https://avatars.githubusercontent.com/u/95026699?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=anasimmons">anasimmons</a>
</td>
<td align="center">
<a href="https://github.com/andrewl776">
<img src="https://avatars.githubusercontent.com/u/64008720?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=andrewl776">andrewl776</a>
</td>
<td align="center">
<a href="https://github.com/HAlexander23">
<img src="https://avatars.githubusercontent.com/u/70958859?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=HAlexander23">HAlexander23</a>
</td>
<td align="center">
<a href="https://github.com/jsta">
<img src="https://avatars.githubusercontent.com/u/7844578?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=jsta">jsta</a>
</td>
<td align="center">
<a href="https://github.com/LiNk-NY">
<img src="https://avatars.githubusercontent.com/u/4392950?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=LiNk-NY">LiNk-NY</a>
</td>
<td align="center">
<a href="https://github.com/nfultz">
<img src="https://avatars.githubusercontent.com/u/418638?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=nfultz">nfultz</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/russHyde">
<img src="https://avatars.githubusercontent.com/u/7734886?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=russHyde">russHyde</a>
</td>
<td align="center">
<a href="https://github.com/marberts">
<img src="https://avatars.githubusercontent.com/u/62676717?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/commits?author=marberts">marberts</a>
</td>
</tr>

</table>


### Issue Authors

<table>

<tr>
<td align="center">
<a href="https://github.com/peterhurford">
<img src="https://avatars.githubusercontent.com/u/5100840?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Apeterhurford">peterhurford</a>
</td>
<td align="center">
<a href="https://github.com/stillmatic">
<img src="https://avatars.githubusercontent.com/u/4743676?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Astillmatic">stillmatic</a>
</td>
<td align="center">
<a href="https://github.com/eribul">
<img src="https://avatars.githubusercontent.com/u/7790927?u=3b760b9b319fe68178c9b19a56196028b311f743&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Aeribul">eribul</a>
</td>
<td align="center">
<a href="https://github.com/richelbilderbeek">
<img src="https://avatars.githubusercontent.com/u/2098230?u=eb0321e52ef07f95057edef9a69350878b4e7a0d&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Arichelbilderbeek">richelbilderbeek</a>
</td>
<td align="center">
<a href="https://github.com/nathaneastwood">
<img src="https://avatars.githubusercontent.com/u/9799530?u=59f2fcbe9ba5ec17471672f351e10014fbd92068&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Anathaneastwood">nathaneastwood</a>
</td>
<td align="center">
<a href="https://github.com/daroczig">
<img src="https://avatars.githubusercontent.com/u/495736?u=ec8564ccd2a077fa010b2f3b10a90f4830608f61&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Adaroczig">daroczig</a>
</td>
<td align="center">
<a href="https://github.com/mdozmorov">
<img src="https://avatars.githubusercontent.com/u/864945?u=73720ea73dedcc007bc666956113fc22d3d3cac2&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Amdozmorov">mdozmorov</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/vdicolab">
<img src="https://avatars.githubusercontent.com/u/5879492?u=47750b7543a8a588494ef89aa7db563a3a8c6384&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Avdicolab">vdicolab</a>
</td>
<td align="center">
<a href="https://github.com/barryrowlingson">
<img src="https://avatars.githubusercontent.com/u/888980?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Abarryrowlingson">barryrowlingson</a>
</td>
<td align="center">
<a href="https://github.com/HenrikBengtsson">
<img src="https://avatars.githubusercontent.com/u/1616850?u=3db13be6479d854fd363b262ae8d379dbd982f91&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3AHenrikBengtsson">HenrikBengtsson</a>
</td>
<td align="center">
<a href="https://github.com/kbenoit">
<img src="https://avatars.githubusercontent.com/u/2182246?u=7b5ff3ddf6c621c156bad611654bbeca6cbe128c&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Akbenoit">kbenoit</a>
</td>
<td align="center">
<a href="https://github.com/erleholgersen">
<img src="https://avatars.githubusercontent.com/u/5060086?u=4ab0621f32b784ff78ed98ba2c6c763601db1aa0&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Aerleholgersen">erleholgersen</a>
</td>
<td align="center">
<a href="https://github.com/adfi">
<img src="https://avatars.githubusercontent.com/u/1760262?u=9d618125bc4ade9569e7122611735eb3673c6b2c&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Aadfi">adfi</a>
</td>
<td align="center">
<a href="https://github.com/maurolepore">
<img src="https://avatars.githubusercontent.com/u/5856545?u=640bf30b4798fd06becb5a22d56f38d63cab78d2&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Amaurolepore">maurolepore</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/jasonserviss">
<img src="https://avatars.githubusercontent.com/u/11668021?u=f4e519e3edf5f12658b0e440403afd29b0ac39bf&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Ajasonserviss">jasonserviss</a>
</td>
<td align="center">
<a href="https://github.com/jackwasey">
<img src="https://avatars.githubusercontent.com/u/7832270?u=5d02a5e29f6fc02f17ad8277f8b6caa33bd143ad&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Ajackwasey">jackwasey</a>
</td>
<td align="center">
<a href="https://github.com/dragosmg">
<img src="https://avatars.githubusercontent.com/u/13176361?u=971be341a6706a4dc6e6e68458dc3eddee5efbac&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Adragosmg">dragosmg</a>
</td>
<td align="center">
<a href="https://github.com/Bisaloo">
<img src="https://avatars.githubusercontent.com/u/10783929?u=92fdf68eafbcdcf8da283b460ab65a2cafb8c7fe&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3ABisaloo">Bisaloo</a>
</td>
<td align="center">
<a href="https://github.com/bfgray3">
<img src="https://avatars.githubusercontent.com/u/20310144?u=12e6c4b62ad37dc6ff193cbaec61a7828a11b91a&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Abfgray3">bfgray3</a>
</td>
<td align="center">
<a href="https://github.com/wlandau">
<img src="https://avatars.githubusercontent.com/u/1580860?u=6ed1edc717e0853259312206ae59a3aa81fe3bbc&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Awlandau">wlandau</a>
</td>
<td align="center">
<a href="https://github.com/florianm">
<img src="https://avatars.githubusercontent.com/u/762815?u=4cc24043df2493b6395f3d6c1dbc83491ef23163&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Aflorianm">florianm</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/wibeasley">
<img src="https://avatars.githubusercontent.com/u/1372890?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Awibeasley">wibeasley</a>
</td>
<td align="center">
<a href="https://github.com/dpprdan">
<img src="https://avatars.githubusercontent.com/u/1423562?u=641a09e8d193d9a34951e623a97a8ab67e8bf3e4&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Adpprdan">dpprdan</a>
</td>
<td align="center">
<a href="https://github.com/kwstat">
<img src="https://avatars.githubusercontent.com/u/484260?u=31a8a31b1c2877c2abda463a8332684f69ec6341&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Akwstat">kwstat</a>
</td>
<td align="center">
<a href="https://github.com/HenningLorenzen-ext-bayer">
<img src="https://avatars.githubusercontent.com/u/89191115?u=c3f1c23c102201cfc09c9365b6f0f5e8a608f291&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3AHenningLorenzen-ext-bayer">HenningLorenzen-ext-bayer</a>
</td>
<td align="center">
<a href="https://github.com/mccroweyclinton-EPA">
<img src="https://avatars.githubusercontent.com/u/37708338?u=58f68265725538054769fe2d2cd39a01bacb3161&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Amccroweyclinton-EPA">mccroweyclinton-EPA</a>
</td>
<td align="center">
<a href="https://github.com/njtierney">
<img src="https://avatars.githubusercontent.com/u/6488485?u=3eacd57f61342d1c3cecd5c8ac741b1c4897e1de&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Anjtierney">njtierney</a>
</td>
<td align="center">
<a href="https://github.com/abichat">
<img src="https://avatars.githubusercontent.com/u/30300749?u=42f51df6ad8d29d7d61c6b640bc1588422a1d517&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Aabichat">abichat</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/damianooldoni">
<img src="https://avatars.githubusercontent.com/u/33662631?u=7c81d46771e348bf1e7b0f2b8ec31b7cdbf72ea3&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Adamianooldoni">damianooldoni</a>
</td>
<td align="center">
<a href="https://github.com/wkmor1">
<img src="https://avatars.githubusercontent.com/u/1680870?u=1171ebc0a6b1af5eb5f19f62247cb93bc0db8622&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Awkmor1">wkmor1</a>
</td>
<td align="center">
<a href="https://github.com/christian-million">
<img src="https://avatars.githubusercontent.com/u/34721595?u=0868ba5b181f86dee3a53e93602efb5f76307e84&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3Achristian-million">christian-million</a>
</td>
<td align="center">
<a href="https://github.com/TanguyBarthelemy">
<img src="https://avatars.githubusercontent.com/u/112711017?u=7317b68f88a56107649c09bb8e2eefdc94c2d70e&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3ATanguyBarthelemy">TanguyBarthelemy</a>
</td>
<td align="center">
<a href="https://github.com/Aariq">
<img src="https://avatars.githubusercontent.com/u/25404783?u=bf39b8163e91fb40423676c1806a9fc1ed665c0c&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+author%3AAariq">Aariq</a>
</td>
</tr>

</table>


### Issue Contributors

<table>

<tr>
<td align="center">
<a href="https://github.com/drisso">
<img src="https://avatars.githubusercontent.com/u/8451432?u=e5a0e3d08f829497bdfc50b368902698ac3ab9d2&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+commenter%3Adrisso">drisso</a>
</td>
<td align="center">
<a href="https://github.com/joelnitta">
<img src="https://avatars.githubusercontent.com/u/13459362?u=205e0d1ae66884c3b361a78dd6802cd10160b041&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+commenter%3Ajoelnitta">joelnitta</a>
</td>
<td align="center">
<a href="https://github.com/annakrystalli">
<img src="https://avatars.githubusercontent.com/u/5583057?u=1e992c8f98d38a959fa064afaeb182bee6e9f666&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+commenter%3Aannakrystalli">annakrystalli</a>
</td>
<td align="center">
<a href="https://github.com/sda030">
<img src="https://avatars.githubusercontent.com/u/13221371?v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+commenter%3Asda030">sda030</a>
</td>
<td align="center">
<a href="https://github.com/etiennebacher">
<img src="https://avatars.githubusercontent.com/u/52219252?u=66331618d799d2d4567ecab2812236c9928be368&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+commenter%3Aetiennebacher">etiennebacher</a>
</td>
<td align="center">
<a href="https://github.com/TimTaylor">
<img src="https://avatars.githubusercontent.com/u/43499035?u=db4f4432cbb6c914ee30b1ebffdf1b2af1acd316&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+commenter%3ATimTaylor">TimTaylor</a>
</td>
<td align="center">
<a href="https://github.com/jonthegeek">
<img src="https://avatars.githubusercontent.com/u/33983824?u=ac41aa539f5825f35c157ae3d9098a80fd0a9fec&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+commenter%3Ajonthegeek">jonthegeek</a>
</td>
</tr>


<tr>
<td align="center">
<a href="https://github.com/atsyplenkov">
<img src="https://avatars.githubusercontent.com/u/34775595?u=40fe99a840fe2fb6cf1c3540b84a0619618e6f4b&v=4" width="100px;" alt=""/>
</a><br>
<a href="https://github.com/ropensci-review-tools/goodpractice/issues?q=is%3Aissue+commenter%3Aatsyplenkov">atsyplenkov</a>
</td>
</tr>

</table>

<!-- markdownlint-enable -->

<!-- prettier-ignore-end -->

<!-- ALL-CONTRIBUTORS-LIST:END -->

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## -----------------------------------------------------------------------------
library(goodpractice)

# make a simple version of the T/F check
check_simple_tf <- make_check(
  
  description = "TRUE and FALSE is used, not T and F",
  gp = "avoid 'T' and 'F', use 'TRUE' and 'FALSE' instead.",
  check = function(state) {
      length(tools::checkTnF(dir = state$path)) == 0
  }
)

## -----------------------------------------------------------------------------
# get path to example package
pkg_path <- system.file("bad1", package = "goodpractice")
gp(pkg_path, checks = "simple_tf",
   extra_checks = list(simple_tf = check_simple_tf))

## -----------------------------------------------------------------------------
# prep: process DESCRIPTION file
desc_fun <- function(path, quiet) {
  desc::description$new(path)
}

prep_desc <- make_prep(name = "desc", func = desc_fun)

# check for an URL field
check_url <- make_check(
  description = "URL field in DESCRIPTION",
  preps = "desc",
  gp = "have a URL field in DESCRIPTION",
  check = function(state) state$desc$has_fields("URL")
)

# check for a BugReport field
check_bugreports <- make_check(
  description = "BugReports in DESCRIPTION",
  preps = "desc",
  gp = "add a BugReports field to DESCRIPTION",
  check = function(state) state$desc$has_fields('BugReports')
)

# run the two checks with their corresponding prep step
gp(pkg_path, checks = c("url", "bugreports"),
   extra_preps = list("desc" = prep_desc),
   extra_checks = list("url" = check_url, "bugreports" = check_bugreports))

D1 <-
"Package: foo
Title: Foo Package
Maintainer: foo@foofoo.com
Description: What the package does.
Depends: R (>= 2.15)
URL: https://www.foo.com
BugReports: https://www.foo.com/bugs
"

D2 <-
"Package: foo
Title: Foo Package
Maintainer: foo@foofoo.com
Description: What the package does.
Depends: foobar (>= 2.15)
Date: 2018-03-22
"

D3 <-
"Package: foo
Title: Foo Package
Maintainer: foo@foofoo.com
Description: This package does things.
"

D3_this_is_a <-
"Package: foo
Title: Foo Package
Maintainer: foo@foofoo.com
Description: This is a package to do things.
"

D3_the_pkg <-
"Package: foo
Title: Foo Package
Maintainer: foo@foofoo.com
Description: The foo package does things.
"

D_url_good <-
"Package: foo
Title: Foo Package
Maintainer: foo@foofoo.com
Description: Tools for stuff. See <https://example.com> for details.
    Based on Smith (2020) <doi:10.1234/example>.
"

D_url_bare <-
"Package: foo
Title: Foo Package
Maintainer: foo@foofoo.com
Description: Tools for stuff. See https://example.com for details.
"

D_doi_as_url <-
"Package: foo
Title: Foo Package
Maintainer: foo@foofoo.com
Description: Based on Smith (2020) <https://doi.org/10.1234/example>.
"

D_http <-
"Package: foo
Title: Foo Package
Maintainer: foo@foofoo.com
Description: Tools for stuff. See <http://example.com> for details.
"

D_dup_deps <-
"Package: foo
Title: Foo Package
Maintainer: foo@foofoo.com
Description: Does things.
Imports: dplyr
Suggests: dplyr
"

D_no_dup_deps <-
"Package: foo
Title: Foo Package
Maintainer: foo@foofoo.com
Description: Does things.
Imports: dplyr
Suggests: testthat
"

D_valid_roles <-
"Package: foo
Title: Foo Package
Authors@R: c(
    person('A', 'B', role = c('aut', 'cre')),
    person('C', 'D', role = 'ctb'))
Description: Does things.
"

D_invalid_roles <-
"Package: foo
Title: Foo Package
Authors@R: c(
    person('A', 'B', role = c('aut', 'cre')),
    person('C', 'D', role = 'inventor'))
Description: Does things.
"

D_pkg_quoted <-
"Package: foo
Title: A 'dplyr' Extension
Description: Extends 'dplyr' with more verbs.
Imports: dplyr
"

D_pkg_unquoted <-
"Package: foo
Title: A dplyr Extension
Description: Extends dplyr with more verbs.
Imports: dplyr
"

D_base_pkg_unquoted <-
"Package: foo
Title: Foo with tools
Description: Uses tools for stuff.
Imports: tools
"

D_no_deps <-
"Package: foo
Title: Foo Package
Maintainer: foo@foofoo.com
Description: Does things.
"

D_bad_authors_at_r <-
"Package: foo
Title: Foo Package
Authors@R: stop('broken')
Description: Does things.
"

test_that("all checks return NA on try-error state", {
  state <- list(description = structure("error", class = "try-error",
    condition = simpleError("test")))

  check_names <- c(
    "no_description_depends", "no_description_date",
    "description_url", "description_not_start_with_package",
    "description_urls_in_angle_brackets", "description_doi_format",
    "description_urls_not_http", "no_description_duplicate_deps",
    "description_valid_roles", "description_pkgname_single_quoted",
    "description_bugreports"
  )

  for (nm in check_names) {
    result <- CHECKS[[nm]]$check(state)
    expect_identical(result$status, NA, label = paste(nm, "on try-error"))
    expect_identical(
      result$positions, list(),
      label = paste(nm, "positions on try-error")
    )
  }
})

test_that("Depends: R is OK", {

  state <- list(description = desc::description$new(text = D1))
  expect_true(
    CHECKS$no_description_depends$check(state)$status
  )

  state <- list(description = desc::description$new(text = D2))
  expect_false(
    CHECKS$no_description_depends$check(state)$status
  )
})

test_that("Date", {

  state <- list(description = desc::description$new(text = D1))
  expect_true(
    CHECKS$no_description_date$check(state)$status
  )

  state <- list(description = desc::description$new(text = D2))
  expect_false(
    CHECKS$no_description_date$check(state)$status
  )
})

test_that("URL", {

  state <- list(description = desc::description$new(text = D1))
  expect_true(
    CHECKS$description_url$check(state)$status
  )

  state <- list(description = desc::description$new(text = D2))
  expect_false(
    CHECKS$description_url$check(state)$status
  )
})

test_that("Description not starting with package name reference", {

  state <- list(description = desc::description$new(text = D1))
  expect_true(
    CHECKS$description_not_start_with_package$check(state)$status
  )

  state <- list(description = desc::description$new(text = D3))
  expect_false(
    CHECKS$description_not_start_with_package$check(state)$status
  )

  state <- list(description = desc::description$new(text = D3_this_is_a))
  expect_false(
    CHECKS$description_not_start_with_package$check(state)$status
  )

  state <- list(description = desc::description$new(text = D3_the_pkg))
  expect_false(
    CHECKS$description_not_start_with_package$check(state)$status
  )
})

test_that("URLs in Description enclosed in angle brackets", {

  state <- list(description = desc::description$new(text = D_url_good))
  expect_true(
    CHECKS$description_urls_in_angle_brackets$check(state)$status
  )

  state <- list(description = desc::description$new(text = D_url_bare))
  expect_false(
    CHECKS$description_urls_in_angle_brackets$check(state)$status
  )
})

test_that("DOIs use <doi:...> format", {

  state <- list(description = desc::description$new(text = D_url_good))
  expect_true(
    CHECKS$description_doi_format$check(state)$status
  )

  state <- list(description = desc::description$new(text = D_doi_as_url))
  expect_false(
    CHECKS$description_doi_format$check(state)$status
  )
})

test_that("URLs use https not http", {

  state <- list(description = desc::description$new(text = D_url_good))
  expect_true(
    CHECKS$description_urls_not_http$check(state)$status
  )

  state <- list(description = desc::description$new(text = D_http))
  expect_false(
    CHECKS$description_urls_not_http$check(state)$status
  )
})

test_that("No duplicate dependencies", {

  state <- list(description = desc::description$new(text = D_no_dup_deps))
  expect_true(
    CHECKS$no_description_duplicate_deps$check(state)$status
  )

  state <- list(description = desc::description$new(text = D_dup_deps))
  expect_false(
    CHECKS$no_description_duplicate_deps$check(state)$status
  )
})

test_that("Valid author roles", {

  state <- list(description = desc::description$new(text = D_valid_roles))
  expect_true(
    CHECKS$description_valid_roles$check(state)$status
  )

  state <- list(description = desc::description$new(text = D_invalid_roles))
  expect_false(
    CHECKS$description_valid_roles$check(state)$status
  )

  state <- list(description = desc::description$new(text = D_bad_authors_at_r))
  result <- CHECKS$description_valid_roles$check(state)
  expect_identical(result$status, NA)
  expect_identical(result$positions, list())
})

test_that("Package names single-quoted in Title/Description", {

  state <- list(description = desc::description$new(text = D_pkg_quoted))
  expect_true(
    CHECKS$description_pkgname_single_quoted$check(state)$status
  )

  state <- list(description = desc::description$new(text = D_pkg_unquoted))
  expect_false(
    CHECKS$description_pkgname_single_quoted$check(state)$status
  )

  state <- list(description = desc::description$new(text = D_no_deps))
  expect_true(
    CHECKS$description_pkgname_single_quoted$check(state)$status
  )
})

test_that("Base R packages are excluded from single-quote check", {
  state <- list(description = desc::description$new(text = D_base_pkg_unquoted))
  expect_true(
    CHECKS$description_pkgname_single_quoted$check(state)$status
  )
})

test_that("BugReports", {

  state <- list(description = desc::description$new(text = D1))
  expect_true(
    CHECKS$description_bugreports$check(state)$status
  )

  state <- list(description = desc::description$new(text = D2))
  expect_false(
    CHECKS$description_bugreports$check(state)$status
  )
})

test_that("namespace checks return NA on try-error state", {
  state <- list(namespace = structure("error", class = "try-error",
    condition = simpleError("test")))

  for (nm in c("no_import_package_as_a_whole", "no_export_pattern")) {
    result <- CHECKS[[nm]]$check(state)
    expect_identical(
      result$status, NA, label = paste(nm, "status on try-error")
    )
    expect_identical(
      result$positions, list(),
      label = paste(nm, "positions on try-error")
    )
  }
})

test_that("no_import_package_as_a_whole returns status and positions", {
  state <- list(namespace = list(
    imports = list(c("pkg1", "fun1"), c("pkg2", "fun2", "fun3"))
  ))
  result <- CHECKS$no_import_package_as_a_whole$check(state)
  expect_true(result$status)
  expect_identical(result$positions, list())

  state <- list(namespace = list(
    imports = list(c("pkg1", "fun1"), "pkg2")
  ))
  result <- CHECKS$no_import_package_as_a_whole$check(state)
  expect_false(result$status)
  expect_identical(result$positions, list())
})

test_that("no_export_pattern returns list with status and positions", {
  state <- list(namespace = list(exportPatterns = character()))
  result <- CHECKS$no_export_pattern$check(state)
  expect_true(result$status)
  expect_identical(result$positions, list())

  state <- list(namespace = list(exportPatterns = "^[^.]"))
  result <- CHECKS$no_export_pattern$check(state)
  expect_false(result$status)
  expect_identical(result$positions, list())
})

test_that("importing package as a whole is not okay", {

  bad1 <- system.file("bad1", package = "goodpractice")
  x <- gp(bad1, checks = "no_import_package_as_a_whole")
  expect_false(results(x)$passed)

  x <- gp("bad2", checks = "no_import_package_as_a_whole")
  expect_true(results(x)$passed)

})

test_that("exportPattern is not okay", {

  # TODO expectation/example for "no_export_pattern" failing

  x <- gp("bad2", checks = "no_export_pattern")
  expect_true(results(x)$passed)

})

# "lintr_semicolon_linter" fails in bad1
# "lintr_attach_detach_linter"  fails in bad2
# "lintr_setwd_linter"  fails in bad2
# "lintr_sapply_linter"  fails in bad2
# "lintr_library_require_linter" fails in bad2
# "lintr_seq_linter" fails in bad3
# not failing yet:
# "lintr_assignment_linter"
# "lintr_line_length_linter"

gp_lintrs <- c(
  "lintr_assignment_linter",
  "lintr_line_length_linter",
  "lintr_semicolon_linter",
  "lintr_attach_detach_linter",
  "lintr_setwd_linter",
  "lintr_sapply_linter",
  "lintr_library_require_linter",
  "installed_packages_linter",
  "lintr_seq_linter"
)

bad1 <- system.file("bad1", package = "goodpractice")
gp_bad1 <- gp(bad1, checks = gp_lintrs)
res_bad1 <- results(gp_bad1)

gp_bad2 <- gp("bad2", checks = gp_lintrs)
res_bad2 <- results(gp_bad2)

gp_bad3 <- gp("bad3", checks = gp_lintrs)
res_bad3 <- results(gp_bad3)

get_result <- function(res, check) res$passed[res$check == check]


test_that("lintr_assignment_linter", {
  expect_true(get_result(res_bad1, "lintr_assignment_linter"))
  # TODO expectation/example where the check fails
})
test_that("lintr_line_length_linter", {
  expect_true(get_result(res_bad1, "lintr_line_length_linter"))
  # TODO expectation/example where the check fails
})
test_that("lintr_semicolon_linter", {
  expect_true(get_result(res_bad2, "lintr_semicolon_linter"))
  expect_false(get_result(res_bad1, "lintr_semicolon_linter"))
})
test_that("lintr_attach_detach_linter", {
  expect_true(get_result(res_bad1, "lintr_attach_detach_linter"))
  expect_false(get_result(res_bad2, "lintr_attach_detach_linter"))
})
test_that("lintr_setwd_linter", {
  expect_true(get_result(res_bad1, "lintr_setwd_linter"))
  expect_false(get_result(res_bad2, "lintr_setwd_linter"))
})
test_that("lintr_sapply_linter", {
  expect_true(get_result(res_bad1, "lintr_sapply_linter"))
  expect_false(get_result(res_bad2, "lintr_sapply_linter"))
})
test_that("lintr_library_require_linter", {
  # library/require is OK in vignettes and examples
  expect_true(get_result(res_bad1, "lintr_library_require_linter"))
  # but not in the R/ folder
  expect_false(get_result(res_bad2, "lintr_library_require_linter"))
})
test_that("installed_packages_linter", {
  expect_true(get_result(res_bad1, "installed_packages_linter"))
  expect_false(get_result(res_bad2, "installed_packages_linter"))
})
test_that("lintr_seq_linter", {
  expect_true(get_result(res_bad1, "lintr_seq_linter"))
  expect_false(get_result(res_bad3, "lintr_seq_linter"))
})

test_that("lintr_implicit_assignment_linter ignores test files", {
  gp_ia <- gp("bad2", checks = "lintr_implicit_assignment_linter")
  res_ia <- results(gp_ia)
  # bad2 has implicit assignment in R/ (should fail)
  expect_false(get_result(res_ia, "lintr_implicit_assignment_linter"))

  # Verify only R/ positions are reported, not tests/
  state <- gp_ia$checks$lintr_implicit_assignment_linter
  expect_true(all(vapply(
    state$positions, function(x) grepl("^R[/\\\\]", x$filename), logical(1)
  )))
})

test_that("get_lintr_state returns NA on try-error", {
  state <- list(lintr = structure("error", class = "try-error"))
  result <- get_lintr_state(state, "assignment_linter")
  expect_true(is.na(result$status))
  expect_identical(result$positions, list())
  expect_true("positions" %in% names(result))
  expect_false("position" %in% names(result))
})

test_that("lintr check fns return positions (not position) on try-error", {
  state <- list(lintr = structure("error", class = "try-error"))

  res_assignment <- CHECKS$lintr_assignment_linter$check(state)
  expect_true(is.na(res_assignment$status))
  expect_identical(res_assignment$positions, list())
  expect_true("positions" %in% names(res_assignment))
  expect_false("position" %in% names(res_assignment))

  res_library <- CHECKS$lintr_library_require_linter$check(state)
  expect_true(is.na(res_library$status))
  expect_identical(res_library$positions, list())
  expect_true("positions" %in% names(res_library))
  expect_false("position" %in% names(res_library))

  res_implicit <- CHECKS$lintr_implicit_assignment_linter$check(state)
  expect_true(is.na(res_implicit$status))
  expect_identical(res_implicit$positions, list())
  expect_true("positions" %in% names(res_implicit))
  expect_false("position" %in% names(res_implicit))
})

test_that("all new lintr checks run and return results", {
  all_lintr <- grep("^lintr_", all_checks(), value = TRUE)
  g <- gp("good", checks = all_lintr)
  res <- results(g)
  expect_gt(nrow(res), 0)
  expect_true(all(all_lintr %in% res$check))
  expect_true(all(vapply(
    res$passed, function(x) is.logical(x) || is.na(x), logical(1)
  )))
})

test_that("make_lintr_check creates valid check", {
  chk <- make_lintr_check(
    "seq_linter", "test desc", "test gp message"
  )
  expect_identical(chk$description, "test desc")
  expect_identical(chk$preps, "lintr")
  expect_true(is.function(chk$check))
})

test_that("cyclocomp check returns NA on try-error state", {
  state <- list(cyclocomp = structure("error", class = "try-error",
    condition = simpleError("test")))

  result <- CHECKS$cyclocomp$check(state)
  expect_identical(result$status, NA)
  expect_identical(result$positions, list())
})

test_that("cyclocomp check passes when all functions are simple", {
  state <- list(cyclocomp = data.frame(
    name = c("f1", "f2"),
    cyclocomp = c(1, 10),
    stringsAsFactors = FALSE
  ))

  result <- CHECKS$cyclocomp$check(state)
  expect_true(result$status)
  expect_identical(result$positions, list())
})

test_that("cyclocomp check fails when a function exceeds limit", {
  state <- list(cyclocomp = data.frame(
    name = c("f1", "complex_fun"),
    cyclocomp = c(1, 50),
    stringsAsFactors = FALSE
  ))

  result <- CHECKS$cyclocomp$check(state)
  expect_false(result$status)
  expect_identical(result$positions, list())
})

test_that("cyclocomp limit respects option", {
  state <- list(cyclocomp = data.frame(
    name = "f1",
    cyclocomp = 20,
    stringsAsFactors = FALSE
  ))

  withr::local_options(goodpractice.cyclocomp_limit = 25)
  result <- CHECKS$cyclocomp$check(state)
  expect_true(result$status)

  withr::local_options(goodpractice.cyclocomp_limit = 10)
  result <- CHECKS$cyclocomp$check(state)
  expect_false(result$status)
})

test_that("has_readme passes when README.md exists", {
  pkg <- withr::local_tempdir()
  file.copy(
    list.files("good", full.names = TRUE, recursive = TRUE),
    pkg
  )
  file.create(file.path(pkg, "README.md"))

  state <- list(path = pkg)
  result <- CHECKS$has_readme$check(state)
  expect_true(result$status)
  expect_type(result$positions, "list")
})

test_that("has_readme passes when README.Rmd exists", {
  pkg <- withr::local_tempdir()
  file.copy(
    list.files("good", full.names = TRUE, recursive = TRUE),
    pkg
  )
  file.create(file.path(pkg, "README.Rmd"))

  state <- list(path = pkg)
  result <- CHECKS$has_readme$check(state)
  expect_true(result$status)
  expect_type(result$positions, "list")
})

test_that("has_readme fails when no README", {
  state <- list(path = "good")
  result <- CHECKS$has_readme$check(state)
  expect_false(result$status)
  expect_type(result$positions, "list")
})

test_that("has_news passes when NEWS.md exists", {
  pkg <- withr::local_tempdir()
  file.copy(
    list.files("good", full.names = TRUE, recursive = TRUE),
    pkg
  )
  file.create(file.path(pkg, "NEWS.md"))

  state <- list(path = pkg)
  result <- CHECKS$has_news$check(state)
  expect_true(result$status)
  expect_type(result$positions, "list")
})

test_that("has_news passes when inst/NEWS.Rd exists", {
  pkg <- withr::local_tempdir()
  file.copy(
    list.files("good", full.names = TRUE, recursive = TRUE),
    pkg
  )
  dir.create(file.path(pkg, "inst"), showWarnings = FALSE)
  file.create(file.path(pkg, "inst", "NEWS.Rd"))

  state <- list(path = pkg)
  result <- CHECKS$has_news$check(state)
  expect_true(result$status)
  expect_type(result$positions, "list")
})

test_that("has_news fails when no NEWS", {
  state <- list(path = "good")
  result <- CHECKS$has_news$check(state)
  expect_false(result$status)
  expect_type(result$positions, "list")
})

test_that("warning if prep step fails", {
  bad1 <- system.file("bad1", package = "goodpractice")

  expect_warning(gp(bad1, checks = "covr"),
                 "Prep step for.*covr.*failed")
})

test_that("covr check passes with NaN coverage (no testable code)", {
  state <- list(covr = list(
    coverage = structure(list(), class = "coverage"),
    zero = data.frame(
      filename = character(), functions = character(),
      first_line = integer(), last_line = integer(),
      first_column = integer(), last_column = integer(),
      value = numeric(), stringsAsFactors = FALSE
    ),
    pct_by_line = NaN,
    pct_by_expr = NaN
  ))
  result <- CHECKS$covr$check(state)
  expect_true(result$status)
})

test_that("covr gp message handles NaN coverage", {
  state <- list(covr = list(pct_by_line = NaN))
  msg <- CHECKS$covr$gp(state)
  expect_match(msg, "No testable code")
})

test_that("run_prep_step passes multiple ... args to fn", {
  state <- list()
  state <- run_prep_step(state, "test", function(a, b, c) {
    list(sum = a + b + c)
  }, a = 1, b = 2, c = 3)
  expect_identical(state$test$sum, 6)
})

#' @include lists.R

CHECKS$has_readme <- make_check(

  description = "README file exists",
  tags = c("info", "documentation"),
  preps = "package_structure",

  gp = 'add a {.file README.md} (or {.file README.Rmd}) file to the top-level
        directory. A good README describes what the package does,
        how to install it, and includes a short example.',

  check = function(state) {
    readme_patterns <- c(
      "README.md", "README.Rmd", "README.qmd", "README"
    )
    found <- any(file.exists(file.path(state$path, readme_patterns)))
    check_result(found)
  }
)

## --------------------------------------------------------------------

CHECKS$has_news <- make_check(

  description = "NEWS file exists",
  tags = c("info", "documentation"),
  preps = "package_structure",

  gp = 'add a {.file NEWS.md} file to track user-visible changes between
        releases. See {.url https://style.tidyverse.org/news.html}
        for formatting guidance.',

  check = function(state) {
    news_patterns <- c("NEWS.md", "NEWS", "NEWS.Rd", "inst/NEWS.Rd")
    found <- any(file.exists(file.path(state$path, news_patterns)))
    check_result(found)
  }
)

## --------------------------------------------------------------------

CHECKS$r_file_extension <- make_check(

  description = "R scripts use .R file extension, not .r or .q",
  tags = c("CRAN", "warning"),
  preps = "package_structure",

  gp = paste(
    "use the {.file .R} file extension for R scripts,",
    "not {.file .r} or {.file .q}.",
    "CRAN requires the uppercase {.file .R} extension."
  ),

  check = function(state) {
    path <- state$path
    rdir <- file.path(path, "R")

    if (!dir.exists(rdir)) {
      return(check_result(TRUE))
    }

    files <- list.files(rdir, pattern = "\\.(r|q)$", ignore.case = FALSE)

    if (length(files) == 0) {
      check_result(TRUE)
    } else {
      problems <- lapply(files, function(f) {
        check_position(file.path("R", f), line = f)
      })
      check_result(FALSE, problems)
    }
  }
)

#' @include lists.R

#' @noRd
spelling_positions <- function(word, locations) {
  unlist(lapply(locations, function(loc) {
    parts <- strsplit(loc, ":")[[1]]
    line_nums <- as.integer(strsplit(parts[2], ",")[[1]])
    lapply(line_nums, function(ln) {
      check_position(parts[1], ln, line = word)
    })
  }), recursive = FALSE)
}

CHECKS$spelling <- make_check(

  description = "No misspelled words in documentation",
  tags = c("info", "documentation"),
  preps = "spelling",

  gp = paste(
    "fix spelling errors in package documentation.",
    "Use {.code spelling::spell_check_package()} to review misspelled words",
    "and add false positives to {.file inst/WORDLIST}."
  ),

  check = function(state) {
    if (identical(state$spelling, "no_wordlist") ||
        inherits(state$spelling, "try-error")) {
      return(na_result())
    }

    res <- state$spelling
    if (nrow(res) == 0) {
      return(check_result(TRUE))
    }

    positions <- unlist(
      Map(spelling_positions, res$word, res$found),
      recursive = FALSE
    )
    check_result(FALSE, positions)
  }
)

linters_to_lint <- function() {
  list(
    # -- existing checks --
    assignment_linter = lintr::assignment_linter(),
    line_length_linter = lintr::line_length_linter(80),
    package_hooks_linter = lintr::package_hooks_linter(),
    semicolon_linter = lintr::semicolon_linter(allow_compound = TRUE),
    attach_detach_linter = lintr::undesirable_function_linter(
      fun = lintr::default_undesirable_functions[c("attach", "detach")]
    ),
    setwd_linter = lintr::undesirable_function_linter(
      fun = lintr::default_undesirable_functions["setwd"]
    ),
    sapply_linter = lintr::undesirable_function_linter(
      fun = lintr::default_undesirable_functions["sapply"]
    ),
    library_require_linter = lintr::undesirable_function_linter(
      fun = lintr::default_undesirable_functions[c("library", "require")]
    ),
    installed_packages_linter = lintr::undesirable_function_linter(
      fun = c(
        installed.packages = paste(
          "it can be very slow; use find.package() or system.file() instead."
        )
      )
    ),
    seq_linter = lintr::seq_linter(),

    # -- correctness / performance --
    any_duplicated_linter = lintr::any_duplicated_linter(),
    any_is_na_linter = lintr::any_is_na_linter(),
    class_equals_linter = lintr::class_equals_linter(),
    condition_message_linter = lintr::condition_message_linter(),
    duplicate_argument_linter = lintr::duplicate_argument_linter(),
    equals_na_linter = lintr::equals_na_linter(),
    fixed_regex_linter = lintr::fixed_regex_linter(),
    for_loop_index_linter = lintr::for_loop_index_linter(),
    length_test_linter = lintr::length_test_linter(),
    matrix_apply_linter = lintr::matrix_apply_linter(),
    missing_argument_linter = lintr::missing_argument_linter(),
    nrow_subset_linter = lintr::nrow_subset_linter(),
    redundant_equals_linter = lintr::redundant_equals_linter(),
    redundant_ifelse_linter = lintr::redundant_ifelse_linter(),
    regex_subset_linter = lintr::regex_subset_linter(),
    sort_linter = lintr::sort_linter(),
    system_file_linter = lintr::system_file_linter(),
    terminal_close_linter = lintr::terminal_close_linter(),
    which_grepl_linter = lintr::which_grepl_linter(),

    # -- readability / idiom --
    boolean_arithmetic_linter = lintr::boolean_arithmetic_linter(),
    comparison_negation_linter = lintr::comparison_negation_linter(),
    consecutive_assertion_linter = lintr::consecutive_assertion_linter(),
    if_not_else_linter = lintr::if_not_else_linter(),
    ifelse_censor_linter = lintr::ifelse_censor_linter(),
    implicit_assignment_linter = lintr::implicit_assignment_linter(),
    inner_combine_linter = lintr::inner_combine_linter(),
    length_levels_linter = lintr::length_levels_linter(),
    literal_coercion_linter = lintr::literal_coercion_linter(),
    nested_ifelse_linter = lintr::nested_ifelse_linter(),
    nested_pipe_linter = lintr::nested_pipe_linter(),
    numeric_leading_zero_linter = lintr::numeric_leading_zero_linter(),
    outer_negation_linter = lintr::outer_negation_linter(),
    paste_linter = lintr::paste_linter(),
    scalar_in_linter = lintr::scalar_in_linter(),
    undesirable_operator_linter = lintr::undesirable_operator_linter(),
    unnecessary_concatenation_linter = lintr::unnecessary_concatenation_linter(),
    unnecessary_lambda_linter = lintr::unnecessary_lambda_linter(),
    unreachable_code_linter = lintr::unreachable_code_linter(),

    # -- testthat expectations --
    conjunct_test_linter = lintr::conjunct_test_linter(),
    expect_comparison_linter = lintr::expect_comparison_linter(),
    expect_identical_linter = lintr::expect_identical_linter(),
    expect_length_linter = lintr::expect_length_linter(),
    expect_named_linter = lintr::expect_named_linter(),
    expect_not_linter = lintr::expect_not_linter(),
    expect_null_linter = lintr::expect_null_linter(),
    expect_s3_class_linter = lintr::expect_s3_class_linter(),
    expect_s4_class_linter = lintr::expect_s4_class_linter(),
    expect_true_false_linter = lintr::expect_true_false_linter(),
    expect_type_linter = lintr::expect_type_linter()
  )
}

#' @include lists.R prep_utils.R
#' @importFrom lintr lint_package

PREPS$lintr <- function(state, path = state$path, quiet) {
  path <- normalizePath(path)
  excl <- as.list(state$exclude_path %||% character())
  run_prep_step(state, "lintr", function(path) {
    suppressMessages(lint_package(
      path, linters = linters_to_lint(), exclusions = excl
    ))
  }, path = path, silent = quiet)
}

#' @include lists.R

rd_find_topic <- function(rd_data, alias) {
  for (topic in rd_data) {
    if (alias %in% topic$aliases) return(topic)
  }
  NULL
}

rd_exported_aliases <- function(state) {
  ns <- state$namespace
  if (inherits(ns, "try-error")) return(character())

  setdiff(ns$exports, ns_s3_method_names(ns))
}

rd_check_field <- function(state, field, skip_internal = FALSE) {
  if (inherits(state$rd, "try-error")) return(na_result())

  rd_data <- state$rd
  if (length(rd_data) == 0) return(na_result())

  exports <- rd_exported_aliases(state)
  problems <- list()

  for (alias in exports) {
    topic <- rd_find_topic(rd_data, alias)
    if (is.null(topic)) next
    if (isTRUE(topic$is_reexport)) next
    if (skip_internal && isTRUE(topic$has_keyword_internal)) next

    if (!isTRUE(topic[[field]])) {
      problems[[length(problems) + 1]] <- check_position(
        file.path("man", topic$file),
        line = alias
      )
    }
  }

  check_result(length(problems) == 0, problems)
}

make_rd_check <- function(description, gp, field, tags = NULL,
                          skip_internal = FALSE) {
  make_check(
    description = description,
    tags = c("style", "documentation", tags),
    preps = c("rd", "namespace"),
    gp = gp,
    check = function(state) rd_check_field(state, field, skip_internal)
  )
}

CHECKS$rd_has_examples <- make_rd_check(
  description = "Exported functions have \\examples in .Rd",
  gp = "Add examples to all exported functions.",
  field = "has_examples"
)

CHECKS$rd_has_return <- make_rd_check(
  description = "Exported functions have \\value in .Rd",
  gp = paste(
    "Document return values for exported (non-method)",
    "functions using {.code \\\\value}."
  ),
  field = "has_value",
  skip_internal = TRUE
)

get_result <- function(res, check) res$passed[res$check == check]

# -- helper: rd_find_topic ----------------------------------------------------

test_that("rd_find_topic returns topic when alias matches", {
  rd_data <- list(
    list(aliases = c("foo", "bar"), file = "foo.Rd"),
    list(aliases = "baz", file = "baz.Rd")
  )
  result <- rd_find_topic(rd_data, "baz")
  expect_identical(result$file, "baz.Rd")
})

test_that("rd_find_topic returns NULL when no alias matches", {
  rd_data <- list(
    list(aliases = "foo", file = "foo.Rd")
  )
  expect_null(rd_find_topic(rd_data, "missing"))
})

# -- helper: rd_exported_aliases ----------------------------------------------

test_that("rd_exported_aliases returns exports minus S3 methods", {
  state <- list(namespace = list(
    exports = c("foo", "print.myclass"),
    S3methods = matrix(c("print", "myclass", "print.myclass"), ncol = 3)
  ))
  result <- rd_exported_aliases(state)
  expect_identical(result, "foo")
})

test_that("rd_exported_aliases returns empty on try-error namespace", {
  state <- list(namespace = structure("error", class = "try-error"))
  expect_identical(rd_exported_aliases(state), character())
})

test_that("rd_exported_aliases handles empty S3methods matrix", {
  state <- list(namespace = list(
    exports = c("foo", "bar"),
    S3methods = matrix(character(0), ncol = 3)
  ))
  result <- rd_exported_aliases(state)
  expect_identical(result, c("foo", "bar"))
})

# -- make_rd_check: direct unit tests ----------------------------------------

test_that("make_rd_check passes when all exported topics have the field", {
  state <- list(
    rd = list(
      list(aliases = "myfun", file = "myfun.Rd", has_examples = TRUE)
    ),
    namespace = list(
      exports = "myfun",
      S3methods = matrix(character(0), ncol = 3)
    )
  )
  result <- CHECKS$rd_has_examples$check(state)
  expect_true(result$status)
  expect_length(result$positions, 0)
})

test_that("make_rd_check fails when exported topic lacks field", {
  state <- list(
    rd = list(
      list(aliases = "myfun", file = "myfun.Rd", has_examples = FALSE)
    ),
    namespace = list(
      exports = "myfun",
      S3methods = matrix(character(0), ncol = 3)
    )
  )
  result <- CHECKS$rd_has_examples$check(state)
  expect_false(result$status)
  expect_length(result$positions, 1)
  expect_identical(result$positions[[1]]$filename, "man/myfun.Rd")
  expect_identical(result$positions[[1]]$line, "myfun")
})

test_that("make_rd_check returns NA on empty rd_data", {
  state <- list(
    rd = list(),
    namespace = list(
      exports = "myfun",
      S3methods = matrix(character(0), ncol = 3)
    )
  )
  result <- CHECKS$rd_has_examples$check(state)
  expect_true(is.na(result$status))
})

test_that("make_rd_check skips exports with no matching Rd topic", {
  state <- list(
    rd = list(
      list(aliases = "documented", file = "documented.Rd", has_value = TRUE)
    ),
    namespace = list(
      exports = c("documented", "undocumented"),
      S3methods = matrix(character(0), ncol = 3)
    )
  )
  result <- CHECKS$rd_has_return$check(state)
  expect_true(result$status)
  expect_length(result$positions, 0)
})

# -- make_rd_check factory ----------------------------------------------------

test_that("make_rd_check creates a working check", {
  chk <- make_rd_check(
    description = "test check",
    gp = "test advice",
    field = "has_examples"
  )
  expect_identical(chk$description, "test check")
  expect_true("documentation" %in% chk$tags)

  state <- list(
    rd = list(list(aliases = "fn", file = "fn.Rd", has_examples = TRUE)),
    namespace = list(
      exports = "fn",
      S3methods = matrix(character(0), ncol = 3)
    )
  )
  expect_true(chk$check(state)$status)
})

# -- rd_has_examples ----------------------------------------------------------

test_that("rd_has_examples fails when exported function has no examples", {
  gp_res <- gp("bad_rd", checks = "rd_has_examples")
  res <- results(gp_res)
  expect_false(get_result(res, "rd_has_examples"))

  pos <- failed_positions(gp_res)$rd_has_examples
  lines <- vapply(pos, `[[`, "", "line")
  expect_true("no_examples" %in% lines)
})

test_that("rd_has_examples passes when all exports have examples", {
  gp_res <- gp("good", checks = "rd_has_examples")
  res <- results(gp_res)
  expect_true(get_result(res, "rd_has_examples"))
})

test_that("rd_has_examples skips S3 methods", {
  gp_res <- gp("bad_rd", checks = "rd_has_examples")
  pos <- failed_positions(gp_res)$rd_has_examples
  lines <- vapply(pos, `[[`, "", "line")
  expect_false("print.badrd" %in% lines)
})

# -- rd_has_return ------------------------------------------------------------

test_that("rd_has_return fails when exported function has no value", {
  gp_res <- gp("bad_rd", checks = "rd_has_return")
  res <- results(gp_res)
  expect_false(get_result(res, "rd_has_return"))

  pos <- failed_positions(gp_res)$rd_has_return
  lines <- vapply(pos, `[[`, "", "line")
  expect_true("no_value" %in% lines)
})

test_that("rd_has_return passes when all exports have value", {
  gp_res <- gp("good", checks = "rd_has_return")
  res <- results(gp_res)
  expect_true(get_result(res, "rd_has_return"))
})

test_that("rd_has_return skips keyword internal functions", {
  state <- list(
    rd = list(
      list(
        aliases = "internal_fn", file = "internal_fn.Rd",
        has_value = FALSE, has_keyword_internal = TRUE
      ),
      list(
        aliases = "public_fn", file = "public_fn.Rd",
        has_value = TRUE, has_keyword_internal = FALSE
      )
    ),
    namespace = list(
      exports = c("internal_fn", "public_fn"),
      S3methods = matrix(character(0), ncol = 3)
    )
  )
  result <- CHECKS$rd_has_return$check(state)
  expect_true(result$status)
  expect_length(result$positions, 0)
})

test_that("rd_has_return still flags exported non-internal without value", {
  state <- list(
    rd = list(
      list(
        aliases = "internal_fn", file = "internal_fn.Rd",
        has_value = FALSE, has_keyword_internal = TRUE
      ),
      list(
        aliases = "public_fn", file = "public_fn.Rd",
        has_value = FALSE, has_keyword_internal = FALSE
      )
    ),
    namespace = list(
      exports = c("internal_fn", "public_fn"),
      S3methods = matrix(character(0), ncol = 3)
    )
  )
  result <- CHECKS$rd_has_return$check(state)
  expect_false(result$status)
  expect_length(result$positions, 1)
  expect_identical(result$positions[[1]]$line, "public_fn")
})

test_that("rd_has_return does not flag keyword internal in bad_rd fixture", {
  gp_res <- gp("bad_rd", checks = "rd_has_return")
  pos <- failed_positions(gp_res)$rd_has_return
  lines <- vapply(pos, `[[`, "", "line")
  expect_true("no_value" %in% lines)
  expect_false("internal_func" %in% lines)
})

# -- reexports are skipped -----------------------------------------------------

test_that("rd_has_examples skips reexported functions", {
  state <- list(
    rd = list(
      list(
        aliases = c("reexports", "reexported_fn"),
        file = "reexports.Rd",
        has_examples = FALSE, has_value = FALSE,
        has_keyword_internal = TRUE, is_reexport = TRUE
      ),
      list(
        aliases = "public_fn", file = "public_fn.Rd",
        has_examples = TRUE, has_value = TRUE,
        has_keyword_internal = FALSE, is_reexport = FALSE
      )
    ),
    namespace = list(
      exports = c("reexported_fn", "public_fn"),
      S3methods = matrix(character(0), ncol = 3)
    )
  )
  result <- CHECKS$rd_has_examples$check(state)
  expect_true(result$status)
  expect_length(result$positions, 0)
})

test_that("rd_has_return skips reexported functions", {
  state <- list(
    rd = list(
      list(
        aliases = c("reexports", "reexported_fn"),
        file = "reexports.Rd",
        has_examples = FALSE, has_value = FALSE,
        has_keyword_internal = TRUE, is_reexport = TRUE
      ),
      list(
        aliases = "public_fn", file = "public_fn.Rd",
        has_examples = TRUE, has_value = TRUE,
        has_keyword_internal = FALSE, is_reexport = FALSE
      )
    ),
    namespace = list(
      exports = c("reexported_fn", "public_fn"),
      S3methods = matrix(character(0), ncol = 3)
    )
  )
  result <- CHECKS$rd_has_return$check(state)
  expect_true(result$status)
  expect_length(result$positions, 0)
})

test_that("rd_has_examples does not flag reexports in bad_rd fixture", {
  gp_res <- gp("bad_rd", checks = "rd_has_examples")
  pos <- failed_positions(gp_res)$rd_has_examples
  lines <- vapply(pos, `[[`, "", "line")
  expect_false("reexported_fn" %in% lines)
})

test_that("parse_rd_files detects docType import as reexport", {
  result <- parse_rd_files("bad_rd/man")
  reex <- result[[which(vapply(result, function(x) {
    "reexported_fn" %in% x$aliases
  }, logical(1)))]]
  expect_true(reex$is_reexport)

  good <- result[[which(vapply(result, function(x) {
    "good_func" %in% x$aliases
  }, logical(1)))]]
  expect_false(good$is_reexport)
})

# -- prep returns NA on missing man/ ------------------------------------------

test_that("rd checks return NA when man/ directory is missing", {
  gp_res <- gp("bad2", checks = "rd_has_examples")
  res <- results(gp_res)
  expect_true(is.na(get_result(res, "rd_has_examples")))
})

# -- prep error handling ------------------------------------------------------

test_that("all rd checks return na_result when prep failed", {
  state <- list(rd = structure("error", class = "try-error"))

  for (check_name in c("rd_has_examples", "rd_has_return")) {
    result <- CHECKS[[check_name]]$check(state)
    expect_true(is.na(result$status), label = paste(check_name, "status"))
    expect_type(result$positions, "list")
  }
})

# -- parse_rd_files -----------------------------------------------------------

test_that("parse_rd_files returns empty list when mandir does not exist", {
  expect_identical(parse_rd_files(tempfile()), list())
})

test_that("parse_rd_files returns empty list when mandir has no .Rd files", {
  d <- withr::local_tempdir()
  expect_identical(parse_rd_files(d), list())
})

test_that("parse_rd_files parses Rd files correctly", {
  result <- parse_rd_files("bad_rd/man")
  aliases <- unlist(lapply(result, `[[`, "aliases"))
  expect_true("good_func" %in% aliases)

  good <- result[[which(vapply(result, function(x) {
    "good_func" %in% x$aliases
  }, logical(1)))]]
  expect_true(good$has_examples)
  expect_true(good$has_value)
})

test_that("parse_rd_files detects keyword internal", {
  result <- parse_rd_files("bad_rd/man")
  internal <- result[[which(vapply(result, function(x) {
    "internal_func" %in% x$aliases
  }, logical(1)))]]
  expect_true(internal$has_keyword_internal)

  good <- result[[which(vapply(result, function(x) {
    "good_func" %in% x$aliases
  }, logical(1)))]]
  expect_false(good$has_keyword_internal)
})

test_that("parse_rd_files handles Rd elements with no Rd_tag attribute", {
  el_no_tag <- list("bare text")
  el_alias <- list("myfun")
  attr(el_alias, "Rd_tag") <- "\\alias"
  fake_parsed <- list(el_no_tag, el_alias)
  class(fake_parsed) <- "Rd"

  local_mocked_bindings(
    parse_Rd = function(...) fake_parsed,
    .package = "tools"
  )
  d <- withr::local_tempdir()
  writeLines("placeholder", file.path(d, "myfun.Rd"))
  result <- parse_rd_files(d)
  expect_length(result, 1)
  expect_identical(result[[1]]$aliases, "myfun")
})

# -- PREPS$rd -----------------------------------------------------------------

test_that("PREPS$rd warns when parsing fails", {
  state <- list(path = "bad_rd")
  local_mocked_bindings(
    parse_rd_files = function(...) stop("forced error")
  )
  expect_warning(PREPS$rd(state, quiet = TRUE), "Prep step for")
})

PREPS <- list()
CHECKS <- list()

#' List the names of all checks
#'
#' @return Character vector of checks
#' @family check_groups
#' @export
#' @examples
#' all_checks()

all_checks <- function() {
  names(CHECKS)
}

#' List the names of default checks (excludes optional check sets)
#'
#' @return Character vector of default check names
#' @family check_groups
#' @export
#' @examples
#' default_checks()

default_checks <- function() {
  setdiff(all_checks(), tidyverse_checks())
}

#' List the names of tidyverse style checks
#'
#' These checks are optional and not included in the default set.
#' They are powered by \code{\link[lintr]{lint_package}} using lintr's
#' default linter set and respect any \code{.lintr} configuration file
#' in the package root (e.g. to disable specific linters or add exclusions).
#' Add them via \code{checks = c(default_checks(), tidyverse_checks())}.
#'
#' @return Character vector of tidyverse check names
#' @family check_groups
#' @export
#' @examples
#' tidyverse_checks()

tidyverse_checks <- function() {
  grep("^tidyverse_", all_checks(), value = TRUE)
}

#' Describe one or more checks
#'
#' @param check_name Names of checks to be described.
#' @return List of character descriptions for each \code{check_name}
#' @family check_groups
#' @export
#' @examples 
#' describe_check("rcmdcheck_non_portable_makevars")
#' check_name <- c("no_description_depends",
#'                 "lintr_assignment_linter",
#'                 "no_import_package_as_a_whole",
#'                 "rcmdcheck_missing_docs")
#' describe_check(check_name)
#' # Or to see all checks:
#' \dontrun{
#'   describe_check(all_checks())
#' }

describe_check <- function(check_name = NULL) {
  check_name <- intersect(check_name, names(CHECKS))
  lapply(CHECKS[check_name], function(i) i$description)
}

#' List available check group names
#'
#' Returns the names of all registered check groups.
#' Use these names with [checks_by_group()] to select checks by group,
#' or with \code{options(goodpractice.exclude_check_groups = ...)} to skip
#' groups.
#' Full descriptions of each check group are return by
#' [describe_check_groups()].
#'
#' @return A character vector of check group names.
#' @family check_groups
#' @export
#' @examples
#' # Names of all check groups:
#' all_check_groups()
#'
#' # List individual checks by group:
#' chks <- lapply(all_check_groups(), checks_by_group)
#' names(chks) <- all_check_groups()
#' chks

all_check_groups <- function() {
  names(PREPS)
}

#' Describe available check groups
#'
#' Returns full descriptions of all registered check groups.
#'
#' @return A named list of each check group defined in [all_check_groups()],
#' with text descriptions of each group.
#' @family check_groups
#' @export
#' @examples
#' # Names of all check groups:
#' all_check_groups()
#'
#' # And corresponding descriptions:
#' describe_check_groups()

describe_check_groups <- function() {
  num_linters <- length(grep("linter", all_checks()))
  d <- list(
    covr = "Test coverage report from 'covr' package.",
    cyclocomp = paste0(
      "Function cyclocomplexity with the 'cyclocomp' ",
      "package; default limit of 50."
    ),
    description = paste0(
      "Check common issues with DESCRIPTION files, including ",
      "formatting issues with URLs, DOIs, BugReports, package ",
      "names, and author and contributor roles"
    ),
    lintr = paste0(
      "Check package linting with the 'lintr' package (",
      num_linters,
      " linters in total)."
    ),
    namespace = "Check import and export patterns in NAMESPACE file",
    rcmdcheck = paste0(
      "Run 'R CMD check' via the 'rcmdcheck' package. ~200 checks for ",
      "documentation, namespace, compilation, tests, vignettes, ",
      "CRAN compliance."
    ),
    rd = paste0(
      "Check whether or not 'man/*.Rd' function documentation includes ",
      "both example code and return values (regardless of whether or not ",
      "documentation files are generated by 'Roxygen2')."
    ),
    revdep = paste0(
      "Check whether package has reverse dependencies, and ",
      "recommending running 'reddev' package if so."
    ),
    roxygen2 = paste0(
      "Only for packages within use 'Roxygen2' to generate documentation. ",
      "Checks for best practices in Roxygen2 tag usage, flags any ",
      "unknown tags, and ensures 'inheritParams' is used correctly."
    ),
    code_structure = paste0(
      "Common issues like duplicated or unused function bodies, ",
      "that 'print()' returns insivibly, that 'on.exit()' uses ",
      "'add = TRUE', and checks on function length (default max",
      "50 lines)."
    ),
    package_structure = paste0(
      "Generic checks like whether a package has a README, a NEWS file, ",
      "or whether all files use a '.R' extension, and not '.r'."
    ),
    spelling = "Check spelling with the 'spelling' package.",
    tidyverse = paste0(
      "Check compliance with the Tidyverse style guide; mostly via 'lintr' package. ",
      "(These checks are not run by default; and only if 'tidyverse_checks()' are ",
      "added to 'checks_by_group()'."
    ),
    urlchecker = paste0(
      "Check whether all URLs are valid, including",
      "identifying any redirects."
    ),
    vignette = "Check that vignette code does not use either 'rm()' or 'setwd()'."
  )
  if (!identical(names(d), all_check_groups())) {
    cli::cli_abort("Group names in 'describe_check_groups' differ from 'all_check_groups'.")
  }
  d
}

#' Select checks by check group
#'
#' Returns the names of all checks that belong to the given group(s).
#' This makes it easy to run or inspect a specific category of checks
#' without knowing individual check names.
#'
#' @param ... Group names as character strings. Use [all_check_groups()] to
#'   see available names.
#' @return Character vector of check names
#' @family check_groups
#' @export
#' @examples
#' # run only DESCRIPTION and namespace checks
#' checks_by_group("description", "namespace")
#'
#' # see what the lintr group covers
#' checks_by_group("lintr")
#' # See all checks by group:
#' lapply(all_check_groups(), checks_by_group)
#' # use directly in gp()
#' \dontrun{
#'   gp(".", checks = checks_by_group("description", "lintr"))
#' }

checks_by_group <- function(...) {
  group <- c(...)
  if (length(group) == 0) return(character(0))
  unknown <- setdiff(group, names(PREPS))
  if (length(unknown) > 0) {
    cli::cli_warn(c(
      "Unknown check group{?s}: {.val {unknown}}.",
      i = "Use {.fn all_check_groups} to see available groups."
    ))
  }
  names(Filter(function(ch) any(ch$preps %in% group), CHECKS))
}

test_that("describe_check", {
  expect_length(describe_check(), 0L)
  chk1 <- "rcmdcheck_non_portable_makevars"
  d1 <- describe_check(chk1)
  expect_type(d1, "list")
  expect_length(d1, length(chk1))
  expect_named(d1, chk1)

  chks4 <- c(
    "no_description_depends",
    "lintr_assignment_linter",
    "no_import_package_as_a_whole",
    "rcmdcheck_missing_docs"
  )
  d4 <- describe_check(chks4)
  expect_type(d4, "list")
  expect_length(d4, length(chks4))
  expect_named(d4, chks4)

  chks6 <- c(chks4, "a", "b") # "a", "b" are not checks, so ignored
  d6 <- describe_check(chks6)
  expect_type(d6, "list")
  expect_false(length(d6) == length(chks6))
  expect_identical(d6, d4)

  d_all <- describe_check(all_checks())
  expect_length(d_all, length(all_checks()))
  expect_named(d_all, all_checks())

  expect_snapshot(describe_check(all_checks()))
})

test_that("all_check_groups returns registered group names", {
  groups <- all_check_groups()
  expect_true(is.character(groups))
  expect_length(groups, 15L)
  current_groups <- c(
        "covr",
        "cyclocomp",
        "description",
        "lintr",
        "namespace",
        "rcmdcheck",
        "rd",
        "revdep",
        "roxygen2",
        "code_structure",
        "package_structure",
        "spelling",
        "tidyverse",
        "urlchecker",
        "vignette"
  )
  expect_true(all(current_groups %in% groups))
})

test_that("describe_check_groups", {
  groups <- all_check_groups()
  descs <- describe_check_groups()
  expect_true(is.list(descs))
  expect_equal(length(groups), length(descs))
  expect_identical(groups, names(descs))
  desc_lens <- vapply(descs, length, integer(1L))
  expect_true(all(desc_lens == 1L))
  desc_chars <- vapply(descs, nchar, integer(1L))
  expect_true(all(desc_chars > 40))
})

test_that("checks_by_group returns checks for a single group", {
  desc_checks <- checks_by_group("description")
  expect_gt(length(desc_checks), 0)
  expect_true("description_url" %in% desc_checks)
  expect_false("lintr_assignment_linter" %in% desc_checks)
  expect_true(all(desc_checks %in% all_checks()))
})

test_that("checks_by_group returns checks for multiple groups", {
  combined <- checks_by_group("description", "lintr")
  desc_only <- checks_by_group("description")
  lintr_only <- checks_by_group("lintr")
  expect_true(all(desc_only %in% combined))
  expect_true(all(lintr_only %in% combined))
  expect_false("covr" %in% combined)
})

test_that("checks_by_group returns character(0) for no args", {
  expect_identical(checks_by_group(), character(0))
})

test_that("checks_by_group warns on unknown group", {
  expect_warning(
    res <- checks_by_group("nonexistent_group"),
    "Unknown check group"
  )
  expect_identical(res, character(0))
})

test_that("every check belongs to at least one group", {
  lens <- vapply(CHECKS, function(ch) length(ch$preps), integer(1L))
  expect_true(all(lens > 0))
})

test_that("all check groups map to registered preps", {
  groups <- unlist(lapply(CHECKS, function(ch) ch$preps))
  expect_true(all(groups %in% names(PREPS)))
})

test_that("checks with multiple groups are known", {
  lens <- vapply(CHECKS, function(ch) length(ch$preps), integer(1L))
  multi <- names(lens[lens >= 2])
  expect_gte(length(multi), 2)
  expect_true("rd_has_examples" %in% multi)
  expect_true("rd_has_return" %in% multi)
})

test_that("all tidyverse checks belong to the tidyverse group", {
  tv_checks <- grep("^tidyverse_", all_checks(), value = TRUE)
  tv_group <- checks_by_group("tidyverse")
  expect_true(all(tv_checks %in% tv_group))
})

test_that("code_structure and package_structure are distinct", {
  cs <- checks_by_group("code_structure")
  ps <- checks_by_group("package_structure")
  expect_gt(length(cs), 0)
  expect_gt(length(ps), 0)
  expect_length(intersect(cs, ps), 0)
  expect_true("print_return_invisible" %in% cs)
  expect_true("has_readme" %in% ps)
})

test_that("checks_by_group works in gp()", {
  pkg_path <- system.file("bad1", package = "goodpractice")
  g <- gp(pkg_path, checks = checks_by_group("description"))
  res <- results(g)
  expect_gt(nrow(res), 0)
  expect_true(all(res$check %in% checks_by_group("description")))
})

tv_checks <- tidyverse_checks()
tv_lintr_checks <- grep("_linter$", tv_checks, value = TRUE)

get_result <- function(res, check) res$passed[res$check == check]

write_dot_lintr("bad_tidyverse")
write_dot_lintr("good_tidyverse")

gp_bad <- gp("bad_tidyverse", checks = tv_checks)
res_bad <- results(gp_bad)

gp_good <- gp("good_tidyverse", checks = tv_checks)
res_good <- results(gp_good)

unlink(file.path("bad_tidyverse", ".lintr"))
unlink(file.path("good_tidyverse", ".lintr"))

test_that("tidyverse_checks() returns only tidyverse_ prefixed checks", {
  expect_gt(length(tv_checks), 0)
  expect_true(all(grepl("^tidyverse_", tv_checks)))
})

test_that("default_checks() excludes tidyverse checks", {
  dc <- default_checks()
  expect_false(any(grepl("^tidyverse_", dc)))
  expect_identical(sort(c(dc, tv_checks)), sort(all_checks()))
})

test_that("tidyverse lintr checks pass on good fixture", {
  for (chk in tv_lintr_checks) {
    expect_true(
      get_result(res_good, chk),
      label = paste(chk, "should pass on good_tidyverse")
    )
  }
})

test_that("tidyverse brace_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_brace_linter"))
})

test_that("tidyverse commas_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_commas_linter"))
})

test_that("tidyverse commented_code_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_commented_code_linter"))
})

test_that("tidyverse equals_na_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_equals_na_linter"))
})

test_that("tidyverse function_left_parentheses_linter fails on bad fixture", {
  expect_false(
    get_result(res_bad, "tidyverse_function_left_parentheses_linter")
  )
})

test_that("tidyverse indentation_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_indentation_linter"))
})

test_that("tidyverse infix_spaces_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_infix_spaces_linter"))
})

test_that("tidyverse object_length_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_object_length_linter"))
})

test_that("tidyverse object_name_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_object_name_linter"))
})

test_that("tidyverse object_usage_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_object_usage_linter"))
})

test_that("tidyverse paren_body_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_paren_body_linter"))
})

test_that("tidyverse pipe_consistency_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_pipe_consistency_linter"))
})

test_that("tidyverse quotes_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_quotes_linter"))
})

test_that("tidyverse return_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_return_linter"))
})

test_that("tidyverse spaces_inside_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_spaces_inside_linter"))
})

test_that("tidyverse spaces_left_parentheses_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_spaces_left_parentheses_linter"))
})

test_that("tidyverse trailing_blank_lines_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_trailing_blank_lines_linter"))
})

test_that("tidyverse trailing_whitespace_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_trailing_whitespace_linter"))
})

test_that("tidyverse vector_logic_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_vector_logic_linter"))
})

test_that("tidyverse whitespace_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_whitespace_linter"))
})

test_that("tidyverse assignment_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_assignment_linter"))
})

test_that("tidyverse line_length_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_line_length_linter"))
})

test_that("tidyverse semicolon_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_semicolon_linter"))
})

test_that("tidyverse seq_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_seq_linter"))
})

test_that("tidyverse T_and_F_symbol_linter fails on bad fixture", {
  expect_false(get_result(res_bad, "tidyverse_T_and_F_symbol_linter"))
})

test_that("tidyverse_r_file_names catches bad file names", {
  expect_false(get_result(res_bad, "tidyverse_r_file_names"))
  expect_true(get_result(res_good, "tidyverse_r_file_names"))
})

test_that("tidyverse_test_file_names catches missing test files", {
  expect_false(get_result(res_bad, "tidyverse_test_file_names"))
  expect_true(get_result(res_good, "tidyverse_test_file_names"))
})

test_that("tidyverse_no_missing fails when missing() is used", {
  write_dot_lintr("bad_missing")
  on.exit(unlink(file.path("bad_missing", ".lintr")))

  gp_res <- gp("bad_missing", checks = "tidyverse_no_missing")
  res <- results(gp_res)
  expect_false(get_result(res, "tidyverse_no_missing"))

  pos <- failed_positions(gp_res)$tidyverse_no_missing
  names <- vapply(pos, `[[`, "", "line")
  expect_true("my_func" %in% names)
})

test_that("tidyverse_no_missing passes when missing() is not used", {
  expect_true(get_result(res_good, "tidyverse_no_missing"))
})

test_that("tidyverse_export_order fails when internal before exported", {
  write_dot_lintr("bad_export_order")
  on.exit(unlink(file.path("bad_export_order", ".lintr")))

  gp_res <- gp("bad_export_order", checks = "tidyverse_export_order")
  res <- results(gp_res)
  expect_false(get_result(res, "tidyverse_export_order"))

  pos <- failed_positions(gp_res)$tidyverse_export_order
  names <- vapply(pos, `[[`, "", "line")
  expect_true("internal_helper" %in% names)
})

test_that("tidyverse_export_order passes with exportPattern", {
  expect_true(get_result(res_good, "tidyverse_export_order"))
})

test_that("tidyverse lintr checks respect .lintr config", {
  pkg <- withr::local_tempdir()
  file.copy(
    list.files("bad_tidyverse", full.names = TRUE, recursive = TRUE),
    pkg, recursive = TRUE
  )
  writeLines(
    "linters: linters_with_defaults(brace_linter = NULL)",
    file.path(pkg, ".lintr")
  )

  gp_res <- gp(pkg, checks = "tidyverse_brace_linter")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "tidyverse_brace_linter"])
})

test_that("get_tidyverse_lintr_state returns NA on try-error", {
  state <- list(tidyverse_lintr = structure("error", class = "try-error"))
  result <- get_tidyverse_lintr_state(state, "brace_linter")
  expect_true(is.na(result$status))
  expect_identical(result$positions, list())
})

test_that("tidyverse_no_missing ignores missing() inside nested functions", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(
    c(
      "Package: nesttest", "Title: Test", "Version: 1.0.0",
      "Author: Test", "Maintainer: Test <test@test.com>",
      "Description: Test.", "License: GPL-2"
    ),
    file.path(pkg, "DESCRIPTION")
  )
  writeLines(
    c(
      "outer <- function(x) {",
      "  inner <- function(y) missing(y)",
      "  inner(x)",
      "}"
    ),
    file.path(pkg, "R", "funcs.R")
  )
  write_dot_lintr(pkg)

  gp_res <- gp(pkg, checks = "tidyverse_no_missing")
  res <- results(gp_res)
  expect_true(get_result(res, "tidyverse_no_missing"))
})

test_that("treesitter checks pass when no functions are defined", {
  pkg <- withr::local_tempdir()
  writeLines(
    c(
      "Package: edgetest", "Title: Test", "Version: 1.0.0",
      "Author: Test", "Maintainer: Test <test@test.com>",
      "Description: Test.", "License: GPL-2"
    ),
    file.path(pkg, "DESCRIPTION")
  )
  write_dot_lintr(pkg)

  gp_res <- gp(pkg, checks = "tidyverse_no_missing")
  expect_true(results(gp_res)$passed)

  dir.create(file.path(pkg, "R"))
  writeLines(character(), file.path(pkg, "R", "empty.R"))
  gp_res <- gp(pkg, checks = "tidyverse_no_missing")
  expect_true(results(gp_res)$passed)

  writeLines(
    c("x <- 42", "library(stats)", "y <- sum(1:10)"),
    file.path(pkg, "R", "misc.R")
  )
  gp_res <- gp(pkg, checks = "tidyverse_no_missing")
  expect_true(results(gp_res)$passed)
})

test_that("default lintr_assignment_linter ignores = inside setMethod()", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(
    c("Package: s4default", "Title: Test", "Version: 1.0.0",
      "Description: Test.", "License: MIT"),
    file.path(pkg, "DESCRIPTION")
  )
  writeLines(c(
    'setMethod("t",',
    '  signature = (x = "dfmSparse"),',
    '  definition = function(x) x',
    ')'
  ), file.path(pkg, "R", "methods.R"))
  write_dot_lintr(pkg)

  gp_res <- gp(pkg, checks = "lintr_assignment_linter")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "lintr_assignment_linter"])
})

test_that("tidyverse assignment_linter ignores = inside setMethod()", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(
    c("Package: s4pkg", "Title: Test", "Version: 1.0.0",
      "Description: Test.", "License: MIT"),
    file.path(pkg, "DESCRIPTION")
  )
  writeLines(c(
    'setMethod("t",',
    '  signature = (x = "dfmSparse"),',
    '  definition = function(x) x',
    ')'
  ), file.path(pkg, "R", "methods.R"))
  write_dot_lintr(pkg)

  gp_res <- gp(pkg, checks = "tidyverse_assignment_linter")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "tidyverse_assignment_linter"])
})

test_that("tidyverse prep passes exclude_path as exclusions", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(
    c("Package: excltest", "Title: Test", "Version: 1.0.0",
      "Description: Test.", "License: MIT"),
    file.path(pkg, "DESCRIPTION")
  )
  writeLines("x<-1", file.path(pkg, "R", "bad.R"))
  writeLines("x <- 1", file.path(pkg, "R", "good.R"))
  write_dot_lintr(pkg)

  withr::local_options(goodpractice.exclude_path = "R/bad.R")
  gp_res <- gp(pkg, checks = "tidyverse_assignment_linter")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "tidyverse_assignment_linter"])
})

test_that("tidyverse prep returns try-error with warning on lint failure", {
  state <- list(path = withr::local_tempdir(), exclude_path = character())
  local_mocked_bindings(lint_package = function(...) stop("lint failure"))
  expect_warning(
    result <- PREPS$tidyverse(state, quiet = FALSE),
    "tidyverse_lintr"
  )
  expect_true(inherits(result$tidyverse_lintr, "try-error"))
})

no_value <- function(x) x + 1
no_examples <- function(x) x + 2
dontrun_example <- function(x) x + 3
all_wrapped <- function(x) x + 4
good_func <- function(x) x + 5
print.badrd <- function(x, ...) cat("badrd\n")
internal_func <- function(x) x + 6

#' @include lists.R prep_utils.R
#' @importFrom desc description

PREPS$description <- function(state, path = state$path, quiet) {
  run_prep_step(state, "description", function(path) {
    desc <- description$new(file.path(path, "DESCRIPTION"))
    if (!desc$has_fields("Encoding")) {
      desc$set("Encoding", "UTF-8")
    }
    desc
  }, path = path, silent = quiet)
}

#' @include lists.R
#' @importFrom urlchecker url_check

#' @noRd
run_url_check <- function(path, quiet) {
  urlchecker::url_check(path, progress = !quiet, fail = FALSE)
}

PREPS$urlchecker <- function(state, path = state$path, quiet) {
  if (!has_internet()) {
    cli::cli_warn("Skipping URL checks: no internet connection.")
    state$urlchecker <- try(stop("offline"), silent = TRUE)
    return(state)
  }

  run_prep_step(state, "urlchecker", function(path, quiet) {
    run_url_check(path, quiet)
  }, path = path, quiet = quiet, silent = quiet)
}

#' Run a prep step with error handling
#'
#' Executes `fn(...)` inside [try()], stores the result in
#' `state[[prep_name]]`, and emits a warning if the step fails.
#' Use this when writing new prep functions to avoid duplicating
#' the try/warning boilerplate.
#'
#' @section Writing a new prep:
#' 1. Create `R/prep_<name>.R`
#' 2. Register the prep and call `run_prep_step()`:
#' ```
#' PREPS$myprep <- function(state, path = state$path, quiet) {
#'   run_prep_step(state, "myprep", function(path) {
#'     do_work(path)
#'   }, path = path, silent = quiet)
#' }
#' ```
#' 3. If you need post-processing on success, capture the returned
#'    state and check `!inherits(state$myprep, "try-error")` before
#'    doing additional work (see `prep_covr.R` for an example).
#' 4. Preps must be independent: when parallel execution is active,
#'    each prep receives the initial state snapshot and cannot read
#'    another prep's output.
#'
#' @param state The goodpractice state list, passed through the prep
#'   pipeline.
#' @param prep_name Character scalar. Name for this prep step, used as
#'   the key in `state` and in the warning message on failure.
#' @param fn A function containing the actual work. Receives
#'   prep-specific arguments via `...`.
#'   Its return value is stored in `state[[prep_name]]`.
#' @param ... Arguments forwarded to `fn`.
#' @param silent Logical. Passed to [try()] as `silent`.
#' @return The updated `state` list.
#' @keywords internal

#' @noRd
run_prep_step <- function(state, prep_name, fn, ..., silent = FALSE) {
  state[[prep_name]] <- try(do.call(fn, list(...)), silent = silent)
  if (inherits(state[[prep_name]], "try-error")) {
    cli::cli_warn(
      "Prep step for {.val {prep_name}} failed:
      {conditionMessage(attr(state[[prep_name]], 'condition'))}"
    )
  }
  state
}

#' `Collate` field from `DESCRIPTION`
#'
#' `NULL` is returned if there is no such field.
#'
#' @param path Path to the package root.
#' @return Character scalar or `NULL`.
#' @keywords internal
#'
#' @importFrom desc desc_get_collate

#' @noRd
package_collate <- function(path = ".") {
  col <- desc_get_collate(file = file.path(path, "DESCRIPTION"))
  if (length(col)) col else NULL
}

#' Get all source files of a package, in the right order
#'
#' It uses the `Collate` entry in the `DESCRIPTION` file,
#' if there is one. Otherwise the order is alphabetical.
#'
#' @param path Path to the root of the R package.
#' @return A character vector of (relative) file
#'   names in the current collation order.
#' @keywords internal

#' @noRd
r_package_files <- function(path, exclude_path = character()) {
  files <- package_collate(path)
  if (is.null(files)) {
    files <- list.files(
      file.path(path, "R"),
      pattern = default_r_file_pattern()
    )
  }

  result <- file.path(path, "R", files)
  filter_excluded_paths(result, path, exclude_path)
}

test_that("package_collate returns NULL when no Collate field", {
  result <- package_collate("good")
  expect_null(result)
})

test_that("package_collate returns file list when Collate exists", {
  pkg <- withr::local_tempdir("collate_pkg")
  dir.create(file.path(pkg, "R"))
  writeLines("f <- function() 1", file.path(pkg, "R", "f.R"))
  writeLines(
    c(
      "Package: collatepkg",
      "Title: Test",
      "Version: 0.0.1",
      "Description: Test.",
      "Collate: 'f.R'"
    ),
    file.path(pkg, "DESCRIPTION")
  )

  result <- package_collate(pkg)
  expect_identical(result, "f.R")
})

test_that("r_package_files returns R files for fixture package", {
  files <- r_package_files("good")
  expect_gt(length(files), 0)
  expect_true(all(grepl("\\.R$", files, ignore.case = TRUE)))
})

test_that("r_package_files includes full path", {
  files <- r_package_files("good")
  expect_true(all(grepl("good/R/", files)))
})

#' @include lists.R
#' @importFrom utils getParseData

#' @noRd
vignette_files <- function(path) {
  vigdir <- file.path(path, "vignettes")
  if (!dir.exists(vigdir)) return(character())
  list.files(vigdir, pattern = "\\.(Rmd|Rnw|qmd)$",
             full.names = TRUE, recursive = TRUE)
}

is_skipped_chunk <- function(lines, start, end) {
  header <- lines[start]
  if (grepl("eval\\s*=\\s*(FALSE|F)\\b", header) ||
      grepl("purl\\s*=\\s*(FALSE|F)\\b", header)) {
    return(TRUE)
  }
  if (end <= start + 1L) return(FALSE)
  body <- lines[seq(start + 1L, end - 1L)]
  hashpipe <- body[grepl("^#\\|", body)]
  any(grepl("eval\\s*:\\s*(false|FALSE|F)\\b", hashpipe)) ||
    any(grepl("purl\\s*:\\s*(false|FALSE|F)\\b", hashpipe))
}

match_chunk_pairs <- function(starts, ends) {
  if (length(starts) == 0) return(matrix(integer(0), ncol = 2))
  if (length(starts) != length(ends)) {
    cli::cli_inform("Chunk start/end indices failed sanity checks in vignette.")
    return(matrix(integer(0), ncol = 2))
  }

  chunks <- cbind(start = starts, end = ends)

  ok <- all(chunks[, 2] > chunks[, 1])
  if (nrow(chunks) > 1) {
    ok <- ok &&
      all(diff(chunks[, 1]) > 0) &&
      all(diff(chunks[, 2]) > 0) &&
      all(chunks[-1, 1] > chunks[-nrow(chunks), 2])
  }

  if (!ok) {
    cli::cli_inform("Chunk start/end indices failed sanity checks in vignette.")
    return(matrix(integer(0), ncol = 2))
  }

  chunks
}

extract_vignette_code <- function(f) {
  lines <- tryCatch(readLines(f, warn = FALSE), error = function(e) NULL)
  if (is.null(lines)) return(NULL)
  n <- length(lines)
  ext <- tolower(tools::file_ext(f))

  if (ext %in% c("rmd", "qmd")) {
    all_starts <- grep("^```\\s*\\{\\s*r\\b", lines)
    fence_ends <- grep("^```\\s*$", lines)
  } else if (ext == "rnw") {
    all_starts <- grep("^<<.*>>=\\s*$", lines)
    fence_ends <- grep("^@\\s*$", lines)
  } else {
    return(NULL)
  }

  all_chunks <- match_chunk_pairs(all_starts, fence_ends)
  if (nrow(all_chunks) == 0) return(NULL)

  keep <- !vapply(seq_len(nrow(all_chunks)), function(i) {
    is_skipped_chunk(lines, all_chunks[i, 1], all_chunks[i, 2])
  }, logical(1))
  chunks <- all_chunks[keep, , drop = FALSE]
  if (nrow(chunks) == 0) return(NULL)

  output <- rep("", n)
  indices <- unlist(apply(chunks, 1, function(i) {
    if (i[2] - i[1] <= 1) return(integer(0))
    seq(i[1] + 1L, i[2] - 1L)
  }))
  output[indices] <- lines[indices]

  if (all(output == "")) return(NULL)
  output
}

vignette_parse_data <- function(f) {
  code_lines <- extract_vignette_code(f)
  if (is.null(code_lines)) return(NULL)

  parsed <- safe_parse(text = code_lines, keep_source = TRUE)
  if (is.null(parsed) || length(parsed) == 0) return(NULL)

  getParseData(parsed)
}

call_descendants <- function(pd, fn_call_id) {
  name_expr <- pd$parent[pd$id == fn_call_id]
  call_expr <- pd$parent[pd$id == name_expr]

  ids <- call_expr
  queue <- call_expr
  while (length(queue) > 0) {
    children <- pd$id[pd$parent %in% queue]
    if (length(children) == 0) break
    ids <- c(ids, children)
    queue <- children
  }
  ids
}

check_vignette_calls <- function(state, fn_name, nested_fn = NULL) {
  problems <- list()

  for (f in names(state$vignette)) {
    entry <- state$vignette[[f]]
    pd <- entry$parse_data
    orig_lines <- entry$lines

    fn_rows <- pd[pd$token == "SYMBOL_FUNCTION_CALL" & pd$text == fn_name, ,
                  drop = FALSE]
    if (nrow(fn_rows) == 0) next

    for (i in seq_len(nrow(fn_rows))) {
      if (!is.null(nested_fn)) {
        desc_ids <- call_descendants(pd, fn_rows$id[i])
        desc <- pd[pd$id %in% desc_ids, , drop = FALSE]
        if (!any(desc$token == "SYMBOL_FUNCTION_CALL" &
                 desc$text == nested_fn)) {
          next
        }
      }

      ln <- fn_rows$line1[i]
      line_text <- if (ln <= length(orig_lines)) orig_lines[ln] else ""

      problems[[length(problems) + 1]] <- check_position(
        file.path("vignettes", basename(f)),
        ln,
        fn_rows$col1[i],
        line = trimws(line_text)
      )
    }
  }

  if (length(problems) == 0) {
    check_result(TRUE)
  } else {
    check_result(FALSE, problems)
  }
}

CHECKS$vignette_no_rm_list <- make_check(

  description = "Vignettes do not use rm(list = ls())",
  tags = c("best practice", "warning"),
  preps = "vignette",

  gp = paste(
    "do not use {.code rm(list = ls())} in vignettes.",
    "Vignettes run in their own environment;",
    "clearing the workspace is unnecessary and confusing for users."
  ),

  check = function(state) {
    check_vignette_calls(state, "rm", nested_fn = "ls")
  }
)

CHECKS$vignette_no_setwd <- make_check(

  description = "Vignettes do not use setwd()",
  tags = c("best practice", "warning"),
  preps = "vignette",

  gp = paste(
    "do not use {.fn setwd} in vignettes.",
    "Changing the working directory makes vignettes fragile",
    "and non-reproducible on other machines."
  ),

  check = function(state) {
    check_vignette_calls(state, "setwd")
  }
)

get_result <- function(res, check) res$passed[res$check == check]

# -- vignette_no_rm_list ------------------------------------------------------

test_that("vignette_no_rm_list fails when vignette has rm(list = ls())", {
  gp_res <- gp("bad_vignettes", checks = "vignette_no_rm_list")
  res <- results(gp_res)
  expect_false(get_result(res, "vignette_no_rm_list"))

  pos <- failed_positions(gp_res)$vignette_no_rm_list
  lines <- vapply(pos, `[[`, "", "line")
  expect_true(any(grepl("rm\\(list = ls\\(\\)\\)", lines)))
})

test_that("vignette_no_rm_list passes when no vignettes directory", {
  gp_res <- gp("good", checks = "vignette_no_rm_list")
  res <- results(gp_res)
  expect_true(get_result(res, "vignette_no_rm_list"))
})

test_that("vignette_no_rm_list ignores rm() without ls()", {
  pkg <- withr::local_tempdir()
  file.copy(
    list.files("good", full.names = TRUE, recursive = TRUE),
    pkg
  )
  dir.create(file.path(pkg, "vignettes"), showWarnings = FALSE)
  writeLines(c(
    "---",
    "title: test",
    "---",
    "",
    "```{r}",
    "rm(x)",
    "```"
  ), file.path(pkg, "vignettes", "demo.Rmd"))

  state <- PREPS$vignette(list(path = pkg), quiet = TRUE)
  expect_true(CHECKS$vignette_no_rm_list$check(state)$status)
})

# -- vignette_no_setwd --------------------------------------------------------

test_that("vignette_no_setwd fails when vignette has setwd()", {
  gp_res <- gp("bad_vignettes", checks = "vignette_no_setwd")
  res <- results(gp_res)
  expect_false(get_result(res, "vignette_no_setwd"))

  pos <- failed_positions(gp_res)$vignette_no_setwd
  lines <- vapply(pos, `[[`, "", "line")
  expect_true(any(grepl("setwd", lines)))
})

test_that("vignette_no_setwd passes when no vignettes directory", {
  gp_res <- gp("good", checks = "vignette_no_setwd")
  res <- results(gp_res)
  expect_true(get_result(res, "vignette_no_setwd"))
})

# -- shared behaviour ---------------------------------------------------------

test_that("vignette checks ignore non-evaluated chunks", {
  pkg <- withr::local_tempdir()
  file.copy(
    list.files("good", full.names = TRUE, recursive = TRUE),
    pkg
  )
  dir.create(file.path(pkg, "vignettes"), showWarnings = FALSE)
  writeLines(c(
    "---",
    "title: test",
    "---",
    "",
    "```{r eval=FALSE}",
    "rm(list = ls())",
    "setwd('/tmp')",
    "```"
  ), file.path(pkg, "vignettes", "demo.Rmd"))

  state <- PREPS$vignette(list(path = pkg), quiet = TRUE)
  expect_true(CHECKS$vignette_no_rm_list$check(state)$status)
  expect_true(CHECKS$vignette_no_setwd$check(state)$status)
})

test_that("vignette checks ignore #| eval: false chunks in qmd and Rmd", {
  pkg <- withr::local_tempdir()
  file.copy(
    list.files("good", full.names = TRUE, recursive = TRUE),
    pkg
  )
  dir.create(file.path(pkg, "vignettes"), showWarnings = FALSE)

  chunk <- c(
    "---",
    "title: test",
    "---",
    "",
    "```{r}",
    "#| eval: false",
    "rm(list = ls())",
    "setwd('/tmp')",
    "```"
  )
  writeLines(chunk, file.path(pkg, "vignettes", "demo.qmd"))
  writeLines(chunk, file.path(pkg, "vignettes", "demo.Rmd"))

  state <- PREPS$vignette(list(path = pkg), quiet = TRUE)
  expect_true(CHECKS$vignette_no_rm_list$check(state)$status)
  expect_true(CHECKS$vignette_no_setwd$check(state)$status)
})

# -- match_chunk_pairs -------------------------------------------------------

test_that("match_chunk_pairs returns empty matrix for no starts", {
  pair_fn <- match_chunk_pairs
  result <- pair_fn(integer(0), c(5L, 10L))
  expect_identical(nrow(result), 0L)
  expect_identical(ncol(result), 2L)
})

test_that("match_chunk_pairs pairs starts with nearest available end", {
  pair_fn <- match_chunk_pairs
  result <- pair_fn(c(1L, 6L), c(4L, 9L))
  expect_identical(result[, "start"], c(1L, 6L))
  expect_identical(result[, "end"], c(4L, 9L))
})

test_that("match_chunk_pairs messages and returns empty on length mismatch", {
  pair_fn <- match_chunk_pairs
  expect_message(
    result <- pair_fn(c(3L), c(2L, 5L, 8L)),
    "sanity checks"
  )
  expect_identical(nrow(result), 0L)

  expect_message(
    result <- pair_fn(c(1L, 6L, 20L), c(4L, 9L)),
    "sanity checks"
  )
  expect_identical(nrow(result), 0L)
})

test_that("match_chunk_pairs skips ends consumed by earlier chunks", {
  pair_fn <- match_chunk_pairs
  result <- pair_fn(c(1L, 3L), c(2L, 5L))
  expect_identical(result[, "start"], c(1L, 3L))
  expect_identical(result[, "end"], c(2L, 5L))
})

test_that("match_chunk_pairs messages when end < start (ordering sanity)", {
  pair_fn <- match_chunk_pairs
  expect_message(
    result <- pair_fn(c(5L), c(3L)),
    "sanity checks"
  )
  expect_identical(nrow(result), 0L)
  expect_identical(ncol(result), 2L)
})

test_that("match_chunk_pairs messages on overlapping chunks", {
  pair_fn <- match_chunk_pairs
  expect_message(
    result <- pair_fn(c(1L, 3L), c(5L, 7L)),
    "sanity checks"
  )
  expect_identical(nrow(result), 0L)
  expect_identical(ncol(result), 2L)
})

test_that("match_chunk_pairs messages on unsorted starts", {
  pair_fn <- match_chunk_pairs
  expect_message(
    result <- pair_fn(c(6L, 1L), c(3L, 9L)),
    "sanity checks"
  )
  expect_identical(nrow(result), 0L)
  expect_identical(ncol(result), 2L)
})

# -- is_skipped_chunk --------------------------------------------------------

test_that("is_skipped_chunk detects eval=FALSE and purl=FALSE in header", {
  skip_fn <- is_skipped_chunk
  expect_true(skip_fn(c("```{r eval=FALSE}", "x <- 1", "```"), 1, 3))
  expect_true(skip_fn(c("```{r eval=F}", "x <- 1", "```"), 1, 3))
  expect_true(skip_fn(c("```{r eval = FALSE}", "x <- 1", "```"), 1, 3))
  expect_true(skip_fn(c("```{r purl=FALSE}", "x <- 1", "```"), 1, 3))
  expect_false(skip_fn(c("```{r}", "x <- 1", "```"), 1, 3))
  expect_false(skip_fn(c("```{r setup}", "x <- 1", "```"), 1, 3))
})

test_that("is_skipped_chunk detects quarto #| eval: false", {
  skip_fn <- is_skipped_chunk
  expect_true(skip_fn(c("```{r}", "#| eval: false", "x <- 1", "```"), 1, 4))
  expect_true(skip_fn(c("```{r}", "#| eval: FALSE", "x <- 1", "```"), 1, 4))
  expect_true(skip_fn(c("```{r}", "#| eval: F", "x <- 1", "```"), 1, 4))
  expect_true(skip_fn(c("```{r}", "#| purl: false", "x <- 1", "```"), 1, 4))
  expect_true(skip_fn(c("```{r}", "#| purl: FALSE", "x <- 1", "```"), 1, 4))
  expect_false(skip_fn(c("```{r}", "#| label: setup", "x <- 1", "```"), 1, 4))
  expect_true(skip_fn(
    c("```{r}", "#| label: skip", "#| eval: false", "```"), 1, 4
  ))
})

test_that("is_skipped_chunk handles empty chunk body", {
  skip_fn <- is_skipped_chunk
  expect_false(skip_fn(c("```{r}", "```"), 1, 2))
  expect_true(skip_fn(c("```{r eval=FALSE}", "```"), 1, 2))
})

# -- vignette_files -----------------------------------------------------------

test_that("vignette_files returns empty for missing vignettes dir", {
  pkg <- withr::local_tempdir()
  expect_identical(vignette_files(pkg), character())
})

test_that("vignette_files finds Rmd, qmd, and Rnw files", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "vignettes"))
  file.create(file.path(pkg, "vignettes", "a.Rmd"))
  file.create(file.path(pkg, "vignettes", "b.qmd"))
  file.create(file.path(pkg, "vignettes", "c.Rnw"))
  file.create(file.path(pkg, "vignettes", "d.txt"))
  result <- vignette_files(pkg)
  expect_length(result, 3)
  expect_true(all(grepl("\\.(Rmd|qmd|Rnw)$", result)))
})

# -- extract_vignette_code ---------------------------------------------------

test_that("extract_vignette_code handles multiple chunks with mixed skip", {
  f <- tempfile(fileext = ".Rmd")
  on.exit(unlink(f))
  writeLines(c(
    "---",
    "title: test",
    "---",
    "",
    "```{r}",
    "x <- 1",
    "```",
    "",
    "```{r eval=FALSE}",
    "y <- 2",
    "```",
    "",
    "```{r}",
    "#| eval: false",
    "z <- 3",
    "```",
    "",
    "```{r}",
    "w <- 4",
    "```"
  ), f)
  result <- extract_vignette_code(f)
  expect_identical(result[6], "x <- 1")
  expect_identical(result[10], "")
  expect_identical(result[15], "")
  expect_identical(result[19], "w <- 4")
})

test_that("extract_vignette_code handles empty chunk body", {
  f <- tempfile(fileext = ".Rmd")
  on.exit(unlink(f))
  writeLines(c(
    "---",
    "title: test",
    "---",
    "",
    "```{r}",
    "```",
    "",
    "```{r}",
    "x <- 1",
    "```"
  ), f)
  result <- extract_vignette_code(f)
  expect_identical(result[9], "x <- 1")
  expect_identical(result[5], "")
})

test_that("extract_vignette_code returns NULL when all chunks are empty", {
  f <- tempfile(fileext = ".Rmd")
  on.exit(unlink(f))
  writeLines(c(
    "---",
    "title: test",
    "---",
    "",
    "```{r}",
    "```"
  ), f)
  expect_null(extract_vignette_code(f))
})

test_that("extract_vignette_code returns NULL for unreadable file", {
  expect_null(suppressWarnings(
    extract_vignette_code(tempfile(fileext = ".Rmd"))
  ))
})

# -- vignette_parse_data ----------------------------------------------------

test_that("vignette_parse_data returns NULL for empty code chunks", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "vignettes"), showWarnings = FALSE)
  writeLines(c(
    "---",
    "title: test",
    "---",
    "",
    "Just prose, no code."
  ), file.path(pkg, "vignettes", "empty.Rmd"))

  result <- vignette_parse_data(
    file.path(pkg, "vignettes", "empty.Rmd")
  )
  expect_null(result)
})

test_that("vignette_parse_data returns NULL for nonexistent file", {
  expect_warning(
    result <- vignette_parse_data(tempfile(fileext = ".Rmd"))
  )
  expect_null(result)
})

test_that("extract_vignette_code handles Rnw files", {
  f <- tempfile(fileext = ".Rnw")
  on.exit(unlink(f))
  writeLines(c(
    "\\documentclass{article}",
    "\\begin{document}",
    "<<setup>>=",
    "x <- 1",
    "@",
    "\\end{document}"
  ), f)
  result <- extract_vignette_code(f)
  expect_identical(result[4], "x <- 1")
  expect_identical(result[1], "")
})

test_that("extract_vignette_code returns NULL for unknown extension", {
  f <- tempfile(fileext = ".txt")
  on.exit(unlink(f))
  writeLines("some text", f)
  expect_null(extract_vignette_code(f))
})

test_that("extract_vignette_code returns NULL for .md extension", {
  f <- tempfile(fileext = ".md")
  on.exit(unlink(f))
  writeLines(c(
    "---",
    "title: test",
    "---",
    "",
    "```{r}",
    "x <- 1",
    "```"
  ), f)
  expect_null(extract_vignette_code(f))
})

test_that("extract_vignette_code skips chunk with no closing fence", {
  f <- tempfile(fileext = ".Rmd")
  on.exit(unlink(f))
  writeLines(c(
    "---",
    "title: test",
    "---",
    "",
    "```{r}",
    "x <- 1"
  ), f)
  expect_message(
    result <- extract_vignette_code(f),
    "sanity checks"
  )
  expect_null(result)
})

test_that("vignette_parse_data returns NULL for unparseable code", {
  f <- tempfile(fileext = ".Rmd")
  on.exit(unlink(f))
  writeLines(c(
    "---",
    "title: test",
    "---",
    "",
    "```{r}",
    "if (TRUE {",
    "```"
  ), f)
  expect_null(vignette_parse_data(f))
})

# -- safe_parse ---------------------------------------------------------------

test_that("safe_parse parses from file path", {
  f <- tempfile(fileext = ".R")
  on.exit(unlink(f))
  writeLines("x <- 1 + 2", f)
  result <- safe_parse(file = f)
  expect_gt(length(result), 0)
})

test_that("safe_parse returns NULL when keep_source FALSE also fails", {
  result <- safe_parse(text = "if (TRUE {", keep_source = FALSE)
  expect_null(result)
})

# -- PREPS$vignette ----------------------------------------------------------

test_that("PREPS$vignette stores parse_data and lines", {
  pkg <- withr::local_tempdir()
  file.copy(
    list.files("good", full.names = TRUE, recursive = TRUE),
    pkg
  )
  dir.create(file.path(pkg, "vignettes"), showWarnings = FALSE)
  writeLines(c(
    "---",
    "title: test",
    "---",
    "",
    "```{r}",
    "x <- 1",
    "```"
  ), file.path(pkg, "vignettes", "demo.Rmd"))

  state <- PREPS$vignette(list(path = pkg), quiet = TRUE)
  expect_length(state$vignette, 1)
  entry <- state$vignette[[1]]
  expect_true("parse_data" %in% names(entry))
  expect_true("lines" %in% names(entry))
  expect_true(is.data.frame(entry$parse_data))
  expect_true(any(grepl("x <- 1", entry$lines)))
})

test_that("PREPS$vignette skips files with no parseable code", {
  pkg <- withr::local_tempdir()
  file.copy(
    list.files("good", full.names = TRUE, recursive = TRUE),
    pkg
  )
  dir.create(file.path(pkg, "vignettes"), showWarnings = FALSE)
  writeLines(c(
    "---",
    "title: prose only",
    "---",
    "",
    "No code chunks here."
  ), file.path(pkg, "vignettes", "nochunks.Rmd"))

  state <- PREPS$vignette(list(path = pkg), quiet = TRUE)
  expect_length(state$vignette, 0)
})

# -- call_descendants --------------------------------------------------------

test_that("call_descendants returns all nested IDs", {
  code <- "rm(list = ls())"
  parsed <- parse(text = code, keep.source = TRUE)
  pd <- getParseData(parsed)

  ls_row <- pd[pd$token == "SYMBOL_FUNCTION_CALL" & pd$text == "ls", ]
  expect_identical(nrow(ls_row), 1L)

  desc_ids <- call_descendants(pd, ls_row$id[1])
  desc <- pd[pd$id %in% desc_ids, ]
  expect_gt(nrow(desc), 0)
})

# -- check_vignette_calls ---------------------------------------------------

test_that("check_vignette_calls skips rm() when nested ls() not found", {
  code <- "rm(x)"
  parsed <- parse(text = code, keep.source = TRUE)
  pd <- getParseData(parsed)

  state <- list(vignette = list(
    "test.Rmd" = list(parse_data = pd, lines = code)
  ))
  result <- check_vignette_calls(state, "rm", nested_fn = "ls")
  expect_true(result$status)
  expect_length(result$positions, 0)
})

test_that("check_vignette_calls passes with empty vignette state", {
  state <- list(vignette = list())
  result <- check_vignette_calls(state, "setwd")
  expect_true(result$status)
  expect_length(result$positions, 0)
})

test_that("vignette checks report correct positions", {
  gp_res <- gp("bad_vignettes",
                checks = c("vignette_no_rm_list", "vignette_no_setwd"))

  rm_pos <- failed_positions(gp_res)$vignette_no_rm_list
  expect_true(all(vapply(rm_pos, function(p) {
    grepl("^vignettes/", p$filename)
  }, logical(1))))
  expect_identical(rm_pos[[1]]$line_number, 10L)

  setwd_pos <- failed_positions(gp_res)$vignette_no_setwd
  expect_true(all(vapply(setwd_pos, function(p) {
    grepl("^vignettes/", p$filename)
  }, logical(1))))
  expect_identical(setwd_pos[[1]]$line_number, 11L)
})

#' An exported function
#' @return TRUE
#' @export
exported_func <- function() {
  TRUE
}

#' Internal with noRd but missing keywords internal
#' @noRd
internal_no_keywords <- function() {
  FALSE
}

#' Properly tagged internal function
#' @noRd
#' @keywords internal
internal_proper <- function() {
  NULL
}

untagged_func <- function() {
  1
}

#' Documented but missing export/noRd tag
#' @return A number.
documented_no_tag <- function() {
  42
}

#' Contradictory tags
#' @export
#' @keywords internal
#' @return A number.
confused_func <- function() {
  2
}

#' Uses a deprecated tag
#' @S3method print foo
#' @noRd
#' @keywords internal
deprecated_tag_func <- function() {
  3
}

#' Inherits from nonexistent function
#' @inheritParams nonexistent_function
#' @return Nothing.
#' @export
#' @keywords internal
bad_inherit_func <- function(x) {
  x
}

myFunc <- function(x ,y)
{
    if(x == NA & y > 0)return( x )
    z <- x+y
    # z <- z * 2
    w <- f (z)
    msg <- 'hello world'   
    x %>% sum() |> print()
    a_very_long_and_excessively_named_function_variable_name_that_goes_well_beyond_the_one_hundred_twenty_char_limit_for_line_length <- w
    result = T
    idx <- 1:length(z); w <- z
	return(z)
}

#' Print goodpractice results
#'
#' @param x Object of class `goodPractice`, as returned by [gp()].
#' @param groups Name of check groups for which to print results, as vector of
#' one or more of [all_check_groups()].
#' @param positions_limit How many positions to print at most.
#' @param ... Unused, for compatibility with [base::print()] generic method.
#'
#' @importFrom rstudioapi hasFun
#' @importFrom praise praise
#' @importFrom stats setNames
#'
#' @family output
#' @export

print.goodPractice <- function(x, groups = NULL, positions_limit = 5, ...) {

  failure <- FALSE
  has_info <- FALSE

  checks <- gp_print_groups(names(x$checks), groups = groups)

  for (check in checks) {
    type <- check_type(x$checks[[check]])
    if (!check_passed(x$checks[[check]], na_as_passed = TRUE)) {
      if (!failure) {
        failure <- TRUE
        gp_header(x)
      }
      gp_advice(x, check, positions_limit, type = "error")
    } else if (type == "info") {
      if (!failure && !has_info) {
        has_info <- TRUE
        gp_header(x)
      }
      gp_advice(x, check, positions_limit, type = "info")
    }
  }

  if (failure || has_info) {
    cli::cat_rule()
    if (failure &&
        getOption("goodpractice.rstudio_source_markers", TRUE) &&
        hasFun("sourceMarkers")) {
      rstudio_source_markers(x)
    }
  }

  if (!failure) {
    msg <- praise(
      '${Exclamation}! ${Adjective} package! Keep up the ${adjective} work!'
    )
    cli::cli_text("{cli::col_red(cli::symbol$heart)} {.strong {msg}}")
  }

  invisible(x)
}

gp_header <- function(x) {
  cli::cat_rule("It is good practice to", col = "cyan")
  cli::cli_text()
}

gp_advice <- function(state, fail, limit, type = "error") {
  MYCHECKS <- prepare_checks(CHECKS, state$extra_checks)
  chk <- MYCHECKS[[fail]]
  res <- state$checks[[fail]]

  str <- if (is.function(chk$gp)) chk$gp(state) else chk$gp
  str <- cli::format_inline(gsub("\n\\s*", " ", str))
  str <- gsub("[{]", "{{", gsub("[}]", "}}", str))

  bullet <- switch(type, info = "i", warning = "!", "x")
  cli::cli_bullets(setNames(str, bullet))

  if ("positions" %in% names(res)) {
    cli::cli_text()
    withr::with_dir(state$path, gp_positions(res[["positions"]], limit))
  }

  cli::cli_text()
}

gp_positions <- function(pos, limit) {
  num <- length(pos)
  if (num > limit) pos <- pos[seq_len(limit)]

  id <- cli::cli_div(theme = list(div = list("margin-left" = 4)))

  for (p in pos) {
    loc <- if (is.na(p$line_number)) "" else paste0(":", p$line_number)
    if (nzchar(loc) && !is.na(p$column_number)) {
      loc <- paste0(loc, ":", p$column_number)
    }
    cli::cli_text("{.path {p$filename}{loc}}")
  }

  if (num > limit) {
    cli::cli_text("{.emph ... and {num - limit} more line{?s}}")
  }

  cli::cli_end(id)
}

gp_print_groups <- function (checks, groups = NULL) {
    if (is.null (groups)) {
        return (checks)
    }
    stopifnot (all (groups %in% all_check_groups ()))
    group_checks <- unlist (lapply (groups, checks_by_group))
    checks [which (checks %in% group_checks)]
}

test_that("print with default and explicit positions.limit", {
  bad1 <- system.file("bad1", package = "goodpractice")
  x <- gp(bad1, checks = "r_file_extension")

  expect_snapshot(print(x))
})

test_that("print shows praise when all checks pass", {
  gp_res <- gp("good", checks = "description_bugreports")
  expect_snapshot(print(gp_res))
})

test_that("gp_positions truncates when exceeding limit", {
  pos <- lapply(1:8, function(i) {
    list(
      filename = paste0("R/file", i, ".R"),
      line_number = i * 10,
      column_number = NA_integer_,
      ranges = list(),
      line = paste0("line ", i)
    )
  })

  tmp <- withr::local_tempdir()
  expect_snapshot(
    withr::with_dir(tmp, gp_positions(pos, limit = 3))
  )
})

test_that("gp_positions handles NA line_number", {
  pos <- list(list(
    filename = "R/test.R",
    line_number = NA_integer_,
    column_number = NA_integer_,
    ranges = list(),
    line = "test line"
  ))

  tmp <- withr::local_tempdir()
  expect_snapshot(
    withr::with_dir(tmp, gp_positions(pos, limit = 5))
  )
})

test_that("gp_positions includes column when available", {
  pos <- list(list(
    filename = "R/test.R",
    line_number = 10L,
    column_number = 5L,
    ranges = list(),
    line = "test line"
  ))

  tmp <- withr::local_tempdir()
  expect_snapshot(
    withr::with_dir(tmp, gp_positions(pos, limit = 5))
  )
})

test_that("gp_print_groups returns all checks when groups is NULL", {
  checks <- c("description_bugreports", "no_description_date", "r_file_extension")
  expect_identical(gp_print_groups(checks, groups = NULL), checks)
})

test_that("gp_print_groups filters checks to a single group", {
  all_desc <- checks_by_group("description")
  other <- "r_file_extension"
  input <- c(all_desc, other)
  result <- gp_print_groups(input, groups = "description")
  expect_true(all(result %in% all_desc))
  expect_false(other %in% result)
})

test_that("gp_print_groups filters checks to multiple groups", {
  desc <- checks_by_group("description")
  ns <- checks_by_group("namespace")
  other <- "r_file_extension"
  input <- c(desc, ns, other)
  result <- gp_print_groups(input, groups = c("description", "namespace"))
  expect_true(all(result %in% c(desc, ns)))
  expect_false(other %in% result)
})

test_that("gp_print_groups errors on invalid group", {
  expect_error(gp_print_groups(character(0), groups = "not_a_real_group"))
})

test_that("print with groups filters output to that group", {
  bad1 <- system.file("bad1", package = "goodpractice")
  x <- gp(bad1, checks = c("r_file_extension", "description_bugreports"))
  expect_snapshot(print(x, groups = "description"))
})

test_that("print with invalid groups errors", {
  bad1 <- system.file("bad1", package = "goodpractice")
  x <- gp(bad1, checks = "r_file_extension")
  expect_error(print(x, groups = "not_a_real_group"))
})

test_that("check_type extracts type from result", {
  expect_identical(check_type(TRUE), "error")
  expect_identical(check_type(list(status = TRUE)), "error")
  expect_identical(check_type(list(status = TRUE, type = "info")), "info")
  expect_identical(check_type(list(status = FALSE, type = "warning")), "warning")
})

test_that("print shows info messages with praise", {
  info_check <- make_check(
    description = "An info check",
    tags = character(),
    preps = "code_structure",
    gp = "consider doing something.",
    check = function(state) list(status = TRUE, type = "info")
  )
  gp_res <- gp(
    "good",
    checks = c("info_test", "description_bugreports"),
    extra_checks = list(info_test = info_check)
  )
  expect_snapshot(print(gp_res))
})

test_that("print calls rstudio_source_markers when hasFun is TRUE", {
  bad1 <- system.file("bad1", package = "goodpractice")
  x <- gp(bad1, checks = "r_file_extension")

  called <- FALSE
  local_mocked_bindings(
    hasFun = function(...) TRUE,
    rstudio_source_markers = function(...) {
      called <<- TRUE
    }
  )
  expect_snapshot(print(x))
  expect_true(called)
})

#' @include treesitter.R

#' @noRd
get_tidyverse_lintr_position <- function(lint) {
  lint[c("filename", "line_number", "column_number", "ranges", "line")]
}

get_tidyverse_lintr_state <- function(state, linter) {
  if (inherits(state$tidyverse_lintr, "try-error")) {
    return(na_result())
  }

  linters <- vapply(state$tidyverse_lintr, "[[", "", "linter")
  check_result(!linter %in% linters, lapply(
      state$tidyverse_lintr[linters == linter],
      get_tidyverse_lintr_position)
  )
}

## --------------------------------------------------------------------
## Lintr-based tidyverse style checks
## --------------------------------------------------------------------

CHECKS$tidyverse_brace_linter <- make_check(

  description = "Opening brace on same line, closing brace on its own line",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'place the opening brace `{{` on the same line as the
        statement and the closing brace `}}` on its own line.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "brace_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_commas_linter <- make_check(

  description = "Space after commas, not before",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'place a space after a comma, never before, just like in
        regular English.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "commas_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_commented_code_linter <- make_check(

  description = "No commented-out code",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'remove commented-out code. Use version control to track
        old code instead of commenting it out.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "commented_code_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_equals_na_linter <- make_check(

  description = "Use is.na() not == NA",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'use {.code is.na(x)} instead of
        {.code x == NA}. Comparing with {.code ==}
        always returns {.code NA}, never {.code TRUE}
        or {.code FALSE}.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "equals_na_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_function_left_parentheses_linter <- make_check(

  description = "No space before ( in function calls",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'not put a space before the opening parenthesis in a
        function call. Write {.code f(x)} not {.code f (x)}.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "function_left_parentheses_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_indentation_linter <- make_check(

  description = "Use 2-space indentation",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'use two spaces for indentation. Do not use tabs.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "indentation_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_infix_spaces_linter <- make_check(

  description = "Spaces around infix operators",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'put spaces around infix operators like {.code +}, {.code -}, {.code <-},
        {.code ==}, etc. Exception: no spaces around
        {.code ::} and {.code :::}.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "infix_spaces_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_object_length_linter <- make_check(

  description = "Object names are not too long",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'keep object names concise. Strive for names that are
        concise and meaningful.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "object_length_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_object_name_linter <- make_check(

  description = "Object names use snake_case",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'use snake_case for variable and function names. Avoid
        camelCase, PascalCase, and dot.case.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "object_name_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_object_usage_linter <- make_check(

  description = "No unused variables or undefined globals",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'avoid defining variables that are never used and ensure
        all referenced objects are defined or imported.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "object_usage_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_paren_body_linter <- make_check(

  description = "No space after opening parenthesis",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'not put a space after {.code (} or before {.code )}. Write {.code f(x)}
        not {.code f( x )}.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "paren_body_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_pipe_consistency_linter <- make_check(

  description = "Consistent pipe operator usage",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'use one pipe operator consistently throughout the package.
        Do not mix {.code %>%} and {.code |>} in the same codebase.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "pipe_consistency_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_pipe_continuation_linter <- make_check(

  description = "Pipe continuation on next line",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'place the pipe operator at the end of the line, not the
        beginning of the next line.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "pipe_continuation_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_quotes_linter <- make_check(

  description = "Use double quotes for strings",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'use double quotes {.code "} for strings, not single quotes {.code \'}.
        The only exception is when the string contains double quotes.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "quotes_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_return_linter <- make_check(

  description = "Use implicit return, not explicit return()",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'rely on implicit return. Only use {.fn return} for early
        returns. The last expression in a function is automatically
        returned.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "return_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_spaces_inside_linter <- make_check(

  description = "No spaces inside parentheses or brackets",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'not put spaces inside parentheses or brackets. Write
        {.code x[1]} and {.code f(x)}, not {.code x[ 1 ]} or {.code f( x )}.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "spaces_inside_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_spaces_left_parentheses_linter <- make_check(

  description = "Space before ( in control flow",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'place a space before {.code (} when used with {.code if}, {.code for},
        {.code while}, etc. Write {.code if (x)} not {.code if(x)}.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "spaces_left_parentheses_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_trailing_blank_lines_linter <- make_check(

  description = "No trailing blank lines",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'end files with a single newline, not multiple blank lines.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "trailing_blank_lines_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_trailing_whitespace_linter <- make_check(

  description = "No trailing whitespace",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'remove trailing whitespace from code lines.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "trailing_whitespace_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_vector_logic_linter <- make_check(

  description = "Use && and || not & and | in if conditions",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'use {.code &&} and {.code ||} (scalar logical operators) in {.code if}
        conditions, not {.code &} and {.code |} (vector logical operators).',

  check = function(state) {
    get_tidyverse_lintr_state(state, "vector_logic_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_whitespace_linter <- make_check(

  description = "No tab characters in code",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'use spaces instead of tabs for indentation.',

  check = function(state) {
    get_tidyverse_lintr_state(state, "whitespace_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_assignment_linter <- make_check(

  description = "Use <- for assignment, not =",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = "use {.code <-} for assignment instead of {.code =}. This is the
        standard convention in R and makes code easier to read.",

  check = function(state) {
    result <- get_tidyverse_lintr_state(state, "assignment_linter")
    filter_s4_assignment_false_positives(state, result)
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_line_length_linter <- make_check(

  description = "Code lines are not too long",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = "keep code lines to a reasonable length for readability.",

  check = function(state) {
    get_tidyverse_lintr_state(state, "line_length_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_semicolon_linter <- make_check(

  description = "No trailing semicolons",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = "omit trailing semicolons from code lines. They are not
        needed in R and most style guides forbid them.",

  check = function(state) {
    get_tidyverse_lintr_state(state, "semicolon_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_seq_linter <- make_check(

  description = "Use seq_len() or seq_along() instead of 1:length(...)",
  tags = c("warning", "tidyverse"),
  preps = "tidyverse",

  gp = "avoid {.code 1:length(...)},
        {.code 1:nrow(...)}, and similar expressions.
        They are error prone when the right hand side
        is zero. Use {.fn seq_len} or {.fn seq_along}
        instead.",

  check = function(state) {
    get_tidyverse_lintr_state(state, "seq_linter")
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_T_and_F_symbol_linter <- make_check(

  description = "Use TRUE and FALSE, not T and F",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = "use {.code TRUE} and {.code FALSE} instead
        of {.code T} and {.code F}. {.code T} and
        {.code F} are not reserved words and can be
        overwritten, leading to unexpected behaviour.",

  check = function(state) {
    get_tidyverse_lintr_state(state, "T_and_F_symbol_linter")
  }
)

## --------------------------------------------------------------------
## Non-lintr structural checks
## --------------------------------------------------------------------

CHECKS$tidyverse_r_file_names <- make_check(

  description = "R file names use snake_case",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = "name R files using {.code -} or {.code _} to delimit
        words and use a {.file .R} extension, e.g. {.file
        prefix-my_function.R}. Avoid capital letters and spaces.",

  check = function(state) {
    r_dir <- file.path(state$path, "R")
    if (!dir.exists(r_dir)) return(check_result(TRUE))

    all_files <- list.files(r_dir, full.names = TRUE)
    file_name_wo_ext <- basename(tools::file_path_sans_ext(all_files))
    bad_file_name <- grepl(pattern = "[A-Z[:space:]]", file_name_wo_ext)
    bad_file_ext <- tools::file_ext(all_files) != "R"
    bad_files <- all_files[bad_file_name | bad_file_ext]

    check_result(length(bad_files) == 0, lapply(bad_files, function(f) {
        check_position(f, line = f)
      }))
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_test_file_names <- make_check(

  description = "Test files mirror R source files",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = 'name test files to mirror the R source file they test,
        e.g. {.file R/my_function.R} should have
        {.file tests/testthat/test-my_function.R}.',

  check = function(state) {
    r_dir <- file.path(state$path, "R")
    if (!dir.exists(r_dir)) return(TRUE)

    r_files <- tools::file_path_sans_ext(
      basename(list.files(r_dir, pattern = "\\.[Rr]$"))
    )
    if (length(r_files) == 0) return(check_result(TRUE))

    test_dir <- file.path(state$path, "tests", "testthat")
    test_files <- if (dir.exists(test_dir)) {
      list.files(test_dir, pattern = "^test-.*\\.[Rr]$")
    } else {
      character()
    }
    tested <- sub("^test-", "", tools::file_path_sans_ext(test_files))

    untested <- setdiff(r_files, tested)
    untested <- untested[!grepl("^(zzz|RcppExports|reexport)", untested)]

    check_result(length(untested) == 0, lapply(untested, function(f) {
        check_position(
          file.path("R", paste0(f, ".R")),
          line = paste0("missing tests/testthat/test-", f, ".R")
          )
      }))
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_no_missing <- make_check(

  description = "Functions do not use missing() to check arguments",
  tags = c("warning", "tidyverse"),
  preps = "tidyverse",

  gp = paste(
    "avoid using {.fn missing} to check whether arguments were supplied.",
    "The tidyverse style guide recommends using {.code NULL} defaults with",
    "{.fn is.null} instead, as {.fn missing} makes functions difficult to",
    "call programmatically."
  ),

  check = function(state) {
    ts <- ts_get(state)
    if (length(ts$functions) == 0) {
      return(check_result(TRUE))
    }
    funcs <- ts$functions
    missing_q <- treesitter::query(ts$language,
      "(call function: (identifier) @fn (#eq? @fn \"missing\"))"
    )
    problems <- list()

    for (fn in funcs) {
      if (ts_body_has_call(fn$fn_node, missing_q)) {
        problems[[length(problems) + 1]] <- check_position(
          file.path("R", basename(fn$file)),
          fn$line,
          line = fn$name
        )
      }
    }

    if (length(problems) == 0) {
      check_result(TRUE)
    } else {
      check_result(FALSE, problems)
    }
  }
)

## --------------------------------------------------------------------

CHECKS$tidyverse_export_order <- make_check(

  description = "Exported functions are defined before internal helpers",
  tags = c("style", "tidyverse"),
  preps = "tidyverse",

  gp = "define exported (user-facing) functions before internal
        helper functions within each R source file.",

  check = function(state) {
    path <- normalizePath(state$path)
    ns <- try(parseNamespaceFile(basename(path), dirname(path)), silent = TRUE)
    if (inherits(ns, "try-error")) {
      return(na_result())
    }

    exported <- c(ns$exports, ns_s3_method_names(ns))

    patterns <- ns$exportPatterns

    is_exported <- function(name) {
      if (name %in% exported) return(TRUE)
      for (p in patterns) {
        if (grepl(p, name)) return(TRUE)
      }
      FALSE
    }

    funcs <- ts_get(state)$functions
    if (length(funcs) == 0) {
      return(check_result(TRUE))
    }

    by_file <- split(funcs, vapply(funcs, `[[`, "", "file"))
    problems <- list()

    for (file_funcs in by_file) {
      if (length(file_funcs) < 2) next

      is_exp <- vapply(
        file_funcs,
        function(fn) is_exported(fn$name),
        logical(1)
      )
      if (!any(is_exp) || all(is_exp)) next

      last_export <- max(which(is_exp))

      for (j in seq_len(last_export - 1)) {
        if (!is_exp[j]) {
          fn <- file_funcs[[j]]
          problems[[length(problems) + 1]] <- check_position(
            file.path("R", basename(fn$file)),
            fn$line,
            line = fn$name
          )
        }
      }
    }

    if (length(problems) == 0) {
      check_result(TRUE)
    } else {
      check_result(FALSE, problems)
    }
  }
)

#' @include lists.R customization.R treesitter.R

## -- print methods return invisible -----------------------------------------

CHECKS$print_return_invisible <- make_check(

  description = "Print methods return the object invisibly",
  tags = c("warning", "best practice"),
  preps = "code_structure",

  gp = paste(
    "print methods should return the input object invisibly,",
    "e.g. {.code invisible(x)}. This allows chaining and consistent behaviour",
    "with base R print methods."
  ),

  check = function(state) {
    ts <- ts_get(state)
    if (length(ts$functions) == 0) {
      return(check_result(TRUE))
    }

    invisible_q <- treesitter::query(ts$language,
      "(call function: (identifier) @fn (#eq? @fn \"invisible\"))"
    )

    problems <- list()
    for (fn in ts$functions) {
      if (!startsWith(fn$name, "print.")) next

      body <- treesitter::node_child_by_field_name(fn$fn_node, "body")
      caps <- treesitter::query_captures(invisible_q, body)
      if (length(caps$name) > 0) next

      problems[[length(problems) + 1]] <- check_position(
        file.path("R", basename(fn$file)),
        fn$line,
        line = fn$name
      )
    }

    if (length(problems) == 0) {
      check_result(TRUE)
    } else {
      check_result(FALSE, problems)
    }
  }
)

## -- on.exit() should have add = TRUE ---------------------------------------

on_exit_call_missing_add <- function(call_node) {
  args_node <- treesitter::node_child_by_field_name(call_node, "arguments")
  if (is.null(args_node)) return(FALSE)

  n_children <- treesitter::node_child_count(args_node)
  if (n_children <= 2L) return(FALSE)

  for (i in seq_len(n_children)) {
    child <- treesitter::node_child(args_node, i)
    if (treesitter::node_type(child) != "argument") next
    name_node <- treesitter::node_child_by_field_name(child, "name")
    if (!is.null(name_node) && treesitter::node_text(name_node) == "add") {
      return(FALSE)
    }
  }

  TRUE
}

CHECKS$on_exit_has_add <- make_check(

  description = "on.exit() calls include add = TRUE",
  tags = c("warning", "best practice"),
  preps = "code_structure",

  gp = paste(
    "always use {.code add = TRUE} in {.fn on.exit} calls.",
    "Without it, each {.fn on.exit} overwrites previous exit handlers,",
    "which is a common source of bugs."
  ),

  check = function(state) {
    ts <- ts_get(state)
    if (length(ts$functions) == 0) {
      return(check_result(TRUE))
    }

    on_exit_q <- treesitter::query(ts$language,
      "(call function: (identifier) @fn (#eq? @fn \"on.exit\"))"
    )

    problems <- list()
    for (fn in ts$functions) {
      if (!ts_body_has_call(fn$fn_node, on_exit_q)) next

      body <- treesitter::node_child_by_field_name(fn$fn_node, "body")
      caps <- treesitter::query_captures(on_exit_q, body)
      for (j in which(caps$name == "fn")) {
        if (ts_inside_nested_function(caps$node[[j]], body)) next
        call_node <- treesitter::node_parent(caps$node[[j]])
        if (on_exit_call_missing_add(call_node)) {
          problems[[length(problems) + 1]] <- check_position(
            file.path("R", basename(fn$file)),
            treesitter::node_start_point(call_node)$row + 1L,
            line = fn$name
          )
        }
      }
    }

    if (length(problems) == 0) {
      check_result(TRUE)
    } else {
      check_result(FALSE, problems)
    }
  }
)

## -- function length --------------------------------------------------------

ts_function_length <- function(fn_node) {
  body <- treesitter::node_child_by_field_name(fn_node, "body")
  if (is.null(body)) return(0L)
  start <- treesitter::node_start_point(body)$row
  end <- treesitter::node_end_point(body)$row
  as.integer(end - start + 1L)
}

CHECKS$complexity_function_length <- make_check(

  description = "Functions are not too long",
  tags = c("warning", "best practice"),
  preps = "code_structure",

  gp = paste(
    "keep functions short and focused. Long functions are harder to",
    "understand, test, and maintain. Consider splitting into smaller helpers.",
    "(Default limit of 50 set with option 'goodpractice.function_length_limit')."
  ),

  check = function(state) {
    ts <- ts_get(state)
    if (length(ts$functions) == 0) {
      return(check_result(TRUE))
    }

    limit <- getOption("goodpractice.function_length_limit", 50L)

    problems <- lapply(ts$functions, function(fn) {
      len <- ts_function_length(fn$fn_node)
      if (len <= limit) return(NULL)
      check_position(
        file.path("R", basename(fn$file)),
        fn$line,
        line = paste0(fn$name, " (", len, " lines)")
      )
    })
    problems <- Filter(Negate(is.null), problems)

    check_result(length(problems) == 0, problems)
  }
)

## -- unused internal functions ----------------------------------------------

# Standard R package hooks: called automatically by R itself, never
# referenced directly from package code.
R_PACKAGE_HOOKS <- c(".onLoad", ".onAttach", ".onUnload", ".onDetach", ".Last.lib")

ts_all_referenced_functions <- function(ts) {
  if (length(ts$trees) == 0) return(character())

  call_q <- treesitter::query(ts$language,
    "(call function: (identifier) @fn)"
  )

  refs <- unlist(lapply(ts$trees, function(entry) {
    if (is.null(entry)) return(NULL)
    root <- entry$root
    caps <- treesitter::query_captures(call_q, root)
    called <- vapply(
      caps$node[caps$name == "fn"],
      treesitter::node_text, character(1)
    )
    body_refs <- ts_body_identifiers(ts$language, entry)
    toplevel_refs <- ts_rhs_identifiers(ts$language, entry)
    c(called, body_refs, toplevel_refs)
  }))

  unique(refs)
}

ts_rhs_identifiers <- function(language, entry) {
  assign_q <- treesitter::query(language,
    "(binary_operator rhs: (identifier) @rhs)"
  )
  caps <- treesitter::query_captures(assign_q, entry$root)
  if (length(caps$name) == 0) return(character())
  vapply(
    caps$node[caps$name == "rhs"],
    treesitter::node_text, character(1)
  )
}

ts_body_identifiers <- function(language, entry) {
  id_q <- treesitter::query(language, "(identifier) @id")
  fn_q <- treesitter::query(language, "(function_definition) @fn")
  infix_q <- treesitter::query(language, "(binary_operator operator: _ @op)")

  fn_caps <- treesitter::query_captures(fn_q, entry$root)
  fn_nodes <- fn_caps$node[fn_caps$name == "fn"]
  if (length(fn_nodes) == 0) return(character())

  ids <- unlist(lapply(fn_nodes, function(fn_node) {
    body <- treesitter::node_child_by_field_name(fn_node, "body")
    if (is.null(body)) return(NULL)
    caps <- treesitter::query_captures(id_q, body)
    vapply(
      caps$node[caps$name == "id"],
      treesitter::node_text, character(1)
    )
  }))

  infix_caps <- treesitter::query_captures(infix_q, entry$root)
  op_nodes <- infix_caps$node[infix_caps$name == "op"]
  is_special <- vapply(op_nodes, treesitter::node_type, character(1)) == "special"
  infix_names <- vapply(op_nodes[is_special], treesitter::node_text, character(1))

  c(ids, infix_names)
}

CHECKS$complexity_unused_internal <- make_check(

  description = "All internal functions are used",
  tags = c("warning", "best practice"),
  preps = "namespace",

  gp = paste(
    "remove or use internal functions that are defined but never called.",
    "Dead code increases maintenance burden."
  ),

  check = function(state) {
    ts <- ts_get(state)
    if (length(ts$functions) == 0) {
      return(check_result(TRUE))
    }

    if (inherits(state$namespace, "try-error")) {
      return(na_result())
    }

    ns <- state$namespace
    exports <- ns$exports
    exported <- c(exports, ns_s3_method_names(ns))

    all_defined <- vapply(ts$functions, `[[`, "", "name")
    internal <- setdiff(all_defined, c(exported, R_PACKAGE_HOOKS))
    if (length(internal) == 0) {
      return(check_result(TRUE))
    }

    called <- ts_all_referenced_functions(ts)
    unused <- setdiff(internal, called)
    if (length(unused) == 0) {
      return(check_result(TRUE))
    }

    fn_lookup <- ts$functions
    names(fn_lookup) <- all_defined

    problems <- lapply(unused, function(name) {
      fn <- fn_lookup[[name]]
      check_position(file.path("R", basename(fn$file)), fn$line, line = name)
    })

    check_result(FALSE, problems)
  }
)

## -- duplicate function bodies ------------------------------------------------

cross_file_duplicates <- function(df, key_col, file_col) {
  duped_keys <- unique(df[[key_col]][duplicated(df[[key_col]])])
  if (length(duped_keys) == 0) return(df[0, , drop = FALSE])

  candidates <- df[df[[key_col]] %in% duped_keys, , drop = FALSE]
  multi_file <- vapply(duped_keys, function(k) {
    matched <- candidates[[file_col]][candidates[[key_col]] == k]
    length(unique(basename(matched))) >= 2
  }, logical(1))

  candidates[candidates[[key_col]] %in% duped_keys[multi_file], , drop = FALSE]
}

normalize_body_text <- function(fn_node) {
  body <- treesitter::node_child_by_field_name(fn_node, "body")
  if (is.null(body)) return("")
  gsub("\\s+", " ", trimws(treesitter::node_text(body)))
}

CHECKS$duplicate_function_bodies <- make_check(

  description = "No functions with identical bodies",
  tags = c("warning", "best practice"),
  preps = "code_structure",

  gp = paste(
    "consolidate functions with identical bodies into a single",
    "shared helper to reduce code duplication."
  ),

  check = function(state) {
    ts <- ts_get(state)
    if (length(ts$functions) < 2) {
      return(check_result(TRUE))
    }

    bodies <- vapply(ts$functions, function(fn) {
      normalize_body_text(fn$fn_node)
    }, character(1))

    min_dup_chars <- 20L
    trivial <- nchar(bodies) < min_dup_chars
    bodies[trivial] <- paste0("__trivial__", seq_along(bodies)[trivial])

    fn_df <- data.frame(
      body = bodies,
      name = vapply(ts$functions, `[[`, "", "name"),
      file = vapply(ts$functions, `[[`, "", "file"),
      line = vapply(ts$functions, function(fn) fn$line, numeric(1))
    )

    dupes <- cross_file_duplicates(fn_df, "body", "file")
    if (nrow(dupes) == 0) return(check_result(TRUE))

    problems <- lapply(seq_len(nrow(dupes)), function(i) {
      check_position(
        file.path("R", basename(dupes$file[i])),
        dupes$line[i],
        line = dupes$name[i]
      )
    })

    check_result(FALSE, problems)
  }
)

# -- complexity_function_length -----------------------------------------------

test_that("complexity_function_length fails on long functions", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: longfn", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "DESCRIPTION"))

  body <- paste(
    paste0("  x", seq_len(60), " <- ", seq_len(60)), collapse = "\n"
  )
  writeLines(
    paste0("big_fn <- function() {\n", body, "\n}"),
    file.path(pkg, "R", "big.R")
  )

  gp_res <- gp(pkg, checks = "complexity_function_length")
  res <- results(gp_res)
  expect_false(res$passed[res$check == "complexity_function_length"])

  pos <- failed_positions(gp_res)$complexity_function_length
  expect_true(any(grepl("big_fn", vapply(pos, `[[`, "", "line"))))
})

test_that("complexity_function_length passes on short functions", {
  gp_res <- gp("good", checks = "complexity_function_length")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "complexity_function_length"])
})

test_that("complexity_function_length respects custom limit option", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: customlimit", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "DESCRIPTION"))

  body <- paste(
    paste0("  x", seq_len(10), " <- ", seq_len(10)), collapse = "\n"
  )
  writeLines(
    paste0("medium_fn <- function() {\n", body, "\n}"),
    file.path(pkg, "R", "medium.R")
  )

  withr::local_options(goodpractice.function_length_limit = 5L)
  gp_res <- gp(pkg, checks = "complexity_function_length")
  res <- results(gp_res)
  expect_false(res$passed[res$check == "complexity_function_length"])
})

test_that("complexity_function_length passes with no R directory", {
  pkg <- withr::local_tempdir()
  writeLines(c(
    "Package: nodir", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "DESCRIPTION"))

  gp_res <- gp(pkg, checks = "complexity_function_length")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "complexity_function_length"])
})

test_that("ts_function_length counts lines correctly", {
  code <- "my_fn <- function(x) {\n  x + 1\n  x + 2\n}"
  lang <- treesitter.r::language()
  p <- treesitter::parser(lang)
  tree <- treesitter::parser_parse(p, code)
  root <- treesitter::tree_root_node(tree)
  fns <- ts_file_functions(root, "test.R")
  expect_identical(ts_function_length(fns[[1]]$fn_node), 4L)
})

test_that("ts_all_referenced_functions returns empty for no trees", {
  pkg <- withr::local_tempdir()
  ts <- ts_parse(pkg)
  expect_identical(ts_all_referenced_functions(ts), character())
})

test_that("ts_all_referenced_functions extracts call names", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines("f <- function() { mean(1:10) }", file.path(pkg, "R", "code.R"))
  ts <- ts_parse(pkg)
  calls <- ts_all_referenced_functions(ts)
  expect_true("mean" %in% calls)
})

test_that("ts_all_referenced_functions finds identifiers in bodies", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "inner <- function(i) i + 1",
    "outer <- function(x) lapply(x, inner)"
  ), file.path(pkg, "R", "code.R"))
  ts <- ts_parse(pkg)
  refs <- ts_all_referenced_functions(ts)
  expect_true("inner" %in% refs)
  expect_true("lapply" %in% refs)
  expect_false("outer" %in% refs)
})

test_that("ts_all_referenced_functions finds RHS assignments", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "MY_LIST <- list()",
    "helper <- function() 42",
    "MY_LIST$fn <- helper"
  ), file.path(pkg, "R", "code.R"))
  ts <- ts_parse(pkg)
  refs <- ts_all_referenced_functions(ts)
  expect_true("helper" %in% refs)
})

# -- complexity_unused_internal ------------------------------------------------

test_that("complexity_unused_internal fails on dead code", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: deadcode", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "DESCRIPTION"))
  writeLines("export(public_fn)", file.path(pkg, "NAMESPACE"))

  writeLines(c(
    "public_fn <- function() helper()",
    "helper <- function() 42",
    "dead_fn <- function() 99"
  ), file.path(pkg, "R", "code.R"))

  gp_res <- gp(pkg, checks = "complexity_unused_internal")
  res <- results(gp_res)
  expect_false(res$passed[res$check == "complexity_unused_internal"])

  pos <- failed_positions(gp_res)$complexity_unused_internal
  names <- vapply(pos, `[[`, "", "line")
  expect_true("dead_fn" %in% names)
  expect_false("helper" %in% names)
})

test_that("complexity_unused_internal passes when all used", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: allused", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "DESCRIPTION"))
  writeLines("export(public_fn)", file.path(pkg, "NAMESPACE"))

  writeLines(c(
    "public_fn <- function() helper()",
    "helper <- function() 42"
  ), file.path(pkg, "R", "code.R"))

  gp_res <- gp(pkg, checks = "complexity_unused_internal")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "complexity_unused_internal"])
})

test_that("complexity_unused_internal passes with no functions", {
  gp_res <- gp("good", checks = "complexity_unused_internal")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "complexity_unused_internal"])
})

test_that("complexity_unused_internal passes when all exported", {
  pkg <- withr::local_tempdir()
  pkg <- file.path(pkg, "allexport")
  dir.create(pkg)
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: allexport", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "DESCRIPTION"))
  writeLines("export(my_fn)", file.path(pkg, "NAMESPACE"))

  writeLines("my_fn <- function() 1", file.path(pkg, "R", "code.R"))

  gp_res <- gp(pkg, checks = "complexity_unused_internal")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "complexity_unused_internal"])
})

test_that("unused_internal detects functions passed as arguments", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: hoftest", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "DESCRIPTION"))
  writeLines("export(public_fn)", file.path(pkg, "NAMESPACE"))

  writeLines(c(
    "public_fn <- function(x) lapply(x, helper)",
    "helper <- function(i) i + 1",
    "dead_fn <- function() 99"
  ), file.path(pkg, "R", "code.R"))

  gp_res <- gp(pkg, checks = "complexity_unused_internal")
  pos <- failed_positions(gp_res)$complexity_unused_internal
  names <- vapply(pos, `[[`, "", "line")
  expect_false("helper" %in% names)
  expect_true("dead_fn" %in% names)
})

test_that("unused_internal detects functions assigned to lists", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: listtest", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "DESCRIPTION"))
  writeLines("export(public_fn)", file.path(pkg, "NAMESPACE"))

  writeLines(c(
    "MY_LIST <- list()",
    "public_fn <- function() MY_LIST[[1]](42)",
    "prep_fn <- function(x) x * 2",
    "MY_LIST$prep <- prep_fn"
  ), file.path(pkg, "R", "code.R"))

  gp_res <- gp(pkg, checks = "complexity_unused_internal")
  pos <- failed_positions(gp_res)$complexity_unused_internal
  if (is.null(pos)) pos <- list()
  names <- vapply(pos, `[[`, "", "line")
  expect_false("prep_fn" %in% names)
})

#' @include lists.R

#' @noRd
ts_nodes_equal <- function(a, b) {
  identical(treesitter::node_start_point(a),
            treesitter::node_start_point(b)) &&
    identical(treesitter::node_end_point(a),
              treesitter::node_end_point(b))
}

ts_inside_nested_function <- function(node, fn_body) {
  parent <- treesitter::node_parent(node)
  while (!is.null(parent) && !ts_nodes_equal(parent, fn_body)) {
    if (treesitter::node_type(parent) == "function_definition") return(TRUE)
    parent <- treesitter::node_parent(parent)
  }
  FALSE
}

ts_body_has_call <- function(fn_node, call_query) {
  body <- treesitter::node_child_by_field_name(fn_node, "body")
  caps <- treesitter::query_captures(call_query, body)
  for (j in which(caps$name == "fn")) {
    if (!ts_inside_nested_function(caps$node[[j]], body)) return(TRUE)
  }
  FALSE
}

ASSIGNMENT_OPERATORS <- c("<-", "=", "<<-")

ts_file_functions <- function(root, file) {
  n_children <- treesitter::node_child_count(root)
  fns <- vector("list", n_children)
  k <- 0L
  for (i in seq_len(n_children)) {
    child <- treesitter::node_child(root, i)
    if (treesitter::node_type(child) != "binary_operator") next
    op <- treesitter::node_child_by_field_name(child, "operator")
    if (is.null(op) || !treesitter::node_text(op) %in% ASSIGNMENT_OPERATORS) next
    lhs <- treesitter::node_child_by_field_name(child, "lhs")
    rhs <- treesitter::node_child_by_field_name(child, "rhs")
    if (is.null(rhs)) next
    if (treesitter::node_type(rhs) != "function_definition") next
    if (treesitter::node_type(lhs) != "identifier") next
    k <- k + 1L
    fns[[k]] <- list(
      name = treesitter::node_text(lhs),
      file = file,
      line = treesitter::node_start_point(lhs)$row + 1L,
      fn_node = rhs
    )
  }
  fns[seq_len(k)]
}

ts_parse <- function(path, exclude_path = character(), encoding = "UTF-8") {
  rdir <- file.path(path, "R")
  if (!dir.exists(rdir)) {
    return(list(trees = list(), functions = list()))
  }

  lang <- treesitter.r::language()
  p <- treesitter::parser(lang)
  rfiles <- list.files(
    rdir, pattern = default_r_file_pattern(), full.names = TRUE
  )
  rfiles <- filter_excluded_paths(rfiles, path, exclude_path)

  trees <- vector("list", length(rfiles))
  names(trees) <- rfiles
  fn_lists <- vector("list", length(rfiles))
  for (i in seq_along(rfiles)) {
    f <- rfiles[[i]]
    code <- tryCatch(
      read_source_file(f, encoding),
      error = function(e) {
        cli::cli_warn("Could not read {.file {f}}: {conditionMessage(e)}")
        NULL
      }
    )
    if (is.null(code)) next
    tree <- treesitter::parser_parse(p, code)
    root <- treesitter::tree_root_node(tree)
    trees[[i]] <- list(tree = tree, root = root)
    fn_lists[[i]] <- ts_file_functions(root, f)
  }

  functions <- unlist(fn_lists, recursive = FALSE)
  list(trees = trees, functions = functions, language = lang)
}

S4_CALL_NAMES <- c(
  "setMethod", "setGeneric", "setClass",
  "setReplaceMethod", "setValidity", "setIs"
)

ts_s4_call_ranges <- function(ts) {
  if (length(ts$trees) == 0) return(list())

  s4_query <- treesitter::query(ts$language, paste0(
    "(call function: (identifier) @fn (#match? @fn \"^(",
    paste(S4_CALL_NAMES, collapse = "|"),
    ")$\"))"
  ))

  ranges <- unlist(lapply(names(ts$trees), function(file) {
    entry <- ts$trees[[file]]
    if (is.null(entry)) return(NULL)
    caps <- treesitter::query_captures(s4_query, entry$root)
    idxs <- which(caps$name == "fn")
    if (length(idxs) == 0) return(NULL)

    lapply(idxs, function(j) {
      call_node <- treesitter::node_parent(caps$node[[j]])
      list(
        file = basename(file),
        start = treesitter::node_start_point(call_node)$row + 1L,
        end = treesitter::node_end_point(call_node)$row + 1L
      )
    })
  }), recursive = FALSE)

  if (is.null(ranges)) list() else ranges
}

ts_get <- function(state) {
  if (is.null(state$.cache$treesitter)) {
    encoding <- get_desc_encoding(state)
    state$.cache$treesitter <- ts_parse(
      state$path, state$exclude_path %||% character(), encoding = encoding
    )
  }
  state$.cache$treesitter
}

filter_s4_assignment_false_positives <- function(state, result) {
  if (isTRUE(result$status) || is.na(result$status)) return(result)

  ts <- ts_get(state)
  s4_ranges <- ts_s4_call_ranges(ts)
  if (length(s4_ranges) == 0) return(result)

  result$positions <- Filter(function(pos) {
    f <- basename(pos$filename)
    ln <- pos$line_number
    !any(vapply(s4_ranges, function(r) {
      f == r$file && ln >= r$start && ln <= r$end
    }, logical(1)))
  }, result$positions)

  result$status <- length(result$positions) == 0
  result
}

# -- ts_parse -----------------------------------------------------------------

test_that("ts_parse extracts functions from a package", {
  ts <- ts_parse("good")
  expect_true(is.list(ts))
  expect_true("trees" %in% names(ts))
  expect_true("functions" %in% names(ts))
  expect_true("language" %in% names(ts))

  expect_gt(length(ts$functions), 0)
  fn <- ts$functions[[1]]
  expect_true(all(c("name", "file", "line", "fn_node") %in% names(fn)))
  expect_identical(fn$name, "func1")
})

test_that("ts_parse returns empty lists when no R directory", {
  pkg <- withr::local_tempdir()
  ts <- ts_parse(pkg)
  expect_identical(ts$trees, list())
  expect_identical(ts$functions, list())
})

# -- ts_file_functions --------------------------------------------------------

test_that("ts_file_functions finds top-level function definitions", {
  code <- "foo <- function(x) x + 1\nbar <- function(y) y * 2\nz <- 42"
  lang <- treesitter.r::language()
  p <- treesitter::parser(lang)
  tree <- treesitter::parser_parse(p, code)
  root <- treesitter::tree_root_node(tree)

  fns <- ts_file_functions(root, "test.R")
  expect_length(fns, 2)
  expect_identical(fns[[1]]$name, "foo")
  expect_identical(fns[[2]]$name, "bar")
  expect_identical(fns[[1]]$file, "test.R")
  expect_identical(fns[[1]]$line, 1)
  expect_identical(fns[[2]]$line, 2)
})

test_that("ts_file_functions skips non-function assignments", {
  code <- "x <- 42\ny <- 'hello'"
  lang <- treesitter.r::language()
  p <- treesitter::parser(lang)
  tree <- treesitter::parser_parse(p, code)
  root <- treesitter::tree_root_node(tree)

  fns <- ts_file_functions(root, "test.R")
  expect_length(fns, 0)
})

test_that("ts_file_functions skips non-assignment operators", {
  code <- "x + function() 5\nreal_fn <- function(y) y"
  lang <- treesitter.r::language()
  p <- treesitter::parser(lang)
  tree <- treesitter::parser_parse(p, code)
  root <- treesitter::tree_root_node(tree)

  fns <- ts_file_functions(root, "test.R")
  expect_length(fns, 1)
  expect_identical(fns[[1]]$name, "real_fn")
})

test_that("ts_file_functions accepts <-, =, and <<- assignments", {
  code <- "a <- function() 1\nb = function() 2\nc <<- function() 3"
  lang <- treesitter.r::language()
  p <- treesitter::parser(lang)
  tree <- treesitter::parser_parse(p, code)
  root <- treesitter::tree_root_node(tree)

  fns <- ts_file_functions(root, "test.R")
  expect_length(fns, 3)
  expect_identical(
    vapply(fns, `[[`, "", "name"),
    c("a", "b", "c")
  )
})

test_that("ts_file_functions skips non-identifier LHS", {
  code <- "env$helper <- function() 1\nreal_fn <- function(x) x"
  lang <- treesitter.r::language()
  p <- treesitter::parser(lang)
  tree <- treesitter::parser_parse(p, code)
  root <- treesitter::tree_root_node(tree)

  fns <- ts_file_functions(root, "test.R")
  expect_length(fns, 1)
  expect_identical(fns[[1]]$name, "real_fn")
})

test_that("ts_parse finds functions in .S files", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines("myfun <- function(x) x + 1", file.path(pkg, "R", "legacy.S"))

  ts <- ts_parse(pkg)
  names <- vapply(ts$functions, `[[`, "", "name")
  expect_true("myfun" %in% names)
})

test_that("ts_parse honours encoding argument", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  f <- file.path(pkg, "R", "latin1.R")
  writeBin(as.raw(c(
    charToRaw("naive <- function() 'r"), 0xe9,
    charToRaw("sum"), 0xe9, charToRaw("'\n")
  )), f)

  ts <- ts_parse(pkg, encoding = "latin1")
  names <- vapply(ts$functions, `[[`, "", "name")
  expect_true("naive" %in% names)
})

test_that("ts_parse skips unreadable files", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines("good_fn <- function() 1", file.path(pkg, "R", "good.R"))
  bad_path <- file.path(pkg, "R", "bad.R")
  writeLines("x <- 1", bad_path)
  Sys.chmod(bad_path, "000")
  on.exit(Sys.chmod(bad_path, "644"), add = TRUE)

  ts <- suppressWarnings(ts_parse(pkg))
  names <- vapply(ts$functions, `[[`, "", "name")
  expect_true("good_fn" %in% names)
})

# -- ts_body_has_call ---------------------------------------------------------

test_that("ts_body_has_call detects a call in function body", {
  code <- "my_fn <- function(x) { invisible(x) }"
  lang <- treesitter.r::language()
  p <- treesitter::parser(lang)
  tree <- treesitter::parser_parse(p, code)
  root <- treesitter::tree_root_node(tree)

  fns <- ts_file_functions(root, "test.R")
  invisible_q <- treesitter::query(lang,
    "(call function: (identifier) @fn (#eq? @fn \"invisible\"))"
  )
  expect_true(ts_body_has_call(fns[[1]]$fn_node, invisible_q))
})

test_that("ts_body_has_call returns FALSE when call absent", {
  code <- "my_fn <- function(x) { x + 1 }"
  lang <- treesitter.r::language()
  p <- treesitter::parser(lang)
  tree <- treesitter::parser_parse(p, code)
  root <- treesitter::tree_root_node(tree)

  fns <- ts_file_functions(root, "test.R")
  invisible_q <- treesitter::query(lang,
    "(call function: (identifier) @fn (#eq? @fn \"invisible\"))"
  )
  expect_false(ts_body_has_call(fns[[1]]$fn_node, invisible_q))
})

# -- ts_inside_nested_function ------------------------------------------------

test_that("ts_body_has_call ignores calls inside nested functions", {
  code <- "outer <- function(x) { inner <- function(y) invisible(y); x }"
  lang <- treesitter.r::language()
  p <- treesitter::parser(lang)
  tree <- treesitter::parser_parse(p, code)
  root <- treesitter::tree_root_node(tree)

  fns <- ts_file_functions(root, "test.R")
  invisible_q <- treesitter::query(lang,
    "(call function: (identifier) @fn (#eq? @fn \"invisible\"))"
  )
  expect_false(ts_body_has_call(fns[[1]]$fn_node, invisible_q))
})

# -- ts_get (lazy cache) -----------------------------------------------------

make_fake_desc <- function(encoding = "UTF-8") {
  list(get_field = function(field, default = NULL) encoding)
}

test_that("ts_get caches treesitter parse result", {
  state <- list(path = "good", description = make_fake_desc(), .cache = list())
  ts1 <- ts_get(state)
  expect_gt(length(ts1$functions), 0)

  state$.cache$treesitter <- ts1
  ts2 <- ts_get(state)
  expect_identical(ts1, ts2)
})

test_that("ts_get reads Encoding from description", {
  state <- list(
    path = "good", description = make_fake_desc("latin1"), .cache = list()
  )
  ts <- ts_get(state)
  expect_gt(length(ts$functions), 0)
})

test_that("prep_description sets Encoding default when missing", {
  pkg <- withr::local_tempdir()
  writeLines(c(
    "Package: fakepkg",
    "Title: Fake",
    "Version: 0.0.1",
    "Description: fake.",
    "License: MIT + file LICENSE",
    "Authors@R: person('A', 'B', email = 'a@b.c', role = c('aut','cre'))"
  ), file.path(pkg, "DESCRIPTION"))

  state <- list(path = pkg, .cache = list())
  state <- PREPS$description(state, path = pkg, quiet = TRUE)
  expect_identical(state$description$get_field("Encoding"), "UTF-8")
})

test_that("prep_description preserves existing Encoding", {
  pkg <- withr::local_tempdir()
  writeLines(c(
    "Package: fakepkg",
    "Title: Fake",
    "Version: 0.0.1",
    "Description: fake.",
    "License: MIT + file LICENSE",
    "Encoding: latin1",
    "Authors@R: person('A', 'B', email = 'a@b.c', role = c('aut','cre'))"
  ), file.path(pkg, "DESCRIPTION"))

  state <- list(path = pkg, .cache = list())
  state <- PREPS$description(state, path = pkg, quiet = TRUE)
  expect_identical(state$description$get_field("Encoding"), "latin1")
})

# -- ts_s4_call_ranges -------------------------------------------------------

test_that("ts_s4_call_ranges finds setMethod call ranges", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    'setMethod("show",',
    '  signature = (object = "MyClass"),',
    '  definition = function(object) cat("hi")',
    ')'
  ), file.path(pkg, "R", "methods.R"))

  ts <- ts_parse(pkg)
  ranges <- ts_s4_call_ranges(ts)
  expect_length(ranges, 1)
  expect_identical(ranges[[1]]$start, 1)
  expect_identical(ranges[[1]]$end, 4)
})

test_that("ts_s4_call_ranges returns empty for no S4 calls", {
  ts <- ts_parse("good")
  ranges <- ts_s4_call_ranges(ts)
  expect_length(ranges, 0)
})

test_that("ts_s4_call_ranges returns empty for no trees", {
  pkg <- withr::local_tempdir()
  ts <- ts_parse(pkg)
  ranges <- ts_s4_call_ranges(ts)
  expect_identical(ranges, list())
})

# -- full workflow (like chk_tidyverse.R) ------------------------------------

test_that("treesitter workflow finds missing() calls in functions", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(
    "check_arg <- function(x) { if (missing(x)) stop('need x') }",
    file.path(pkg, "R", "check.R")
  )

  ts <- ts_parse(pkg)
  expect_length(ts$functions, 1)
  expect_identical(ts$functions[[1]]$name, "check_arg")

  missing_q <- treesitter::query(ts$language,
    "(call function: (identifier) @fn (#eq? @fn \"missing\"))"
  )
  expect_true(ts_body_has_call(ts$functions[[1]]$fn_node, missing_q))
})

#' @include lists.R

# Sentinel preps: checks that perform their own work (via ts_get() or
# direct filesystem calls) still need a registered prep name so that
# required_preps() can resolve the check's `preps` field.

PREPS$code_structure <- function(state, path = state$path, quiet) {
  state
}

PREPS$package_structure <- function(state, path = state$path, quiet) {
  state
}

#' @title Defining custom preparations and checks
#' @family API
#' @name customization
#' @rdname customization
NULL


#' @param name Name of the preparation function.
#' @param func A function that takes two arguments:
#'   The \code{path} to the root directory of the package, and a logical
#'   argument: \code{quiet}. If \code{quiet} is true, the preparation function
#'   may print out diagnostic messages. The output of this function will be
#'   saved as the " \code{name}" entry of \code{state}, i.e. of the input for
#'   the  \code{check}-functions (see example).
#'   
#' @return For \code{make_prep}: a preparation function. For
#'   \code{make_check}: a check object of class \code{"check"}.
#' @describeIn customization Create a preparation function
#' @export
#' @examples 
#' # make a preparation function
#' url_prep <- make_prep(
#'   name = "desc", 
#'   func = function(path, quiet) desc::description$new(path)
#' )
#' # and the corresponding check function
#' url_chk <- make_check(
#'   description = "URL field in DESCRIPTION",
#'   tags = character(),
#'   preps = "desc",
#'   gp = "have a URL field in DESCRIPTION",
#'   check = function(state) state$desc$has_fields("URL")
#' )
#' # use together in gp():
#' # (note that you have to list the name of your custom check in
#' # the checks-argument as well....)
#' bad1 <- system.file("bad1", package = "goodpractice")
#' res <- gp(bad1, checks = c("url", "no_description_depends"),
#'           extra_preps = list("desc" = url_prep),
#'           extra_checks = list("url" = url_chk))

make_prep <- function(name, func) {
  force(name)
  force(func)
  function(state, path = state$path, quiet) {
    state[[name]] <- func(path, quiet)
    state
  }
}

#' @param description A description of the check.
#' @param check A function that takes the \code{state} as an argument.
#' @param gp A short description of what is good practice.
#' @param ... Further arguments. Most important: A \code{preps} argument that 
#'  contains the names of all the preparation functions required for the check.
#' @describeIn customization Create a check function
#' @export

make_check <- function(description, check, gp, ...) {
  structure(
    list(description = description, check = check, gp = gp, ...),
    class = "check"
  )
}

prepare_preps <- function(preps, extra_preps) {
  utils::modifyList(preps, c(list(), extra_preps))
}

prepare_checks <- function(checks, extra_checks) {
  utils::modifyList(checks, c(list(), extra_checks))
}

bad1 <- system.file("bad1", package = "goodpractice")

test_that("extra check", {
  simple_truefalse_not_tf <- make_check(

    description = "TRUE and FALSE is used, not T and F",
    gp = "avoid 'T' and 'F', use 'TRUE' and 'FALSE' instead.",
    check = function(state) {
      length(tools::checkTnF(dir = state$path)) == 0
    }
  )

  res <- gp(bad1, checks = "simple_tf",
            extra_checks = list(simple_tf = simple_truefalse_not_tf))

  expect_identical(checks(res), "simple_tf")
  expect_false(results(res)$passed)

})

test_that("extra prep and check pair", {
  url_prep <- function(path, quiet) {
    desc::description$new(path)
  }
  url_chk <- make_check(
    description = "URL field in DESCRIPTION",
    tags = character(),
    preps = "desc",
    gp = "have a URL field in DESCRIPTION",
    check = function(state) state$desc$has_fields("URL")
  )

  res <- gp(bad1, checks = c("no_description_depends", "url"),
            extra_preps = list(desc = make_prep("desc", url_prep)),
            extra_checks = list(url = url_chk))

  expect_identical(checks(res), c("no_description_depends", "url"))
  expect_identical(results(res)$passed, rep(FALSE, 2))

})

#' @include lists.R
#' @include chk_vignette.R
#' @include prep_utils.R

PREPS$vignette <- function(state, path = state$path, quiet) {
  run_prep_step(state, "vignette", function(path) {
    vfiles <- vignette_files(path)
    pd_list <- list()
    for (f in vfiles) {
      pd <- vignette_parse_data(f)
      if (!is.null(pd)) {
        pd_list[[f]] <- list(
          parse_data = pd,
          lines = readLines(f, warn = FALSE)
        )
      }
    }
    pd_list
  }, path = path, silent = quiet)
}

#' @include lists.R prep_utils.R

PREPS$namespace <- function(state, path = state$path, quiet) {
  path <- normalizePath(path)
  run_prep_step(state, "namespace", function(path) {
    parseNamespaceFile(basename(path), file.path(path, ".."))
  }, path = path, silent = quiet)
}

#' @importFrom rstudioapi callFun

#' @noRd
rstudio_source_markers <- function(gp) {

  markers <- get_markers(gp)

  if (length(markers) == 0) return()

  callFun(
    "sourceMarkers",
    name = "goodpractice",
    markers = markers,
    basePath = NULL,
    autoSelect = "first"
  )
}

get_markers <- function(gp) {
  m <- lapply(names(gp$checks), get_marker, gp = gp)
  m <- drop_nulls(m)
  unlist(m, recursive = FALSE)
}

#' Get a marker from the positions of a check
#'
#' @param gp `gp()` output
#' @param check name of the check to extract
#'
#' @keywords internal
#' @importFrom utils head

#' @noRd
get_marker <- function(gp, check) {

  chk <- CHECKS[[check]]
  res <- gp$checks[[check]]

  ## Check passed
  if (check_passed(gp$checks[[check]], na_as_passed = TRUE)) return(NULL)

  ## Check failed, but alas, no positions
  if (! "positions" %in% names(gp$checks[[check]])) return(NULL)

  my_message <- if (is.function(chk$gp)) chk$gp(gp) else chk$gp

  type <- head(intersect(
    chk$tags,
    c("error", "warning", "info", "style", "usage")
  ), 1)

  lapply(
    res$positions,
    function(p) {
      list(
        type = type,
        file = normalizePath(file.path(gp$path, p$filename)),
        line = if (is.na(p$line_number)) 1L else p$line_number,
        column = if (is.na(p$column_number)) 1L else p$column_number,
        message = my_message
      )
    }
  )
}

bad1 <- system.file("bad1", package = "goodpractice")
gp_obj <- gp(bad1, checks = "r_file_extension")

test_that("get_marker returns NULL for passing checks", {
  passing <- gp("good", checks = "description_bugreports")
  result <- get_marker(passing, "description_bugreports")
  expect_null(result)
})

test_that("get_marker returns NULL when no positions field", {
  state <- gp_obj
  state$checks$r_file_extension <- FALSE
  result <- get_marker(state, "r_file_extension")
  expect_null(result)
})

test_that("get_marker returns position list for failing check", {
  result <- get_marker(gp_obj, "r_file_extension")
  expect_type(result, "list")
  expect_gt(length(result), 0)
  expect_true(all(
    c("type", "file", "line", "column", "message") %in%
      names(result[[1]])
  ))
})

test_that("get_marker handles NA line_number and column_number", {
  state <- gp_obj
  state$checks$r_file_extension$positions[[1]]$line_number <- NA
  state$checks$r_file_extension$positions[[1]]$column_number <- NA
  result <- get_marker(state, "r_file_extension")
  expect_identical(result[[1]]$line, 1L)
  expect_identical(result[[1]]$column, 1L)
})

test_that("get_markers collects markers from all failing checks", {
  markers <- get_markers(gp_obj)
  expect_type(markers, "list")
  expect_gt(length(markers), 0)
})

test_that("get_markers returns empty for all-pass results", {
  passing <- gp("good", checks = "description_bugreports")
  markers <- get_markers(passing)
  expect_length(markers, 0)
})

test_that("rstudio_source_markers calls callFun", {
  called <- FALSE
  local_mocked_bindings(
    callFun = function(...) {
      called <<- TRUE
    }
  )
  rstudio_source_markers(gp_obj)
  expect_true(called)
})

test_that("rstudio_source_markers returns early with no markers", {
  passing <- gp("good", checks = "description_bugreports")
  called <- FALSE
  local_mocked_bindings(
    callFun = function(...) {
      called <<- TRUE
    }
  )
  rstudio_source_markers(passing)
  expect_false(called)
})

print.simple <- function(x, ...) {
  cat("simple object\n")
  invisible(x)
}

print.nested <- function(x, ...) {
  if (length(x) > 0) {
    cat(format(x), sep = "\n")
    invisible(x)
  } else {
    cat("empty\n")
    invisible(x)
  }
}

print.assigned_with_equals = function(x, ...) {
  cat("equals object\n")
  invisible(x)
}

print.with_helper <- function(x, ...) {
  fmt <- function(val) paste0("[", val, "]")
  cat(fmt(x), "\n")
  invisible(x)
}

#' @noRd
`%||%` <- function(l, r) { if (is.null(l)) r else l }

#' @noRd
check_result <- function(status = NULL, positions = NULL, ...) {
  list(
    status = status %||% NA,
    positions = positions %||% list(),
    ...
  )
}

na_result <- function() {
  check_result()
}

#' Construct a source position object
#'
#' General constructor for the position objects returned by check
#' functions. Each position identifies a location in the source code
#' where a check found an issue. Use this instead of building
#' `list(filename, line_number, ...)` by hand.
#'
#' @param filename Path relative to the package root, e.g. `"R/foo.R"`.
#' @param line_number Line number (1-based), or `NA_integer_` if unknown.
#' @param column_number Column number, or `NA_integer_` if unknown.
#' @param ranges List of start/end pairs for multi-line issues.
#' @param line Short display string identifying the problem to the user.
#' @return A named list with the five position fields.
#' @keywords internal
#' @noRd
check_position <- function(filename, line_number = NA_integer_,
                           column_number = NA_integer_,
                           ranges = list(), line = "") {
  list(
    filename = filename,
    line_number = line_number,
    column_number = column_number,
    ranges = ranges,
    line = line
  )
}

ns_s3_method_names <- function(ns) {
  s3m <- ns$S3methods
  if (is.null(s3m) || nrow(s3m) == 0) return(character())
  paste0(s3m[, 1], ".", s3m[, 2])
}

#' Default pattern for R files
#' @return Regular expression.
#' @keywords internal
#' @noRd
default_r_file_pattern <- function() {
  "\\.[RrSs]$"
}

filter_excluded_paths <- function(files, pkg_path, exclude_path) {
  if (length(exclude_path) == 0) return(files)
  abs_excluded <- normalizePath(file.path(pkg_path, exclude_path),
                                mustWork = FALSE)
  abs_files <- normalizePath(files, mustWork = FALSE)
  files[!abs_files %in% abs_excluded]
}

read_source_file <- function(path, encoding = "UTF-8") {
  con <- file(path, encoding = encoding)
  on.exit(close(con))
  paste(readLines(con, warn = FALSE), collapse = "\n")
}

trim_ws <- function(x) {
  sub("\\s+$", "", sub("^\\s+", "", x))
}

loaded_pkg_version <- function(pkg) {
  asNamespace(pkg)$`.__NAMESPACE__.`$spec[["version"]]
}

drop_nulls <- function(l) {
  l[ ! vapply(l, is.null, TRUE) ]
}

get_desc_encoding <- function(state) {
  if (inherits(state$description, "description")) {
    state$description$get_field("Encoding")
  } else {
    "UTF-8"
  }
}

has_internet <- function() {
  curl::has_internet()
}

# Used only to parse code chunks in vignettes;
# all other code is parsed with treesitter.R fns
safe_parse <- function(file = NULL, text = NULL, keep_source = TRUE,
                       encoding = "UTF-8") {
  args <- if (!is.null(file)) {
    list(file = file, keep.source = keep_source, encoding = encoding)
  } else {
    list(text = text, keep.source = keep_source)
  }

  tryCatch(
    do.call(base::parse, args),
    error = function(e) {
      if (keep_source) {
        args$keep.source <- FALSE
        tryCatch(do.call(base::parse, args), error = function(e) NULL)
      } else {
        NULL
      }
    }
  )
}

test_that("%||% returns left when non-NULL", {
  expect_identical(1 %||% 2, 1)
  expect_identical("a" %||% "b", "a")
})

test_that("%||% returns right when left is NULL", {
  expect_identical(NULL %||% 2, 2)
  expect_identical(NULL %||% "fallback", "fallback")
})

test_that("trim_ws removes leading and trailing whitespace", {
  expect_identical(trim_ws("  hello  "), "hello")
  expect_identical(trim_ws("no_ws"), "no_ws")
  expect_identical(trim_ws("\thello\n"), "hello")
})

test_that("loaded_pkg_version returns a version string", {
  v <- loaded_pkg_version("goodpractice")
  expect_type(v, "character")
  expect_true(nzchar(v))
})

test_that("drop_nulls removes NULL elements", {
  expect_identical(drop_nulls(list(1, NULL, 3)), list(1, 3))
  expect_identical(drop_nulls(list(NULL, NULL)), list())
  expect_identical(drop_nulls(list("a", "b")), list("a", "b"))
})

test_that("excluded_paths returns empty by default", {
  withr::local_options(goodpractice.exclude_path = NULL)
  withr::local_envvar(GP_EXCLUDE_PATH = "")
  expect_identical(excluded_paths(), character(0))
})

test_that("excluded_paths reads from option", {
  withr::local_options(goodpractice.exclude_path = c("R/a.R", "R/b.R"))
  result <- excluded_paths()
  expect_identical(result, c("R/a.R", "R/b.R"))
})

test_that("excluded_paths reads from envvar", {
  withr::local_options(goodpractice.exclude_path = NULL)
  withr::local_envvar(GP_EXCLUDE_PATH = "R/a.R,R/b.R")
  result <- excluded_paths()
  expect_identical(result, c("R/a.R", "R/b.R"))
})

test_that("excluded_paths option takes precedence over envvar", {
  withr::local_options(goodpractice.exclude_path = "R/from_option.R")
  withr::local_envvar(GP_EXCLUDE_PATH = "R/from_envvar.R")
  expect_identical(excluded_paths(), "R/from_option.R")
})

test_that("filter_excluded_paths removes matching files", {
  tmp <- withr::local_tempdir()
  dir.create(file.path(tmp, "R"))
  file.create(file.path(tmp, "R", "a.R"))
  file.create(file.path(tmp, "R", "b.R"))
  file.create(file.path(tmp, "R", "c.R"))

  files <- file.path(tmp, "R", c("a.R", "b.R", "c.R"))
  result <- filter_excluded_paths(files, tmp, "R/a.R")
  expect_identical(basename(result), c("b.R", "c.R"))
})

test_that("filter_excluded_paths with empty exclude returns all files", {
  files <- c("/tmp/R/a.R", "/tmp/R/b.R")
  result <- filter_excluded_paths(files, "/tmp", character())
  expect_identical(result, files)
})

test_that("exclude_path filters roxygen2 blocks", {
  withr::local_options(goodpractice.exclude_path = "R/functions.R")
  g <- gp("good", checks = "roxygen2_has_export_or_nord")
  expect_identical(g$exclude_path, "R/functions.R")
})

test_that("gp passes exclude_path to state", {
  withr::local_options(goodpractice.exclude_path = "R/functions.R")
  pkg <- system.file("bad1", package = "goodpractice")
  g <- gp(pkg, checks = "print_return_invisible")
  res <- results(g)
  expect_true("print_return_invisible" %in% res$check)
  expect_identical(g$exclude_path, "R/functions.R")
})

#' @include treesitter.R

#' @noRd
get_lintr_position <- function(linter) {
  linter[c("filename", "line_number", "column_number", "ranges", "line")]
}

get_lintr_state <- function(state, linter) {
  if(inherits(state$lintr, "try-error")) {
    return(na_result())
  }

  linters <- vapply(state$lintr, "[[", "", "linter")
  check_result(! linter %in% linters, lapply(
      state$lintr[linters == linter],
      get_lintr_position)
  )
}

make_lintr_check <- function(linter_name, description, gp, tags = "lintr") {
  make_check(
    description = description,
    tags = c("warning", tags),
    preps = "lintr",
    gp = gp,
    check = function(state) get_lintr_state(state, linter_name)
  )
}

#' @include lists.R

CHECKS$lintr_assignment_linter <- make_check(

  description = "'<-' and not '=' is used for assignment",
  tags = c("style", "lintr"),
  preps = "lintr",

  gp = "use {.code <-} for assignment instead of {.code =}.
        {.code <-} is the standard, and R users and developers are
        used it and it is easier to read your code for them if
        you use {.code <-}.",

  check = function(state) {
    result <- get_lintr_state(state, "assignment_linter")
    filter_s4_assignment_false_positives(state, result)
  }
)

CHECKS$lintr_line_length_linter <- make_check(

  description = "Code lines are short",
  tags = c("style", "lintr"),
  preps = "lintr",

  gp = "avoid long code lines, it is bad for
        readability. Also, many people prefer editor windows
        that are about 80 characters wide. Try making your lines
        shorter than 80 characters",

  check = function(state) {
    get_lintr_state(state, "line_length_linter")
  }
)

CHECKS$lintr_semicolon_linter <- make_check(

  description = "No trailing semicolons",
  tags = c("style", "lintr"),
  preps = "lintr",

  gp = "omit trailing semicolons from code lines.
        They are not needed and most R coding standards
        forbid them",

  check = function(state) {
    get_lintr_state(state, "semicolon_linter")
  }
)

CHECKS$lintr_attach_detach_linter <- make_check(

  description = "Avoid attach and detach",
  tags = c("warning", "lintr"),
  preps = "lintr",

  gp = "avoid the {.fn attach} and {.fn detach} functions,
        they are fragile and code that uses them will
        probably break sooner than later.",

  check = function(state) {
    get_lintr_state(state, "attach_detach_linter")
  }
)

CHECKS$lintr_setwd_linter <- make_check(

  description = "Avoid setwd in R packages",
  tags = c("warning", "lintr"),
  preps = "lintr",

  gp = "avoid calling {.fn setwd}, it changes the global environment.
        If you need it, consider using {.fn on.exit} to restore the
        working directory.",

  check = function(state) {
    get_lintr_state(state, "setwd_linter")
  }
)

CHECKS$lintr_sapply_linter <- make_check(

  description = "Avoid sapply",
  tags = c("warning", "lintr"),
  preps = "lintr",

  gp = "avoid {.fn sapply}, it is not type safe.
        It might return a vector, or a list, depending on the
        input data. Consider using {.fn vapply} instead.",

  check = function(state) {
    get_lintr_state(state, "sapply_linter")
  }
)

CHECKS$lintr_library_require_linter <- make_check(

  description = "Avoid library and require in packages",
  tags = c("warning", "lintr"),
  preps = "lintr",

  gp = "avoid the {.fn library} and {.fn require} functions,
        they change the global search path.
        If you need to use other packages, import them.
        If you need to load them explicitly, then consider
        {.fn loadNamespace} instead, or as a last resort, declare
        them as {.field Depends} dependencies.",

  check = function(state) {
    if(inherits(state$lintr, "try-error")) {
      return(na_result())
    }

    res <- get_lintr_state(state, "library_require_linter")

    ## library() and require() are OK in tests and vignettes
    res$positions <- Filter(
      f = function(x) grepl("^R[/(\\)]", x$filename),
      res$positions
    )
    res$status <- length(res$positions) == 0
    res
  }
)

CHECKS$installed_packages_linter <- make_check(

  description = "Avoid calling installed.packages",
  tags = c("warning", "lintr"),
  preps = "lintr",

  gp = "avoid {.fn installed.packages}, it can be very slow to
        run on some machines, and will be rejected by CRAN.
        Use {.fn find.package} or {.fn system.file} instead.",

  check = function(state) {
    get_lintr_state(state, "installed_packages_linter")
  }
)

CHECKS$lintr_seq_linter <- make_check(

  description = "Avoid 1:length(...) and similar expressions",
  tags = c("warning", "lintr"),
  preps = "lintr",

  gp = "avoid {.code 1:length(...)}, {.code 1:nrow(...)}, {.code 1:ncol(...)},
        {.code 1:NROW(...)} and {.code 1:NCOL(...)} expressions. They are error
        prone and result {.code 1:0} if the expression on the right hand
        side is zero. Use {.fn seq_len} or {.fn seq_along} instead.",

  check = function(state) {
    get_lintr_state(state, "seq_linter")
  }
)

## -- correctness / performance ------------------------------------------------

CHECKS$lintr_any_duplicated_linter <- make_lintr_check(
  "any_duplicated_linter",
  "Use anyDuplicated() instead of any(duplicated())",
  "use {.fn anyDuplicated} instead of {.code any(duplicated(x))}. It is more
        efficient because it stops at the first duplicate."
)

CHECKS$lintr_any_is_na_linter <- make_lintr_check(
  "any_is_na_linter",
  "Use anyNA() instead of any(is.na())",
  "use {.fn anyNA} instead of {.code any(is.na(x))}. It is more efficient
        because it stops at the first {.code NA}."
)

CHECKS$lintr_class_equals_linter <- make_lintr_check(
  "class_equals_linter",
  "Use inherits() instead of class() ==",
  "use {.code inherits(x, 'class')} instead of {.code class(x) == 'class'}.
        Objects can have multiple classes, so {.code ==} may miss matches."
)

CHECKS$lintr_condition_message_linter <- make_lintr_check(
  "condition_message_linter",
  "Build condition messages correctly",
  "build error and warning messages with the condition function itself,
        not with {.fn paste} or {.fn sprintf}.
        Use {.code stop('msg')} not
        {.code stop(paste('msg'))}."
)

CHECKS$lintr_duplicate_argument_linter <- make_lintr_check(
  "duplicate_argument_linter",
  "No duplicate arguments in function calls",
  "avoid passing the same argument twice in a function call.
        Duplicate arguments are silently ignored, which is likely a bug."
)

CHECKS$lintr_equals_na_linter <- make_lintr_check(
  "equals_na_linter",
  "Use is.na() not == NA",
  "use {.fn is.na} instead of {.code x == NA}. Comparing with {.code ==} always
        returns {.code NA}, never {.code TRUE} or {.code FALSE}."
)

CHECKS$lintr_fixed_regex_linter <- make_lintr_check(
  "fixed_regex_linter",
  "Use fixed strings instead of regex where possible",
  "use {.code fixed = TRUE} in {.fn grep}/{.fn grepl}/
        {.fn sub}/{.fn gsub} when the pattern is a plain
        string, not a regular expression.
        It is faster and clearer."
)

CHECKS$lintr_for_loop_index_linter <- make_lintr_check(
  "for_loop_index_linter",
  "Do not overwrite existing variables with for loop index",
  "avoid using a variable name in {.fn for} that already exists in the
        surrounding scope. This overwrites the existing value."
)

CHECKS$lintr_length_test_linter <- make_lintr_check(
  "length_test_linter",
  "Use length() correctly in conditions",
  "avoid {.code length(x) == 0} or
        {.code length(x) > 0} in {.fn if} conditions
        when a type-stable alternative like {.fn is.null}
        or {.fn nzchar} exists."
)

CHECKS$lintr_matrix_apply_linter <- make_lintr_check(
  "matrix_apply_linter",
  "Use colSums/rowSums/colMeans/rowMeans instead of apply()",
  "use {.fn colSums}, {.fn rowSums},
        {.fn colMeans}, or {.fn rowMeans} instead of
        {.fn apply} with {.fn sum} or {.fn mean}.
        The specialized functions are much faster."
)

CHECKS$lintr_missing_argument_linter <- make_lintr_check(
  "missing_argument_linter",
  "No missing arguments in function calls",
  "avoid empty arguments in function calls like {.code f(x, , y)}.
        Missing arguments are a common source of bugs."
)

CHECKS$lintr_nrow_subset_linter <- make_lintr_check(
  "nrow_subset_linter",
  "Use nrow(subset) instead of nrow() after subsetting",
  "avoid patterns like {.code nrow(x[condition, ])}. Subset first, then
        count rows, or use {.code sum(condition)} directly."
)

CHECKS$lintr_redundant_equals_linter <- make_lintr_check(
  "redundant_equals_linter",
  "No redundant comparisons to TRUE or FALSE",
  "avoid {.code x == TRUE} or {.code x == FALSE}.
        Use {.code x} or {.code !x} directly.
        The comparison is redundant and less readable."
)

CHECKS$lintr_redundant_ifelse_linter <- make_lintr_check(
  "redundant_ifelse_linter",
  "No redundant ifelse() calls",
  "avoid {.code ifelse(condition, TRUE, FALSE)} or similar. The condition
        itself is already logical - use it directly."
)

CHECKS$lintr_regex_subset_linter <- make_lintr_check(
  "regex_subset_linter",
  "Use grep() with value = TRUE instead of subsetting",
  "use {.code grep(pattern, x, value = TRUE)} instead of
        {.code x[grep(pattern, x)]} or {.code x[grepl(pattern, x)]}."
)

CHECKS$lintr_sort_linter <- make_lintr_check(
  "sort_linter",
  "Use sort() or is.unsorted() instead of order()",
  "use {.code sort(x)} instead of
        {.code x[order(x)]} and {.fn is.unsorted} instead
        of checking {.code all(x == sort(x))}.
        The dedicated functions are clearer."
)

CHECKS$lintr_system_file_linter <- make_lintr_check(
  "system_file_linter",
  "Use system.file() correctly",
  "use {.code system.file('dir', 'file', package = 'pkg')} instead of
        {.code file.path(system.file(package = 'pkg'), 'dir', 'file')}. The
        single-call form correctly returns {.code ''}
        when the file does not exist."
)

CHECKS$lintr_terminal_close_linter <- make_lintr_check(
  "terminal_close_linter",
  "Do not close connections redundantly",
  "avoid closing connections at the end of a function when they
        will be closed automatically. Use {.code on.exit(close(con))} for
        connections that need explicit cleanup."
)

CHECKS$lintr_which_grepl_linter <- make_lintr_check(
  "which_grepl_linter",
  "Use grep() instead of which(grepl())",
  "use {.code grep(pattern, x)} instead of {.code which(grepl(pattern, x))}.
        {.fn grep} returns indices directly."
)

## -- readability / idiom ------------------------------------------------------

CHECKS$lintr_boolean_arithmetic_linter <- make_lintr_check(
  "boolean_arithmetic_linter",
  "Avoid arithmetic on logical vectors",
  "avoid {.code sum(x == value)} or similar arithmetic on logical vectors.
        Use {.fn sum} with a logical condition only when counting, and prefer
        dedicated functions like {.fn any} or {.fn all} for logical tests."
)

CHECKS$lintr_comparison_negation_linter <- make_lintr_check(
  "comparison_negation_linter",
  "Use natural comparison operators instead of negation",
  "use {.code x != y} instead of
        {.code !(x == y)}, and {.code x >= y} instead
        of {.code !(x < y)}.
        Direct comparisons are clearer."
)

CHECKS$lintr_consecutive_assertion_linter <- make_lintr_check(
  "consecutive_assertion_linter",
  "Combine consecutive stopifnot() calls",
  "combine consecutive {.fn stopifnot} calls into one. Multiple
        conditions can be passed to a single {.fn stopifnot}."
)

CHECKS$lintr_if_not_else_linter <- make_lintr_check(
  "if_not_else_linter",
  "Avoid if (!cond) { a } else { b }",
  "prefer {.code if (cond) {{b}} else {{a}}} over
        {.code if (!cond) {{a}} else {{b}}}.
        Positive conditions are easier to reason about."
)


CHECKS$lintr_ifelse_censor_linter <- make_lintr_check(
  "ifelse_censor_linter",
  "Use pmin()/pmax() instead of ifelse() for clamping",
  "use {.fn pmin} or {.fn pmax} instead of {.code ifelse(x > limit, limit, x)}.
        The dedicated functions are faster and clearer for clamping values."
)

CHECKS$lintr_implicit_assignment_linter <- make_check(

  description = "Avoid implicit assignments in function calls",
  tags = c("warning", "lintr"),
  preps = "lintr",

  gp = "avoid assignments inside function calls like {.code f(x <- value)}.
        Assign first, then call the function. Implicit assignments are
        easy to miss when reading code.",

  check = function(state) {
    if (inherits(state$lintr, "try-error")) {
      return(na_result())
    }

    res <- get_lintr_state(state, "implicit_assignment_linter")

    ## Implicit assignment is a valid testthat pattern, e.g.
    ## expect_warning(tmp <- f()) to capture output and check for warnings.
    ## Only flag files under R/.
    res$positions <- Filter(
      f = function(x) grepl("^R[/(\\\\)]", x$filename),
      res$positions
    )
    res$status <- length(res$positions) == 0
    res
  }
)

CHECKS$lintr_inner_combine_linter <- make_lintr_check(
  "inner_combine_linter",
  "Combine inside c() not outside",
  "use {.code c(paste0('a', x), paste0('b', x))} instead of
        {.code paste0(c('a', 'b'), x)}.
        Combining at the outer level is clearer."
)

CHECKS$lintr_length_levels_linter <- make_lintr_check(
  "length_levels_linter",
  "Use nlevels() instead of length(levels())",
  "use {.fn nlevels} instead of {.code length(levels(x))}. {.fn nlevels} is
        more expressive and works correctly with non-factor inputs."
)

CHECKS$lintr_literal_coercion_linter <- make_lintr_check(
  "literal_coercion_linter",
  "Use typed literals instead of coercion functions",
  "use {.code 1L} instead of
        {.code as.integer(1)}, and {.code 'text'}
        instead of {.code as.character('text')}.
        Typed literals are clearer and faster."
)

CHECKS$lintr_nested_ifelse_linter <- make_lintr_check(
  "nested_ifelse_linter",
  "Avoid deeply nested ifelse() calls",
  "avoid nesting {.fn ifelse} calls. Use {.fn dplyr::case_when} or
        {.fn data.table::fcase} for multi-condition logic, or a lookup table."
)

CHECKS$lintr_nested_pipe_linter <- make_lintr_check(
  "nested_pipe_linter",
  "Avoid pipes inside other pipes",
  "avoid nesting pipe chains inside other pipe chains. Extract the
        inner pipeline into a named intermediate variable."
)

CHECKS$lintr_numeric_leading_zero_linter <- make_lintr_check(
  "numeric_leading_zero_linter",
  "Include leading zero in decimal numbers",
  "write {.code 0.1} instead of {.code .1}.
        The leading zero makes decimal numbers easier
        to spot when reading code."
)

CHECKS$lintr_outer_negation_linter <- make_lintr_check(
  "outer_negation_linter",
  "Negate at the outer level",
  "use {.code !any(x)} instead of
        {.code all(!x)}, and {.code !all(x)} instead of
        {.code any(!x)}. Outer negation short-circuits
        and is clearer."
)

CHECKS$lintr_paste_linter <- make_lintr_check(
  "paste_linter",
  "Use paste0() or file.path() instead of paste() with sep",
  "use {.fn paste0} instead of {.code paste(..., sep = '')} and
        {.fn file.path} instead of {.code paste(..., sep = '/')}."
)

CHECKS$lintr_scalar_in_linter <- make_lintr_check(
  "scalar_in_linter",
  "Use == instead of %in% for scalar comparison",
  "use {.code x == 'value'} instead of {.code x %in% 'value'} when comparing
        against a single scalar. {.code ==} is clearer for single values."
)

CHECKS$lintr_undesirable_operator_linter <- make_lintr_check(
  "undesirable_operator_linter",
  "Avoid undesirable operators",
  "avoid operators like {.code <<-} (global assignment) and {.code :::}
        (accessing internal package functions). These make code harder
        to understand and maintain."
)

CHECKS$lintr_unnecessary_concatenation_linter <- make_lintr_check(
  "unnecessary_concatenation_linter",
  "Avoid unnecessary c() calls",
  "avoid {.fn c} with a single argument or no arguments. {.code c('x')} is
        just {.code 'x'}, and {.code c()} is better written as {.code NULL}."
)

CHECKS$lintr_unnecessary_lambda_linter <- make_lintr_check(
  "unnecessary_lambda_linter",
  "Avoid unnecessary anonymous functions",
  "avoid {.code \\(x) f(x)} when {.code f} alone can be passed directly.
        Unnecessary wrappers add noise without changing behaviour."
)

CHECKS$lintr_unreachable_code_linter <- make_lintr_check(
  "unreachable_code_linter",
  "No unreachable code after return/stop/next",
  "remove code after {.fn return}, {.fn stop}, or {.code next} statements.
        Unreachable code is dead code and likely a bug."
)

## -- testthat expectations ---------------------------------------------------

CHECKS$lintr_conjunct_test_linter <- make_lintr_check(
  "conjunct_test_linter",
  "Use separate expect_ calls instead of &&",
  paste(
    "split {.code expect_true(a && b)} into separate",
    "{.code expect_true(a)} and {.code expect_true(b)}",
    "calls. Separate assertions give clearer failure",
    "messages."
  ),
  tags = c("lintr", "testing")
)

CHECKS$lintr_expect_comparison_linter <- make_lintr_check(
  "expect_comparison_linter",
  "Use expect_gt/lt/gte/lte instead of expect_true with comparison",
  paste(
    "use {.code expect_gt(x, y)} instead of",
    "{.code expect_true(x > y)}. Dedicated",
    "comparison expectations give better failure messages."
  ),
  tags = c("lintr", "testing")
)

CHECKS$lintr_expect_identical_linter <- make_lintr_check(
  "expect_identical_linter",
  "Use expect_identical() instead of expect_equal() where possible",
  paste(
    "prefer {.fn expect_identical} over {.fn expect_equal} when the comparison",
    "should be exact. {.fn expect_equal} uses {.fn all.equal} with tolerance,",
    "which can mask subtle type or attribute differences."
  ),
  tags = c("lintr", "testing")
)

CHECKS$lintr_expect_length_linter <- make_lintr_check(
  "expect_length_linter",
  "Use expect_length() instead of expect_equal(length())",
  paste(
    "use {.code expect_length(x, n)} instead of",
    "{.code expect_equal(length(x), n)}.",
    "The dedicated function gives clearer failure messages."
  ),
  tags = c("lintr", "testing")
)

CHECKS$lintr_expect_named_linter <- make_lintr_check(
  "expect_named_linter",
  "Use expect_named() instead of expect_equal(names())",
  paste(
    "use {.code expect_named(x, expected)} instead of",
    "{.code expect_equal(names(x), expected)}. The dedicated function gives",
    "clearer failure messages."
  ),
  tags = c("lintr", "testing")
)

CHECKS$lintr_expect_not_linter <- make_lintr_check(
  "expect_not_linter",
  "Use expect_false() instead of expect_true(!x)",
  paste(
    "use {.fn expect_false} instead of {.code expect_true(!x)}. The dedicated",
    "function gives clearer failure messages."
  ),
  tags = c("lintr", "testing")
)

CHECKS$lintr_expect_null_linter <- make_lintr_check(
  "expect_null_linter",
  "Use expect_null() instead of expect_equal(x, NULL)",
  paste(
    "use {.fn expect_null} instead of {.code expect_equal(x, NULL)} or",
    "{.code expect_identical(x, NULL)}. The dedicated function gives clearer",
    "failure messages."
  ),
  tags = c("lintr", "testing")
)

CHECKS$lintr_expect_s3_class_linter <- make_lintr_check(
  "expect_s3_class_linter",
  "Use expect_s3_class() instead of expect_equal(class())",
  paste(
    "use {.code expect_s3_class(x, 'class')} instead of",
    "{.code expect_equal(class(x), 'class')}. The dedicated function gives",
    "clearer failure messages and handles multiple classes correctly."
  ),
  tags = c("lintr", "testing")
)

CHECKS$lintr_expect_s4_class_linter <- make_lintr_check(
  "expect_s4_class_linter",
  "Use expect_s4_class() instead of expect_true(is())",
  paste(
    "use {.code expect_s4_class(x, 'Class')} instead of",
    "{.code expect_true(is(x, 'Class'))}. The dedicated function gives",
    "clearer failure messages."
  ),
  tags = c("lintr", "testing")
)

CHECKS$lintr_expect_true_false_linter <- make_lintr_check(
  "expect_true_false_linter",
  "Use expect_true()/expect_false() instead of expect_equal(x, TRUE)",
  paste(
    "use {.fn expect_true} instead of {.code expect_equal(x, TRUE)} and",
    "{.fn expect_false} instead of",
    "{.code expect_equal(x, FALSE)}. The dedicated",
    "functions give clearer failure messages."
  ),
  tags = c("lintr", "testing")
)

CHECKS$lintr_expect_type_linter <- make_lintr_check(
  "expect_type_linter",
  "Use expect_type() instead of expect_equal(typeof())",
  paste(
    "use {.code expect_type(x, 'type')} instead of",
    "{.code expect_equal(typeof(x), 'type')}. The dedicated function gives",
    "clearer failure messages."
  ),
  tags = c("lintr", "testing")
)

#' @include lists.R

CHECKS$no_import_package_as_a_whole <- make_check(

  description = 'Packages are not imported as a whole',
  tags = c("warning", "NAMESPACE"),
  preps = "namespace",

  gp = 'not import packages as a whole, as this can cause name
        clashes between the imported packages,
        especially over time as packages change.
        Instead, import only the specific functions you need.',

  check = function(state) {
    if(inherits(state$namespace, "try-error")) return(na_result())
    imports <- state$namespace$imports
    check_result(all(vapply(imports, length, 1L) > 1))
  }
)

CHECKS$no_export_pattern <- make_check(

  description = "exportPattern in NAMESPACE file",
  tags = c("warning", "NAMESPACE"),
  preps = "namespace",

  gp = 'not use {.code exportPattern} in {.file NAMESPACE}. It can lead to
        exporting functions unintendedly. Instead, export
        functions that constitute the external API of your
        package.',

  check = function(state) {
    if(inherits(state$namespace, "try-error")) return(na_result())
    check_result(length(state$namespace$exportPatterns) == 0)
  }
)

#' @include lists.R prep_utils.R
#' @importFrom rcmdcheck rcmdcheck

PREPS$rcmdcheck <- function(state, path = state$path, quiet) {
  path <- normalizePath(path)
  run_prep_step(state, "rcmdcheck", function(path, quiet) {
    rcmdcheck(path, quiet = quiet)
  }, path = path, quiet = quiet, silent = quiet)
}

#' @include lists.R

block_is_function <- function(block) {
  cl <- block$call
  if (is.null(cl) || !is.call(cl) || length(cl) < 3) return(FALSE)
  op <- as.character(cl[[1]])
  if (!op %in% c("<-", "=", "<<-")) return(FALSE)
  if (!is.name(cl[[2]])) return(FALSE)
  rhs <- cl[[3]]
  is.call(rhs) && identical(rhs[[1]], quote(`function`))
}

block_function_name <- function(block) {
  as.character(block$call[[2]])
}

make_block_position <- function(block) {
  check_position(
    file.path("R", basename(block$file)),
    as.integer(block$line),
    line = deparse(block$call, nlines = 1)
  )
}

# -- export / noRd tagging ----------------------------------------------------

CHECKS$roxygen2_has_export_or_nord <- make_check(

  description = "Documented functions have @export, @noRd, or @rdname",
  tags = c("style", "documentation", "roxygen2"),
  preps = "roxygen2",
  gp = paste(
    "Tag every documented function with either {.code @export},",
    "{.code @noRd}, or {.code @rdname}.",
    "Functions without roxygen2 documentation are implicitly internal",
    "and do not need tagging."
  ),

  check = function(state) {
    if (inherits(state$roxygen2, "try-error")) return(na_result())
    rox <- state$roxygen2
    problems <- list()

    documented_names <- character()
    for (block in rox$blocks) {
      if (!block_is_function(block)) next
      name <- block_function_name(block)
      documented_names <- c(documented_names, name)

      has_tag <- roxygen2::block_has_tags(
        block, c("export", "noRd", "rdname")
      )
      in_ns <- name %in% rox$namespace_exports ||
        name %in% rox$namespace_s3methods
      if (!has_tag && !in_ns) {
        problems[[length(problems) + 1]] <- make_block_position(block)
      }
    }

    check_result(length(problems) == 0, problems)
  }
)

# -- unknown tags -------------------------------------------------------------

CHECKS$roxygen2_unknown_tags <- make_check(

  description = "All roxygen2 tags are recognized",
  tags = c("style", "documentation", "roxygen2"),
  preps = "roxygen2",
  gp = paste(
    "Fix or remove unknown {.pkg roxygen2} tags.",
    "This may indicate a typo, a removed tag like {.code @S3method},",
    "or a custom tag from an unregistered {.pkg roxygen2} extension",
    "(e.g. a roclet or a package that creates new tags)."
  ),

  check = function(state) {
    if (inherits(state$roxygen2, "try-error")) return(na_result())
    rox <- state$roxygen2
    msgs <- if (is.null(rox$parse_messages)) character() else
      rox$parse_messages
    problems <- list()

    ansi_re <- "\033\\[[0-9;]*m"
    tag_re <- "([A-Za-z0-9_./-]+\\.R):(\\d+):.*@(\\S+)\\s+is not a known tag"
    for (msg in msgs) {
      clean <- gsub(ansi_re, "", msg)
      m <- regmatches(clean, regexec(tag_re, clean))[[1]]
      if (length(m) < 4) next
      raw_file <- m[2]
      fname <- if (startsWith(raw_file, "R/")) raw_file else
        file.path("R", raw_file)
      problems[[length(problems) + 1]] <- check_position(
        fname,
        as.integer(m[3]),
        line = paste0("@", m[4])
      )
    }

    check_result(length(problems) == 0, problems)
  }
)

# -- @inheritParams / @inheritDotParams validation ----------------------------

CHECKS$roxygen2_valid_inherit <- make_check(

  description = "@inheritParams/@inheritDotParams reference known functions",
  tags = c("style", "documentation", "roxygen2"),
  preps = "roxygen2",
  gp = paste(
    "Ensure functions referenced by",
    "{.code @inheritParams} and",
    "{.code @inheritDotParams} exist within the",
    "package. Use {.code pkg::func} syntax for",
    "external functions."
  ),

  check = function(state) {
    if (inherits(state$roxygen2, "try-error")) return(na_result())
    rox <- state$roxygen2
    pkg_fns <- rox$function_defs$name
    problems <- list()

    for (block in rox$blocks) {
      inherit_tags <- c(
        roxygen2::block_get_tags(block, "inheritParams"),
        roxygen2::block_get_tags(block, "inheritDotParams")
      )
      if (length(inherit_tags) == 0) next

      for (tag in inherit_tags) {
        ref <- trimws(strsplit(trimws(tag$val), "\\s+")[[1]][1])
        if (grepl("::", ref, fixed = TRUE)) next
        if (!ref %in% pkg_fns) {
          problems[[length(problems) + 1]] <- make_block_position(block)
          break
        }
      }
    }

    check_result(length(problems) == 0, problems)
  }
)

# -- duplicate @param documentation ------------------------------------------

extract_block_params <- function(block) {
  param_tags <- roxygen2::block_get_tags(block, "param")
  if (length(param_tags) == 0) return(NULL)

  lapply(param_tags, function(tag) {
    list(
      name = tag$val$name,
      description = trimws(tag$val$description),
      file = block$file,
      line = block$line
    )
  })
}

CHECKS$roxygen2_duplicate_params <- make_check(

  description = "Avoid duplicated @param documentation across functions",
  tags = c("style", "documentation", "roxygen2"),
  preps = "roxygen2",

  gp = paste(
    "use {.code @inheritParams} to avoid duplicating parameter documentation.",
    "Identical {.code @param} descriptions across files suggest shared docs",
    "should be inherited from a single source."
  ),

  check = function(state) {
    if (inherits(state$roxygen2, "try-error")) return(na_result())

    all_params <- unlist(
      lapply(state$roxygen2$blocks, extract_block_params),
      recursive = FALSE
    )
    if (is.null(all_params) || length(all_params) == 0) {
      return(check_result(TRUE))
    }

    keys <- vapply(all_params, function(p) {
      paste(p$name, p$description, sep = "|||")
    }, character(1))
    files <- vapply(all_params, function(p) p$file, character(1))

    duped_keys <- unique(keys[duplicated(keys)])
    if (length(duped_keys) == 0) {
      return(check_result(TRUE))
    }

    problems <- list()
    for (key in duped_keys) {
      idxs <- which(keys == key)
      dup_files <- unique(basename(files[idxs]))
      if (length(dup_files) < 2) next

      for (i in idxs) {
        p <- all_params[[i]]
        problems[[length(problems) + 1]] <- check_position(
          file.path("R", basename(p$file)),
          as.integer(p$line),
          line = paste0("@param ", p$name)
        )
      }
    }

    check_result(length(problems) == 0, problems)
  }
)

#' @include lists.R prep_utils.R

#' @noRd
parse_rd_files <- function(mandir) {
  if (!dir.exists(mandir)) {
    return(list())
  }

  rd_files <- list.files(mandir, pattern = "\\.Rd$", full.names = TRUE)
  if (length(rd_files) == 0) {
    return(list())
  }

  lapply(rd_files, function(rd_file) {
    parsed <- tools::parse_Rd(rd_file)
    tags <- vapply(
      parsed,
      function(x) {
        tag <- attr(x, "Rd_tag")
        if (is.null(tag)) "" else tag
      },
      ""
    )

    aliases <- vapply(
      parsed[tags == "\\alias"],
      function(x) trimws(paste(unlist(x), collapse = "")),
      ""
    )

    keywords <- vapply(
      parsed[tags == "\\keyword"],
      function(x) trimws(paste(unlist(x), collapse = "")),
      ""
    )

    doctypes <- vapply(
      parsed[tags == "\\docType"],
      function(x) trimws(paste(unlist(x), collapse = "")),
      ""
    )

    list(
      file = basename(rd_file),
      aliases = aliases,
      has_examples = any(tags == "\\examples"),
      has_value = any(tags == "\\value"),
      has_keyword_internal = "internal" %in% keywords,
      is_reexport = "import" %in% doctypes
    )
  })
}

PREPS$rd <- function(state, path = state$path, quiet) {
  run_prep_step(state, "rd", function(path) {
    parse_rd_files(file.path(path, "man"))
  }, path = path, silent = quiet)
}

#' List all checks performed
#'
#' @param gp \code{\link{gp}} output.
#' @return  Character vector of check names.
#'
#' @family API
#' @export
#' @examples 
#' path <- system.file("bad1", package = "goodpractice")
#' # Run a subset of all checks available
#' g <- gp(path, checks = all_checks()[9:16])
#' checks(g)
#' # Or run with named check groups
#' g <- gp(path, checks = checks_by_group("description", "namespace"))
#' checks(g)

checks <- function(gp) {
  names(gp$checks)
}

#' Return all check results in a data frame
#'
#' @param gp \code{\link{gp}} output.
#' @return Data frame, with columns:
#' \item{check}{The name of the check.}
#' \item{passed}{Logical, whether the check passed.}
#'
#' @family output
#' @export
#' @examples 
#' path <- system.file("bad1", package = "goodpractice")
#' # Run a subset of all checks available
#' g <- gp(path, checks = all_checks()[9:16])
#' results(g)
#' # Or run with named check groups
#' g <- gp(path, checks = checks_by_group("description", "namespace"))
#' results(g)

results <- function(gp) {
  data.frame(
    row.names = NULL,
    check = names(gp$checks),
    passed = vapply(gp$checks, check_passed, TRUE)
  )
}

#' Names of the failed checks
#'
#' @param gp \code{\link{gp}} output.
#' @return Names of the failed checks.
#'
#' @family API
#' @export
#' @examples 
#' path <- system.file("bad1", package = "goodpractice")
#' # run a subset of all checks available
#' g <- gp(path, checks = all_checks()[9:16])
#' failed_checks(g)
#' # Or run with named check groups
#' g <- gp(path, checks = checks_by_group("description", "namespace"))
#' failed_checks(g)

failed_checks <- function(gp) {
  names(Filter(check_failed, gp$checks))
}

#' Positions of check failures in the source code
#'
#' Note that not all checks refer to the source code.
#' For these the result will be \code{NULL}.
#'
#' For the ones that do, the results is a list, one for each failure.
#' Since the same check can fail multiple times. A single failure
#' is a list with entries: \code{filename}, \code{line_number},
#' \code{column_number}, \code{ranges}. \code{ranges} is a list of
#' pairs of start and end positions for each line involved in the
#' check.
#'
#' @param gp \code{\link{gp}} output.
#' @return A list of lists of positions. See details below.
#'
#' @family API
#' @export
#' @examples
#' path <- system.file("bad1", package = "goodpractice")
#' g <- gp(path, checks = "description_url")
#' failed_positions(g)

failed_positions <- function(gp) {
  failed <- Filter(check_failed, gp$checks)
  lapply(failed, get_position)
}

get_position <- function(chk) {
  if ("positions" %in% names(chk)) chk$positions else NULL
}

#' Export failed checks to JSON
#'
#' @param gp \code{\link{gp}} output.
#' @param file Output connection or file.
#' @param pretty Whether to pretty-print the JSON.
#' @return Invisibly returns the path to the output file.
#'
#' @family output
#' @export
#' @examples
#' path <- system.file("bad1", package = "goodpractice")
#' g <- gp(path, checks = "description_url")
#' tmp <- tempfile(fileext = ".json")
#' export_json(g, tmp)
#' unlink(tmp)
#' @importFrom jsonlite toJSON
#' @importFrom whoami username fullname

export_json <- function(gp, file, pretty = FALSE) {

  pkg <- if (!is.null(gp$description)) gp$description$get("Package") else gp$package
  obj <- list(
    package = pkg,
    path = gp$path,
    failures = Filter(check_failed, gp$checks),
    gp_version = loaded_pkg_version("goodpractice"),
    date = as.character(Sys.time()),
    user = username(fallback = "<unknown>"),
    name = fullname(fallback = "<unknown>")
  )

  cat(toJSON(obj, pretty = pretty), file = file)
  invisible(file)
}

#' Download a local copy of 'goodpractice4agents.md' to tell AI agents how to
#' use this package.
#'
#' The 'goodpractice4agents.md' file contains sufficient instructions for most
#' AI agents to edit package code so that it passes most 'goodpractice'
#' checks. The file can be directly edited and tweaked for personal use cases.
#' An agent can be instructed to simply "run the file goodpractice4agents.md".
#' Alternatively, the file can be directly downloaded from GitHub at
#' \url{https://github.com/ropensci-review-tools/goodpractice/blob/main/inst/skills/goodpractice4agents.md}.
#'
#' You can also instruct agents to call the companion function,
#' \link{learn_skill_gp}, directly in order to learn and use the skill itself.
#'
#' @param path Local path where the file should be written. If \code{path} is
#' an existing directory, the file is written as \code{goodpractice4agents.md}
#' within it; otherwise \code{path} is treated as the full target file name.
#' @param use_skills_subdir If \code{TRUE}, the \code{path} will be edited to
#' write the file in to a \code{skills/} sub-directory.
#' @param overwrite If \code{FALSE} (default) and file specified by \code{path}
#' already exists, function will error and not overwrite the existing file.
#'
#' @return (Invisibly) the full path to the written file.
#'
#' @family skills
#' @export
#' @examples
#' path <- file.path(tempdir(), "skill.md")
#' f <- use_skill_gp(path)
#' file.exists(f)
#' unlink(f)

use_skill_gp <- function(
  path = "goodpractice4agents.md",
  use_skills_subdir = FALSE,
  overwrite = FALSE) {
    if (dir.exists(path)) {
        path <- file.path(path, "goodpractice4agents.md")
    } else if (!dir.exists(dirname(path))) {
        cli::cli_abort("Directory {.path {dirname(path)}} does not exist.")
    }

    if (file.exists(path) && !overwrite) {
        cli::cli_abort("File {.file {path}} already exists.")
    }
    if (use_skills_subdir) {
        path <- file.path(dirname(path), "skills", basename(path))
    }

    src <- system.file(
        "skills", "goodpractice4agents.md", package = "goodpractice"
    )
    if (!file.exists(src)) {
        cli::cli_abort("File {.file {src}} not found.")
    }
    if (!file.copy(src, path, overwrite = overwrite)) {
        cli::cli_abort("Failed to copy file to {.file {path}}.")
    }

    invisible(path)
}

#' Function for AI agents to run and learn how to use the 'goodpractice'
#' package.
#'
#' You tell agents to directly use this function in order for them to learn the
#' skill. If you want to use the skill yourself to guide an agent, then use the
#' \link{use_skill_gp} function.
#'
#' @return The content of the file \code{system.file("skills",
#' "goodpractice4agents.md", package = "goodpractice"))}
#'
#' @family skills
#' @export
#' @examples
#' \dontrun{
#' learn_gp_skill()
#' }
learn_skill_gp <- function() {

    src <- system.file(
        "skills", "goodpractice4agents.md", package = "goodpractice"
    )
    writeLines(readLines(src))
}

test_that("use_skill_gp writes to a full file path", {
  tmp <- withr::local_tempdir()
  path <- file.path(tmp, "skill.md")
  on.exit(unlink(path), add = TRUE)
  f <- use_skill_gp(path)
  expect_identical(f, path)
  expect_true(file.exists(path))
})

test_that("use_skill_gp returns the path invisibly", {
  tmp <- withr::local_tempdir()
  path <- file.path(tmp, "skill.md")
  on.exit(unlink(path), add = TRUE)
  expect_invisible(use_skill_gp(path))
})

test_that("use_skill_gp writes into an existing directory", {
  tmp <- withr::local_tempdir()
  f <- use_skill_gp(tmp)
  expect_identical(f, file.path(tmp, "goodpractice4agents.md"))
  expect_true(file.exists(f))
  on.exit(unlink(f), add = TRUE)
})

test_that("use_skill_gp defaults to the working directory", {
  tmp <- withr::local_tempdir()
  withr::local_dir(tmp)
  use_skill_gp()
  f <- file.path(tmp, "goodpractice4agents.md")
  expect_true(file.exists(f))
  on.exit(unlink(f), add = TRUE)
})

test_that("use_skill_gp errors when the target file already exists", {
  tmp <- withr::local_tempdir()
  path <- file.path(tmp, "skill.md")
  use_skill_gp(path)
  expect_error(use_skill_gp(path), "already exists")
  on.exit(unlink(path), add = TRUE)
})

test_that("use_skill_gp errors when the parent directory is missing", {
  tmp <- withr::local_tempdir()
  path <- file.path(tmp, "nope", "skill.md")
  expect_error(use_skill_gp(path), "does not exist")
})

test_that("written file matches the packaged source", {
  tmp <- withr::local_tempdir()
  f <- use_skill_gp(file.path(tmp, "skill.md"))
  src <- system.file(
    "skills",
    "goodpractice4agents.md",
    package = "goodpractice"
  )
  expect_identical(readLines(f), readLines(src))
})

test_that("learn_gp_skill prints the bundled skill file", {
  out <- capture.output(learn_skill_gp())
  src <- system.file(
    "skills",
    "goodpractice4agents.md",
    package = "goodpractice"
  )
  expect_identical(out, readLines(src))
})

test_that("duplicate_function_bodies fails on identical fns across files", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: dupecode", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "DESCRIPTION"))

  writeLines(
    "helper_a <- function(x) { result <- x + 1; log(result); result }",
    file.path(pkg, "R", "a.R")
  )
  writeLines(
    "helper_b <- function(x) { result <- x + 1; log(result); result }",
    file.path(pkg, "R", "b.R")
  )

  gp_res <- gp(pkg, checks = "duplicate_function_bodies")
  res <- results(gp_res)
  expect_false(res$passed[res$check == "duplicate_function_bodies"])

  pos <- failed_positions(gp_res)$duplicate_function_bodies
  names <- vapply(pos, `[[`, "", "line")
  expect_true("helper_a" %in% names)
  expect_true("helper_b" %in% names)
})

test_that("duplicate_function_bodies passes with different bodies", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: nodupecode", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "DESCRIPTION"))

  writeLines(
    "add_one <- function(x) { x + 1 }",
    file.path(pkg, "R", "a.R")
  )
  writeLines(
    "times_two <- function(x) { x * 2 }",
    file.path(pkg, "R", "b.R")
  )

  gp_res <- gp(pkg, checks = "duplicate_function_bodies")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "duplicate_function_bodies"])
})

test_that("duplicate_function_bodies ignores trivial one-liners", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: trivial", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "DESCRIPTION"))

  writeLines("f1 <- function() NULL", file.path(pkg, "R", "a.R"))
  writeLines("f2 <- function() NULL", file.path(pkg, "R", "b.R"))

  gp_res <- gp(pkg, checks = "duplicate_function_bodies")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "duplicate_function_bodies"])
})

test_that("duplicate_function_bodies ignores duplicates within same file", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: samefile", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "DESCRIPTION"))

  writeLines(c(
    "helper_a <- function(x) { result <- x + 1; log(result); result }",
    "helper_b <- function(x) { result <- x + 1; log(result); result }"
  ), file.path(pkg, "R", "funcs.R"))

  gp_res <- gp(pkg, checks = "duplicate_function_bodies")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "duplicate_function_bodies"])
})

get_result <- function(res, check) res$passed[res$check == check]

test_that("print_return_invisible fails when print method lacks invisible()", {
  gp_res <- gp("bad_print", checks = "print_return_invisible")
  res <- results(gp_res)
  expect_false(get_result(res, "print_return_invisible"))

  pos <- failed_positions(gp_res)$print_return_invisible
  names <- vapply(pos, `[[`, "", "line")
  expect_true("print.myclass" %in% names)
  expect_true("print.another" %in% names)
})

test_that("print_return_invisible passes with correct invisible() return", {
  gp_res <- gp("good_print", checks = "print_return_invisible")
  res <- results(gp_res)
  expect_true(get_result(res, "print_return_invisible"))
})

test_that("print_return_invisible passes when no print methods exist", {
  gp_res <- gp("good", checks = "print_return_invisible")
  res <- results(gp_res)
  expect_true(get_result(res, "print_return_invisible"))
})

test_that("print_return_invisible passes when no R directory exists", {
  state <- list(path = tempdir())
  result <- CHECKS$print_return_invisible$check(state)
  expect_true(result$status)
  expect_length(result$positions, 0)
})

test_that("no_obsolete_deps fails when DESCRIPTION depends on obsolete pkg", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(
    c(
      "Package: badpkg",
      "Title: Test",
      "Version: 1.0.0",
      "Description: Test.",
      "License: MIT",
      "Imports: sp"
    ),
    file.path(pkg, "DESCRIPTION")
  )
  writeLines("f <- function() 1", file.path(pkg, "R", "code.R"))

  gp_res <- gp(pkg, checks = "no_obsolete_deps")
  res <- results(gp_res)
  expect_false(res$passed[res$check == "no_obsolete_deps"])

  pos <- failed_positions(gp_res)$no_obsolete_deps
  lines <- vapply(pos, `[[`, "", "line")
  expect_true(any(grepl("sp", lines)))
})

test_that("no_obsolete_deps passes with no obsolete dependencies", {
  gp_res <- gp("good", checks = "no_obsolete_deps")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "no_obsolete_deps"])
})

test_that("no_obsolete_deps catches multiple obsolete packages", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(
    c(
      "Package: badpkg2",
      "Title: Test",
      "Version: 1.0.0",
      "Description: Test.",
      "License: MIT",
      "Imports: RCurl, XML"
    ),
    file.path(pkg, "DESCRIPTION")
  )
  writeLines("f <- function() 1", file.path(pkg, "R", "code.R"))

  gp_res <- gp(pkg, checks = "no_obsolete_deps")
  res <- results(gp_res)
  expect_false(res$passed[res$check == "no_obsolete_deps"])

  pos <- failed_positions(gp_res)$no_obsolete_deps
  expect_length(pos, 2)
})

test_that("no_obsolete_deps returns NA on description error", {
  state <- list(description = structure("error", class = "try-error"))
  result <- CHECKS$no_obsolete_deps$check(state)
  expect_true(is.na(result$status))
})

test_that("no_obsolete_deps gp message lists found packages", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(
    c(
      "Package: msgtest",
      "Title: Test",
      "Version: 1.0.0",
      "Description: Test.",
      "License: MIT",
      "Imports: rjson"
    ),
    file.path(pkg, "DESCRIPTION")
  )
  writeLines("f <- function() 1", file.path(pkg, "R", "code.R"))

  gp_res <- gp(pkg, checks = "no_obsolete_deps")
  chk <- CHECKS$no_obsolete_deps
  msg <- chk$gp(gp_res)
  expect_match(msg, "rjson")
  expect_match(msg, "jsonlite")
})

# "rcmdcheck_package_dependencies_present"  fails in bad1
# "rcmdcheck_undeclared_library_require" fails in bad2
# "rcmdcheck_undefined_globals" fails in bad2
# "rcmdcheck_missing_docs" fails in baddoc

# all take fairly long to run... commenting out for now

# bad1 <- system.file("bad1", package = "goodpractice")
# gp_bad1 <- gp(bad1, checks = c("rcmdcheck_missing_docs",
#                                "rcmdcheck_package_dependencies_present",
#                                "rcmdcheck_undeclared_library_require",
#                                "rcmdcheck_undefined_globals"))
# res_bad1 <- results(gp_bad1)
#
# gp_bad2 <- gp("bad2", checks = c("rcmdcheck_missing_docs",
#                                  "rcmdcheck_package_dependencies_present",
#                                  "rcmdcheck_undeclared_library_require",
#                                  "rcmdcheck_undefined_globals"))
# res_bad2 <- results(gp_bad2)
#
# get_result <- function(res, check) res$passed[res$check == check]
#
#
# test_that("rcmdcheck_package_dependencies_present", {
#
#   expect_true(get_result(res_bad2, "rcmdcheck_package_dependencies_present"))
#   expect_false(get_result(res_bad1, "rcmdcheck_package_dependencies_present"))
#
# })
#
# test_that("rcmdcheck_undeclared_library_require", {
#
#   expect_true(get_result(res_bad1, "rcmdcheck_undeclared_library_require"))
#   expect_false(get_result(res_bad2, "rcmdcheck_undeclared_library_require"))
#
# })
#
# test_that("rcmdcheck_undefined_globals", {
#
#   expect_true(get_result(res_bad1, "rcmdcheck_undefined_globals"))
#   expect_false(get_result(res_bad2, "rcmdcheck_undefined_globals"))
#
# })
#
# test_that("rcmdcheck_missing_docs", {
#
#   expect_true(get_result(res_bad2, "rcmdcheck_missing_docs"))
#
#   x <- gp("baddoc", checks = "rcmdcheck_missing_docs")
#   expect_false(results(x)$passed)
#
# })

get_result <- function(res, check) res$passed[res$check == check]

test_that("r_file_extension fails when .r or .q files exist", {
  gp_res <- gp("bad_r_ext", checks = "r_file_extension")
  res <- results(gp_res)
  expect_false(get_result(res, "r_file_extension"))

  pos <- failed_positions(gp_res)$r_file_extension
  filenames <- vapply(pos, `[[`, "", "filename")
  expect_true(any(grepl("bad_func\\.r$", filenames)))
  expect_true(any(grepl("old_syntax\\.q$", filenames)))
})

test_that("r_file_extension passes when only .R files exist", {
  gp_res <- gp("good", checks = "r_file_extension")
  res <- results(gp_res)
  expect_true(get_result(res, "r_file_extension"))
})

test_that("r_file_extension passes when no R directory exists", {
  state <- list(path = tempdir())
  result <- CHECKS$r_file_extension$check(state)
  expect_true(result$status)
  expect_length(result$positions, 0)
})

test_that("on_exit_has_add fails when add is missing", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: badonexit", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "DESCRIPTION"))
  writeLines(
    "cleanup <- function() { on.exit(close(con)) }",
    file.path(pkg, "R", "cleanup.R")
  )

  gp_res <- gp(pkg, checks = "on_exit_has_add")
  res <- results(gp_res)
  expect_false(res$passed[res$check == "on_exit_has_add"])
})

test_that("on_exit_has_add passes when add = TRUE", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: goodonexit", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "R/../DESCRIPTION"))
  writeLines(
    "cleanup <- function() { on.exit(close(con), add = TRUE) }",
    file.path(pkg, "R", "cleanup.R")
  )

  gp_res <- gp(pkg, checks = "on_exit_has_add")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "on_exit_has_add"])
})

test_that("on_exit_has_add passes when add = FALSE (explicitly set)", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: explicitfalse", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "DESCRIPTION"))
  writeLines(
    "cleanup <- function() { on.exit(close(con), add = FALSE) }",
    file.path(pkg, "R", "cleanup.R")
  )

  gp_res <- gp(pkg, checks = "on_exit_has_add")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "on_exit_has_add"])
})

test_that("on_exit_has_add passes when on.exit() has no args (reset)", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: resetonexit", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT"
  ), file.path(pkg, "DESCRIPTION"))
  writeLines(
    "cleanup <- function() { on.exit() }",
    file.path(pkg, "R", "cleanup.R")
  )

  gp_res <- gp(pkg, checks = "on_exit_has_add")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "on_exit_has_add"])
})

test_that("on_exit_has_add passes with no functions", {
  gp_res <- gp("good", checks = "on_exit_has_add")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "on_exit_has_add"])
})

bad1 <- system.file("bad1", package = "goodpractice")

# Warning already tested in test-coverage.R
x <- suppressWarnings(
  gp(bad1, checks = c("covr", "description_bugreports"))
)

test_that("checks", {
  expect_identical(checks(x), c("covr", "description_bugreports"))
})

test_that("results", {
  res <- data.frame(
    check = c("covr", "description_bugreports"),
    passed = c(NA, FALSE),
    stringsAsFactors = FALSE
  )
  expect_identical(results(x), res)
})

test_that("failed_checks returns names of failed checks", {
  fc <- failed_checks(x)
  expect_true("description_bugreports" %in% fc)
  expect_false("covr" %in% fc)
})

test_that("failed_positions returns positions for failed checks", {
  fp <- failed_positions(x)
  expect_type(fp, "list")
  expect_true("description_bugreports" %in% names(fp))
})

test_that("failed_positions returns empty list when check has no positions", {
  fp <- failed_positions(x)
  expect_identical(fp$description_bugreports, list())
})

test_that("failed_positions returns positions when check has them", {
  y <- gp(bad1, checks = "r_file_extension")
  fp <- failed_positions(y)
  expect_type(fp$r_file_extension, "list")
  expect_gt(length(fp$r_file_extension), 0)
})

test_that("get_position returns NULL when check has no positions", {
  chk <- list(passed = FALSE)
  expect_null(goodpractice:::get_position(chk))
})

test_that("export_json writes valid JSON", {
  tmp <- withr::local_tempfile(fileext = ".json")
  export_json(x, file = tmp)
  expect_true(file.exists(tmp))
  obj <- jsonlite::fromJSON(tmp)
  expect_identical(obj$package, "badpackage")
  expect_true("failures" %in% names(obj))
  expect_true("gp_version" %in% names(obj))
})

test_that("export_json pretty-prints when requested", {
  tmp <- withr::local_tempfile(fileext = ".json")
  export_json(x, file = tmp, pretty = TRUE)
  content <- expect_warning(
    readLines(tmp),
    "incomplete final line"
  )
  expect_gt(length(content), 1)
})

test_that("export_json works for a single non-description check group", {
  y <- gp(bad1, checks = checks_by_group("lintr"))
  tmp <- withr::local_tempfile(fileext = ".json")
  export_json(y, file = tmp)
  obj <- jsonlite::fromJSON(tmp)
  expect_identical(obj$package, "badpackage")
})

get_result <- function(res, check) res$passed[res$check == check]

test_that("spelling fails when misspelled words exist", {
  gp_res <- gp("bad_spelling", checks = "spelling")
  res <- results(gp_res)
  expect_false(get_result(res, "spelling"))

  pos <- failed_positions(gp_res)$spelling
  words <- vapply(pos, `[[`, "", "line")
  expect_true(any(grepl("incremnet", words)))
})

test_that("spelling passes when no misspellings exist", {
  gp_res <- gp("good", checks = "spelling")
  res <- results(gp_res)
  expect_true(get_result(res, "spelling"))
})

test_that("spelling returns NA when no inst/WORDLIST exists", {
  state <- list(spelling = "no_wordlist")
  result <- CHECKS$spelling$check(state)
  expect_true(is.na(result$status))
  expect_identical(result$positions, list())
})

#' @include lists.R customization.R

CHECKS$covr <- make_check(

  description = "All code is unit tested",
  tags = c("warning", "code coverage", "covr"),
  preps = "covr",

  gp = function(state) {
    percent <- state$covr$pct_by_line
    if (is.nan(percent)) {
      return("write unit tests. No testable code was found in this package.")
    }
    paste0(
      "write unit tests for all functions, and all package code in ",
      "general. ", trunc(percent), "% of code lines are covered by ",
      "test cases."
    )
  },

  check = function(state) {
    if(inherits(state$covr, "try-error"))
      return(na_result())

    if (is.nan(state$covr$pct_by_line))
      return(check_result(TRUE))

    zero <- state$covr$zero
    if (NROW(zero) == 0) return(check_result(TRUE))
    
    positions <- lapply(seq_len(NROW(zero)), function(i) {
      check_position(zero$filename[i], zero$line[i], line = NA_character_)
    })
    
    check_result(FALSE, positions)
  }
)

#' goodpractice: Advice on R Package Building
#'
#' Give advice about good practices when building R packages. Advice includes
#' functions and syntax to avoid, package structure, code complexity, code
#' formatting, etc.
#' @family main
#' @name goodpractice-package
#' @aliases goodpractice-package
"_PACKAGE"

#' @include lists.R customization.R

AVOIDED_PACKAGES <- list(
  multicore = "Use the 'parallel' package instead.",
  RCurl     = "Use 'httr2', 'curl', or 'crul' instead.",
  rjson     = "Use 'jsonlite' instead.",
  RJSONIO   = "Use 'jsonlite' instead.",
  XML       = "Use 'xml2' instead.",
  sp        = "Use 'sf' instead. 'sp' is deprecated.",
  rgdal     = paste0(
    "Use 'sf', 'terra', or alternatives described at ",
    "https://www.hypertidy.org/posts/",
    "2026-03-24_check-out-wk/#check-it-out-series. ",
    "'rgdal' was retired in 2023."
  ),
  rgeos     = paste0(
    "Use 'sf', 'terra', or alternatives described at ",
    "https://www.hypertidy.org/posts/",
    "2026-03-24_check-out-wk/#check-it-out-series. ",
    "'rgeos' was retired in 2023."
  ),
  maptools  = "Use 'sf' instead. 'maptools' was retired in 2023."
)

CHECKS$no_obsolete_deps <- make_check(

  description = "No obsolete or retired package dependencies",
  tags = c("warning", "best practice"),
  preps = "description",

  gp = function(state) {
    if (inherits(state$description, "try-error")) {
      return("avoid depending on obsolete packages.")
    }
    deps <- state$description$get_deps()
    found <- intersect(deps$package, names(AVOIDED_PACKAGES))
    reasons <- vapply(found, function(p) {
      paste0(p, ": ", AVOIDED_PACKAGES[[p]])
    }, character(1))
    paste(
      "avoid depending on obsolete or retired packages.",
      paste(reasons, collapse = " ")
    )
  },

  check = function(state) {
    if (inherits(state$description, "try-error")) {
      return(na_result())
    }

    deps <- tryCatch(state$description$get_deps(), error = function(e) NULL)
    if (is.null(deps)) return(check_result(TRUE))

    found <- intersect(deps$package, names(AVOIDED_PACKAGES))
    if (length(found) == 0) return(check_result(TRUE))

    problems <- lapply(found, function(pkg) {
      dep_row <- deps[deps$package == pkg, ]
      check_position("DESCRIPTION", line = paste0(dep_row$type[1], ": ", pkg))
    })

    check_result(FALSE, problems)
  }
)

#' @include lists.R prep_utils.R
#' @importFrom covr package_coverage zero_coverage percent_coverage
#' @importFrom withr with_options

PREPS$covr <- function(state, path = state$path, quiet) {
  state <- run_prep_step(state, "covr", function(path, quiet) {
    list(coverage = package_coverage(path, quiet = quiet))
  }, path = path, quiet = quiet, silent = quiet)

  if (!inherits(state$covr, "try-error")) {
    state$covr$zero <- with_options(
      list(covr.rstudio_source_markers = FALSE),
      zero_coverage(state$covr$coverage)
    )
    state$covr$pct_by_line <- percent_coverage(state$covr$coverage, by = "line")
    state$covr$pct_by_expr <- percent_coverage(
      state$covr$coverage, by = "expression"
    )
  }

  state
}

#' Write a `.lintr` file into a test fixture directory
#'
#' lintr falls back to a user's global `~/.lintr` whenever a project has no
#' `.lintr` of its own, so a developer's personal lintr config can silently
#' change which tidyverse/lintr checks pass or fail on these fixtures. This
#' function sive every fixture package its own `.lintr` so tests are isolated
#' from that.
#'
#' (`.lintr` files can't be committed to the fixture packages themselves
#' because CRAN forbids all dotfiles.)
#'
#' @param path Directory of the fixture package (or temp dir) to write into.
#' @param linters Text of the `linters:` value in the generated `.lintr`
#'   file; defaults to lintr's own defaults.
#' @return Invisibly, the path of the `.lintr` file written.
write_dot_lintr <- function(path, linters = "linters_with_defaults()") {
  lintr_path <- file.path(path, ".lintr")
  writeLines(paste0("linters: ", linters), lintr_path)
  invisible(lintr_path)
}

#' @include lists.R

urlchecker_make_positions <- function(db) {
  lapply(seq_len(nrow(db)), function(i) {
    from <- db$From[[i]]
    check_position(
      if (length(from) > 0) from[[1]] else "unknown",
      line = db$URL[i]
    )
  })
}

make_urlchecker_check <- function(description, gp, filter, tags = NULL) {
  make_check(
    description = description,
    tags = c("style", "documentation", "url", tags),
    preps = "urlchecker",
    gp = gp,

    check = function(state) {
      if (inherits(state$urlchecker, "try-error")) {
        return(na_result())
      }

      db <- state$urlchecker
      if (is.null(db) || nrow(db) == 0) {
        return(check_result(TRUE))
      }

      problems <- filter(db)
      if (nrow(problems) == 0) {
        return(check_result(TRUE))
      }

      check_result(FALSE, urlchecker_make_positions(problems))
    }
  )
}

CHECKS$urlchecker_ok <- make_urlchecker_check(
  description = "All URLs are reachable",
  gp = "Fix or remove broken URLs in documentation and {.file DESCRIPTION}.",
  filter = function(db) db[!db$Status %in% c("200", "301", "302"), ]
)

CHECKS$urlchecker_no_redirects <- make_urlchecker_check(
  description = "No URLs redirect to a different location",
  tags = "CRAN",
  gp = "Update URLs that redirect to their final destination.",
  filter = function(db) db[nzchar(db$New), ]
)

make_urlchecker_db <- function(urls = character(),
                              from = list(),
                              status = character(),
                              message = character(),
                              new = character()) {
  structure(
    data.frame(
      URL = urls,
      Status = status,
      Message = message,
      New = new,
      stringsAsFactors = FALSE
    ),
    class = c("urlchecker_db", "check_url_db", "data.frame")
  )
}

# We need to add From as a list column separately
add_from <- function(db, from) {
  db$From <- from
  db
}

# -- helpers -------------------------------------------------------------------

test_that("na_result returns NA status with empty positions", {
  res <- na_result()
  expect_true(is.na(res$status))
  expect_length(res$positions, 0)
})

test_that("urlchecker_make_positions extracts first From entry as filename", {
  db <- make_urlchecker_db(
    urls = c("https://a.com", "https://b.com"),
    status = c("404", "500"),
    message = c("Not Found", "Error"),
    new = c("", "")
  )
  db <- add_from(db, list(c("DESCRIPTION", "R/foo.R"), "man/bar.Rd"))
  pos <- urlchecker_make_positions(db)
  expect_length(pos, 2)
  expect_identical(pos[[1]]$filename, "DESCRIPTION")
  expect_identical(pos[[2]]$filename, "man/bar.Rd")
  expect_identical(pos[[1]]$line, "https://a.com")
  expect_identical(pos[[2]]$line, "https://b.com")
  expect_true(is.na(pos[[1]]$line_number))
  expect_true(is.na(pos[[1]]$column_number))
})

test_that("urlchecker_make_positions handles empty From with 'unknown'", {
  db <- make_urlchecker_db(
    urls = "https://a.com",
    status = "404",
    message = "Not Found",
    new = ""
  )
  db <- add_from(db, list(character()))
  pos <- urlchecker_make_positions(db)
  expect_identical(pos[[1]]$filename, "unknown")
})

# -- factory structure ---------------------------------------------------------

test_that("make_urlchecker_check produces a valid check object", {
  chk <- CHECKS$urlchecker_ok
  expect_s3_class(chk, "check")
  expect_identical(chk$description, "All URLs are reachable")
  expect_true("urlchecker" %in% chk$preps)
  expect_true("url" %in% chk$tags)
  expect_true(is.function(chk$check))
})

test_that("urlchecker_no_redirects has CRAN tag", {
  chk <- CHECKS$urlchecker_no_redirects
  expect_true("CRAN" %in% chk$tags)
})

test_that("make_urlchecker_check factory produces working checks", {
  custom <- make_urlchecker_check(
    description = "test check",
    gp = "test advice",
    filter = function(db) db[db$Status == "999", ],
    tags = "test"
  )
  expect_s3_class(custom, "check")
  expect_identical(custom$description, "test check")
  expect_true("test" %in% custom$tags)
  expect_true("urlchecker" %in% custom$preps)

  db <- make_urlchecker_db(
    urls = c("https://a.com", "https://b.com"),
    status = c("999", "200"),
    message = c("Custom", "OK"),
    new = c("", "")
  )
  db <- add_from(db, list("DESCRIPTION", "DESCRIPTION"))

  result <- custom$check(list(urlchecker = db))
  expect_false(result$status)
  expect_length(result$positions, 1)
  expect_identical(result$positions[[1]]$line, "https://a.com")

  result_pass <- custom$check(list(urlchecker = make_urlchecker_db()))
  expect_true(result_pass$status)

  result_na <- custom$check(list(
    urlchecker = structure("err", class = "try-error")
  ))
  expect_true(is.na(result_na$status))

  result_null <- custom$check(list(urlchecker = NULL))
  expect_true(result_null$status)

  db_no_match <- make_urlchecker_db(
    urls = "https://ok.com",
    status = "200",
    message = "OK",
    new = ""
  )
  db_no_match <- add_from(db_no_match, list("DESCRIPTION"))
  result_filtered <- custom$check(list(urlchecker = db_no_match))
  expect_true(result_filtered$status)
})

# -- NULL state ----------------------------------------------------------------

test_that("urlchecker_ok passes when state$urlchecker is NULL", {
  state <- list(urlchecker = NULL)
  result <- CHECKS$urlchecker_ok$check(state)
  expect_true(result$status)
})

test_that("urlchecker_no_redirects passes when state$urlchecker is NULL", {
  state <- list(urlchecker = NULL)
  result <- CHECKS$urlchecker_no_redirects$check(state)
  expect_true(result$status)
})

# -- urlchecker_ok ------------------------------------------------------------

test_that("urlchecker_ok passes when no broken URLs", {
  db <- make_urlchecker_db(
    urls = "https://example.com",
    status = "200",
    message = "OK",
    new = "https://example.org"
  )
  db <- add_from(db, list("DESCRIPTION"))
  state <- list(urlchecker = db)
  result <- CHECKS$urlchecker_ok$check(state)
  expect_true(result$status)
})

test_that("urlchecker_ok fails on 404", {
  db <- make_urlchecker_db(
    urls = "https://example.com/gone",
    status = "404",
    message = "Not Found",
    new = ""
  )
  db <- add_from(db, list("DESCRIPTION"))
  state <- list(urlchecker = db)
  result <- CHECKS$urlchecker_ok$check(state)
  expect_false(result$status)
  expect_length(result$positions, 1)
  expect_identical(result$positions[[1]]$line, "https://example.com/gone")
})

test_that("urlchecker_ok passes when all URLs return 200 or redirect", {
  db <- make_urlchecker_db(
    urls = c("https://a.com", "https://b.com"),
    status = c("200", "301"),
    message = c("OK", "Moved Permanently"),
    new = c("", "https://b.org")
  )
  db <- add_from(db, list("DESCRIPTION", "man/foo.Rd"))
  state <- list(urlchecker = db)
  result <- CHECKS$urlchecker_ok$check(state)
  expect_true(result$status)
})

test_that("urlchecker_ok passes with empty result", {
  db <- make_urlchecker_db()
  state <- list(urlchecker = db)
  result <- CHECKS$urlchecker_ok$check(state)
  expect_true(result$status)
})

test_that("urlchecker_ok returns NA on prep failure", {
  state <- list(urlchecker = structure("error", class = "try-error"))
  result <- CHECKS$urlchecker_ok$check(state)
  expect_true(is.na(result$status))
})

# -- urlchecker_no_redirects --------------------------------------------------

test_that("urlchecker_no_redirects fails when URLs redirect", {
  db <- make_urlchecker_db(
    urls = c("https://old.com", "https://fine.com"),
    status = c("200", "200"),
    message = c("OK", "OK"),
    new = c("https://new.com", "")
  )
  db <- add_from(db, list("DESCRIPTION", "man/bar.Rd"))
  state <- list(urlchecker = db)
  result <- CHECKS$urlchecker_no_redirects$check(state)
  expect_false(result$status)
  expect_length(result$positions, 1)
  expect_identical(result$positions[[1]]$line, "https://old.com")
})

test_that("urlchecker_no_redirects passes when no redirects", {
  db <- make_urlchecker_db(
    urls = "https://fine.com",
    status = "404",
    message = "Not Found",
    new = ""
  )
  db <- add_from(db, list("DESCRIPTION"))
  state <- list(urlchecker = db)
  result <- CHECKS$urlchecker_no_redirects$check(state)
  expect_true(result$status)
})

test_that("urlchecker_no_redirects passes with empty result", {
  db <- make_urlchecker_db()
  state <- list(urlchecker = db)
  result <- CHECKS$urlchecker_no_redirects$check(state)
  expect_true(result$status)
})

test_that("urlchecker_no_redirects returns NA on prep failure", {
  state <- list(urlchecker = structure("error", class = "try-error"))
  result <- CHECKS$urlchecker_no_redirects$check(state)
  expect_true(is.na(result$status))
})

# -- position reporting -------------------------------------------------------

test_that("positions report filename from From column", {
  db <- make_urlchecker_db(
    urls = "https://broken.com",
    status = "500",
    message = "Server Error",
    new = ""
  )
  db <- add_from(db, list(c("man/foo.Rd", "R/bar.R")))
  state <- list(urlchecker = db)
  result <- CHECKS$urlchecker_ok$check(state)
  expect_identical(result$positions[[1]]$filename, "man/foo.Rd")
})

# -- multiple failures ---------------------------------------------------------

test_that("urlchecker_ok reports all broken URLs", {
  db <- make_urlchecker_db(
    urls = c("https://a.com", "https://b.com", "https://c.com"),
    status = c("404", "200", "500"),
    message = c("Not Found", "OK", "Server Error"),
    new = c("", "", "")
  )
  db <- add_from(db, list("DESCRIPTION", "DESCRIPTION", "man/foo.Rd"))
  state <- list(urlchecker = db)
  result <- CHECKS$urlchecker_ok$check(state)
  expect_false(result$status)
  expect_length(result$positions, 2)
  urls <- vapply(result$positions, `[[`, "", "line")
  expect_setequal(urls, c("https://a.com", "https://c.com"))
})

test_that("urlchecker_no_redirects reports all redirecting URLs", {
  db <- make_urlchecker_db(
    urls = c("https://a.com", "https://b.com", "https://c.com"),
    status = c("200", "200", "200"),
    message = c("OK", "OK", "OK"),
    new = c("https://a.org", "", "https://c.org")
  )
  db <- add_from(db, list("DESCRIPTION", "DESCRIPTION", "man/foo.Rd"))
  state <- list(urlchecker = db)
  result <- CHECKS$urlchecker_no_redirects$check(state)
  expect_false(result$status)
  expect_length(result$positions, 2)
})

# -- filter edge cases ---------------------------------------------------------

test_that("urlchecker_ok treats 302 as acceptable", {
  db <- make_urlchecker_db(
    urls = "https://a.com",
    status = "302",
    message = "Found",
    new = "https://a.org"
  )
  db <- add_from(db, list("DESCRIPTION"))
  state <- list(urlchecker = db)
  result <- CHECKS$urlchecker_ok$check(state)
  expect_true(result$status)
})

test_that("urlchecker_ok catches timeout status", {
  db <- make_urlchecker_db(
    urls = "https://slow.com",
    status = "Timeout",
    message = "Connection timed out",
    new = ""
  )
  db <- add_from(db, list("DESCRIPTION"))
  state <- list(urlchecker = db)
  result <- CHECKS$urlchecker_ok$check(state)
  expect_false(result$status)
})

# -- prep tests ----------------------------------------------------------------

test_that("PREPS$urlchecker stores result in state", {
  local_mocked_bindings(has_internet = function() TRUE)
  local_mocked_bindings(
    url_check = function(path, ...) {
      make_urlchecker_db(
        urls = "https://example.com",
        status = "200",
        message = "OK",
        new = ""
      )
    },
    .package = "urlchecker"
  )
  state <- list(path = "good")
  state <- PREPS$urlchecker(state, quiet = TRUE)
  expect_false(inherits(state$urlchecker, "try-error"))
  expect_s3_class(state$urlchecker, "data.frame")
})

test_that("PREPS$urlchecker warns on failure", {
  local_mocked_bindings(has_internet = function() TRUE)
  local_mocked_bindings(
    url_check = function(path, ...) stop("url_check failed"),
    .package = "urlchecker"
  )
  state <- list(path = "good")
  expect_warning(
    state <- PREPS$urlchecker(state, quiet = TRUE),
    "Prep step for"
  )
  expect_true(inherits(state$urlchecker, "try-error"))
})

# -- gp() integration with mocked prep ----------------------------------------

mock_url_check <- function(db) {
  function(path, ...) db
}

test_that("urlchecker_ok passes through gp() with no problems", {
  db <- make_urlchecker_db(
    urls = "https://example.com",
    status = "200",
    message = "OK",
    new = ""
  )
  db <- add_from(db, list("DESCRIPTION"))
  local_mocked_bindings(has_internet = function() TRUE)
  local_mocked_bindings(url_check = mock_url_check(db), .package = "urlchecker")
  gp_res <- gp("good", checks = "urlchecker_ok")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "urlchecker_ok"])
})

test_that("urlchecker_ok fails through gp() with broken URLs", {
  db <- make_urlchecker_db(
    urls = "https://broken.com",
    status = "404",
    message = "Not Found",
    new = ""
  )
  db <- add_from(db, list("DESCRIPTION"))
  local_mocked_bindings(has_internet = function() TRUE)
  local_mocked_bindings(url_check = mock_url_check(db), .package = "urlchecker")
  gp_res <- gp("good", checks = "urlchecker_ok")
  res <- results(gp_res)
  expect_false(res$passed[res$check == "urlchecker_ok"])
  pos <- failed_positions(gp_res)$urlchecker_ok
  expect_length(pos, 1)
  expect_identical(pos[[1]]$line, "https://broken.com")
})

test_that("urlchecker_no_redirects fails through gp() with redirects", {
  db <- make_urlchecker_db(
    urls = "https://old.com",
    status = "200",
    message = "OK",
    new = "https://new.com"
  )
  db <- add_from(db, list("DESCRIPTION"))
  local_mocked_bindings(has_internet = function() TRUE)
  local_mocked_bindings(url_check = mock_url_check(db), .package = "urlchecker")
  gp_res <- gp("good", checks = "urlchecker_no_redirects")
  res <- results(gp_res)
  expect_false(res$passed[res$check == "urlchecker_no_redirects"])
})

test_that("urlchecker_no_redirects passes through gp() with no redirects", {
  db <- make_urlchecker_db(
    urls = "https://fine.com",
    status = "200",
    message = "OK",
    new = ""
  )
  db <- add_from(db, list("DESCRIPTION"))
  local_mocked_bindings(has_internet = function() TRUE)
  local_mocked_bindings(url_check = mock_url_check(db), .package = "urlchecker")
  gp_res <- gp("good", checks = "urlchecker_no_redirects")
  res <- results(gp_res)
  expect_true(res$passed[res$check == "urlchecker_no_redirects"])
})

test_that("urlchecker checks return NA through gp() on prep failure", {
  local_mocked_bindings(has_internet = function() TRUE)
  local_mocked_bindings(
    url_check = function(path, ...) stop("boom"),
    .package = "urlchecker"
  )
  expect_warning(
    gp_res <- gp(
      "good", checks = c("urlchecker_ok", "urlchecker_no_redirects")
    ),
    "Prep step for"
  )
  res <- results(gp_res)
  expect_true(is.na(res$passed[res$check == "urlchecker_ok"]))
  expect_true(is.na(res$passed[res$check == "urlchecker_no_redirects"]))
})

test_that("urlchecker checks pass through gp() with empty db", {
  db <- make_urlchecker_db()
  local_mocked_bindings(has_internet = function() TRUE)
  local_mocked_bindings(url_check = mock_url_check(db), .package = "urlchecker")
  gp_res <- gp("good", checks = c("urlchecker_ok", "urlchecker_no_redirects"))
  res <- results(gp_res)
  expect_true(res$passed[res$check == "urlchecker_ok"])
  expect_true(res$passed[res$check == "urlchecker_no_redirects"])
})

# -- offline gate --------------------------------------------------------------

test_that("urlchecker checks return NA when offline", {
  local_mocked_bindings(has_internet = function() FALSE)
  expect_warning(
    gp_res <- gp("good", checks = "urlchecker_ok"),
    "Skipping URL checks: no internet connection"
  )
  res <- results(gp_res)
  expect_true(is.na(res$passed[res$check == "urlchecker_ok"]))
})

# -- integration tests (network) ----------------------------------------------

test_that("urlchecker prep runs on good fixture", {
  skip_on_cran()
  skip_if_offline()
  gp_res <- gp("good", checks = "urlchecker_no_redirects")
  res <- results(gp_res)
  result <- res$passed[res$check == "urlchecker_no_redirects"]
  expect_false(result)

  pos <- failed_positions(gp_res)$urlchecker_no_redirects
  urls <- vapply(pos, `[[`, "", "line")
  expect_true(any(grepl("mangothecat", urls)))
})

test_that("urlchecker_ok passes on package with valid URLs", {
  skip_on_cran()
  skip_if_offline()
  gp_res <- gp("good", checks = "urlchecker_ok")
  res <- results(gp_res)
  result <- res$passed[res$check == "urlchecker_ok"]
  expect_true(result)
})

#' @include lists.R prep_utils.R
#' @importFrom lintr lint_package

PREPS$tidyverse <- function(state, path = state$path, quiet) {
  path <- normalizePath(path)
  excl <- as.list(state$exclude_path %||% character())
  run_prep_step(state, "tidyverse_lintr", function(path) {
    suppressMessages(lint_package(path, exclusions = excl))
  }, path = path, silent = quiet)
}

#' @include lists.R

CHECKS$no_description_depends <- make_check(

  description = 'No "Depends" in DESCRIPTION',
  tags = c("info", "DESCRIPTION"),
  preps = "description",

  gp = 'not use {.field Depends} in {.file DESCRIPTION}, as it can cause name
        clashes, and poor interaction with other packages. Use {.field Imports}
        instead.',

  check = function(state) {
    if(inherits(state$description, "try-error")) return(na_result())

    deps <- state$description$get_deps()
    ## Remove 'methods' and R, these are OK
    deps <- deps[! deps$package %in% c("methods", "R"), , drop = FALSE]
    check_result(! 'Depends' %in% deps$type)
  }
)

## --------------------------------------------------------------------

CHECKS$no_description_date <- make_check(

  description = 'No "Date" in DESCRIPTION',
  tags = c("info", "DESCRIPTION"),
  preps = "description",

  gp = 'omit the {.field Date} field in {.file DESCRIPTION}.
        It is not required and it gets invalid quite often.
        A build date will be added to the package when you
        perform {.code R CMD build} on it.',

  check = function(state) {
    if(inherits(state$description, "try-error")) return(na_result())

    check_result(! state$description$has_fields('Date'))
  }
)

## --------------------------------------------------------------------

CHECKS$description_url <- make_check(

  description = 'URL in DESCRIPTION',
  tags = c("info", 'DESCRIPTION'),
  preps = 'description',

  gp = 'add a {.field URL} field to {.file DESCRIPTION}.
        It helps users find information about your package
        online. If your package does not have a homepage,
        add an URL to GitHub, or the CRAN package page.',

  check = function(state) {
    if(inherits(state$description, "try-error")) return(na_result())

    check_result(state$description$has_fields("URL"))
  }
)

## --------------------------------------------------------------------

CHECKS$description_not_start_with_package <- make_check(

  description = 'Description does not start with package name reference',
  tags = c("info", "DESCRIPTION"),
  preps = "description",

  gp = 'not start the {.field Description} field by referring to the package
        itself, e.g. "This package ...", "This is a package ...",
        or "The FooBar package ...". Describe what the package does
        directly, e.g. "Provides tools for ..." instead of
        "This package provides tools for ...".',

  check = function(state) {
    if(inherits(state$description, "try-error")) return(na_result())

    desc_text <- state$description$get_field("Description")
    pkg_name <- state$description$get_field("Package")
    starts_this_pkg <- grepl(
      "^This\\s+(is\\s+a\\s+)?package\\b",
      desc_text, ignore.case = TRUE
    )
    starts_the_pkg <- grepl(
      paste0("^The\\s+", pkg_name, "\\s+package\\b"),
      desc_text, ignore.case = TRUE
    )
    check_result(!(starts_this_pkg || starts_the_pkg))
  }
)

## --------------------------------------------------------------------

CHECKS$description_urls_in_angle_brackets <- make_check(

  description = "URLs in Description are enclosed in angle brackets",
  tags = c("info", "DESCRIPTION"),
  preps = "description",

  gp = 'enclose URLs in the {.field Description} field in angle brackets,
        e.g. <https://example.com>. CRAN requires angle brackets
        around URLs for auto-linking.',

  check = function(state) {
    if(inherits(state$description, "try-error")) return(na_result())

    desc_text <- state$description$get_field("Description")
    bare_url <- "(?<!<)(https?://[^\\s,)>]+)"
    check_result(!grepl(bare_url, desc_text, perl = TRUE))
  }
)

## --------------------------------------------------------------------

CHECKS$description_doi_format <- make_check(

  description = "DOIs in Description use <doi:...> format",
  tags = c("info", "DESCRIPTION"),
  preps = "description",

  gp = 'use the {.code <doi:10.prefix/suffix>} format for DOIs in the
        {.field Description} field instead of a full URL like
        {.code <https://doi.org/10.prefix/suffix>}. CRAN requires
        the {.code <doi:...>} notation for auto-linking.',

  check = function(state) {
    if(inherits(state$description, "try-error")) return(na_result())

    desc_text <- state$description$get_field("Description")
    check_result(!grepl("https?://doi\\.org/", desc_text))
  }
)

## --------------------------------------------------------------------

CHECKS$description_urls_not_http <- make_check(

  description = "URLs in Description use https not http",
  tags = c("info", "DESCRIPTION"),
  preps = "description",

  gp = 'consider using {.code https://} instead of
        {.code http://} for URLs in the {.field Description}
        field. CRAN prefers secure URLs where available.
        Check if there might be a secure
        {.code https://} alternative.',

  check = function(state) {
    if(inherits(state$description, "try-error")) return(na_result())

    desc_text <- state$description$get_field("Description")
    check_result(!grepl("http://", desc_text, fixed = TRUE))
  }
)

## --------------------------------------------------------------------

CHECKS$no_description_duplicate_deps <- make_check(

  description = "No duplicate dependencies in DESCRIPTION",
  tags = c("info", "DESCRIPTION"),
  preps = "description",

  gp = 'not list the same package in multiple dependency fields
        (e.g. both {.field Imports} and
        {.field Suggests}). Each dependency should appear
        only once.',

  check = function(state) {
    if(inherits(state$description, "try-error")) return(na_result())

    deps <- state$description$get_deps()
    deps <- deps[deps$package != "R" & deps$type != "LinkingTo", , drop = FALSE]
    check_result(!anyDuplicated(deps$package))
  }
)

## --------------------------------------------------------------------

CHECKS$description_valid_roles <- make_check(

  description = "Authors have valid roles in DESCRIPTION",
  tags = c("info", "DESCRIPTION"),
  preps = "description",

  gp = 'use only standard R author roles in {.field Authors@R}. Valid roles are:
        aut (author), com (compiler), cph (copyright holder),
        cre (maintainer), ctb (contributor), ctr (contractor),
        dtc (data contributor), fnd (funder), rev (reviewer),
        ths (thesis advisor), trl (translator).',

  check = function(state) {
    if(inherits(state$description, "try-error")) return(na_result())

    authors_r <- state$description$get_field("Authors@R", default = NA)
    if (is.na(authors_r)) return(na_result())

    valid_roles <- tryCatch(utils:::MARC_relator_db$code, error = function(e) NULL)
    if (is.null(valid_roles)) return(na_result())

    roles_env <- new.env(parent = emptyenv())
    roles_env$roles <- character()
    capture_person <- function(..., role = NULL) {
      roles_env$roles <- c(roles_env$roles, role)
      NULL
    }
    env <- new.env(parent = baseenv())
    env$person <- capture_person
    env$c <- c

    ok <- tryCatch({
      eval(parse(text = authors_r), envir = env)
      TRUE
    }, error = function(e) FALSE)
    if (!ok) return(na_result())

    check_result(all(roles_env$roles %in% valid_roles))
  }
)

## --------------------------------------------------------------------

CHECKS$description_pkgname_single_quoted <- make_check(

  description = "Package names in Title/Description are single-quoted",
  tags = c("info", "DESCRIPTION"),
  preps = "description",

  gp = 'single-quote package names in the
        {.field Title} and {.field Description} fields,
        e.g. {.pkg dplyr} not dplyr. CRAN requires that
        software names are single-quoted.',

  check = function(state) {
    if(inherits(state$description, "try-error")) return(na_result())

    deps <- state$description$get_deps()
    pkg_names <- deps$package[deps$package != "R"]
    # Base packages shipped with R (alphabetical, matching r-source).
    # Excluded: "translations" (not a loadable package).
    # https://github.com/wch/r-source/tree/trunk/src/library
    base_pkgs <- c(
      "base", "compiler", "datasets", "grDevices",
      "graphics", "grid", "methods", "parallel",
      "splines", "stats", "stats4", "tcltk",
      "tools", "utils"
    )
    pkg_names <- setdiff(pkg_names, base_pkgs)
    if (length(pkg_names) == 0) return(check_result(TRUE))

    title <- state$description$get_field("Title")
    desc_text <- state$description$get_field("Description")
    text <- paste(title, desc_text)

    for (pkg in pkg_names) {
      unquoted <- paste0("(?<!')\\b", pkg, "\\b(?!')")
      if (grepl(unquoted, text, perl = TRUE)) return(check_result(FALSE))
    }
    check_result(TRUE)
  }
)

## --------------------------------------------------------------------

CHECKS$description_bugreports <- make_check(

  description = "BugReports in DESCRIPTION",
  tags = c("info", 'DESCRIPTION'),
  preps = 'description',

  gp = 'add a {.field BugReports} field to
        {.file DESCRIPTION}, and point it to a bug tracker.
        Many online code hosting services provide bug
        trackers for free, {.url https://github.com},
        {.url https://gitlab.com}, etc.',

  check = function(state) {
    if(inherits(state$description, "try-error")) return(na_result())

    check_result(state$description$has_fields('BugReports'))
  }
)

## --------------------------------------------------------------------

#' Run good practice checks
#'
#' To see the results, just print it to the screen.
#'
#' @param path Path to a package root.
#' @param checks Character vector, the checks to run. Defaults to
#'   \code{\link{default_checks}}. Use \code{\link{all_checks}} to list all
#'   checks, or add optional sets like \code{\link{tidyverse_checks}}.
#'   When \code{NULL}, all registered checks are run, subject to any
#'   exclusions from \code{goodpractice.exclude_check_groups} or
#'   \code{GP_EXCLUDE_CHECK_GROUPS}.
#' @param extra_preps Custom preparation functions. See
#'   \code{\link{make_prep}} on creating preparation functions.
#' @param extra_checks Custom checks. See \code{\link{make_check}} on
#'   creating checks.
#' @param quiet Whether to suppress output from the preparation
#'   functions. Note that not all preparation functions produce output,
#'   even if this option is set to \code{FALSE}.
#' @return A goodpractice object that you can query
#'   with a simple API. See \code{\link{results}} to start.
#'
#' @section Excluding check groups:
#' When using the default \code{checks = all_checks()}, entire groups of
#' checks can be excluded by group name via the
#' \code{goodpractice.exclude_check_groups} option or the
#' \code{GP_EXCLUDE_CHECK_GROUPS} environment variable (comma-separated).
#' The option takes precedence.
#'
#' \preformatted{
#' # Skip URL and coverage checks:
#' options(goodpractice.exclude_check_groups = c("urlchecker", "covr"))
#'
#' # Or via environment variable:
#' Sys.setenv(GP_EXCLUDE_CHECK_GROUPS = "urlchecker,covr")
#' }
#'
#' Exclusion only applies when \code{checks = NULL} (the default).
#' Explicit \code{checks} arguments are never filtered.
#'
#' @section Excluding files:
#' Specific files can be excluded from checks via the
#' \code{goodpractice.exclude_path} option or the \code{GP_EXCLUDE_PATH}
#' environment variable (comma-separated). Paths are relative to the
#' package root.
#'
#' \preformatted{
#' options(goodpractice.exclude_path = c("R/RcppExports.R", "R/generated.R"))
#'
#' # Or via environment variable:
#' Sys.setenv(GP_EXCLUDE_PATH = "R/RcppExports.R,R/generated.R")
#' }
#'
#' Excluded files are skipped by lintr, treesitter, expression, and
#' roxygen2 checks.
#'
#' @section Parallel preparation:
#' Preparation steps run sequentially by default. To run them in
#' parallel, install \pkg{future.apply} and set a
#' \code{\link[future]{plan}}:
#'
#' \preformatted{
#' future::plan("multisession")
#' gp(".")
#' }
#'
#' Preps run in parallel only when a non-sequential plan is active.
#' Prep functions must be independent: in parallel mode each prep
#' receives the initial state snapshot, so a prep cannot read another
#' prep's output. Only new state fields are merged back; if two preps
#' write the same field, the second is dropped with a warning.
#'
#' @family main
#' @export
#' @aliases goodpractice
#' @importFrom desc desc_get
#' @examples
#' path <- system.file("bad1", package = "goodpractice")
#' # Run a subset of all checks available
#' g <- gp(path, checks = all_checks()[9:16])
#' g
#' # Or run with named check groups
#' g <- gp(path, checks = checks_by_group("description", "namespace"))

gp <- function(
  path = ".",
  checks = default_checks(),
  extra_preps = NULL,
  extra_checks = NULL,
  quiet = TRUE
) {
  pkgname <- validate_pkg_path(path)
  cli::cat_rule(
    paste("Preparing goodpractice for", pkgname),
    col = "cyan"
  )

  MYCHECKS <- prepare_checks(CHECKS, extra_checks)
  MYPREPS <- prepare_preps(PREPS, extra_preps)
  checks <- resolve_checks(checks, MYCHECKS)
  preps <- required_preps(checks, MYCHECKS)

  # No pipe here:
  state <- init_state(path, pkgname, extra_preps, extra_checks)
  state <- run_preps(state, preps, MYPREPS, quiet)
  state <- run_checks(state, checks, MYCHECKS)

  state$.cache <- NULL
  class(state) <- "goodPractice"
  state
}

validate_pkg_path <- function(path) {
  if (!file.exists(file.path(path, "DESCRIPTION"))) {
    cli::cli_abort(c(
      "{.path {path}} must be a package.",
      i = "Can't find {.file DESCRIPTION}."
    ))
  }
  desc_get("Package", file = file.path(path, "DESCRIPTION"))
}

resolve_checks <- function(checks, mychecks) {
  if (identical(checks, default_checks())) {
    checks <- exclude_checks_by_group(checks, mychecks)
  }
  checks
}

required_preps <- function(checks, mychecks) {
  unique(unlist(lapply(mychecks[checks], "[[", "preps")))
}

init_state <- function(path, pkgname, extra_preps, extra_checks) {
  list(
    path = path,
    package = pkgname,
    extra_preps = extra_preps,
    extra_checks = extra_checks,
    exclude_path = excluded_paths(),
    .cache = new.env(parent = emptyenv())
  )
}

run_preps <- function(state, preps, mypreps, quiet) {
  use_future <-
    requireNamespace("future.apply", quietly = TRUE) &&
    requireNamespace("future", quietly = TRUE) &&
    !inherits(future::plan(), "sequential")

  # nocov start - requires non-sequential future::plan()
  apply_fn <- if (use_future) {
    future.apply::future_lapply
  } else { # nocov end
    lapply
  }

  results <- apply_fn(preps, function(prep) {
    cli::cli_progress_step("Preparing: {prep}")
    result <- mypreps[[prep]](state, quiet = quiet)
    cli::cli_progress_done()
    result
  })

  # Preps must be independent: each receives the original state snapshot,
  # and only NEW fields (not already in state) are merged back. If two
  # preps write the same field, the second is dropped with a warning.
  for (res in results) {
    new_fields <- setdiff(names(res), names(state))
    conflict_fields <- intersect(names(res), names(state))
    init_fields <- c(
      "path", "package", "extra_preps",
      "extra_checks", "exclude_path", ".cache"
    )
    conflict_fields <- setdiff(conflict_fields, init_fields)
    if (length(conflict_fields) > 0) {
      cli::cli_warn(
        "Parallel prep conflict: field{?s} \\
        {.val {conflict_fields}} already set, skipping."
      )
    }
    for (field in new_fields) {
      state[[field]] <- res[[field]]
    }
  }

  state
}

run_checks <- function(state, checks, mychecks) {
  state$checks <- list()
  for (check in checks) {
    state$checks[[check]] <- mychecks[[check]]$check(state)
  }
  state
}

excluded_paths <- function() {
  opt <- getOption("goodpractice.exclude_path")
  if (!is.null(opt)) return(opt)
  env <- Sys.getenv("GP_EXCLUDE_PATH", "")
  parts <- strsplit(env, ",\\s*")[[1]]
  parts[nzchar(parts)]
}

excluded_check_groups <- function() {
  opt <- getOption("goodpractice.exclude_check_groups")
  if (!is.null(opt)) return(opt)
  env <- Sys.getenv("GP_EXCLUDE_CHECK_GROUPS", "")
  parts <- strsplit(env, ",\\s*")[[1]]
  parts[nzchar(parts)]
}

exclude_checks_by_group <- function(checks, mychecks) {
  exclude <- excluded_check_groups()
  if (length(exclude) == 0) return(checks)

  dominated <- vapply(checks, function(ch) {
    any(mychecks[[ch]]$preps %in% exclude)
  }, logical(1))

  if (any(dominated)) {
    cli::cli_alert_info("Excluding checks that depend on: {.val {exclude}}")
  }
  checks[!dominated]
}

check_passed <- function(chk, na_as_passed = FALSE) {
  status <- if ("status" %in% names(chk)) {
    chk$status
  } else {
    chk
  }

  if (na_as_passed) {
    isTRUE(status) || is.na(status)
  } else if (is.na(status)) {
    NA
  } else {
    isTRUE(status)
  }
}

check_failed <- function(chk, na_as_passed = FALSE) {
  !check_passed(chk, na_as_passed = na_as_passed)
}

check_type <- function(chk) {
  if (is.list(chk) && "type" %in% names(chk)) chk$type else "error"
}

#' @family main
#' @export goodpractice
goodpractice <- gp

test_that("gp errors when DESCRIPTION is missing", {
  fake_pkg <- withr::local_tempdir("not_a_pkg")
  expect_error(gp(fake_pkg), "must be a package")
})

# -- validate_pkg_path --------------------------------------------------------

test_that("validate_pkg_path returns package name", {
  bad1 <- system.file("bad1", package = "goodpractice")
  expect_identical(
    unname(validate_pkg_path(bad1)), "badpackage"
  )
})

test_that("validate_pkg_path errors for non-package path", {
  fake <- withr::local_tempdir()
  expect_error(validate_pkg_path(fake), "must be a package")
})

# -- resolve_checks -----------------------------------------------------------

test_that("resolve_checks returns checks when provided", {
  mychecks <- prepare_checks(CHECKS, NULL)
  result <- resolve_checks("has_readme", mychecks)
  expect_identical(result, "has_readme")
})

test_that("resolve_checks applies exclusions when NULL", {
  mychecks <- prepare_checks(CHECKS, NULL)
  withr::local_options(
    goodpractice.exclude_check_groups = "covr"
  )
  result <- suppressWarnings(
    resolve_checks(NULL, mychecks)
  )
  expect_false("covr" %in% result)
})

# -- required_preps -----------------------------------------------------------

test_that("required_preps returns unique prep names", {
  mychecks <- prepare_checks(CHECKS, NULL)
  preps <- required_preps("description_url", mychecks)
  expect_identical(preps, "description")
})

test_that("required_preps deduplicates across checks", {
  mychecks <- prepare_checks(CHECKS, NULL)
  checks <- c("description_url", "no_description_depends")
  preps <- required_preps(checks, mychecks)
  expect_identical(preps, "description")
})

# -- init_state ---------------------------------------------------------------

test_that("init_state creates proper state list", {
  bad1 <- system.file("bad1", package = "goodpractice")
  state <- init_state(bad1, "badpackage", NULL, NULL)
  expect_identical(state$path, bad1)
  expect_identical(unname(state$package), "badpackage")
  expect_true(is.environment(state$.cache))
})

# -- run_preps ----------------------------------------------------------------

test_that("run_preps executes prep and adds field to state", {
  bad1 <- system.file("bad1", package = "goodpractice")
  state <- init_state(bad1, "badpackage", NULL, NULL)
  mypreps <- prepare_preps(PREPS, NULL)
  state <- run_preps(state, "description", mypreps, quiet = TRUE)
  expect_false(is.null(state$description))
})

# -- run_checks ---------------------------------------------------------------

test_that("run_checks populates state$checks", {
  bad1 <- system.file("bad1", package = "goodpractice")
  state <- init_state(bad1, "badpackage", NULL, NULL)
  mypreps <- prepare_preps(PREPS, NULL)
  mychecks <- prepare_checks(CHECKS, NULL)
  state <- run_preps(
    state, "description", mypreps, quiet = TRUE
  )
  state <- run_checks(state, "description_url", mychecks)
  expect_true("description_url" %in% names(state$checks))
})

test_that("check_passed handles list with status field", {
  expect_true(check_passed(list(status = TRUE)))
  expect_false(check_passed(list(status = FALSE)))
  expect_true(is.na(check_passed(list(status = NA))))
})

test_that("check_passed handles bare logical", {
  expect_true(check_passed(TRUE))
  expect_false(check_passed(FALSE))
})

test_that("check_passed na_as_passed treats NA as pass", {
  expect_true(check_passed(list(status = NA), na_as_passed = TRUE))
  expect_true(check_passed(NA, na_as_passed = TRUE))
})

test_that("check_failed is the inverse of check_passed", {
  expect_true(check_failed(list(status = FALSE)))
  expect_false(check_failed(list(status = TRUE)))
})

# -- exclude_checks_by_group -------------------------------------------------

test_that("option excludes checks by group name", {
  bad1 <- system.file("bad1", package = "goodpractice")
  excl <- c("covr", "rcmdcheck", "cyclocomp")
  withr::local_options(goodpractice.exclude_check_groups = excl)
  gp_res <- gp(bad1)
  ptn <- paste0("^(", paste0(excl, collapse = "|"), ")")
  expect_false(any(grepl(ptn, checks(gp_res))))
})

test_that("envvar excludes checks by group name", {
  bad1 <- system.file("bad1", package = "goodpractice")
  withr::local_options(goodpractice.exclude_check_groups = NULL)
  excl <- c("covr", "rcmdcheck", "cyclocomp")
  withr::local_envvar(GP_EXCLUDE_CHECK_GROUPS = paste0(excl,collapse=","))
  gp_res <- gp(bad1)
  ptn <- paste0("^(", paste0(excl, collapse = "|"), ")")
  expect_false(any(grepl(ptn, checks(gp_res))))
})

test_that("option takes precedence over envvar", {
  bad1 <- system.file("bad1", package = "goodpractice")
  withr::local_options(goodpractice.exclude_check_groups = c("covr", "cyclocomp", "rcmdcheck"))
  withr::local_envvar(GP_EXCLUDE_CHECK_GROUPS = "lintr")
  gp_res <- gp(bad1)
  expect_false("covr" %in% checks(gp_res))
  expect_false("cyclocomp" %in% checks(gp_res))
  expect_true(any(grepl("^lintr_", checks(gp_res))))
})

test_that("non-default checks argument overrides exclusion", {
  bad1 <- system.file("bad1", package = "goodpractice")
  withr::local_options(goodpractice.exclude_check_groups = "description")
  gp_res <- gp(bad1, checks = "no_description_depends")
  expect_true("no_description_depends" %in% checks(gp_res))
})

test_that("empty exclusion returns checks unchanged", {
  withr::local_options(goodpractice.exclude_check_groups = NULL)
  withr::local_envvar(GP_EXCLUDE_CHECK_GROUPS = "")
  checks <- c("no_description_depends", "covr")
  result <- exclude_checks_by_group(checks, CHECKS)
  expect_identical(result, checks)
})

test_that(".cache is removed from gp result", {
  gp_res <- gp("good", checks = "has_news")
  expect_null(gp_res$.cache)
})

test_that("parallel merge warns on conflicting fields", {
  state <- list(path = "good", package = "goodpackage",
                extra_preps = NULL, extra_checks = NULL,
                exclude_path = character(), .cache = NULL)

  prep_a <- function(state, quiet) {
    state$my_field <- "from_a"
    state
  }
  prep_b <- function(state, quiet) {
    state$my_field <- "from_b"
    state
  }

  preps <- list(a = prep_a, b = prep_b)
  expect_warning(
    result <- run_preps(state, c("a", "b"), preps, quiet = TRUE),
    "Parallel prep conflict"
  )
  expect_equal(result$my_field, "from_a")
})

## ----setup, include = FALSE---------------------------------------------------
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)

## ----windowns-flag, echo = FALSE----------------------------------------------
# The bash commands are likely to fail on winddows:
on_windows <- identical(.Platform$OS.type, "windows")

## ----echo = FALSE, eval = TRUE, output = TRUE---------------------------------
cli::cli_alert_success("No problems found")

#' @include lists.R
#' @importFrom spelling spell_check_package

PREPS$spelling <- function(state, path = state$path, quiet) {
  wordlist <- file.path(path, "inst", "WORDLIST")
  if (!file.exists(wordlist)) {
    state$spelling <- "no_wordlist"
    if (!quiet) {
      cli::cli_inform(
        "Skipping spelling check: no {.file inst/WORDLIST} found."
      )
    }
    return(state)
  }
  run_prep_step(state, "spelling", function(path) {
    suppressMessages(spell_check_package(path))
  }, path = path, silent = quiet)
}

#' @include lists.R

#' @noRd
cyclocomp_limit <- function() {
  getOption("goodpractice.cyclocomp_limit", 15)
}

CHECKS$cyclocomp <- make_check(

  description = "Functions are simple",
  tags = c("info", "code complexity"),
  preps = "cyclocomp",

  gp = function(state) {
    limit <- cyclocomp_limit()
    cyc <- state$cyclocomp
    long <- which(cyc$cyclocomp > limit)
    funcs <- paste0(
      cyc$name[long],
      " (", cyc$cyclocomp[long], ")",
      collapse = ", "
    )
    paste0(
      "write short and simple functions.",
      " These functions have high",
      " cyclomatic complexity (>", limit, "): ",
      funcs, ". ",
      "You can make them easier to reason about",
      " by encapsulating distinct steps",
      " of your function into subfunctions."
    )
  },

  check = function(state) {
    if (inherits(state$cyclocomp, "try-error")) return(na_result())
    check_result(all(state$cyclocomp$cyclocomp <= cyclocomp_limit()))
  }
)

#' @include lists.R

PREPS$revdep <- function(state, path = state$path, quiet) {
  if (!curl::has_internet()) {
    state$revdep <- NA
    if (!quiet) {
      cli::cli_warn(
        "Prep step for {.val revdep} skipped: no internet connection."
      )
    }
    return(state)
  }

  repos <- getOption("repos")["CRAN"]
  if (is.na(repos) || repos == "@CRAN@") {
    repos <- "https://cloud.r-project.org"
  }

  run_prep_step(state, "revdep", function(repos) {
    utils::available.packages(repos = repos)
  }, repos = repos, silent = quiet)
}

#' @include lists.R prep_utils.R treesitter.R
#' @importFrom roxygen2 parse_package
#' @details See #295 for updates to Roxygen2 DESC configs, thanks to @wkmor1
#' @noRd
uses_roxygen2 <- function(path) {
  desc_path <- file.path(path, "DESCRIPTION")
  if (!file.exists(desc_path)) return(FALSE)
  fields <- names(read.dcf(desc_path)[1, ])
  roxygen_ptn_to_v7 <- "^(\\s*?)Roxygen"
  roxygen_ptn_v8 <- "^(\\s*?)Config\\/roxygen"
  ptn <- paste0 (roxygen_ptn_to_v7, "|", roxygen_ptn_v8)
  any(grepl(ptn, fields))
}

find_function_defs <- function(path, exclude_path = character()) {
  ts <- ts_parse(path, exclude_path)
  if (length(ts$functions) == 0) {
    return(data.frame(name = character(), file = character(), line = integer()))
  }
  data.frame(
    name = vapply(ts$functions, `[[`, "", "name"),
    file = vapply(ts$functions, `[[`, "", "file"),
    line = vapply(ts$functions, function(fn) fn$line, numeric(1))
  )
}

parse_roxygen2 <- function(path, exclude_path = character()) {
  if (!uses_roxygen2(path)) {
    cli::cli_abort("Package does not use {.pkg roxygen2}.")
  }

  rfiles <- r_package_files(path, exclude_path)

  parse_messages <- character()
  blocks <- withCallingHandlers(
    {
      bl <- lapply(rfiles, roxygen2::parse_file)
      do.call(c, bl)
    },
    message = function(m) {
      msg <- conditionMessage(m)
      if (grepl("is not a known tag", msg)) {
        parse_messages <<- c(parse_messages, msg)
        invokeRestart("muffleMessage")
      }
    }
  )

  ns <- tryCatch(
    parseNamespaceFile(basename(path), dirname(path)),
    error = function(e) {
      list(
        exports = character(),
        S3methods = matrix(character(), ncol = 3, nrow = 0)
      )
    }
  )

  s3methods <- ns_s3_method_names(ns)

  list(
    blocks = blocks,
    namespace_exports = ns$exports,
    namespace_s3methods = s3methods,
    function_defs = find_function_defs(path, exclude_path),
    parse_messages = parse_messages
  )
}

PREPS$roxygen2 <- function(state, path = state$path, quiet) {
  run_prep_step(state, "roxygen2", function(path) {
    parse_roxygen2(path, state$exclude_path %||% character())
  }, path = path, silent = quiet)
}

# -- uses_roxygen2 detection --------------------------------------------------

test_that("uses_roxygen2 returns TRUE for roxygen2 packages", {
  expect_true(uses_roxygen2("bad_roxygen"))
})

test_that("uses_roxygen2 returns FALSE for non-roxygen2 packages", {
  expect_false(uses_roxygen2("no_roxygen"))
})

# -- prep skips non-roxygen2 packages -----------------------------------------

test_that("roxygen2 checks return NA for non-roxygen2 packages", {
  expect_warning(
    gp_res <- gp("no_roxygen", checks = "roxygen2_unknown_tags"),
    "Prep step for"
  )
  expect_true(is.na(results(gp_res)$passed))
})

# -- roxygen2_has_export_or_nord ----------------------------------------------

test_that("roxygen2_has_export_or_nord flags documented fns missing tags", {
  gp_res <- gp("bad_roxygen", checks = "roxygen2_has_export_or_nord")
  expect_false(results(gp_res)$passed)

  pos <- failed_positions(gp_res)$roxygen2_has_export_or_nord
  lines <- vapply(pos, `[[`, "", "line")
  expect_match(lines, "documented_no_tag", all = FALSE)
})

test_that("roxygen2_has_export_or_nord ignores undocumented functions", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: undoctest", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT",
    "Roxygen: list(markdown = TRUE)",
    "RoxygenNote: 7.3.3"
  ), file.path(pkg, "DESCRIPTION"))
  writeLines("export(public_fn)", file.path(pkg, "NAMESPACE"))

  writeLines(c(
    "#' @export",
    "public_fn <- function() 1",
    "",
    "internal_fn <- function() 2"
  ), file.path(pkg, "R", "code.R"))

  gp_res <- gp(pkg, checks = "roxygen2_has_export_or_nord")
  expect_true(results(gp_res)$passed)
})

test_that("roxygen2_has_export_or_nord passes when all tagged", {
  gp_res <- gp("good", checks = "roxygen2_has_export_or_nord")
  expect_true(results(gp_res)$passed)
})

test_that("roxygen2_has_export_or_nord skips @noRd operator functions", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: optest", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT",
    "Roxygen: list(markdown = TRUE)",
    "RoxygenNote: 7.3.3"
  ), file.path(pkg, "DESCRIPTION"))
  writeLines("export(my_fn)", file.path(pkg, "NAMESPACE"))

  writeLines(c(
    "#' @export",
    "my_fn <- function() 1",
    "",
    "#' @noRd",
    "`%op%` <- function(a, b) a + b"
  ), file.path(pkg, "R", "code.R"))

  gp_res <- gp(pkg, checks = "roxygen2_has_export_or_nord")
  expect_true(results(gp_res)$passed)
})


# -- roxygen2_unknown_tags ----------------------------------------------------

test_that("roxygen2_unknown_tags fails on deprecated/unknown tags", {
  gp_res <- gp("bad_roxygen", checks = "roxygen2_unknown_tags")
  expect_false(results(gp_res)$passed)

  pos <- failed_positions(gp_res)$roxygen2_unknown_tags
  lines <- vapply(pos, `[[`, "", "line")
  expect_match(lines, "@S3method", all = FALSE)
})

test_that("roxygen2_unknown_tags passes when all tags are valid", {
  gp_res <- gp("good", checks = "roxygen2_unknown_tags")
  expect_true(results(gp_res)$passed)
})

# -- roxygen2_valid_inherit ---------------------------------------------------

test_that("roxygen2_valid_inherit fails on nonexistent reference", {
  gp_res <- gp("bad_roxygen", checks = "roxygen2_valid_inherit")
  expect_false(results(gp_res)$passed)

  pos <- failed_positions(gp_res)$roxygen2_valid_inherit
  lines <- vapply(pos, `[[`, "", "line")
  expect_match(lines, "bad_inherit_func", all = FALSE)
})

test_that("roxygen2_valid_inherit passes with valid references", {
  gp_res <- gp("good", checks = "roxygen2_valid_inherit")
  expect_true(results(gp_res)$passed)
})

# -- prep error handling ------------------------------------------------------

test_that("all roxygen2 checks return na_result on prep failure", {
  state <- list(roxygen2 = structure("error", class = "try-error"))

  for (check_name in c(
    "roxygen2_has_export_or_nord",
    "roxygen2_unknown_tags",
    "roxygen2_valid_inherit",
    "roxygen2_duplicate_params"
  )) {
    result <- CHECKS[[check_name]]$check(state)
    expect_true(is.na(result$status), label = paste(check_name, "status"))
    expect_type(result$positions, "list")
  }
})

# -- block_is_function edge cases ---------------------------------------------

test_that("block_is_function returns FALSE for non-assignment operator", {
  block <- list(call = quote(x + y))
  expect_false(block_is_function(block))
})

test_that("block_is_function returns FALSE when LHS is not a name", {
  block <- list(call = quote(x$y <- function() {}))
  expect_false(block_is_function(block))
})

# -- find_function_defs -------------------------------------------------------

test_that("find_function_defs returns data.frame with correct columns", {
  pkg <- withr::local_tempdir("has_fns")
  dir.create(file.path(pkg, "R"))
  writeLines(
    c("alpha <- function(x) x + 1",
      "beta <- function(y) y * 2"),
    file.path(pkg, "R", "fns.R")
  )
  writeLines(
    c("Package: hasfnspkg", "Title: Test", "Version: 1.0.0",
      "Description: Test.", "License: MIT"),
    file.path(pkg, "DESCRIPTION")
  )

  result <- find_function_defs(pkg)
  expect_s3_class(result, "data.frame")
  expect_named(result, c("name", "file", "line"))
  expect_identical(nrow(result), 2L)
  expect_true(all(c("alpha", "beta") %in% result$name))
  expect_true(all(grepl("fns\\.R$", result$file)))
  expect_true(is.numeric(result$line))
  expect_identical(result$line[result$name == "alpha"], 1)
  expect_identical(result$line[result$name == "beta"], 2)
})

test_that("find_function_defs returns empty data.frame when no R files exist", {
  pkg <- withr::local_tempdir("no_r_files")
  dir.create(file.path(pkg, "R"))
  writeLines(
    c("Package: norfpkg", "Title: Test", "Version: 1.0.0",
      "Description: Test.", "License: MIT"),
    file.path(pkg, "DESCRIPTION")
  )

  result <- find_function_defs(pkg)
  expect_s3_class(result, "data.frame")
  expect_identical(nrow(result), 0L)
  expect_named(result, c("name", "file", "line"))
})

# -- find_function_defs edge cases --------------------------------------------

test_that("find_function_defs returns empty data.frame when no fns found", {
  pkg <- withr::local_tempdir("no_fns")
  dir.create(file.path(pkg, "R"))
  writeLines("x <- 1", file.path(pkg, "R", "data.R"))
  writeLines(
    c("Package: nofnspkg", "Title: Test", "Version: 1.0.0",
      "Description: Test.", "License: MIT"),
    file.path(pkg, "DESCRIPTION")
  )

  result <- find_function_defs(pkg)
  expect_identical(nrow(result), 0L)
  expect_named(result, c("name", "file", "line"))
})

test_that("find_function_defs skips non-identifier LHS assignments", {
  pkg <- withr::local_tempdir("non_id_lhs")
  dir.create(file.path(pkg, "R"))
  writeLines(
    c("env <- new.env()",
      "env$helper <- function() 1",
      "real_fn <- function(x) x + 1"),
    file.path(pkg, "R", "code.R")
  )
  writeLines(
    c("Package: nonidlhspkg", "Title: Test", "Version: 1.0.0",
      "Description: Test.", "License: MIT"),
    file.path(pkg, "DESCRIPTION")
  )

  result <- find_function_defs(pkg)
  expect_identical(nrow(result), 1L)
  expect_identical(result$name, "real_fn")
})

test_that("find_function_defs skips non-existent files", {
  pkg <- withr::local_tempdir("missing_file")
  dir.create(file.path(pkg, "R"))
  writeLines("my_fn <- function() 1", file.path(pkg, "R", "real.R"))
  writeLines(
    c("Package: missingpkg", "Title: Test", "Version: 1.0.0",
      "Description: Test.", "License: MIT",
      "Collate: 'ghost.R' 'real.R'"),
    file.path(pkg, "DESCRIPTION")
  )

  result <- find_function_defs(pkg)
  expect_identical(nrow(result), 1L)
  expect_identical(result$name, "my_fn")
})

# -- parse_roxygen2 NAMESPACE error fallback ----------------------------------

test_that("parse_roxygen2 handles broken NAMESPACE gracefully", {
  pkg <- withr::local_tempdir("broken_ns")
  dir.create(file.path(pkg, "R"))
  writeLines(
    c("#' A function", "#' @export", "hello <- function() 1"),
    file.path(pkg, "R", "hello.R")
  )
  writeLines(
    c("Package: brokenns", "Title: Test", "Version: 1.0.0",
      "Description: Test.", "License: MIT", "RoxygenNote: 7.3.3"),
    file.path(pkg, "DESCRIPTION")
  )
  writeLines(
    "this is not valid namespace content!!!",
    file.path(pkg, "NAMESPACE")
  )

  result <- parse_roxygen2(pkg)
  expect_identical(result$namespace_exports, character())
  expect_identical(result$namespace_s3methods, character())
})

# -- parse_roxygen2 S3 methods ------------------------------------------------

test_that("parse_roxygen2 extracts S3 methods from NAMESPACE", {
  pkg <- withr::local_tempdir("s3methods")
  dir.create(file.path(pkg, "R"))
  writeLines(
    c("#' Print method", "#' @export", "print.myclass <- function(x, ...) x"),
    file.path(pkg, "R", "print.R")
  )
  writeLines(
    c("Package: s3pkg", "Title: Test", "Version: 1.0.0",
      "Description: Test.", "License: MIT", "RoxygenNote: 7.3.3"),
    file.path(pkg, "DESCRIPTION")
  )
  writeLines(
    c("# Generated by roxygen2: do not edit by hand",
      "S3method(print,myclass)"),
    file.path(pkg, "NAMESPACE")
  )

  result <- parse_roxygen2(pkg)
  expect_true("print.myclass" %in% result$namespace_s3methods)
})

# -- roxygen2_duplicate_params ------------------------------------------------

test_that("roxygen2_duplicate_params fails on identical params across files", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: dupeparam", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT", "RoxygenNote: 7.3.3"
  ), file.path(pkg, "DESCRIPTION"))

  writeLines(c(
    "#' Function A",
    "#' @param x A numeric value.",
    "#' @export",
    "func_a <- function(x) x + 1"
  ), file.path(pkg, "R", "a.R"))

  writeLines(c(
    "#' Function B",
    "#' @param x A numeric value.",
    "#' @export",
    "func_b <- function(x) x * 2"
  ), file.path(pkg, "R", "b.R"))

  gp_res <- gp(pkg, checks = "roxygen2_duplicate_params")
  expect_false(results(gp_res)$passed)

  pos <- failed_positions(gp_res)$roxygen2_duplicate_params
  lines <- vapply(pos, `[[`, "", "line")
  expect_match(lines, "@param x", all = TRUE)
})

test_that("roxygen2_duplicate_params passes with different descriptions", {
  pkg <- withr::local_tempdir()
  dir.create(file.path(pkg, "R"))
  writeLines(c(
    "Package: nodupe", "Title: Test", "Version: 1.0.0",
    "Description: Test.", "License: MIT", "RoxygenNote: 7.3.3"
  ), file.path(pkg, "DESCRIPTION"))

  writeLines(c(
    "#' Function A",
    "#' @param x A numeric input.",
    "#' @export",
    "func_a <- function(x) x + 1"
  ), file.path(pkg, "R", "a.R"))

  writeLines(c(
    "#' Function B",
    "#' @param x A character string.",
    "#' @export",
    "func_b <- function(x) paste(x)"
  ), file.path(pkg, "R", "b.R"))

  gp_res <- gp(pkg, checks = "roxygen2_duplicate_params")
  expect_true(results(gp_res)$passed)
})

test_that("roxygen2_duplicate_params passes for non-roxygen2 packages", {
  expect_warning(
    gp_res <- gp("no_roxygen", checks = "roxygen2_duplicate_params"),
    "Prep step for"
  )
  expect_true(is.na(results(gp_res)$passed))
})

#' Wrapper on make_check, specific to R CMD check
#'
#' @inheritParams make_check
#' @param type Type of notification, one of "warnings", "notes" or "errors".
#' @param pattern The text pattern identifying the check.
#' @param tags Tags to be passed on to make_check.
#' @param preps Preps to be passed on to make_check.
#' @param ... Currently not supported.
#'
#' @keywords internal
#' @include lists.R

#' @noRd
make_rcmd_check <- function(
  description, pattern, gp = NULL, type = c("warnings", "notes", "errors"),
  tags = NULL, preps = NULL, ...) {

  type <- match.arg(type)
  kw <- c(warnings = "WARNING", notes = "NOTE", errors = "ERROR")
  chk_pat <- paste0("^.* \\.\\.\\. ", kw[type])

  etag <- c(warnings = "warning", notes = "info", errors = "error")[type]

  make_check(
    description = description,
    tags = c(etag, "rcmdcheck", tags),
    preps = c("rcmdcheck", preps),
    gp = function(state) {
      paste(
        sub(
          chk_pat, paste0("fix this R CMD check ", kw[type], ":"),
          grep(pattern, state$rcmdcheck[[type]], value = TRUE)
        ),
        collapse = "\n"
      )
    },
    check = function(state) {
      if (inherits(state$rcmdcheck, "try-error")) return(na_result())
      matched <- any(grep(pattern, state$rcmdcheck[[type]]))
      check_result(!matched)
    }
  )
}

CHECKS$rcmdcheck_package_directory_exists <- make_rcmd_check(

  "Package directory exists",
  type = "errors",
  pattern = "Package directory .* does not exist"
)

CHECKS$rcmdcheck_vignette_leftovers <- make_rcmd_check(

  "No vignette leftovers",
  pattern = "for left-overs from vignette generation"
)

CHECKS$rcmdcheck_qpdf_present_for_checking <- make_rcmd_check(

  "'qpdf' must be installed for checks",
  type = "warnings",
  pattern = "'qpdf' is needed for checks on size reduction of PDFs"
)

CHECKS$rcmdcheck_vignettes_at_right_place <- make_rcmd_check(

  "Vignettes are in 'vignettes'",
  type = "notes",
  pattern = "Vignette sources only in 'inst/doc':"
)

CHECKS$rcmdcheck_portable_file_names <- make_rcmd_check(

  "File names are portable",
  type = "errors",
  pattern = "Please rename the files and try again"
)

CHECKS$rcmdcheck_duplicate_file_names <- make_rcmd_check(

  "No file names that only differ in case",
  type = "errors",
  pattern = "Found the following files with duplicate lower-cased file names"
)

CHECKS$rcmdcheck_ascii_file_names <- make_rcmd_check(

  "File names are portable (ASCII)",
  type = "warnings",
  pattern = "These are not fully portable file names"
)

CHECKS$rcmdcheck_portable_paths <- make_rcmd_check(

  "Paths are short enough to be portable",
  type = "errors",
  pattern = "Found the following non-portable file path"
)

CHECKS$rcmdcheck_proper_permissions <- make_rcmd_check(

  "Files have correct permissions",
  type = "errors",
  pattern = "Found the following files with insufficient permissions"
)

CHECKS$rcmdcheck_executable_files <- make_rcmd_check(

  "configure and cleanup files must be executable",
  type = "warnings",
  pattern = "The following files should most likely be executable"
)

CHECKS$rcmdcheck_description_meta <- make_rcmd_check(

  "DESCRIPTION meta information is correct",
  type = "error",
  pattern = "DESCRIPTION meta-information .* ERROR"
)

CHECKS$rcmdcheck_malformed_title_or_description <- make_rcmd_check(

  "Title and Description fields are valid",
  type = "notes",
  pattern = "Malformed (Title|Description)"
)

CHECKS$rcmdcheck_portable_description_encoding <- make_rcmd_check(

  "Portable DESCRIPTION Encoding",
  type = "warnings",
  pattern = "Encoding '.*' is not portable"
)

CHECKS$rcmdcheck_description_encoding_known <- make_rcmd_check(

  "Encoding of DESCRIPTION is defined",
  type = "warnings",
  pattern = "Unknown encoding with non-ASCII data"
)

CHECKS$rcmdcheck_description_ascii_tags <- make_rcmd_check(

  "All tags in DESCRIPTION are ASCII",
  type = "warnings",
  pattern = "All field tags must be ASCII"
)

CHECKS$rcmdcheck_description_ascii_fields <- make_rcmd_check(

  "All tags in DESCRIPTION are ASCII, or Encoding is defined",
  type = "warnings",
  pattern = "Fields with non-ASCII values"
)

CHECKS$rcmdcheck_standardizable_license <- make_rcmd_check(

  "Package license is standardizable",
  type = "warnings",
  pattern = "Standardizable: FALSE"
)

CHECKS$rcmdcheck_license_file_pointers <- make_rcmd_check(

  "License file pointers are correct",
  type = "warnings",
  pattern = "Invalid license file pointers"
)

CHECKS$rcmdcheck_non_standard_license <- make_rcmd_check(

  "Non-standard, but standardizable license",
  type = "notes",
  pattern = "Standardized license specification"
)

CHECKS$rcmdcheck_deprecated_license <- make_rcmd_check(

  "Deprecated license",
  type = "notes",
  pattern = "Deprecated license"
)

CHECKS$rcmdcheck_not_permitted_license_restrictions <- make_rcmd_check(

  "Not permitted license restrictions",
  type = "notes",
  pattern = "License components with restrictions not permitted"
)

CHECKS$rcmdcheck_template_licenses <- make_rcmd_check(

  "Template licenses have a LICENSE file",
  type = "notes",
  pattern = "License components which are templates and need"
)

CHECKS$rcmdcheck_authors_at_r_field <- make_rcmd_check(

  "Valid Authors@R field",
  type = "notes",
  pattern = "Malformed Authors@R field"
)

CHECKS$rcmdcheck_valid_author_in_authors_at_r <- make_rcmd_check(

  "Can extract Author from Authors@R",
  type = "notes",
  pattern = "Cannot extract Author field from Authors@R field:"
)

CHECKS$rcmdcheck_author_in_authors_at_r <- make_rcmd_check(

  "Author is in Authors@R",
  type = "notes",
  pattern = "Authors@R field gives no person with author role."
)

CHECKS$rcmdcheck_authors_at_r_valid_roles <- make_rcmd_check(

  "Authors@R has valid roles for all authors",
  type = "notes",
  pattern = "Authors@R field gives persons with no valid roles:"
)

CHECKS$rcmdcheck_maintainer_in_authors_at_r <- make_rcmd_check(

  "Authors@R has package Maintainer",
  type = "notes",
  pattern = "Cannot extract Maintainer field from Authors@R field"
)

CHECKS$rcmdcheck_single_maintainer_in_authors_at_r <- make_rcmd_check(

  "Authors@R has one package Maintainer only",
  type = "notes",
  pattern = "Authors@R field gives more than one person with maintainer role"
)

CHECKS$rcmdcheck_valid_maintainer_in_authors_at_t <- make_rcmd_check(

  "Authors@R has a valid Maintainer role person",
  type = "notes",
  pattern = "Authors@R field gives no person with maintainer role"
)

CHECKS$rcmdcheck_stale_author <- make_rcmd_check(

  "No stale Author field if Authors@R is present",
  type = "notes",
  pattern = "Author field differs from that derived from Authors@R"
)

CHECKS$rcmdcheck_stale_maintainer <- make_rcmd_check(

  "No state Maintainer field if Authors@R is present",
  type = "notes",
  pattern = "Maintainer field differs from that derived from Authors@R"
)

CHECKS$rcmdcheck_needscompilation_field_valid <- make_rcmd_check(

  "NeedsCompilation field must be 'yes' or 'no'",
  type = "notes",
  pattern = "NeedsCompilation field must take value 'yes' or 'no'"
)

CHECKS$rcmdcheck_multiple_dependency_modes <- make_rcmd_check(

  "Packages should be listed as one type of dependency (except for LinkingTo)",
  type = "notes",
  pattern = paste0(
    "Packages? listed in more than one of ",
    "Depends, Imports, Suggests, Enhances"
  )
)

CHECKS$rcmdcheck_unused_linkingto <- make_rcmd_check(

  "LinkingTo field makes sense, there are headers to include",
  type = "notes",
  pattern = "are unused as (it|they) (has|have) no 'include' directory"
)

CHECKS$rcmdcheck_versioned_linkingto <- make_rcmd_check(

  "Versioned 'LinkingTo' needs R >= 3.0.2",
  type = "notes",
  pattern = "Versioned 'LinkingTo' values for"
)

CHECKS$rcmdcheck_linkingto_without_compiled_code <- make_rcmd_check(

  "'LinkingTo' field only makes sense if package has compiled code",
  type = "notes",
  pattern = "'LinkingTo' field is unused: package has no 'src' directory"
)

CHECKS$rcmdcheck_broken_vignettes_1 <- make_rcmd_check(

  "Vignette output not in build/vignette.rds",
  type = "warnings",
  pattern = "Vignette(s) without any output listed in 'build/vignette.rds'"
)

CHECKS$rcmdcheck_broken_vignettes_2 <- make_rcmd_check(

  "All outputs from build/vignette.rds are in the package",
  type = "warnings",
  pattern = "Output(s) listed in 'build/vignette.rds' but not in package:"
)

CHECKS$rcmdcheck_defunct_top_level <- make_rcmd_check(

  "Some top level files are defunct: install.R, R_PROFILE.R",
  type = "warnings",
  pattern = "These files are defunct."
)

CHECKS$rcmdcheck_valid_readme_and_news <- make_rcmd_check(

  "README.md and NEWS.md must work with pandoc",
  type = "warnings",
  pattern = "Conversion of '.*(README|NEWS)\\.md.*' failed"
)

CHECKS$rcmdcheck_stale_license_file <- make_rcmd_check(

  "LICENSE (and LICENCE) must be mentioned in DESCRIPTION",
  type = "notes",
  pattern = "not mentioned in the DESCRIPTION file"
)

CHECKS$rcmdcheck_stale_inst_license_file <- make_rcmd_check(

  "inst/LICENSE (and inst/LICENCE) must be mentioned in DESCRIPTION",
  type = "notes",
  pattern = paste0(
    "will install at top-level and (is|are) not mentioned ",
    "in the DESCRIPTION file"
  )
)

CHECKS$rcmdcheck_non_standard_top_level_files <- make_rcmd_check(

  "Top-level files are standardized",
  type = "notes",
  pattern = "Non-standard files?/director(ies|y) found at top level"
)

CHECKS$rcmdcheck_copyright_file_in_inst <- make_rcmd_check(

  "Copyright file should be in inst/",
  type = "notes",
  pattern = "Copyright information should be in file inst/COPYRIGHTS"
)

CHECKS$rcmdcheck_authors_in_inst <- make_rcmd_check(

  "AUTHORS file should be in inst/",
  type = "notes",
  pattern = "Authors information should be in file inst/AUTHORS"
)

CHECKS$rcmdcheck_leftover_files <- make_rcmd_check(

  "Check for left-over files",
  type = "notes",
  pattern = "The following files look like leftovers"
)

CHECKS$rcmdcheck_empty_index_file <- make_rcmd_check(

  "INDEX file must not be empty if present",
  type = "warnings",
  pattern = "Empty file 'INDEX'"
)

CHECKS$rcmdcheck_empty_demo_index_file <- make_rcmd_check(

  "demo/00Index must not be empty if present",
  type = "warnings",
  pattern = "Empty or missing file demo.*00Index"
)

CHECKS$rcmdcheck_proper_demo_index_file <- make_rcmd_check(

  "demo/00Index file is valid if it exists",
  type = "warnings",
  pattern = "cannot read index information in file"
)

CHECKS$rcmdcheck_demos_are_indexed <- make_rcmd_check(

  "All demos are in demo/00Index",
  type = "warnings",
  pattern = "Demos with missing or empty index information"
)

CHECKS$rcmdcheck_stale_demos_in_index <- make_rcmd_check(

  "Demos in the index exist",
  type = "warnings",
  pattern = "Demo index entries without corresponding demo:"
)

CHECKS$rcmdcheck_vignette_index <- make_rcmd_check(

  "Vignettes have index entries",
  type = "warnings",
  pattern = "Vignettes with missing or empty"
)

CHECKS$rcmdcheck_has_r_source_files <- make_rcmd_check(

  paste0(
    "Package must have R source files (or R/sysdata.rda) if R ",
    "directory is present"
  ),
  type = "warnings",
  pattern = "Found directory 'R' with no source files"
)

CHECKS$rcmdcheck_r_directory_uppercase <- make_rcmd_check(

  "R directory must be upper case",
  type = "warnings",
  pattern = "Most likely, this should be 'R'"
)

CHECKS$rcmdcheck_man_directory_lowercase <- make_rcmd_check(

  "man directory must be lower case",
  type = "warnings",
  pattern = "Most likely, this should be 'man'"
)

CHECKS$rcmdcheck_data_directory_lowercase <- make_rcmd_check(

  "data directory must be lower case",
  type = "warnings",
  pattern = "Most likely, this should be 'data'"
)

CHECKS$rcmdcheck_rcheck_directories <- make_rcmd_check(

  "No leftover .Rcheck directories",
  type = "warnings",
  pattern = paste0(
    "(Found the following directory with the name of a check directory",
    "|Found the following directories with names of check directories)")
)

CHECKS$rcmdcheck_leftover_dvi_pdf_build_directories <- make_rcmd_check(

  "No leftover Rd2dvi or Rd2pdf directories",
  type = "warnings",
  pattern = paste0(
    "Found the following director(y|ies) with the name ",
    "of a Rd2pdf director"
  )
)

CHECKS$rcmdcheck_leftover_vs_directories <- make_rcmd_check(

  "No leftover .git, .svn (etc.) version control directories",
  type = "warnings",
  pattern = paste0(
    "Found the following director(y|ies) with the name of a ",
    "version control director"
  )
)

CHECKS$rcmdcheck_invalid_file_names <- make_rcmd_check(

  "File names in some directories are restricted",
  type = "warnings",
  pattern = "Subdirectory '.*' contains invalid file names"
)

CHECKS$rcmdcheck_empty_data_directory <- make_rcmd_check(

  "data directory must not be empty, if it exists",
  type = "warnings",
  pattern = "Subdirectory 'data' contains no data sets"
)

CHECKS$rcmdcheck_empty_demo_directory <- make_rcmd_check(

  "demo directory must not be empty, if it exists",
  type = "warnings",
  pattern = "Subdirectory 'demo' contains no demos."
)

CHECKS$rcmdcheck_demos_are_ascii <- make_rcmd_check(

  "Demos should use ASCII characters",
  type = "warnings",
  pattern = "Demos with non-ASCII characters:"
)

CHECKS$rcmdcheck_demos_have_valid_code <- make_rcmd_check(

  "Demos should contain valid R code",
  type = "warnings",
  pattern = "Demos which do not contain valid R code"
)

CHECKS$rcmdcheck_empty_exec_directory <- make_rcmd_check(

  "exec directory must not be empty, if it exists",
  type = "warnings",
  pattern = "Subdirectory 'exec' contains no files."
)

CHECKS$rcmdcheck_empty_inst_directory <- make_rcmd_check(

  "inst directory must not be empty, if it exists",
  type = "warnings",
  pattern = "Subdirectory 'inst' contains no files."
)

CHECKS$rcmdcheck_src_without_sources <- make_rcmd_check(

  "src directory must contain some source files, if it exists",
  type = "warnings",
  pattern = "Subdirectory 'src' contains no source files."
)

CHECKS$rcmdcheck_inst_interference <- make_rcmd_check(

  "Some directory names in inst are not allowed",
  type = "warnings",
  pattern = "with package subdirectories used by R"
)

CHECKS$rcmdcheck_news_rd_is_valid <- make_rcmd_check(

  "inst/NEWS.Rd is valid if present",
  type = "warnings",
  pattern = "Problems with news in 'inst/NEWS.Rd':"
)

CHECKS$rcmdcheck_citation_file_is_valid <- make_rcmd_check(

  "inst/CITATION is valid if present",
  type = "warnings",
  pattern = "Invalid citation information in 'inst/CITATION'"
)

CHECKS$rcmdcheck_citation_file_at_standard_place <- make_rcmd_check(

  "CITATION is in inst/, not somewhere else",
  type = "notes",
  pattern = "Found the following CITATION file in a non-standard place"
)

CHECKS$rcmdcheck_r_files_are_ascii <- make_rcmd_check(

  "R files should be ASCII",
  type = "warnings",
  pattern = "Found the following files? with non-ASCII characters"
)

CHECKS$rcmdcheck_errors_in_r_files <- make_rcmd_check(

  "R files do not have syntax errors",
  type = "errors",
  pattern = "R files for syntax errors .*Error in file"
)

CHECKS$rcmdcheck_warnings_in_r_files <- make_rcmd_check(

  "R files do not throw warnings",
  type = "warnings",
  pattern = "R files for syntax errors .*Warnings? in file"
)

CHECKS$rcmdcheck_undeclared_imports <- make_rcmd_check(

  "'::' and ':::' imports must be declared",
  type = "warnings",
  pattern = "'::' or ':::' imports? not declared from:"
)

CHECKS$rcmdcheck_undeclared_library_require <- make_rcmd_check(

  "'library' and 'require' calls must be declared",
  type = "warnings",
  pattern = "'library' or 'require' calls? not declared from:"
)

CHECKS$rcmdcheck_undeclared_loadnamespace_requirenamespace <- make_rcmd_check(

  "'loadNamespace' and 'requireNamespace' calls must be declared",
  type = "warnings",
  pattern = "'loadNamespace' or 'requireNamespace' calls? not declared from:"
)

CHECKS$rcmdcheck_library_require_to_attached <- make_rcmd_check(

  "'library' and 'require' calls to attached packages are not needed",
  type = "notes",
  pattern = paste0(
    "('library' or 'require' calls to packages already attached by Depends:",
    "|'library' or 'require' call to .* which was already attached by Depends)")
)

CHECKS$rcmdcheck_library_require_in_package_code <- make_rcmd_check(

  paste("Do not use 'library' or 'require' in package code. Use '::' or",
        "'requireNamespace' instead"),
  type = "notes",
  pattern = paste0(
    "'library' or 'require' calls in package code:",
    "'library' or 'require' call to .* in package code.")
)

CHECKS$rcmdcheck_imports_not_imported_from <- make_rcmd_check(

  "Packages declared in 'Imports' must be imported from",
  type = "notes",
  pattern = "Namespaces? in Imports field not imported from:"
)

CHECKS$rcmdcheck_depends_not_imported_from <- make_rcmd_check(

  "Packages in 'Depends' must be imported from",
  type = "notes",
  pattern = "Packages? in Depends field not imported from:"
)

CHECKS$rcmdcheck_missing_or_unexported_objects <- make_rcmd_check(

  "Exported objects must be present",
  type = "notes",
  pattern = "Missing or unexported object"
)

CHECKS$rcmdcheck_unneeded_triple_colon <- make_rcmd_check(

  "Use '::' to call exported functions, not ':::'",
  type = "notes",
  pattern = "':::' calls? which should be '::':"
)

CHECKS$rcmdcheck_triple_colon_imported_objects_exist <- make_rcmd_check(

  "All objects imported via ':::' must exist",
  type = "notes",
  pattern = "Missing objects? imported by ':::' calls:"
)

CHECKS$rcmdcheck_unexported_base_objects_imported <- make_rcmd_check(

  "':::' imports to unexported objects, from base packages",
  type = "warnings",
  pattern = "Unexported objects? imported by ':::' calls"
)

CHECKS$rcmdcheck_unexported_objects_imported <- make_rcmd_check(

  "':::' imports to unexported objects",
  type = "notes",
  pattern = "Unexported objects? imported by ':::' calls"
)

CHECKS$rcmdcheck_triple_colon_to_itself <- make_rcmd_check(

  "':::' imports to the package itself are usually not needed",
  type = "notes",
  pattern = "There are ::: calls to the package's namespace in its code."
)

CHECKS$rcmdcheck_triple_colon_to_unknown <- make_rcmd_check(

  "':::' imports to unavailable packages",
  type = "notes",
  pattern = "Unavailable namespaces? imported from by ':::' calls:"
)

CHECKS$rcmdcheck_data_call_to_undeclared_package <- make_rcmd_check(

  "Packages used in 'data(package=)' must be declared",
  type = "notes",
  pattern = "data(package=)' calls? not declared from"
)

CHECKS$rcmdcheck_exported_s3_methods_are_registered <- make_rcmd_check(

  "Exported S3 methods are registered",
  type = "warnings",
  pattern = paste0(
    "Found the following apparent S3 methods ",
    "exported but not registered"
  )
)

CHECKS$rcmdcheck_replacement_function_arg_names <- make_rcmd_check(

  "Replacement functions must have a 'value' argument",
  type = "warnings",
  pattern = "The argument of a replacement function"
)

CHECKS$rcmdcheck_foreign_calls_have_package_argument <- make_rcmd_check(

  "Foreign calls must have a 'PACKAGE' argument",
  type = "warnings",
  pattern = "Foreign function calls? without 'PACKAGE' argument:"
)

CHECKS$rcmdcheck_foreign_call_empty_package_argument <- make_rcmd_check(

  "'PACKAGE' argument of foreign calls must not be empty",
  type = "warnings",
  pattern = "Foreign function calls? with empty 'PACKAGE' argument:"
)

CHECKS$rcmdcheck_foreign_call_to_base_package <- make_rcmd_check(

  "Foreign calls should not call base packages",
  type = "warnings",
  pattern = "Foreign function calls? to a base package"
)

CHECKS$rcmdcheck_foreign_calls_to_another_package <- make_rcmd_check(

  "Should not call another package with foreign calls",
  type = "notes",
  pattern = "Foreign function calls? to a different package"
)

CHECKS$rcmdcheck_foreign_calls_to_unknown_package <- make_rcmd_check(

  "Packages called from foreign calls must be declared",
  type = "warnings",
  pattern = "Undeclared packages? in foreign function calls"
)

CHECKS$rcmdcheck_foreign_registration_problems <- make_rcmd_check(

  "Check for registration problems of foreign function calls",
  type = "notes",
  pattern = "Registration problem"
)

CHECKS$rcmdcheck_foreign_calls_with_dup <- make_rcmd_check(

  "The 'DUP' argument of foreign calls is ignored",
  type = "notes",
  pattern = "Calls? with DUP"
)

CHECKS$rcmdcheck_loading_and_unloading <- make_rcmd_check(

  "Check loading and unloading the package",
  type = "errors",
  pattern = "Incorrect [(]un[)]loading of package"
)

CHECKS$rcmdcheck_first_lib_obsolete <- make_rcmd_check(

  ".First.lib is obsolete",
  type = "notes",
  pattern = ".First.lib is obsolete and will not be used in R >= 3.0.0"
)

CHECKS$rcmdcheck_startup_function_arguments <- make_rcmd_check(

  "Package startup function arguments should be lib* and pkg*",
  type = "notes",
  pattern = paste0(
    "Package startup functions should have two ",
    "arguments with names starting with"
  )
)

CHECKS$rcmdcheck_startup_function_change_search_path <- make_rcmd_check(

  paste("Package startup functions should not change the search path.",
        "I.e. do not call 'library' or 'require'"),
  type = "notes",
  pattern = "Package startup functions should not change the search path"
)

CHECKS$rcmdcheck_startup_function_messages <- make_rcmd_check(

  paste0(
    "Package startup functions should use 'packageStartupMessage' ",
    "to generate messages"
  ),
  type = "notes",
  pattern = paste0(
    "Package startup functions should use .packageStartupMessage. ",
    "to generate messages"
  )
)

CHECKS$rcmdcheck_statup_function_unsafe_calls <- make_rcmd_check(

  "Package startup functions should not call installed.packages",
  type = "notes",
  pattern = "Package startup functions should not call"
)

CHECKS$rcmdcheck_last_lib_needs_to_be_exported <- make_rcmd_check(

  ".Last.lib will not be used unless it is exported",
  type = "notes",
  pattern = ".Last.lib will not be used unless it is exported"
)

CHECKS$rcmdcheck_detach_function_arguments <- make_rcmd_check(

  paste0(
    "Package detach functions should have one argument with ",
    "name starting with 'lib'"
  ),
  type = "notes",
  pattern = paste0(
    "Package detach functions should have one argument ",
    "with name starting with"
  )
)

CHECKS$rcmdcheck_detach_no_library_dynam_unload <- make_rcmd_check(

  "Package detach functions should not call 'library.dynam.unload'",
  type = "notes",
  pattern = "Package detach functions should not call"
)

CHECKS$rcmdcheck_unsafe_calls <- make_rcmd_check(

  "Check for possibly unsafe calls: unlockBinding, assignInNamespace",
  type = "notes",
  pattern = "Found the following possibly unsafe calls"
)

CHECKS$rcmdcheck_partial_argument_match <- make_rcmd_check(

  "Partial argument matches are fragile",
  type = "notes",
  pattern = "partial argument match of"
)

CHECKS$rcmdcheck_undefined_globals <- make_rcmd_check(

  "Check for undefined globals",
  type = "notes",
  pattern = "Undefined global functions or variables"
)

CHECKS$rcmdcheck_avoid_internal_calls <- make_rcmd_check(

  "No .Internal calls are allowed",
  type = "notes",
  pattern = paste0(
    "(Found a .Internal call in the following function",
    "|Found .Internal calls in the following functions)")
)

CHECKS$rcmdcheck_avoid_internal_calls_s4 <- make_rcmd_check(

  "No .Internal calls are allowed in S4 methods",
  type = "notes",
  pattern = paste0(
    "(Found a.Internal call in methods for the following S4 generic",
    "|Found .Internal calls in methods for the following S4 generics)")
)

CHECKS$rcmdcheck_avoid_internal_calls_rc <- make_rcmd_check(

  "No .Internal calls are allowed in reference class methods",
  type = "notes",
  pattern = paste0(
    "(Found a .Internal call in methods for the following reference class",
    "|Found .Internal calls in methods for the following reference classes)")
)

CHECKS$rcmdcheck_assignment_to_globalenv <- make_rcmd_check(

  "Assignment to .GlobalEnv is not allowed",
  type = "notes",
  pattern = "Found the following assignments to the global environment"
)

CHECKS$rcmdcheck_avoid_using_attach <- make_rcmd_check(

  "It is better to avoid attach()",
  type = "notes",
  pattern = "Found the following calls to attach"
)

CHECKS$rcmdcheck_data_into_globalenv <- make_rcmd_check(

  "Avoid loading data into the global environment",
  type = "notes",
  pattern = paste0(
    "Found the following calls to data.. loading into the global ",
    "environment"
  )
)

CHECKS$rcmdcheck_obsolete_platform_specific <- make_rcmd_check(

  "Avoid calling obsolete and platform specific functions",
  type = "notes",
  pattern = "Found an obsolete/platform-specific call in the following function"
)

CHECKS$rcmdcheck_obsolete_platform_specific_s4 <- make_rcmd_check(

  "Avoid calling obsolete and platform specific functions from S4 methods",
  type = "notes",
  pattern = paste0(
    "Found an obsolete/platform-specific call in methods ",
    "for the following S4 generic"
  )
)

CHECKS$rcmdcheck_obsolete_platform_specific_rc <- make_rcmd_check(

  paste0(
    "Avoid calling obsolete and platform specific functions ",
    "from reference class methods"
  ),
  type = "notes",
  pattern = paste0(
    "Found an obsolete/platform-specific call in methods ",
    "for the following reference class"
  )
)

CHECKS$rcmdcheck_deprecated_functions <- make_rcmd_check(

  "Avoid calling deprecated functions",
  type = "notes",
  pattern = "Found the deprecated function"
)

CHECKS$rcmdcheck_defunct_removed_functions <- make_rcmd_check(

  "Avoid calling defunct and removed functions",
  type = "warnings",
  pattern = "Found the defunct/removed function"
)

CHECKS$rcmdcheck_avoid_platform_specific_devices <- make_rcmd_check(

  "Avoid using platform specific devices",
  type = "notes",
  pattern = "Found the platform-specific device"
)

CHECKS$rcmdcheck_rd_empty_sections <- make_rcmd_check(

  "Check Rd files for empty sections",
  type = "warnings",
  pattern = "Dropping empty section"
)

CHECKS$rcmdcheck_rd_problems <- make_rcmd_check(

  "Check Rd files for problems",
  type = "notes",
  pattern = "Rd files"
)

CHECKS$rcmdcheck_rd_duplicated_name <- make_rcmd_check(

  "Check for Rd files with duplicated \\name",
  type = "notes",
  pattern = "Rd files with duplicated name"
)

CHECKS$rcmdcheck_rd_duplicated_alias <- make_rcmd_check(

  "Check for Rd files with duplicated \\alias",
  type = "notes",
  pattern = "Rd files with duplicated alias"
)

CHECKS$rcmdcheck_rd_long_code_lines <- make_rcmd_check(

  "Check for long code lines in Rd files",
  type = "notes",
  pattern = "These lines will be truncated in the PDF manual"
)

CHECKS$rcmdcheck_rd_cross_references <- make_rcmd_check(

  "Check Rd links",
  type = "warnings",
  pattern = "Missing link or links in documentation object"
)

CHECKS$rcmdcheck_missing_docs <- make_rcmd_check(

  "Check for undocumented exported objects",
  type = "warnings",
  pattern = "All user-level objects"
)

CHECKS$rcmdcheck_code_docs_mismatch <- make_rcmd_check(

  "Check for code/documentation mismatches",
  type = "warnings",
  pattern = "for code/documentation mismatches"
)

CHECKS$rcmdcheck_rd_usage <- make_rcmd_check(

  "Check Rd \\usage sections",
  type = "warnings",
  pattern = "Rd \\usage sections"
)

CHECKS$rcmdcheck_rd_s3_usage <- make_rcmd_check(

  "S3 methods should use the \\method markup, not their full name",
  type = "notes",
  pattern = "entries for S3 methods should use.*markup and not their full name."
)

CHECKS$rcmdcheck_rd_contents <- make_rcmd_check(

  "Check Rd contents",
  type = "warnings",
  pattern = "Rd contents"
)

CHECKS$rcmdcheck_unstated_dependencies_in_examples <- make_rcmd_check(

  "Check for unstated dependencies in examples",
  type = "warnings",
  pattern = "for unstated dependencies in examples"
)

CHECKS$rcmdcheck_data_contents <- make_rcmd_check(

  "Check files in 'data' directory, if exists",
  type = "warnings",
  pattern = "contents of 'data' directory"
)

CHECKS$rcmdcheck_undeclared_non_ascii_characters_in_data <- make_rcmd_check(

  "Check for undeclared non-ASCII characters in data",
  type = "warnings",
  pattern = "Warning: found non-ASCII string"
)

CHECKS$rcmdcheck_non_ascii_characters_in_data <- make_rcmd_check(

  "Check for non-ASCII characters in data",
  type = "notes",
  pattern = "data for non-ASCII characters"
)

CHECKS$rcmdcheck_uncompressed_data_files <- make_rcmd_check(

  "Check for data files that could be compressed better",
  type = "warnings",
  pattern = "data for ASCII and uncompressed saves"
)

CHECKS$rcmdcheck_bzip2_xz_requirement <- make_rcmd_check(

  "bzip2 or xz compression needs more recent R version",
  type = "warnings",
  pattern = "Warning: package needs dependence on R [(]>= 2.10[)]"
)

CHECKS$rcmdcheck_uncompressed_sysdata <- make_rcmd_check(

  "Check if sysdata.rda could be compressed better",
  type = "notes",
  pattern = "R/sysdata.rda"
)

CHECKS$rcmdcheck_unneeded_doc_extra <- make_rcmd_check(

  "Check for unneeded extra style files for documentation",
  type = "notes",
  pattern = "The following files are already in R:"
)

CHECKS$rcmdcheck_doc_extra_licenses <- make_rcmd_check(

  paste0(
    "Check if extra doc styles have license that requires ",
    "the distribution of its original sources"
  ),
  type = "notes",
  pattern = "The following files contain a license that requires"
)

CHECKS$rcmdcheck_doc_tex_leftovers <- make_rcmd_check(

  "Check for possible leftover files for docs",
  type = "notes",
  pattern = "The following files look like leftovers/mistakes"
)

CHECKS$rcmdcheck_extra_files_in_docs <- make_rcmd_check(

  "Check for extra files in docs",
  type = "notes",
  pattern = "The following files should probably not be installed"
)

CHECKS$rcmdcheck_extra_directories_in_docs <- make_rcmd_check(

  "Check for extra directories in docs",
  type = "notes",
  pattern = "The following directories should probably not be installed"
)

CHECKS$rcmdcheck_vignettes_in_vignettes <- make_rcmd_check(

  "Vignettes must be vignettes/",
  type = "warnings",
  pattern = paste0(
    "Vignette sources in 'inst/doc' missing ",
    "from the 'vignettes' directory"
  )
)

CHECKS$rcmdcheck_knitr_leftovers <- make_rcmd_check(

  "Check for possible knitr leftover files",
  type = "notes",
  pattern = "The following directories look like leftovers from 'knitr'"
)

CHECKS$rcmdcheck_pdf_file_sizes <- make_rcmd_check(

  "Check if PDF files could be made signicantly smaller",
  type = "notes",
  pattern = "consider running tools::compactPDF.. on these files"
)

CHECKS$rcmdcheck_pdf_file_sizes_gs <- make_rcmd_check(

  "Check if PDF files could be made much smaller with GhostScript and qpdf",
  type = "warnings",
  pattern = paste0(
    "consider running tools::compactPDF.gs_quality = \"ebook\". ",
    "on these files"
  )
)

CHECKS$rcmdcheck_source_line_endings <- make_rcmd_check(

  "Source files have LF (Unix) line endings",
  type = "warnings",
  pattern = "Found the following sources/headers with CR or CRLF line endings"
)

CHECKS$rcmdcheck_makefile_line_endings <- make_rcmd_check(

  "Makefile(s) have LF (Unix) line endings",
  type = "warnings",
  pattern = "Found the following Makefile.s. with CR or CRLF line endings"
)

CHECKS$rcmdcheck_makefile_with_final_lf <- make_rcmd_check(

  "Makefile(s) have a final LF (newline character)",
  type = "notes",
  pattern = "Found the following Makefile.s. without a final LF"
)

CHECKS$rcmdcheck_non_portable_makevars <- make_rcmd_check(

  "Check for non-portable Makevars flags",
  type = "warnings",
  pattern = "Non-portable flags in variable"
)

CHECKS$rcmdcheck_makevars_overriding_user_site <- make_rcmd_check(

  "Check for Makevars flags overriding user/site settings",
  type = "warnings",
  pattern = "Variables overriding user/site settings"
)

CHECKS$rcmdcheck_makevars_and_makevars_in <- make_rcmd_check(

  "Check if package has both Makevars and Makevars.in",
  type = "notes",
  pattern = "Package has both .*Makevars.in.* and .*Makevars"
)

CHECKS$rcmdcheck_gnu_make_required <- make_rcmd_check(

  "Check if GNU make is a system requirement",
  type = "notes",
  pattern = "GNU make is a SystemRequirements"
)

CHECKS$rcmdcheck_gnu_makefile_extensions <- make_rcmd_check(

  "Check if GNU make is needed (and is undeclared)",
  type = "warnings",
  pattern = "Found the following file.s. containing GNU extensions"
)

CHECKS$rcmdcheck_blas_and_lapack_flags <- make_rcmd_check(

  "Check if $(BLAS_LIBS) are used when $(LAPACK_LIBS) are",
  type = "warnings",
  pattern = "apparently using \\$.LAPACK_LIBS. without \\$.BLAS_LIBS. in"
)

CHECKS$rcmdcheck_missing_flibs <- make_rcmd_check(

  "Check if $(FLIBS) is required and missing",
  type = "warnings",
  pattern = "apparently PKG_LIBS is missing \\$.FLIBS. in "
)

CHECKS$rcmdcheck_unsafe_calls_in_compiled_code <- make_rcmd_check(

  "Check for unsafe calls in compiled code",
  type = "notes",
  pattern = "Compiled code should not call entry points which"
)

CHECKS$rcmdcheck_loading_package <- make_rcmd_check(

  "Check loading the package",
  type = "errors",
  pattern = paste0(
    "(Loading this package had a fatal error",
    "|has a loading problem: see the messages)")
)

CHECKS$rcmdcheck_unloading_package <- make_rcmd_check(

  "Check unloading the package",
  type = "warnings",
  pattern = "whether the package can be unloaded cleanly"
)

CHECKS$rcmdcheck_namespace_can_be_loaded <- make_rcmd_check(

  "Check if namespace can be loaded",
  type = "warnings",
  pattern = "whether the namespace can be loaded with stated dependencies"
)

CHECKS$rcmdcheck_namespace_can_be_loaded_safely <- make_rcmd_check(

  "Check if namespace can be loaded safely",
  type = "notes",
  pattern = "whether the namespace can be loaded with stated dependencies"
)

CHECKS$rcmdcheck_namespace_can_be_unloaded <- make_rcmd_check(

  "Check if namespace can be unloaded",
  type = "warnings",
  pattern = "whether the namespace can be unloaded cleanly"
)

CHECKS$rcmdcheck_loading_when_not_on_search_path <- make_rcmd_check(

  "Packages must be able to load when not on the search path",
  type = "warnings",
  pattern = "loading without being on the library search path"
)

CHECKS$rcmdcheck_s3_method_registration <- make_rcmd_check(

  "S3 methods are properly registered",
  type = "warnings",
  pattern = "use of S3 registration"
)

CHECKS$rcmdcheck_encoding_in_ascii_locale <- make_rcmd_check(

  paste0(
    "Check if we are checking a package with non-ascii ",
    "encoding in an ASCII locale"
  ),
  type = "warnings",
  pattern = "checking a package with encoding .* in an ASCII locale"
)

CHECKS$rcmdcheck_examples_run <- make_rcmd_check(

  "Examples must run",
  type = "errors",
  pattern = "Running examples in .* failed"
)

CHECKS$rcmdcheck_examples_run_without_warnings <- make_rcmd_check(

  "Examples shoudl run without warnings",
  type = "warnings",
  pattern = "Found the following significant warnings"
)

CHECKS$rcmdcheck_two_many_cores_used <- make_rcmd_check(

  "Package should not use more than two cores during checks",
  type = "warnings",
  pattern = "Note that CRAN packages must never use more than two"
)

CHECKS$rcmdcheck_can_collect_examples <- make_rcmd_check(

  "Can create a single file with all example",
  type = "errors",
  pattern = "Running massageExamples to create"
)

CHECKS$rcmdcheck_unstated_dependencies_in_tests <- make_rcmd_check(

  "All packages needed for tests are declared",
  type = "warnings",
  pattern = "for unstated dependencies in .*tests"
)

CHECKS$rcmdcheck_tests_pass <- make_rcmd_check(

  "Tests",
  type = "errors",
  pattern = "Last 13 lines of output"
)

CHECKS$rcmdcheck_unstated_dependencies_in_vignettes <- make_rcmd_check(

  "All packages needed for vignettes are declared",
  type = "notes",
  pattern = "for unstated dependencies in vignettes"
)

CHECKS$rcmdcheck_vignette_output_present <- make_rcmd_check(

  "Output from all vignettes are present",
  type = "warnings",
  pattern = "Package vignettes? without corresponding PDF/HTML"
)

CHECKS$rcmdcheck_encoding_defined_in_vignettes <- make_rcmd_check(

  "Non-ASCII vignettes must define their encoding",
  type = "warnings",
  pattern = "Non-ASCII package vignette without specified encoding"
)

CHECKS$rcmdcheck_doc_makefile_uppercase <- make_rcmd_check(

  "inst/doc/Makefile is uppercase",
  type = "warnings",
  pattern = paste0(
    "Found 'inst/doc/makefile': should be 'Makefile' ",
    "and will be ignored"
  )
)

CHECKS$rcmdcheck_calling_r_from_makefile <- make_rcmd_check(

  "R is called properly from Makefile",
  type = "warnings",
  pattern = "Found 'R CMD' in Makefile: should be"
)

CHECKS$rcmdcheck_makefile_line_endings_2 <- make_rcmd_check(

  "Makefile line ending should be LF (Unix)",
  type = "warnings",
  pattern = "Found Makefile with CR or CRLF line endings"
)

CHECKS$rcmdcheck_calling_rscript_from_makefile <- make_rcmd_check(

  "Rscript is called properly from Makefile",
  type = "warnings",
  pattern = "Found 'Rscript' in Makefile: should be"
)

CHECKS$rcmdcheck_correct_vignette_encodings <- make_rcmd_check(

  "Vignettes use the encoding they define",
  type = "warnings",
  pattern = paste0(
    "(Package vignette which is not in its specified encoding",
    "|Package vignettes which are not in their specified encoding)")
)

CHECKS$rcmdcheck_vignettes_load_dependencies <- make_rcmd_check(

  "Vignettes can load dependencies",
  type = "warnings",
  pattern = "Errors in running code in vignettes"
)

CHECKS$rcmdcheck_vignettes_run <- make_rcmd_check(

  "Vignettes run",
  type = "errors",
  pattern = "Errors in running code in vignettes"
)

CHECKS$rcmdcheck_vignettes_build <- make_rcmd_check(

  "Vignettes build",
  type = "warnings",
  pattern = "Warning in re-building vignettes"
)

CHECKS$rcmdcheck_vignettes_build_2 <- make_rcmd_check(

  "Vignettes build",
  type = "notes",
  pattern = "(Warning|Error) in re-building vignettes"
)

CHECKS$rcmdcheck_can_convert_rd_to_pdf <- make_rcmd_check(

  "Can convert Rd manual to PDF",
  type = "errors",
  pattern = "Rd conversion errors"
)

CHECKS$rcmdcheck_can_convert_rd_to_pdf_2 <- make_rcmd_check(

  "Can convert Rd manual to PDF",
  type = "warnings",
  pattern = "LaTeX errors when creating PDF version"
)

CHECKS$rcmdcheck_pdf_without_hyperref <- make_rcmd_check(

  "PDF manual builds without hyperref",
  type = "errors",
  pattern = "PDF version of manual without hyperrefs or index"
)

CHECKS$rcmdcheck_executable_files_in_packages <- make_rcmd_check(

  "Package should not contain executable files",
  type = "warnings",
  pattern = "Found the following executable file"
)

CHECKS$rcmdcheck_hidden_files_and_directories <- make_rcmd_check(

  "Package does not have hidden files and directories",
  type = "notes",
  pattern = "Found the following hidden files and directories"
)

CHECKS$rcmdcheck_installs <- make_rcmd_check(

  "Package installs",
  type = "errors",
  pattern = "Installation failed\\."
)

CHECKS$rcmdcheck_significant_compilation_warnings <- make_rcmd_check(

  "Check for significant warnings when compiling C/C++/Fortran code",
  type = "warnings",
  pattern = "Found the following significant warnings"
)

CHECKS$rcmdcheck_other_compilation_warnings <- make_rcmd_check(

  "Check for significant warnings when compiling C/C++/Fortran code",
  type = "notes",
  pattern = "Found the following warnings"
)

CHECKS$rcmdcheck_reasonable_installed_size <- make_rcmd_check(

  "Package size is reasonable when installed",
  type = "notes",
  pattern = "installed package size"
)

CHECKS$rcmdcheck_description_required_fields <- make_rcmd_check(

  "DESCRIPTION has required fields",
  type = "errors",
  pattern = "Required fields? missing or empty"
)

CHECKS$rcmdcheck_package_name_portable <- make_rcmd_check(

  "Package name is portable",
  type = "warnings",
  pattern = "Package name is not portable"
)

CHECKS$rcmdcheck_description_right_case <- make_rcmd_check(

  "DESCRIPTION is all upper case",
  type = "errors",
  pattern = paste0(
    "File DESCRIPTION does not exist but there ",
    "is a case-insensitive match"
  )
)

CHECKS$rcmdcheck_not_package_bundle <- make_rcmd_check(

  "Package bundles are defunct",
  type = "errors",
  pattern = "is a package bundle -- they are defunct"
)

CHECKS$rcmdcheck_cran_incoming_feasibility_1 <- make_rcmd_check(

  "CRAN incoming feasibility",
  type = "errors",
  pattern = "CRAN incoming feasibility"
)

CHECKS$rcmdcheck_cran_incoming_feasibility_2 <- make_rcmd_check(

  "CRAN incoming feasibility",
  type = "warnings",
  pattern = "CRAN incoming feasibility"
)

CHECKS$rcmdcheck_cran_incoming_feasibility_3 <- make_rcmd_check(

  "CRAN incoming feasibility",
  type = "notes",
  pattern = "CRAN incoming feasibility"
)

CHECKS$rcmdcheck_valid_namespace <- make_rcmd_check(

  "Valid NAMESPACE file",
  type = "errors",
  pattern = "Invalid NAMESPACE file, parsing gives"
)

CHECKS$rcmdcheck_empty_importfrom_in_namespace <- make_rcmd_check(

  "Check for empty importFrom command in NAMESPACE",
  type = "notes",
  pattern = "Namespace with empty importFrom"
)

CHECKS$rcmdcheck_too_many_s3_methods <- make_rcmd_check(

  "R < 3.0.2 had a limit of 500 registered S3 methods",
  type = "notes",
  pattern = "R < 3.0.2 had a limit of 500 registered S3 methods: found"
)

CHECKS$rcmdcheck_package_dependencies_present <- make_rcmd_check(

  "Package dependencies are present for the check",
  type = "errors",
  pattern = "package dependencies"
)

CHECKS$rcmdcheck_if_source_package <- make_rcmd_check(

  "Only *source* packages can be checked",
  type = "errors",
  pattern = "Only \\*source\\* packages can be checked"
)

CHECKS$rcmdcheck_object_files_in_source_package <- make_rcmd_check(

  "Source packages should not contain object files",
  type = "warnings",
  pattern = "Subdirectory .* contains apparent object files/libraries"
)

CHECKS$rcmdcheck_multi_arch_build_dir <- make_rcmd_check(

  "Check for leftover multi-arch build directory",
  type = "warnings",
  pattern = paste0(
    "Found the following directory with a name of ",
    "a multi-arch build directory"
  )
)

CHECKS$rcmdcheck_compilation_leftover_files <- make_rcmd_check(

  "Possible leftover files from a compilation",
  type = "warnings",
  pattern = "These are unlikely file names for src files"
)

CHECKS$rcmdcheck_object_files_in_source_package_2 <- make_rcmd_check(

  "Source packages should not contain object files",
  type = "notes",
  pattern = "Found the following apparent object files/libraries"
)

CHECKS$rcmdcheck_installed_version_included <- make_rcmd_check(

  "Check if an installed version is included by mistake",
  type = "notes",
  pattern = paste0(
    "Subdirectory .* seems to contain an installed ",
    "version of the package"
  )
)

CHECKS$rcmdcheck_some_code_in_docs <- make_rcmd_check(

  "*Some* form of documentation should contain some code to run ",
  type = "warnings",
  pattern = "No examples, no tests, no vignettes"
)

make_rcmd_state <- function(
  warnings = character(),
  notes = character(),
  errors = character()
) {
  list(
    rcmdcheck = list(
      warnings = warnings,
      notes = notes,
      errors = errors
    )
  )
}

test_that("make_rcmd_check creates a working check", {
  chk <- make_rcmd_check(
    description = "Test check",
    pattern = "test pattern",
    type = "notes"
  )
  expect_true(is.function(chk$check))
  expect_true(is.function(chk$gp))
  expect_true("rcmdcheck" %in% chk$preps)
  expect_true("info" %in% chk$tags)
})

test_that("make_rcmd_check check passes when no match", {
  chk <- make_rcmd_check(
    description = "Test",
    pattern = "bad stuff",
    type = "notes"
  )
  state <- make_rcmd_state(notes = "Everything is fine")
  result <- chk$check(state)
  expect_true(result$status)
  expect_type(result$positions, "list")
})

test_that("make_rcmd_check check fails when pattern matches", {
  chk <- make_rcmd_check(
    description = "Test",
    pattern = "bad stuff",
    type = "notes"
  )
  state <- make_rcmd_state(notes = "found bad stuff in code")
  result <- chk$check(state)
  expect_false(result$status)
  expect_type(result$positions, "list")
})

test_that("make_rcmd_check check returns na_result on try-error", {
  chk <- make_rcmd_check(
    description = "Test",
    pattern = "anything",
    type = "warnings"
  )
  state <- list(rcmdcheck = structure("error", class = "try-error"))
  result <- chk$check(state)
  expect_true(is.na(result$status))
  expect_type(result$positions, "list")
})

test_that("make_rcmd_check gp returns advice", {
  chk <- make_rcmd_check(
    description = "Test",
    pattern = "bad stuff",
    type = "notes"
  )
  state <- make_rcmd_state(
    notes = "checking ... NOTE found bad stuff"
  )
  msg <- chk$gp(state)
  expect_type(msg, "character")
  expect_true(grepl("fix this R CMD check NOTE", msg))
})

test_that("make_rcmd_check handles warning type", {
  chk <- make_rcmd_check(
    description = "Test",
    pattern = "oops",
    type = "warnings"
  )
  expect_true("warning" %in% chk$tags)

  state <- make_rcmd_state(warnings = "checking ... WARNING oops")
  result <- chk$check(state)
  expect_false(result$status)
  expect_true(grepl("WARNING", chk$gp(state)))
})

test_that("make_rcmd_check handles error type", {
  chk <- make_rcmd_check(
    description = "Test",
    pattern = "fatal",
    type = "errors"
  )
  expect_true("error" %in% chk$tags)

  state <- make_rcmd_state(errors = "checking ... ERROR fatal crash")
  result <- chk$check(state)
  expect_false(result$status)
  expect_true(grepl("ERROR", chk$gp(state)))
})

#' @include lists.R prep_utils.R
#' @importFrom cyclocomp cyclocomp_package_dir

PREPS$cyclocomp <- function(state, path = state$path, quiet) {
  run_prep_step(state, "cyclocomp", function(path) {
    cyclocomp_package_dir(path)
  }, path = path, silent = quiet)
}

#' @include lists.R

#' @noRd
query_reverse_deps <- function(pkg_name, db) {
  deps <- tools::package_dependencies(pkg_name, db = db, reverse = TRUE)
  deps[[pkg_name]]
}

revdep_gp_message <- function(revdeps) {
  n <- length(revdeps)
  dep_list <- paste(utils::head(revdeps, 10), collapse = ", ")
  suffix <- if (n > 10) paste0(", ... and ", n - 10L, " more") else ""
  paste0(
    "run revdepcheck::revdep_check() before CRAN submission. ",
    "This package has ", n, " reverse ",
    ifelse(n == 1, "dependency", "dependencies"),
    " on CRAN: ", dep_list, suffix, "."
  )
}

CHECKS$reverse_dependencies <- make_check(

  description = "Check for reverse dependencies on CRAN",
  tags = c("info", "CRAN"),
  preps = c("description", "revdep"),

  gp = function(state) {
    revdeps <- state$checks$reverse_dependencies$revdeps
    revdep_gp_message(revdeps)
  },

  check = function(state) {
    if (inherits(state$description, "try-error")) return(na_result())
    if (identical(state$revdep, NA) ||
        inherits(state$revdep, "try-error")) return(na_result())

    pkg_name <- state$description$get_field("Package", default = NA_character_)
    if (is.na(pkg_name)) return(na_result())

    revdeps <- tryCatch(
      query_reverse_deps(pkg_name, state$revdep),
      error = function(e) NA
    )

    if (identical(revdeps, NA)) return(na_result())
    if (is.null(revdeps) || length(revdeps) == 0) {
      return(check_result(TRUE))
    }

    check_result(TRUE, type = "info", revdeps = revdeps)
  }
)

fake_db <- matrix(
  character(0), nrow = 0, ncol = 3,
  dimnames = list(NULL, c("Package", "Version", "Depends"))
)

make_desc_state <- function(pkg_name = "testpkg") {
  d <- desc::desc("!new")
  d$set(Package = pkg_name)
  list(description = d, revdep = fake_db)
}

describe("reverse_dependencies check", {

  it("passes when package has no reverse deps", {
    local_mocked_bindings(query_reverse_deps = function(pkg_name, db) NULL)
    state <- make_desc_state()
    result <- CHECKS$reverse_dependencies$check(state)
    expect_true(result$status)
    expect_type(result$positions, "list")
  })

  it("returns info result when package has reverse deps", {
    local_mocked_bindings(
      query_reverse_deps = function(pkg_name, db) c("pkgA", "pkgB")
    )
    state <- make_desc_state()
    result <- CHECKS$reverse_dependencies$check(state)
    expect_true(result$status)
    expect_identical(result$type, "info")
    expect_identical(result$revdeps, c("pkgA", "pkgB"))
  })

  it("returns na_result on query error", {
    local_mocked_bindings(
      query_reverse_deps = function(pkg_name, db) stop("no internet")
    )
    state <- make_desc_state()
    result <- CHECKS$reverse_dependencies$check(state)
    expect_true(is.na(result$status))
    expect_type(result$positions, "list")
  })

  it("returns na_result when description is a try-error", {
    state <- list(
      description = try(stop("fail"), silent = TRUE),
      revdep = fake_db
    )
    result <- CHECKS$reverse_dependencies$check(state)
    expect_true(is.na(result$status))
    expect_type(result$positions, "list")
  })

  it("returns na_result when revdep is NA", {
    state <- make_desc_state()
    state$revdep <- NA
    result <- CHECKS$reverse_dependencies$check(state)
    expect_true(is.na(result$status))
    expect_type(result$positions, "list")
  })

  it("returns na_result when package name is missing", {
    d <- desc::desc("!new")
    d$del("Package")
    state <- list(description = d, revdep = fake_db)
    result <- CHECKS$reverse_dependencies$check(state)
    expect_true(is.na(result$status))
    expect_type(result$positions, "list")
  })

  it("passes when package has zero reverse deps (empty character)", {
    local_mocked_bindings(
      query_reverse_deps = function(pkg_name, db) character(0)
    )
    state <- make_desc_state()
    result <- CHECKS$reverse_dependencies$check(state)
    expect_true(result$status)
    expect_type(result$positions, "list")
  })

  it("returns na_result when query_reverse_deps returns NA (no internet)", {
    local_mocked_bindings(query_reverse_deps = function(pkg_name, db) NA)
    state <- make_desc_state()
    result <- CHECKS$reverse_dependencies$check(state)
    expect_true(is.na(result$status))
    expect_type(result$positions, "list")
  })
})

describe("revdep_gp_message", {

  it("lists reverse deps", {
    msg <- revdep_gp_message(c("pkgA", "pkgB"))
    expect_match(msg, "pkgA, pkgB")
  })

  it("uses singular for 1 reverse dep", {
    msg <- revdep_gp_message("pkgA")
    expect_match(msg, "1 reverse dependency")
  })

  it("truncates when more than 10 deps", {
    deps <- paste0("pkg", seq_len(12))
    msg <- revdep_gp_message(deps)
    expect_match(msg, "and 2 more")
  })

  it("mentions revdep_check", {
    msg <- revdep_gp_message("pkgA")
    expect_match(msg, "revdep_check")
  })
})

describe("query_reverse_deps", {

  it("returns reverse deps from provided db", {
    fake_db <- matrix(
      c("pkgA", "1.0", "testpkg"), nrow = 1,
      dimnames = list("pkgA", c("Package", "Version", "Depends"))
    )
    local_mocked_bindings(
      package_dependencies = function(pkg, db, reverse) {
        list(testpkg = "pkgA")
      },
      .package = "tools"
    )
    result <- query_reverse_deps("testpkg", fake_db)
    expect_identical(result, "pkgA")
  })

  it("returns NULL when no reverse deps", {
    fake_db <- matrix(
      character(0), nrow = 0, ncol = 3,
      dimnames = list(NULL, c("Package", "Version", "Depends"))
    )
    local_mocked_bindings(
      package_dependencies = function(pkg, db, reverse) list(testpkg = NULL),
      .package = "tools"
    )
    result <- query_reverse_deps("testpkg", fake_db)
    expect_null(result)
  })
})

describe("revdep prep", {

  it("returns NA when no internet", {
    local_mocked_bindings(has_internet = function() FALSE, .package = "curl")
    state <- list(path = ".")
    state <- PREPS$revdep(state, quiet = TRUE)
    expect_true(identical(state$revdep, NA))
  })
})

att <- function() {
  attach(iris)
  Sepal.Width
  detach(iris)
}

sw <- function() {
  setwd("foobar")
}

sapp <- function() {
  sapply(1:10, print)
  sapply(list(), print)
}

lib <- function() {
  library(igraph)
  require(pkgconfig)
}

ip <- function() {
  installed.packages()
}

ia <- function() {
  system(x <- "echo hello")
}

test_that("checks are not overwritten", {
  skip_on_cran()

  pkgdir <- file.path("..", "..", "R")
  rfiles <- list.files(pkgdir, full.names = TRUE, pattern = "\\.R$")
  skip_if(length(rfiles) == 0, "no R source files found (not running against source tree)")
  rlines <- as.character(unlist(lapply(rfiles, readLines)))
  rwords <- unlist(strsplit(rlines, "\\s+"))
  checks <- grep("CHECKS$", rwords, fixed = TRUE, value = TRUE)
  expect_identical(anyDuplicated(checks), 0L)
})

---
title: "Library calls are OK here"
author: "Gabor Csardi"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Library calls are OK here}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Chapter

## Section 1

```{r}
library(stats)
"This is totally fine"
```

## Section 2

---
title: "Custom Checks"
author: "Hannah Frick"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Custom Checks}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

By default, the `goodpractice()` function and its alias `gp()` run all the checks available in the package (use `all_checks()` to show all checks implemented). In addition, users can provide their own custom checks.

## What's happening inside of `gp()`?

The `gp()` function essentially carries out two types of steps:

- First, it runs preparation steps for the checks, such as calculating test coverage or cyclomatic complexity.
- Then it carries out the checks, e.g., if cyclomatic complexity exceeds a threshold.

The results of the preparation steps and the checks are added to the return object, also referred to as the state. The print method accesses the check results and prints the advice for the failed checks - or praise if none fail.


## Writing custom checks

### Checks without corresponding preparation steps

Custom checks can be created with the `make_check()` function. As inputs it takes a short `description` of the check, the `check` itself, and the `gp` advice to be given in case the check fails. To illustrate this, let's use a simplified version of the check on `T` and `F` instead of `TRUE` and `FALSE`. 

The `check` argument is a function which takes the state as the input and carries out the check, returning `TRUE` if the check was successful. The state includes the path to the source code of the package and the `checkTnF()` function of the **tools** package can be used to check if `T` or `F` was used. 

```{r}
library(goodpractice)

# make a simple version of the T/F check
check_simple_tf <- make_check(
  
  description = "TRUE and FALSE is used, not T and F",
  gp = "avoid 'T' and 'F', use 'TRUE' and 'FALSE' instead.",
  check = function(state) {
      length(tools::checkTnF(dir = state$path)) == 0
  }
)
```

Additional checks can be used in `gp()` via the `extra_checks` argument. This should be a named list of `check` objects as returned by the `make_check()` function. All checks to be carried out, regardless of whether they are provided by the **goodpractice** package or custom checks, must be named in the `checks` argument to `gp()`.

```{r}
# get path to example package
pkg_path <- system.file("bad1", package = "goodpractice")
gp(pkg_path, checks = "simple_tf",
   extra_checks = list(simple_tf = check_simple_tf))
```

### Including a preparation step

Including a preparation step is optional but can be helpful if several checks require the same preparations upfront. In the following example we check for two different fields to be present in the DESCRIPTION file, the URL field and the BugReports field. Both checks can be carried out easily building on a preparation step with the [**desc**](https://cran.r-project.org/package=desc) package for handling DESCRIPTION files.

The checks are linked to the preparation via the prep name: it appears as the `name` argument to `make_prep()`, as the `preps` argument to `make_check()` and finally as the name in the list for the `extra_preps` argument to `gp()`.

```{r}
# prep: process DESCRIPTION file
desc_fun <- function(path, quiet) {
  desc::description$new(path)
}

prep_desc <- make_prep(name = "desc", func = desc_fun)

# check for an URL field
check_url <- make_check(
  description = "URL field in DESCRIPTION",
  preps = "desc",
  gp = "have a URL field in DESCRIPTION",
  check = function(state) state$desc$has_fields("URL")
)

# check for a BugReport field
check_bugreports <- make_check(
  description = "BugReports in DESCRIPTION",
  preps = "desc",
  gp = "add a BugReports field to DESCRIPTION",
  check = function(state) state$desc$has_fields('BugReports')
)

# run the two checks with their corresponding prep step
gp(pkg_path, checks = c("url", "bugreports"),
   extra_preps = list("desc" = prep_desc),
   extra_checks = list("url" = check_url, "bugreports" = check_bugreports))
```

More examples for using custom checks can be found in the [rOpenSci unconf 2017](https://unconf17.ropensci.org/) project [**checkers**](https://github.com/ropensci-archive/checkers) for automated checking of best practices for research compendia.

---
title: "goodpractice"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{goodpractice}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Why goodpractice?

A robust R package needs more than passing `R CMD check`.
Tests give you confidence that things work.
A consistent coding style makes your code readable to others.
Keeping functions short and focused reduces the risk of bugs.
Good metadata — a BugReports URL, properly declared dependencies, documented return values — helps users and reviewers understand what they're working with.

goodpractice checks all of this in one pass.
It bundles `R CMD check` with code coverage ([covr](https://cran.r-project.org/package=covr)), source linting ([lintr](https://cran.r-project.org/package=lintr)), cyclomatic complexity ([cyclocomp](https://cran.r-project.org/package=cyclocomp)), and its own checks for documentation, package structure, and common pitfalls.
Each message tells you _what_ to fix, _why_ it matters, and _where_ in your code to look.
You can also add [custom checks](custom_checks.html) for your team's own conventions.

## Quick start

The main function is `gp()` (short for `goodpractice()`).
Point it at a package directory and it runs the default set of checks:

```{r library}
library(goodpractice)

# goodpractice ships with an example package that has some issues
pkg_path <- system.file("bad1", package = "goodpractice")
```

```{r initial-gp-call-dummy, echo = TRUE, eval = FALSE}
g <- gp(pkg_path)
```
```{r initial-gp-call, echo = FALSE, eval = TRUE}
g <- suppressWarnings(gp(pkg_path))
```

Printing the result shows only the checks that failed:

```{r gp-print}
g
```

Each line starting with a cross is one failed check.
The text after the cross explains what to fix and why.
The indented file paths below it show exactly where in your code the problem was found — in terminals and IDEs that support it, these paths are clickable.
If every check passes, you get a short praise message instead.

## What gp() actually does

When you call `gp()`, two things happen in sequence:

1. **Gather data** — goodpractice runs a set of preparation steps, one per check group.
   One step runs `R CMD check`.
   Another computes code coverage by installing your package and exercising the tests.
   Another lints your source files.
   Each step runs exactly once and stores its results for the checks to use.

2. **Run checks** — each check reads from the stored results and returns pass, fail, or skip.
   A single preparation step can feed many checks — the `rcmdcheck` step alone powers over 200 individual checks, all drawn from a single `R CMD check` run.

This two-step design is why goodpractice can run 250+ checks without repeating expensive work.
It also means you can skip an entire category of checks by excluding its group — more on that below.

## Choosing what to run

By default, `gp()` runs everything in `default_checks()` — about 250 checks covering documentation, code style, test coverage, namespace hygiene, and CRAN compliance:

```{r default-checks-length}
length(default_checks())
```

If you only need a specific check, pass its name to the `checks` argument.
You can find check names with `all_checks()` and filter with `grep()`:

```{r find-url-checks}
# find checks related to URLs
grep("url", all_checks(), value = TRUE)
```

Then run just the ones you care about:

```{r gp-desc-url-only}
g_url <- gp(pkg_path, checks = "description_url")
g_url
```

To go the other direction and run _more_ than the defaults, combine check sets.
For example, to add the optional tidyverse style checks on top of the defaults:

```{r gp-tidyverse-extra-dummy, eval = FALSE}
g <- gp(".", checks = c(default_checks(), tidyverse_checks()))
```

Three helper functions give you the available check names:

- `default_checks()` — the standard set, run by default
- `tidyverse_checks()` — opt-in style checks following [tidyverse conventions](https://style.tidyverse.org/)
- `all_checks()` — both combined

```{r check-lengths}
length(tidyverse_checks())
length(all_checks())
```

## Check groups

Every check belongs to at least one _check group_.
Groups let you work with categories of checks instead of individual names:

```{r all-check-groups}
all_check_groups()
```

To see which checks belong to a group, use `checks_by_group()`:

```{r desc-check-group}
checks_by_group("description")
```

You can select multiple groups at once:

```{r desc-namespace-only-dummy, eval = FALSE}
# run only DESCRIPTION and namespace checks
gp(".", checks = checks_by_group("description", "namespace"))
```

There is also a `describe_check_groups()` function to return full descriptions
of all checks (modified here from default screen output to look nice):

```{r describe-check-groups-dummy, eval = FALSE}
describe_check_groups()
```


```{r kable, results="asis", echo = FALSE}
descs <- describe_check_groups()
df <- data.frame(Group = names(descs), Description = unlist(descs), row.names = NULL)
tbl <- knitr::kable(df)
if (knitr::is_html_output()) {
  tbl <- kableExtra::column_spec(tbl, 1L, extra_css = "white-space: nowrap")
  tbl <- kableExtra::column_spec(tbl, 2L, width = "100%")
}
tbl
```

## Excluding check groups

The checks themselves run in milliseconds — what takes time is the data gathering.
Computing code coverage with `covr` requires installing your package and running all your tests.
Running `R CMD check` via `rcmdcheck` exercises documentation, examples, vignettes, and compilation.
On a large package, these two steps alone can take several minutes.

If you only care about code style or DESCRIPTION metadata right now, you can skip the slow groups entirely.
Set the `goodpractice.exclude_check_groups` option to a character vector of group names:

```{r gp-skip-dummy, eval = FALSE}
# skip coverage and R CMD check — just run the fast checks
options(goodpractice.exclude_check_groups = c("covr", "rcmdcheck"))
gp(".")
```

Every check that depends on a skipped group is automatically excluded too.

For CI/CD pipelines, you can set this via the `GP_EXCLUDE_CHECK_GROUPS` environment variable instead of modifying R code:

```bash
GP_EXCLUDE_CHECK_GROUPS=covr,rcmdcheck Rscript -e 'goodpractice::gp(".")'
```

When both the R option and the environment variable are set, the R option wins.
Exclusion only applies when you use the default check set — if you explicitly pass check names to the `checks` argument, exclusion settings are ignored and those checks always run.

```{r options-dummy, eval = FALSE}
options(goodpractice.exclude_check_groups = "covr")

# covr checks are excluded here (using defaults)
gp(".")

# the "covr" check runs here because we asked for it by name
gp(".", checks = "covr")
```

## Excluding files

If your package contains generated code or vendored files that should not be checked, you can exclude specific files.
Set the `goodpractice.exclude_path` option to a character vector of paths relative to the package root:

```{r options-files-dummy, eval = FALSE}
options(goodpractice.exclude_path = c("R/RcppExports.R", "R/generated_bindings.R"))
gp(".")
```

Or via the `GP_EXCLUDE_PATH` environment variable:

```bash
GP_EXCLUDE_PATH=R/RcppExports.R,R/generated.R Rscript -e 'goodpractice::gp(".")'
```

Excluded files are skipped by lintr, treesitter, expression, and roxygen2 checks.

## Parallel preparation

Preparation steps run sequentially by default.
If you have the [future.apply](https://cran.r-project.org/package=future.apply) package installed, you can run them in parallel by setting a `future::plan()`:

```{r future-dummy, eval = FALSE}
future::plan("multisession")
gp(".")
```

This can significantly speed up runs on large packages where multiple slow preps (covr, rcmdcheck, lintr) would otherwise run one after another.
Preps run in parallel only when a non-sequential plan is active — the default behaviour is unchanged.

## Tidyverse style checks

The default checks deliberately stay away from style preferences — they focus on things that are good practice regardless of how you format your code.

If you or your team follows the [tidyverse style guide](https://style.tidyverse.org/), you can opt into an additional set of style checks:

```{r tidyverse-only-dummy, eval = FALSE}
# add tidyverse checks to the defaults
gp(".", checks = c(default_checks(), tidyverse_checks()))

# or run only the tidyverse checks
gp(".", checks = tidyverse_checks())
```

Most of these are powered by [lintr](https://cran.r-project.org/package=lintr) — brace placement, spacing, naming conventions, and so on.
They run `lintr::lint_package()` once and share the results across all lintr-based checks.
If your package has a `.lintr` configuration file, those settings are respected: disabled linters stay disabled, exclusions are honoured.

A few tidyverse checks go beyond linting and look at package structure directly — whether R file names use snake_case, whether test files mirror source files, whether functions use `missing()` where `NULL` defaults would be clearer, and whether exported functions appear before internal helpers.

All tidyverse checks belong to the `tidyverse` group:

```{r checks-by-group-tidyverse}
checks_by_group("tidyverse")
```

## Working with results

Beyond printing, the object returned by `gp()` gives you programmatic access to the results.

`checks()` returns the names of all checks that were run:

```{r check-names}
checks(g_url)
```

`failed_checks()` returns only the names of the checks that failed:

```{r failed-checks}
failed_checks(g)
```

`results()` gives you a data frame with one row per check and a `passed` column that is `TRUE`, `FALSE`, or `NA` (skipped):

```{r results}
results(g)[1L:5L, ]
```

A check shows `NA` when its preparation step failed or was excluded — it was not evaluated, so it neither passed nor failed.

To see all file positions for failed checks (not just the first five that `print()` shows), use `print()` with a higher limit:

```{r print-inf-dummy, eval = FALSE}
print(g, positions_limit = Inf)
```

You can also export the full results to JSON for use in other tools:

```{r export-json-dummy, eval = FALSE}
export_json(g, "gp_results.json")
```

## Custom checks

goodpractice is extensible — you can define your own checks and preparation steps to enforce team-specific conventions.
The `gp` string in each check supports [cli inline markup](https://cli.r-lib.org/reference/inline-markup.html) — use `{.fn func}` for function names, `{.code expression}` for code, `{.file path}` for file paths, `{.field name}` for field names, and `{.url url}` for URLs.

See `vignette("custom_checks")` for the full guide.

---
title: "'goodpractice' for developers"
author: "Mark Padgham"
date: "`r Sys.Date()`"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{'goodpractice' for developers}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

```{r windowns-flag, echo = FALSE}
# The bash commands are likely to fail on winddows:
on_windows <- identical(.Platform$OS.type, "windows")
```


## The 'Makefile'

The 'goodpractice' repository includes a ['Makefile'](https://www.gnu.org/software/make/manual/make.html#Introduction).
This allows many commands to be run directly from a shell console (rather than within an R session).
The default behaviour is to list all options:

```{bash, eval = FALSE}
make
```


```{bash, echo = FALSE, output = TRUE, eval = !on_windows}
cd ..
make help | sed -E 's/\x1b\[[0-9;]*m//g'
```

Any of those options can then be run as `make <command>`.
For example, this command checks for any issues with the `pkgdown` site:
```{bash, eval = FALSE}
make pkgdowncheck
```
```{r, echo = FALSE, eval = TRUE, output = TRUE}
cli::cli_alert_success("No problems found")
```

Note that some of the `Makefile` options depend on additional packages which may need to be installed.
These include:

- [`allcontributors`](https://docs.ropensci.org/allcontributors/)
- [`pkgcheck`](https://docs.ropensci.org/pkgcheck/)
- [`pkgdown`](https://pkgdown.r-lib.org/)

---

## Adding new checks

Checks are defined in groups.
Each group is generally defined within two files:
1. A `prep_<group>.R` file for collecting data needed for the check
2. A `chk_<group>.R` file defining the output structure of the check.

The `R/` directory here consistns almost entirely of paired files defining each check group:
```{bash, echo = FALSE, output = TRUE, eval = !on_windows}
ls ../R
```

An example is the `cyclocomp` check, which is largely a wrapper around the external [`cyclocomp` package](https://github.com/gaborcsardi/cyclocomp).
The `prep_cyclocomp.R` file looks like this:
```{bash, echo = FALSE, output = TRUE, eval = !on_windows}
cat ../R/prep_cyclocomp.R
```

That preparation steps runs the `cyclocomp::cyclocomp_package_dir()` function in the source directory of the package being checked.
The corresponding `check_cyclocomp.R` file then uses a variable called `state` that is accessible to all functions within the package.
The line above `PREPS$cyclocomp <-` creates an entry of `state$cyclocomp` containing the results of the preparation step that can then be accessed to defined the check output in `R/chk_cyclocomp.R`:

```{bash, echo = FALSE, output = TRUE, eval = !on_windows}
cat ../R/chk_cyclocomp.R
```

All checks following this general pattern of defining `PREPS$<group> <- ` entries containing data from the check, then defining `CHECKS$<group>` entries for output format, using `state$<group>` to access the check data.
The best way to learn about check structure is to examine some of the existing files.

---
title: "goodpractice"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{goodpractice}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r setup, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>"
)
```

## Why goodpractice?

A robust R package needs more than passing `R CMD check`.
Tests give you confidence that things work.
A consistent coding style makes your code readable to others.
Keeping functions short and focused reduces the risk of bugs.
Good metadata — a BugReports URL, properly declared dependencies, documented return values — helps users and reviewers understand what they're working with.

goodpractice checks all of this in one pass.
It bundles `R CMD check` with code coverage ([covr](https://cran.r-project.org/package=covr)), source linting ([lintr](https://cran.r-project.org/package=lintr)), cyclomatic complexity ([cyclocomp](https://cran.r-project.org/package=cyclocomp)), and its own checks for documentation, package structure, and common pitfalls.
Each message tells you _what_ to fix, _why_ it matters, and _where_ in your code to look.
You can also add [custom checks](custom_checks.html) for your team's own conventions.

## Quick start

The main function is `gp()` (short for `goodpractice()`).
Point it at a package directory and it runs the default set of checks:

```{r library}
library(goodpractice)

# goodpractice ships with an example package that has some issues
pkg_path <- system.file("bad1", package = "goodpractice")
```

```{r initial-gp-call-dummy, echo = TRUE, eval = FALSE}
g <- gp(pkg_path)
```
```{r initial-gp-call, echo = FALSE, eval = TRUE}
g <- suppressWarnings(gp(pkg_path))
```

Printing the result shows only the checks that failed:

```{r gp-print}
g
```

Each line starting with a cross is one failed check.
The text after the cross explains what to fix and why.
The indented file paths below it show exactly where in your code the problem was found — in terminals and IDEs that support it, these paths are clickable.
If every check passes, you get a short praise message instead.

## What gp() actually does

When you call `gp()`, two things happen in sequence:

1. **Gather data** — goodpractice runs a set of preparation steps, one per check group.
   One step runs `R CMD check`.
   Another computes code coverage by installing your package and exercising the tests.
   Another lints your source files.
   Each step runs exactly once and stores its results for the checks to use.

2. **Run checks** — each check reads from the stored results and returns pass, fail, or skip.
   A single preparation step can feed many checks — the `rcmdcheck` step alone powers over 200 individual checks, all drawn from a single `R CMD check` run.

This two-step design is why goodpractice can run 250+ checks without repeating expensive work.
It also means you can skip an entire category of checks by excluding its group — more on that below.

## Choosing what to run

By default, `gp()` runs everything in `default_checks()` — about 250 checks covering documentation, code style, test coverage, namespace hygiene, and CRAN compliance:

```{r default-checks-length}
length(default_checks())
```

If you only need a specific check, pass its name to the `checks` argument.
You can find check names with `all_checks()` and filter with `grep()`:

```{r find-url-checks}
# find checks related to URLs
grep("url", all_checks(), value = TRUE)
```

Then run just the ones you care about:

```{r gp-desc-url-only}
g_url <- gp(pkg_path, checks = "description_url")
g_url
```

To go the other direction and run _more_ than the defaults, combine check sets.
For example, to add the optional tidyverse style checks on top of the defaults:

```{r gp-tidyverse-extra-dummy, eval = FALSE}
g <- gp(".", checks = c(default_checks(), tidyverse_checks()))
```

Three helper functions give you the available check names:

- `default_checks()` — the standard set, run by default
- `tidyverse_checks()` — opt-in style checks following [tidyverse conventions](https://style.tidyverse.org/)
- `all_checks()` — both combined

```{r check-lengths}
length(tidyverse_checks())
length(all_checks())
```

## Check groups

Every check belongs to at least one _check group_.
Groups let you work with categories of checks instead of individual names:

```{r all-check-groups}
all_check_groups()
```

To see which checks belong to a group, use `checks_by_group()`:

```{r desc-check-group}
checks_by_group("description")
```

You can select multiple groups at once:

```{r desc-namespace-only-dummy, eval = FALSE}
# run only DESCRIPTION and namespace checks
gp(".", checks = checks_by_group("description", "namespace"))
```

There is also a `describe_check_groups()` function to return full descriptions
of all checks (modified here from default screen output to look nice):

```{r describe-check-groups-dummy, eval = FALSE}
describe_check_groups()
```


```{r kable, results="asis", echo = FALSE}
descs <- describe_check_groups()
df <- data.frame(Group = names(descs), Description = unlist(descs), row.names = NULL)
tbl <- knitr::kable(df)
if (knitr::is_html_output()) {
  tbl <- kableExtra::column_spec(tbl, 1L, extra_css = "white-space: nowrap")
  tbl <- kableExtra::column_spec(tbl, 2L, width = "100%")
}
tbl
```

## Excluding check groups

The checks themselves run in milliseconds — what takes time is the data gathering.
Computing code coverage with `covr` requires installing your package and running all your tests.
Running `R CMD check` via `rcmdcheck` exercises documentation, examples, vignettes, and compilation.
On a large package, these two steps alone can take several minutes.

If you only care about code style or DESCRIPTION metadata right now, you can skip the slow groups entirely.
Set the `goodpractice.exclude_check_groups` option to a character vector of group names:

```{r gp-skip-dummy, eval = FALSE}
# skip coverage and R CMD check — just run the fast checks
options(goodpractice.exclude_check_groups = c("covr", "rcmdcheck"))
gp(".")
```

Every check that depends on a skipped group is automatically excluded too.

For CI/CD pipelines, you can exclude check groups by specifying them in comma-separate format via the `GP_EXCLUDE_CHECK_GROUPS` environment variable:

```bash
GP_EXCLUDE_CHECK_GROUPS='covr,lintr,rcmdcheck'
```

When both the R option and the environment variable are set, the R option wins.
Exclusion only applies when you use the default check set — if you explicitly pass check names to the `checks` argument, exclusion settings are ignored and those checks always run.

```{r options-dummy, eval = FALSE}
options(goodpractice.exclude_check_groups = "covr")

# covr checks are excluded here (using defaults)
gp(".")

# the "covr" check runs here because we asked for it by name
gp(".", checks = "covr")
```

## Excluding files

If your package contains generated code or vendored files that should not be checked, you can exclude specific files.
Set the `goodpractice.exclude_path` option to a character vector of paths relative to the package root:

```{r options-files-dummy, eval = FALSE}
options(goodpractice.exclude_path = c("R/RcppExports.R", "R/generated_bindings.R"))
gp(".")
```

Or via the `GP_EXCLUDE_PATH` environment variable:

```bash
GP_EXCLUDE_PATH=R/RcppExports.R,R/generated.R Rscript -e 'goodpractice::gp(".")'
```

Excluded files are skipped by lintr, treesitter, expression, and roxygen2 checks.

## Parallel preparation

Preparation steps run sequentially by default.
If you have the [future.apply](https://cran.r-project.org/package=future.apply) package installed, you can run them in parallel by setting a `future::plan()`:

```{r future-dummy, eval = FALSE}
future::plan("multisession")
gp(".")
```

This can significantly speed up runs on large packages where multiple slow preps (covr, rcmdcheck, lintr) would otherwise run one after another.
Preps run in parallel only when a non-sequential plan is active — the default behaviour is unchanged.

## Tidyverse style checks

The default checks deliberately stay away from style preferences — they focus on things that are good practice regardless of how you format your code.

If you or your team follows the [tidyverse style guide](https://style.tidyverse.org/), you can opt into an additional set of style checks:

```{r tidyverse-only-dummy, eval = FALSE}
# add tidyverse checks to the defaults
gp(".", checks = c(default_checks(), tidyverse_checks()))

# or run only the tidyverse checks
gp(".", checks = tidyverse_checks())
```

Most of these are powered by [lintr](https://cran.r-project.org/package=lintr) — brace placement, spacing, naming conventions, and so on.
They run `lintr::lint_package()` once and share the results across all lintr-based checks.
If your package has a `.lintr` configuration file, those settings are respected: disabled linters stay disabled, exclusions are honoured.

A few tidyverse checks go beyond linting and look at package structure directly — whether R file names use snake_case, whether test files mirror source files, whether functions use `missing()` where `NULL` defaults would be clearer, and whether exported functions appear before internal helpers.

All tidyverse checks belong to the `tidyverse` group:

```{r checks-by-group-tidyverse}
checks_by_group("tidyverse")
```

## Working with results

Beyond printing, the object returned by `gp()` gives you programmatic access to the results.

`checks()` returns the names of all checks that were run:

```{r check-names}
checks(g_url)
```

`failed_checks()` returns only the names of the checks that failed:

```{r failed-checks}
failed_checks(g)
```

`results()` gives you a data frame with one row per check and a `passed` column that is `TRUE`, `FALSE`, or `NA` (skipped):

```{r results}
results(g)[1L:5L, ]
```

A check shows `NA` when its preparation step failed or was excluded — it was not evaluated, so it neither passed nor failed.

To see all file positions for failed checks (not just the first five that `print()` shows), use `print()` with a higher limit:

```{r print-inf-dummy, eval = FALSE}
print(g, positions_limit = Inf)
```

You can also export the full results to JSON for use in other tools:

```{r export-json-dummy, eval = FALSE}
export_json(g, "gp_results.json")
```

## Custom checks

goodpractice is extensible — you can define your own checks and preparation steps to enforce team-specific conventions.
The `gp` string in each check supports [cli inline markup](https://cli.r-lib.org/reference/inline-markup.html) — use `{.fn func}` for function names, `{.code expression}` for code, `{.file path}` for file paths, `{.field name}` for field names, and `{.url url}` for URLs.

See `vignette("custom_checks")` for the full guide.

---
name: goodpractice4agents
description: Use the goodpractice R package to identify package issues, then repair them. Trigger whenever an agent is asked to fix, clean up, reduce, or pass goodpractice, lintr, or R CMD check failures in an R package, or to bring a package up to good-practice standards.
compatibility: Requires the goodpractice R package installed, with read and write access within an R package repository.
---

# Fix issues identified by 'goodpractice' package

**AGENT** These are instructions for you to modify the R package code in the
current directory. You will be using the `goodpractice` tool which flags
several categories (`groups`) of common issues in R package development. Each
"issue" is called a failure, and your job is to modify the code to ensure as
few failures as possible.

Only ever load the package you are checking with `devtools::load_all()`, never
`R CMD install`. `load_all()` reflects your in-progress edits immediately
without a build step, so each re-check sees your latest changes; an installed
copy would mask them until rebuilt.

## Important Note

It may not be possible to fix all failures, and judgement will sometimes be
necessary. This is particularly likely for failures in the `lintr` group. It
may be impossible to get `lintr` checks to run with no failures. You should
nevertheless try hard. For example, line length linters often fail for extended
URLs. If they are part of normal code, they may be broken by using `paste()` or
similar, but if they are embedded in documentation within a `\url{}`, they will
need to be left as is, and consequently fail. Similar kinds of subtleties will
apply to many other linters.

If a linter appears to be giving a false failure, you may append a terminal `#
nolint` to the end of the line. Avoid wrapping blocks between `# nolint start`
and `# nolint end`: a terminal `# nolint` suppresses exactly one line you have
judged, whereas a start/end block silently mutes every linter across a region,
which is how genuine issues get hidden.

## Questions for user

**AGENT**: You are to ask the user the following two questions first, and then
follow these instructions in sequence.

The questions you must pose are in italics here, anything after the italicised
text provides additional context for you, and not for the user. Ask each
question individually, and request a simple "y/n" response for each before
proceeding.

1. _Do you want to include "tidyverse" checks (y/n)?_
2. _Do you want me to report results of each check group and wait? (y/n)?_

If the user answers no to question 2, additionally ask:

2a. _Do you want separate commits for each check group (y), or a single commit
with all changes (n)?_

## Tasks for agent

1. Load the `goodpractice` package and identify check groups by running this R
   code:
    ``` r
    library(goodpractice)
    groups <- all_check_groups()
    ```
    Then exclude `c("covr", "cyclocomp", "rcmdcheck", "revdep")` from `groups`
    If the user wants to exclude "tidyverse" checks, also exclude those from
    `groups`.
2. **Stop and tell user** which check groups are going to be run, and ask
   whether they'd like to exclude any of those groups. Exclude any indicated
   ones from `groups` as well.
3. Unless the user says otherwise, keep groups in default order, but with
   `lintr` as the last group. If the user wants to run `tidyverse` checks,
   append that as the very last check group.

**The following checks are to be run for each `group` in `groups`:**

4. Run `g <- gp(checks = checks_by_group(group))`
5. Export checks to a temp JSON file: `file <- tempfile(fileext = ".json"); export_json(g, file)`.
6. Read all failures as `failures <- jsonlite::read_json(file)$failures` (AGENT:
   use your own tools to read this JSON file). The name of each item in
   `failures` is called `nm` from here on. As you progress through the groups,
   keep tallies of numbers of failures in each group, and the total across all
   groups.

**The following checks are to be run for each `failure` in `failures` (with name `nm`):**

7. Get description of failure and recommended fix with `describe_check(nm)`
8. For each `instance` of that failure type, extract `filename` and `line`.
    8a. If `filename` is in a location ignored by `.Rbuildignore`, ignore that
        failure and proceed to next `instance`.
    8b. If `filename` is in `man/` directory and package uses `roxygen2`
        (indicated in `DESCRIPTION` file), then do not edit `man/` files
        directly, but edit at source in `R/`, and run `roxygen2::roxygenise()`
        to update corresponding `man/` entry.
    8c. If it's a `roxygen2`-group check for `@inheritParams` instead of
        duplicated parameters, only automatically do that if parameter
        descriptions are identical. Use the `get_dup_params.R` code in "Useful
        functions" below to extract those. If they seem very close, use your
        judgement, or ask user if you are unsure.
9. Go to that location and fix the `failure`, then back to 7. for next `failure`

Once you've finished all `failures` in one group:

10. If `group` is one of "lintr", or "package_structure", run
    `devtools::load_all(); testthat::test_package()` to make sure nothing
    breaks. For any test failures, make sure you first understand why your
    changes caused those, and only update test code if it was obviously
    designed to respond to previous conditions that no longer apply.
11. If `group` is `roxygen2`, then run `roxygen2::roxygenise()` to update
    corresponding `man/` entries.
12. Rerun `gp(checks = checks_by_group(group))`, and extract number of
    remaining failures from the resultant JSON file. Provide a summary to user,
    noting initial and final number of failing checks. If the user asked you to
    stop after each `group`, do that, but ensure you remind them to keep going
    once they seem satisfied.
13. Loop back to next `group` in `groups`

Once you've finished all of the initial groups - that is, those minus the
excluded groups:

14. Repeat steps 4 - 13 for `gp(checks=checks_by_group(g))` for each of the
    three remaining groups, `g = c("covr", "cyclocomp", "rcmdcheck")`. Execute
    all steps for each single group. (`revdep` was excluded in step 1 and is
    deliberately not re-run here: it checks reverse dependencies, which is slow,
    network-dependent, and outside the scope of fixing this package's own code.)

### Final steps

15. Remove any terminal blank spaces at any line ends, and terminal blank lines
    in files.
16. Summarise all updates for the user.

---

## Useful functions

**get_dup_params.R**

``` r
library(goodpractice)
ns <- getNamespace("goodpractice")
g <- gp(checks = "roxygen2_duplicate_params")
blocks <- g[["roxygen2"]][["blocks"]]
extract_block_params <- get("extract_block_params", envir = ns)

all_params <- unlist(lapply(blocks, extract_block_params), recursive = FALSE)
keys <- vapply(all_params, function(p) paste(p[["name"]], p[["description"]], sep = "|||"), character(1))
files <- vapply(all_params, function(p) p[["file"]], character(1))
lines_v <- vapply(all_params, function(p) as.character(p[["line"]]), character(1))
duped_keys <- unique(keys[duplicated(keys)])

for (key in duped_keys) {
    idxs <- which(keys == key)
    dup_files <- unique(basename(files[idxs]))
    if (length(dup_files) < 2) next
    pname <- sub("[|][|][|].*", "", key)
    pdesc <- sub(".*[|][|][|]", "", key)
    cat(sprintf("PARAM: %s\n  DESC: %s\n", pname, substr(pdesc, 1, 120)))
    for (i in idxs) {
        cat(sprintf("  -> %s line %s\n", basename(files[i]), lines_v[i]))
    }
    cat("\n")
}
```

# describe_check

    Code
      describe_check(all_checks())
    Output
      $no_obsolete_deps
      [1] "No obsolete or retired package dependencies"
      
      $print_return_invisible
      [1] "Print methods return the object invisibly"
      
      $on_exit_has_add
      [1] "on.exit() calls include add = TRUE"
      
      $complexity_function_length
      [1] "Functions are not too long"
      
      $complexity_unused_internal
      [1] "All internal functions are used"
      
      $duplicate_function_bodies
      [1] "No functions with identical bodies"
      
      $covr
      [1] "All code is unit tested"
      
      $cyclocomp
      [1] "Functions are simple"
      
      $no_description_depends
      [1] "No \"Depends\" in DESCRIPTION"
      
      $no_description_date
      [1] "No \"Date\" in DESCRIPTION"
      
      $description_url
      [1] "URL in DESCRIPTION"
      
      $description_not_start_with_package
      [1] "Description does not start with package name reference"
      
      $description_urls_in_angle_brackets
      [1] "URLs in Description are enclosed in angle brackets"
      
      $description_doi_format
      [1] "DOIs in Description use <doi:...> format"
      
      $description_urls_not_http
      [1] "URLs in Description use https not http"
      
      $no_description_duplicate_deps
      [1] "No duplicate dependencies in DESCRIPTION"
      
      $description_valid_roles
      [1] "Authors have valid roles in DESCRIPTION"
      
      $description_pkgname_single_quoted
      [1] "Package names in Title/Description are single-quoted"
      
      $description_bugreports
      [1] "BugReports in DESCRIPTION"
      
      $has_readme
      [1] "README file exists"
      
      $has_news
      [1] "NEWS file exists"
      
      $r_file_extension
      [1] "R scripts use .R file extension, not .r or .q"
      
      $lintr_assignment_linter
      [1] "'<-' and not '=' is used for assignment"
      
      $lintr_line_length_linter
      [1] "Code lines are short"
      
      $lintr_semicolon_linter
      [1] "No trailing semicolons"
      
      $lintr_attach_detach_linter
      [1] "Avoid attach and detach"
      
      $lintr_setwd_linter
      [1] "Avoid setwd in R packages"
      
      $lintr_sapply_linter
      [1] "Avoid sapply"
      
      $lintr_library_require_linter
      [1] "Avoid library and require in packages"
      
      $installed_packages_linter
      [1] "Avoid calling installed.packages"
      
      $lintr_seq_linter
      [1] "Avoid 1:length(...) and similar expressions"
      
      $lintr_any_duplicated_linter
      [1] "Use anyDuplicated() instead of any(duplicated())"
      
      $lintr_any_is_na_linter
      [1] "Use anyNA() instead of any(is.na())"
      
      $lintr_class_equals_linter
      [1] "Use inherits() instead of class() =="
      
      $lintr_condition_message_linter
      [1] "Build condition messages correctly"
      
      $lintr_duplicate_argument_linter
      [1] "No duplicate arguments in function calls"
      
      $lintr_equals_na_linter
      [1] "Use is.na() not == NA"
      
      $lintr_fixed_regex_linter
      [1] "Use fixed strings instead of regex where possible"
      
      $lintr_for_loop_index_linter
      [1] "Do not overwrite existing variables with for loop index"
      
      $lintr_length_test_linter
      [1] "Use length() correctly in conditions"
      
      $lintr_matrix_apply_linter
      [1] "Use colSums/rowSums/colMeans/rowMeans instead of apply()"
      
      $lintr_missing_argument_linter
      [1] "No missing arguments in function calls"
      
      $lintr_nrow_subset_linter
      [1] "Use nrow(subset) instead of nrow() after subsetting"
      
      $lintr_redundant_equals_linter
      [1] "No redundant comparisons to TRUE or FALSE"
      
      $lintr_redundant_ifelse_linter
      [1] "No redundant ifelse() calls"
      
      $lintr_regex_subset_linter
      [1] "Use grep() with value = TRUE instead of subsetting"
      
      $lintr_sort_linter
      [1] "Use sort() or is.unsorted() instead of order()"
      
      $lintr_system_file_linter
      [1] "Use system.file() correctly"
      
      $lintr_terminal_close_linter
      [1] "Do not close connections redundantly"
      
      $lintr_which_grepl_linter
      [1] "Use grep() instead of which(grepl())"
      
      $lintr_boolean_arithmetic_linter
      [1] "Avoid arithmetic on logical vectors"
      
      $lintr_comparison_negation_linter
      [1] "Use natural comparison operators instead of negation"
      
      $lintr_consecutive_assertion_linter
      [1] "Combine consecutive stopifnot() calls"
      
      $lintr_if_not_else_linter
      [1] "Avoid if (!cond) { a } else { b }"
      
      $lintr_ifelse_censor_linter
      [1] "Use pmin()/pmax() instead of ifelse() for clamping"
      
      $lintr_implicit_assignment_linter
      [1] "Avoid implicit assignments in function calls"
      
      $lintr_inner_combine_linter
      [1] "Combine inside c() not outside"
      
      $lintr_length_levels_linter
      [1] "Use nlevels() instead of length(levels())"
      
      $lintr_literal_coercion_linter
      [1] "Use typed literals instead of coercion functions"
      
      $lintr_nested_ifelse_linter
      [1] "Avoid deeply nested ifelse() calls"
      
      $lintr_nested_pipe_linter
      [1] "Avoid pipes inside other pipes"
      
      $lintr_numeric_leading_zero_linter
      [1] "Include leading zero in decimal numbers"
      
      $lintr_outer_negation_linter
      [1] "Negate at the outer level"
      
      $lintr_paste_linter
      [1] "Use paste0() or file.path() instead of paste() with sep"
      
      $lintr_scalar_in_linter
      [1] "Use == instead of %in% for scalar comparison"
      
      $lintr_undesirable_operator_linter
      [1] "Avoid undesirable operators"
      
      $lintr_unnecessary_concatenation_linter
      [1] "Avoid unnecessary c() calls"
      
      $lintr_unnecessary_lambda_linter
      [1] "Avoid unnecessary anonymous functions"
      
      $lintr_unreachable_code_linter
      [1] "No unreachable code after return/stop/next"
      
      $lintr_conjunct_test_linter
      [1] "Use separate expect_ calls instead of &&"
      
      $lintr_expect_comparison_linter
      [1] "Use expect_gt/lt/gte/lte instead of expect_true with comparison"
      
      $lintr_expect_identical_linter
      [1] "Use expect_identical() instead of expect_equal() where possible"
      
      $lintr_expect_length_linter
      [1] "Use expect_length() instead of expect_equal(length())"
      
      $lintr_expect_named_linter
      [1] "Use expect_named() instead of expect_equal(names())"
      
      $lintr_expect_not_linter
      [1] "Use expect_false() instead of expect_true(!x)"
      
      $lintr_expect_null_linter
      [1] "Use expect_null() instead of expect_equal(x, NULL)"
      
      $lintr_expect_s3_class_linter
      [1] "Use expect_s3_class() instead of expect_equal(class())"
      
      $lintr_expect_s4_class_linter
      [1] "Use expect_s4_class() instead of expect_true(is())"
      
      $lintr_expect_true_false_linter
      [1] "Use expect_true()/expect_false() instead of expect_equal(x, TRUE)"
      
      $lintr_expect_type_linter
      [1] "Use expect_type() instead of expect_equal(typeof())"
      
      $no_import_package_as_a_whole
      [1] "Packages are not imported as a whole"
      
      $no_export_pattern
      [1] "exportPattern in NAMESPACE file"
      
      $rcmdcheck_package_directory_exists
      [1] "Package directory exists"
      
      $rcmdcheck_vignette_leftovers
      [1] "No vignette leftovers"
      
      $rcmdcheck_qpdf_present_for_checking
      [1] "'qpdf' must be installed for checks"
      
      $rcmdcheck_vignettes_at_right_place
      [1] "Vignettes are in 'vignettes'"
      
      $rcmdcheck_portable_file_names
      [1] "File names are portable"
      
      $rcmdcheck_duplicate_file_names
      [1] "No file names that only differ in case"
      
      $rcmdcheck_ascii_file_names
      [1] "File names are portable (ASCII)"
      
      $rcmdcheck_portable_paths
      [1] "Paths are short enough to be portable"
      
      $rcmdcheck_proper_permissions
      [1] "Files have correct permissions"
      
      $rcmdcheck_executable_files
      [1] "configure and cleanup files must be executable"
      
      $rcmdcheck_description_meta
      [1] "DESCRIPTION meta information is correct"
      
      $rcmdcheck_malformed_title_or_description
      [1] "Title and Description fields are valid"
      
      $rcmdcheck_portable_description_encoding
      [1] "Portable DESCRIPTION Encoding"
      
      $rcmdcheck_description_encoding_known
      [1] "Encoding of DESCRIPTION is defined"
      
      $rcmdcheck_description_ascii_tags
      [1] "All tags in DESCRIPTION are ASCII"
      
      $rcmdcheck_description_ascii_fields
      [1] "All tags in DESCRIPTION are ASCII, or Encoding is defined"
      
      $rcmdcheck_standardizable_license
      [1] "Package license is standardizable"
      
      $rcmdcheck_license_file_pointers
      [1] "License file pointers are correct"
      
      $rcmdcheck_non_standard_license
      [1] "Non-standard, but standardizable license"
      
      $rcmdcheck_deprecated_license
      [1] "Deprecated license"
      
      $rcmdcheck_not_permitted_license_restrictions
      [1] "Not permitted license restrictions"
      
      $rcmdcheck_template_licenses
      [1] "Template licenses have a LICENSE file"
      
      $rcmdcheck_authors_at_r_field
      [1] "Valid Authors@R field"
      
      $rcmdcheck_valid_author_in_authors_at_r
      [1] "Can extract Author from Authors@R"
      
      $rcmdcheck_author_in_authors_at_r
      [1] "Author is in Authors@R"
      
      $rcmdcheck_authors_at_r_valid_roles
      [1] "Authors@R has valid roles for all authors"
      
      $rcmdcheck_maintainer_in_authors_at_r
      [1] "Authors@R has package Maintainer"
      
      $rcmdcheck_single_maintainer_in_authors_at_r
      [1] "Authors@R has one package Maintainer only"
      
      $rcmdcheck_valid_maintainer_in_authors_at_t
      [1] "Authors@R has a valid Maintainer role person"
      
      $rcmdcheck_stale_author
      [1] "No stale Author field if Authors@R is present"
      
      $rcmdcheck_stale_maintainer
      [1] "No state Maintainer field if Authors@R is present"
      
      $rcmdcheck_needscompilation_field_valid
      [1] "NeedsCompilation field must be 'yes' or 'no'"
      
      $rcmdcheck_multiple_dependency_modes
      [1] "Packages should be listed as one type of dependency (except for LinkingTo)"
      
      $rcmdcheck_unused_linkingto
      [1] "LinkingTo field makes sense, there are headers to include"
      
      $rcmdcheck_versioned_linkingto
      [1] "Versioned 'LinkingTo' needs R >= 3.0.2"
      
      $rcmdcheck_linkingto_without_compiled_code
      [1] "'LinkingTo' field only makes sense if package has compiled code"
      
      $rcmdcheck_broken_vignettes_1
      [1] "Vignette output not in build/vignette.rds"
      
      $rcmdcheck_broken_vignettes_2
      [1] "All outputs from build/vignette.rds are in the package"
      
      $rcmdcheck_defunct_top_level
      [1] "Some top level files are defunct: install.R, R_PROFILE.R"
      
      $rcmdcheck_valid_readme_and_news
      [1] "README.md and NEWS.md must work with pandoc"
      
      $rcmdcheck_stale_license_file
      [1] "LICENSE (and LICENCE) must be mentioned in DESCRIPTION"
      
      $rcmdcheck_stale_inst_license_file
      [1] "inst/LICENSE (and inst/LICENCE) must be mentioned in DESCRIPTION"
      
      $rcmdcheck_non_standard_top_level_files
      [1] "Top-level files are standardized"
      
      $rcmdcheck_copyright_file_in_inst
      [1] "Copyright file should be in inst/"
      
      $rcmdcheck_authors_in_inst
      [1] "AUTHORS file should be in inst/"
      
      $rcmdcheck_leftover_files
      [1] "Check for left-over files"
      
      $rcmdcheck_empty_index_file
      [1] "INDEX file must not be empty if present"
      
      $rcmdcheck_empty_demo_index_file
      [1] "demo/00Index must not be empty if present"
      
      $rcmdcheck_proper_demo_index_file
      [1] "demo/00Index file is valid if it exists"
      
      $rcmdcheck_demos_are_indexed
      [1] "All demos are in demo/00Index"
      
      $rcmdcheck_stale_demos_in_index
      [1] "Demos in the index exist"
      
      $rcmdcheck_vignette_index
      [1] "Vignettes have index entries"
      
      $rcmdcheck_has_r_source_files
      [1] "Package must have R source files (or R/sysdata.rda) if R directory is present"
      
      $rcmdcheck_r_directory_uppercase
      [1] "R directory must be upper case"
      
      $rcmdcheck_man_directory_lowercase
      [1] "man directory must be lower case"
      
      $rcmdcheck_data_directory_lowercase
      [1] "data directory must be lower case"
      
      $rcmdcheck_rcheck_directories
      [1] "No leftover .Rcheck directories"
      
      $rcmdcheck_leftover_dvi_pdf_build_directories
      [1] "No leftover Rd2dvi or Rd2pdf directories"
      
      $rcmdcheck_leftover_vs_directories
      [1] "No leftover .git, .svn (etc.) version control directories"
      
      $rcmdcheck_invalid_file_names
      [1] "File names in some directories are restricted"
      
      $rcmdcheck_empty_data_directory
      [1] "data directory must not be empty, if it exists"
      
      $rcmdcheck_empty_demo_directory
      [1] "demo directory must not be empty, if it exists"
      
      $rcmdcheck_demos_are_ascii
      [1] "Demos should use ASCII characters"
      
      $rcmdcheck_demos_have_valid_code
      [1] "Demos should contain valid R code"
      
      $rcmdcheck_empty_exec_directory
      [1] "exec directory must not be empty, if it exists"
      
      $rcmdcheck_empty_inst_directory
      [1] "inst directory must not be empty, if it exists"
      
      $rcmdcheck_src_without_sources
      [1] "src directory must contain some source files, if it exists"
      
      $rcmdcheck_inst_interference
      [1] "Some directory names in inst are not allowed"
      
      $rcmdcheck_news_rd_is_valid
      [1] "inst/NEWS.Rd is valid if present"
      
      $rcmdcheck_citation_file_is_valid
      [1] "inst/CITATION is valid if present"
      
      $rcmdcheck_citation_file_at_standard_place
      [1] "CITATION is in inst/, not somewhere else"
      
      $rcmdcheck_r_files_are_ascii
      [1] "R files should be ASCII"
      
      $rcmdcheck_errors_in_r_files
      [1] "R files do not have syntax errors"
      
      $rcmdcheck_warnings_in_r_files
      [1] "R files do not throw warnings"
      
      $rcmdcheck_undeclared_imports
      [1] "'::' and ':::' imports must be declared"
      
      $rcmdcheck_undeclared_library_require
      [1] "'library' and 'require' calls must be declared"
      
      $rcmdcheck_undeclared_loadnamespace_requirenamespace
      [1] "'loadNamespace' and 'requireNamespace' calls must be declared"
      
      $rcmdcheck_library_require_to_attached
      [1] "'library' and 'require' calls to attached packages are not needed"
      
      $rcmdcheck_library_require_in_package_code
      [1] "Do not use 'library' or 'require' in package code. Use '::' or 'requireNamespace' instead"
      
      $rcmdcheck_imports_not_imported_from
      [1] "Packages declared in 'Imports' must be imported from"
      
      $rcmdcheck_depends_not_imported_from
      [1] "Packages in 'Depends' must be imported from"
      
      $rcmdcheck_missing_or_unexported_objects
      [1] "Exported objects must be present"
      
      $rcmdcheck_unneeded_triple_colon
      [1] "Use '::' to call exported functions, not ':::'"
      
      $rcmdcheck_triple_colon_imported_objects_exist
      [1] "All objects imported via ':::' must exist"
      
      $rcmdcheck_unexported_base_objects_imported
      [1] "':::' imports to unexported objects, from base packages"
      
      $rcmdcheck_unexported_objects_imported
      [1] "':::' imports to unexported objects"
      
      $rcmdcheck_triple_colon_to_itself
      [1] "':::' imports to the package itself are usually not needed"
      
      $rcmdcheck_triple_colon_to_unknown
      [1] "':::' imports to unavailable packages"
      
      $rcmdcheck_data_call_to_undeclared_package
      [1] "Packages used in 'data(package=)' must be declared"
      
      $rcmdcheck_exported_s3_methods_are_registered
      [1] "Exported S3 methods are registered"
      
      $rcmdcheck_replacement_function_arg_names
      [1] "Replacement functions must have a 'value' argument"
      
      $rcmdcheck_foreign_calls_have_package_argument
      [1] "Foreign calls must have a 'PACKAGE' argument"
      
      $rcmdcheck_foreign_call_empty_package_argument
      [1] "'PACKAGE' argument of foreign calls must not be empty"
      
      $rcmdcheck_foreign_call_to_base_package
      [1] "Foreign calls should not call base packages"
      
      $rcmdcheck_foreign_calls_to_another_package
      [1] "Should not call another package with foreign calls"
      
      $rcmdcheck_foreign_calls_to_unknown_package
      [1] "Packages called from foreign calls must be declared"
      
      $rcmdcheck_foreign_registration_problems
      [1] "Check for registration problems of foreign function calls"
      
      $rcmdcheck_foreign_calls_with_dup
      [1] "The 'DUP' argument of foreign calls is ignored"
      
      $rcmdcheck_loading_and_unloading
      [1] "Check loading and unloading the package"
      
      $rcmdcheck_first_lib_obsolete
      [1] ".First.lib is obsolete"
      
      $rcmdcheck_startup_function_arguments
      [1] "Package startup function arguments should be lib* and pkg*"
      
      $rcmdcheck_startup_function_change_search_path
      [1] "Package startup functions should not change the search path. I.e. do not call 'library' or 'require'"
      
      $rcmdcheck_startup_function_messages
      [1] "Package startup functions should use 'packageStartupMessage' to generate messages"
      
      $rcmdcheck_statup_function_unsafe_calls
      [1] "Package startup functions should not call installed.packages"
      
      $rcmdcheck_last_lib_needs_to_be_exported
      [1] ".Last.lib will not be used unless it is exported"
      
      $rcmdcheck_detach_function_arguments
      [1] "Package detach functions should have one argument with name starting with 'lib'"
      
      $rcmdcheck_detach_no_library_dynam_unload
      [1] "Package detach functions should not call 'library.dynam.unload'"
      
      $rcmdcheck_unsafe_calls
      [1] "Check for possibly unsafe calls: unlockBinding, assignInNamespace"
      
      $rcmdcheck_partial_argument_match
      [1] "Partial argument matches are fragile"
      
      $rcmdcheck_undefined_globals
      [1] "Check for undefined globals"
      
      $rcmdcheck_avoid_internal_calls
      [1] "No .Internal calls are allowed"
      
      $rcmdcheck_avoid_internal_calls_s4
      [1] "No .Internal calls are allowed in S4 methods"
      
      $rcmdcheck_avoid_internal_calls_rc
      [1] "No .Internal calls are allowed in reference class methods"
      
      $rcmdcheck_assignment_to_globalenv
      [1] "Assignment to .GlobalEnv is not allowed"
      
      $rcmdcheck_avoid_using_attach
      [1] "It is better to avoid attach()"
      
      $rcmdcheck_data_into_globalenv
      [1] "Avoid loading data into the global environment"
      
      $rcmdcheck_obsolete_platform_specific
      [1] "Avoid calling obsolete and platform specific functions"
      
      $rcmdcheck_obsolete_platform_specific_s4
      [1] "Avoid calling obsolete and platform specific functions from S4 methods"
      
      $rcmdcheck_obsolete_platform_specific_rc
      [1] "Avoid calling obsolete and platform specific functions from reference class methods"
      
      $rcmdcheck_deprecated_functions
      [1] "Avoid calling deprecated functions"
      
      $rcmdcheck_defunct_removed_functions
      [1] "Avoid calling defunct and removed functions"
      
      $rcmdcheck_avoid_platform_specific_devices
      [1] "Avoid using platform specific devices"
      
      $rcmdcheck_rd_empty_sections
      [1] "Check Rd files for empty sections"
      
      $rcmdcheck_rd_problems
      [1] "Check Rd files for problems"
      
      $rcmdcheck_rd_duplicated_name
      [1] "Check for Rd files with duplicated \\name"
      
      $rcmdcheck_rd_duplicated_alias
      [1] "Check for Rd files with duplicated \\alias"
      
      $rcmdcheck_rd_long_code_lines
      [1] "Check for long code lines in Rd files"
      
      $rcmdcheck_rd_cross_references
      [1] "Check Rd links"
      
      $rcmdcheck_missing_docs
      [1] "Check for undocumented exported objects"
      
      $rcmdcheck_code_docs_mismatch
      [1] "Check for code/documentation mismatches"
      
      $rcmdcheck_rd_usage
      [1] "Check Rd \\usage sections"
      
      $rcmdcheck_rd_s3_usage
      [1] "S3 methods should use the \\method markup, not their full name"
      
      $rcmdcheck_rd_contents
      [1] "Check Rd contents"
      
      $rcmdcheck_unstated_dependencies_in_examples
      [1] "Check for unstated dependencies in examples"
      
      $rcmdcheck_data_contents
      [1] "Check files in 'data' directory, if exists"
      
      $rcmdcheck_undeclared_non_ascii_characters_in_data
      [1] "Check for undeclared non-ASCII characters in data"
      
      $rcmdcheck_non_ascii_characters_in_data
      [1] "Check for non-ASCII characters in data"
      
      $rcmdcheck_uncompressed_data_files
      [1] "Check for data files that could be compressed better"
      
      $rcmdcheck_bzip2_xz_requirement
      [1] "bzip2 or xz compression needs more recent R version"
      
      $rcmdcheck_uncompressed_sysdata
      [1] "Check if sysdata.rda could be compressed better"
      
      $rcmdcheck_unneeded_doc_extra
      [1] "Check for unneeded extra style files for documentation"
      
      $rcmdcheck_doc_extra_licenses
      [1] "Check if extra doc styles have license that requires the distribution of its original sources"
      
      $rcmdcheck_doc_tex_leftovers
      [1] "Check for possible leftover files for docs"
      
      $rcmdcheck_extra_files_in_docs
      [1] "Check for extra files in docs"
      
      $rcmdcheck_extra_directories_in_docs
      [1] "Check for extra directories in docs"
      
      $rcmdcheck_vignettes_in_vignettes
      [1] "Vignettes must be vignettes/"
      
      $rcmdcheck_knitr_leftovers
      [1] "Check for possible knitr leftover files"
      
      $rcmdcheck_pdf_file_sizes
      [1] "Check if PDF files could be made signicantly smaller"
      
      $rcmdcheck_pdf_file_sizes_gs
      [1] "Check if PDF files could be made much smaller with GhostScript and qpdf"
      
      $rcmdcheck_source_line_endings
      [1] "Source files have LF (Unix) line endings"
      
      $rcmdcheck_makefile_line_endings
      [1] "Makefile(s) have LF (Unix) line endings"
      
      $rcmdcheck_makefile_with_final_lf
      [1] "Makefile(s) have a final LF (newline character)"
      
      $rcmdcheck_non_portable_makevars
      [1] "Check for non-portable Makevars flags"
      
      $rcmdcheck_makevars_overriding_user_site
      [1] "Check for Makevars flags overriding user/site settings"
      
      $rcmdcheck_makevars_and_makevars_in
      [1] "Check if package has both Makevars and Makevars.in"
      
      $rcmdcheck_gnu_make_required
      [1] "Check if GNU make is a system requirement"
      
      $rcmdcheck_gnu_makefile_extensions
      [1] "Check if GNU make is needed (and is undeclared)"
      
      $rcmdcheck_blas_and_lapack_flags
      [1] "Check if $(BLAS_LIBS) are used when $(LAPACK_LIBS) are"
      
      $rcmdcheck_missing_flibs
      [1] "Check if $(FLIBS) is required and missing"
      
      $rcmdcheck_unsafe_calls_in_compiled_code
      [1] "Check for unsafe calls in compiled code"
      
      $rcmdcheck_loading_package
      [1] "Check loading the package"
      
      $rcmdcheck_unloading_package
      [1] "Check unloading the package"
      
      $rcmdcheck_namespace_can_be_loaded
      [1] "Check if namespace can be loaded"
      
      $rcmdcheck_namespace_can_be_loaded_safely
      [1] "Check if namespace can be loaded safely"
      
      $rcmdcheck_namespace_can_be_unloaded
      [1] "Check if namespace can be unloaded"
      
      $rcmdcheck_loading_when_not_on_search_path
      [1] "Packages must be able to load when not on the search path"
      
      $rcmdcheck_s3_method_registration
      [1] "S3 methods are properly registered"
      
      $rcmdcheck_encoding_in_ascii_locale
      [1] "Check if we are checking a package with non-ascii encoding in an ASCII locale"
      
      $rcmdcheck_examples_run
      [1] "Examples must run"
      
      $rcmdcheck_examples_run_without_warnings
      [1] "Examples shoudl run without warnings"
      
      $rcmdcheck_two_many_cores_used
      [1] "Package should not use more than two cores during checks"
      
      $rcmdcheck_can_collect_examples
      [1] "Can create a single file with all example"
      
      $rcmdcheck_unstated_dependencies_in_tests
      [1] "All packages needed for tests are declared"
      
      $rcmdcheck_tests_pass
      [1] "Tests"
      
      $rcmdcheck_unstated_dependencies_in_vignettes
      [1] "All packages needed for vignettes are declared"
      
      $rcmdcheck_vignette_output_present
      [1] "Output from all vignettes are present"
      
      $rcmdcheck_encoding_defined_in_vignettes
      [1] "Non-ASCII vignettes must define their encoding"
      
      $rcmdcheck_doc_makefile_uppercase
      [1] "inst/doc/Makefile is uppercase"
      
      $rcmdcheck_calling_r_from_makefile
      [1] "R is called properly from Makefile"
      
      $rcmdcheck_makefile_line_endings_2
      [1] "Makefile line ending should be LF (Unix)"
      
      $rcmdcheck_calling_rscript_from_makefile
      [1] "Rscript is called properly from Makefile"
      
      $rcmdcheck_correct_vignette_encodings
      [1] "Vignettes use the encoding they define"
      
      $rcmdcheck_vignettes_load_dependencies
      [1] "Vignettes can load dependencies"
      
      $rcmdcheck_vignettes_run
      [1] "Vignettes run"
      
      $rcmdcheck_vignettes_build
      [1] "Vignettes build"
      
      $rcmdcheck_vignettes_build_2
      [1] "Vignettes build"
      
      $rcmdcheck_can_convert_rd_to_pdf
      [1] "Can convert Rd manual to PDF"
      
      $rcmdcheck_can_convert_rd_to_pdf_2
      [1] "Can convert Rd manual to PDF"
      
      $rcmdcheck_pdf_without_hyperref
      [1] "PDF manual builds without hyperref"
      
      $rcmdcheck_executable_files_in_packages
      [1] "Package should not contain executable files"
      
      $rcmdcheck_hidden_files_and_directories
      [1] "Package does not have hidden files and directories"
      
      $rcmdcheck_installs
      [1] "Package installs"
      
      $rcmdcheck_significant_compilation_warnings
      [1] "Check for significant warnings when compiling C/C++/Fortran code"
      
      $rcmdcheck_other_compilation_warnings
      [1] "Check for significant warnings when compiling C/C++/Fortran code"
      
      $rcmdcheck_reasonable_installed_size
      [1] "Package size is reasonable when installed"
      
      $rcmdcheck_description_required_fields
      [1] "DESCRIPTION has required fields"
      
      $rcmdcheck_package_name_portable
      [1] "Package name is portable"
      
      $rcmdcheck_description_right_case
      [1] "DESCRIPTION is all upper case"
      
      $rcmdcheck_not_package_bundle
      [1] "Package bundles are defunct"
      
      $rcmdcheck_cran_incoming_feasibility_1
      [1] "CRAN incoming feasibility"
      
      $rcmdcheck_cran_incoming_feasibility_2
      [1] "CRAN incoming feasibility"
      
      $rcmdcheck_cran_incoming_feasibility_3
      [1] "CRAN incoming feasibility"
      
      $rcmdcheck_valid_namespace
      [1] "Valid NAMESPACE file"
      
      $rcmdcheck_empty_importfrom_in_namespace
      [1] "Check for empty importFrom command in NAMESPACE"
      
      $rcmdcheck_too_many_s3_methods
      [1] "R < 3.0.2 had a limit of 500 registered S3 methods"
      
      $rcmdcheck_package_dependencies_present
      [1] "Package dependencies are present for the check"
      
      $rcmdcheck_if_source_package
      [1] "Only *source* packages can be checked"
      
      $rcmdcheck_object_files_in_source_package
      [1] "Source packages should not contain object files"
      
      $rcmdcheck_multi_arch_build_dir
      [1] "Check for leftover multi-arch build directory"
      
      $rcmdcheck_compilation_leftover_files
      [1] "Possible leftover files from a compilation"
      
      $rcmdcheck_object_files_in_source_package_2
      [1] "Source packages should not contain object files"
      
      $rcmdcheck_installed_version_included
      [1] "Check if an installed version is included by mistake"
      
      $rcmdcheck_some_code_in_docs
      [1] "*Some* form of documentation should contain some code to run "
      
      $rd_has_examples
      [1] "Exported functions have \\examples in .Rd"
      
      $rd_has_return
      [1] "Exported functions have \\value in .Rd"
      
      $reverse_dependencies
      [1] "Check for reverse dependencies on CRAN"
      
      $roxygen2_has_export_or_nord
      [1] "Documented functions have @export, @noRd, or @rdname"
      
      $roxygen2_unknown_tags
      [1] "All roxygen2 tags are recognized"
      
      $roxygen2_valid_inherit
      [1] "@inheritParams/@inheritDotParams reference known functions"
      
      $roxygen2_duplicate_params
      [1] "Avoid duplicated @param documentation across functions"
      
      $spelling
      [1] "No misspelled words in documentation"
      
      $tidyverse_brace_linter
      [1] "Opening brace on same line, closing brace on its own line"
      
      $tidyverse_commas_linter
      [1] "Space after commas, not before"
      
      $tidyverse_commented_code_linter
      [1] "No commented-out code"
      
      $tidyverse_equals_na_linter
      [1] "Use is.na() not == NA"
      
      $tidyverse_function_left_parentheses_linter
      [1] "No space before ( in function calls"
      
      $tidyverse_indentation_linter
      [1] "Use 2-space indentation"
      
      $tidyverse_infix_spaces_linter
      [1] "Spaces around infix operators"
      
      $tidyverse_object_length_linter
      [1] "Object names are not too long"
      
      $tidyverse_object_name_linter
      [1] "Object names use snake_case"
      
      $tidyverse_object_usage_linter
      [1] "No unused variables or undefined globals"
      
      $tidyverse_paren_body_linter
      [1] "No space after opening parenthesis"
      
      $tidyverse_pipe_consistency_linter
      [1] "Consistent pipe operator usage"
      
      $tidyverse_pipe_continuation_linter
      [1] "Pipe continuation on next line"
      
      $tidyverse_quotes_linter
      [1] "Use double quotes for strings"
      
      $tidyverse_return_linter
      [1] "Use implicit return, not explicit return()"
      
      $tidyverse_spaces_inside_linter
      [1] "No spaces inside parentheses or brackets"
      
      $tidyverse_spaces_left_parentheses_linter
      [1] "Space before ( in control flow"
      
      $tidyverse_trailing_blank_lines_linter
      [1] "No trailing blank lines"
      
      $tidyverse_trailing_whitespace_linter
      [1] "No trailing whitespace"
      
      $tidyverse_vector_logic_linter
      [1] "Use && and || not & and | in if conditions"
      
      $tidyverse_whitespace_linter
      [1] "No tab characters in code"
      
      $tidyverse_assignment_linter
      [1] "Use <- for assignment, not ="
      
      $tidyverse_line_length_linter
      [1] "Code lines are not too long"
      
      $tidyverse_semicolon_linter
      [1] "No trailing semicolons"
      
      $tidyverse_seq_linter
      [1] "Use seq_len() or seq_along() instead of 1:length(...)"
      
      $tidyverse_T_and_F_symbol_linter
      [1] "Use TRUE and FALSE, not T and F"
      
      $tidyverse_r_file_names
      [1] "R file names use snake_case"
      
      $tidyverse_test_file_names
      [1] "Test files mirror R source files"
      
      $tidyverse_no_missing
      [1] "Functions do not use missing() to check arguments"
      
      $tidyverse_export_order
      [1] "Exported functions are defined before internal helpers"
      
      $urlchecker_ok
      [1] "All URLs are reachable"
      
      $urlchecker_no_redirects
      [1] "No URLs redirect to a different location"
      
      $vignette_no_rm_list
      [1] "Vignettes do not use rm(list = ls())"
      
      $vignette_no_setwd
      [1] "Vignettes do not use setwd()"
      

---
title: "Introduction"
output: rmarkdown::html_vignette
vignette: >
  %\VignetteIndexEntry{Introduction}
  %\VignetteEngine{knitr::rmarkdown}
---

```{r setup}
rm(list = ls())
setwd("/tmp/myproject")
```

Some text about the package.

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gp.R
\name{gp}
\alias{gp}
\alias{goodpractice}
\title{Run good practice checks}
\usage{
gp(
  path = ".",
  checks = default_checks(),
  extra_preps = NULL,
  extra_checks = NULL,
  quiet = TRUE
)
}
\arguments{
\item{path}{Path to a package root.}

\item{checks}{Character vector, the checks to run. Defaults to
\code{\link{default_checks}}. Use \code{\link{all_checks}} to list all
checks, or add optional sets like \code{\link{tidyverse_checks}}.
When \code{NULL}, all registered checks are run, subject to any
exclusions from \code{goodpractice.exclude_check_groups} or
\code{GP_EXCLUDE_CHECK_GROUPS}.}

\item{extra_preps}{Custom preparation functions. See
\code{\link{make_prep}} on creating preparation functions.}

\item{extra_checks}{Custom checks. See \code{\link{make_check}} on
creating checks.}

\item{quiet}{Whether to suppress output from the preparation
functions. Note that not all preparation functions produce output,
even if this option is set to \code{FALSE}.}
}
\value{
A goodpractice object that you can query
with a simple API. See \code{\link{results}} to start.
}
\description{
To see the results, just print it to the screen.
}
\section{Excluding check groups}{

When using the default \code{checks = all_checks()}, entire groups of
checks can be excluded by group name via the
\code{goodpractice.exclude_check_groups} option or the
\code{GP_EXCLUDE_CHECK_GROUPS} environment variable (comma-separated).
The option takes precedence.

\preformatted{
# Skip URL and coverage checks:
options(goodpractice.exclude_check_groups = c("urlchecker", "covr"))

# Or via environment variable:
Sys.setenv(GP_EXCLUDE_CHECK_GROUPS = "urlchecker,covr")
}

Exclusion only applies when \code{checks = NULL} (the default).
Explicit \code{checks} arguments are never filtered.
}

\section{Excluding files}{

Specific files can be excluded from checks via the
\code{goodpractice.exclude_path} option or the \code{GP_EXCLUDE_PATH}
environment variable (comma-separated). Paths are relative to the
package root.

\preformatted{
options(goodpractice.exclude_path = c("R/RcppExports.R", "R/generated.R"))

# Or via environment variable:
Sys.setenv(GP_EXCLUDE_PATH = "R/RcppExports.R,R/generated.R")
}

Excluded files are skipped by lintr, treesitter, expression, and
roxygen2 checks.
}

\section{Parallel preparation}{

Preparation steps run sequentially by default. To run them in
parallel, install \pkg{future.apply} and set a
\code{\link[future]{plan}}:

\preformatted{
future::plan("multisession")
gp(".")
}

Preps run in parallel only when a non-sequential plan is active.
Prep functions must be independent: in parallel mode each prep
receives the initial state snapshot, so a prep cannot read another
prep's output. Only new state fields are merged back; if two preps
write the same field, the second is dropped with a warning.
}

\examples{
path <- system.file("bad1", package = "goodpractice")
# Run a subset of all checks available
g <- gp(path, checks = all_checks()[9:16])
g
# Or run with named check groups
g <- gp(path, checks = checks_by_group("description", "namespace"))
}
\seealso{
Other main:
\code{\link{goodpractice-package}}
}
\concept{main}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lists.R
\name{all_check_groups}
\alias{all_check_groups}
\title{List available check group names}
\usage{
all_check_groups()
}
\value{
A character vector of check group names.
}
\description{
Returns the names of all registered check groups.
Use these names with \code{\link[=checks_by_group]{checks_by_group()}} to select checks by group,
or with \code{options(goodpractice.exclude_check_groups = ...)} to skip
groups.
Full descriptions of each check group are return by
\code{\link[=describe_check_groups]{describe_check_groups()}}.
}
\examples{
# Names of all check groups:
all_check_groups()

# List individual checks by group:
chks <- lapply(all_check_groups(), checks_by_group)
names(chks) <- all_check_groups()
chks
}
\seealso{
Other check_groups:
\code{\link[=all_checks]{all_checks()}},
\code{\link[=checks_by_group]{checks_by_group()}},
\code{\link[=default_checks]{default_checks()}},
\code{\link[=describe_check]{describe_check()}},
\code{\link[=describe_check_groups]{describe_check_groups()}},
\code{\link[=tidyverse_checks]{tidyverse_checks()}}
}
\concept{check_groups}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{export_json}
\alias{export_json}
\title{Export failed checks to JSON}
\usage{
export_json(gp, file, pretty = FALSE)
}
\arguments{
\item{gp}{\code{\link{gp}} output.}

\item{file}{Output connection or file.}

\item{pretty}{Whether to pretty-print the JSON.}
}
\value{
Invisibly returns the path to the output file.
}
\description{
Export failed checks to JSON
}
\examples{
path <- system.file("bad1", package = "goodpractice")
g <- gp(path, checks = "description_url")
tmp <- tempfile(fileext = ".json")
export_json(g, tmp)
unlink(tmp)
}
\seealso{
Other output:
\code{\link[=print.goodPractice]{print.goodPractice()}},
\code{\link[=results]{results()}}
}
\concept{output}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{failed_positions}
\alias{failed_positions}
\title{Positions of check failures in the source code}
\usage{
failed_positions(gp)
}
\arguments{
\item{gp}{\code{\link{gp}} output.}
}
\value{
A list of lists of positions. See details below.
}
\description{
Note that not all checks refer to the source code.
For these the result will be \code{NULL}.
}
\details{
For the ones that do, the results is a list, one for each failure.
Since the same check can fail multiple times. A single failure
is a list with entries: \code{filename}, \code{line_number},
\code{column_number}, \code{ranges}. \code{ranges} is a list of
pairs of start and end positions for each line involved in the
check.
}
\examples{
path <- system.file("bad1", package = "goodpractice")
g <- gp(path, checks = "description_url")
failed_positions(g)
}
\seealso{
Other API:
\code{\link[=checks]{checks()}},
\code{\link{customization}},
\code{\link[=failed_checks]{failed_checks()}}
}
\concept{API}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.R
\docType{package}
\name{goodpractice-package}
\alias{goodpractice-package}
\title{goodpractice: Advice on R Package Building}
\description{
Give advice about good practices when building R packages. Advice includes
functions and syntax to avoid, package structure, code complexity, code
formatting, etc.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://docs.ropensci.org/goodpractice/}
  \item \url{https://github.com/ropensci-review-tools/goodpractice}
  \item Report bugs at \url{https://github.com/ropensci-review-tools/goodpractice/issues}
}


Other main:
\code{\link[=gp]{gp()}}
}
\author{
\strong{Maintainer}: Mark Padgham \email{mark@ropensci.org} (\href{https://orcid.org/0000-0003-2172-5265}{ORCID})

Authors:
\itemize{
  \item Mark Padgham \email{mark@ropensci.org} (\href{https://orcid.org/0000-0003-2172-5265}{ORCID})
  \item Karina Marks \email{karina.marks@ascent.io} (GitHub: KarinaMarks)
  \item Daniel de Bortoli (GitHub: ddbortoli)
  \item Gabor Csardi \email{csardi.gabor@gmail.com}
  \item Hannah Frick \email{hannah.frick@gmail.com}
  \item Owen Jones \email{owenjonesuob@gmail.com} (GitHub: owenjonesuob)
  \item Hannah Alexander \email{halexander@mango-solutions.com}
  \item Athanasia Mo Mowinckel \email{a.m.mowinckel@psykologi.uio.no} (\href{https://orcid.org/0000-0002-5756-0223}{ORCID}) (GitHub: drmowinckels)
}

Other contributors:
\itemize{
  \item Ascent Digital Services UK Limited (GitHub: MangoTheCat) [copyright holder]
  \item Ana Simmons \email{ana.simmons@ascent.io} (GitHub: anasimmons) [contributor]
  \item Fabian Scheipl (GitHub: fabian-s) [contributor]
}

}
\concept{main}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{failed_checks}
\alias{failed_checks}
\title{Names of the failed checks}
\usage{
failed_checks(gp)
}
\arguments{
\item{gp}{\code{\link{gp}} output.}
}
\value{
Names of the failed checks.
}
\description{
Names of the failed checks
}
\examples{
path <- system.file("bad1", package = "goodpractice")
# run a subset of all checks available
g <- gp(path, checks = all_checks()[9:16])
failed_checks(g)
# Or run with named check groups
g <- gp(path, checks = checks_by_group("description", "namespace"))
failed_checks(g)
}
\seealso{
Other API:
\code{\link[=checks]{checks()}},
\code{\link{customization}},
\code{\link[=failed_positions]{failed_positions()}}
}
\concept{API}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{results}
\alias{results}
\title{Return all check results in a data frame}
\usage{
results(gp)
}
\arguments{
\item{gp}{\code{\link{gp}} output.}
}
\value{
Data frame, with columns:
\item{check}{The name of the check.}
\item{passed}{Logical, whether the check passed.}
}
\description{
Return all check results in a data frame
}
\examples{
path <- system.file("bad1", package = "goodpractice")
# Run a subset of all checks available
g <- gp(path, checks = all_checks()[9:16])
results(g)
# Or run with named check groups
g <- gp(path, checks = checks_by_group("description", "namespace"))
results(g)
}
\seealso{
Other output:
\code{\link[=export_json]{export_json()}},
\code{\link[=print.goodPractice]{print.goodPractice()}}
}
\concept{output}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/customization.R
\name{customization}
\alias{customization}
\alias{make_prep}
\alias{make_check}
\title{Defining custom preparations and checks}
\usage{
make_prep(name, func)

make_check(description, check, gp, ...)
}
\arguments{
\item{name}{Name of the preparation function.}

\item{func}{A function that takes two arguments:
The \code{path} to the root directory of the package, and a logical
argument: \code{quiet}. If \code{quiet} is true, the preparation function
may print out diagnostic messages. The output of this function will be
saved as the " \code{name}" entry of \code{state}, i.e. of the input for
the  \code{check}-functions (see example).}

\item{description}{A description of the check.}

\item{check}{A function that takes the \code{state} as an argument.}

\item{gp}{A short description of what is good practice.}

\item{...}{Further arguments. Most important: A \code{preps} argument that
contains the names of all the preparation functions required for the check.}
}
\value{
For \code{make_prep}: a preparation function. For
\code{make_check}: a check object of class \code{"check"}.
}
\description{
Defining custom preparations and checks
}
\section{Functions}{
\itemize{
\item \code{make_prep()}: Create a preparation function

\item \code{make_check()}: Create a check function

}}
\examples{
# make a preparation function
url_prep <- make_prep(
  name = "desc", 
  func = function(path, quiet) desc::description$new(path)
)
# and the corresponding check function
url_chk <- make_check(
  description = "URL field in DESCRIPTION",
  tags = character(),
  preps = "desc",
  gp = "have a URL field in DESCRIPTION",
  check = function(state) state$desc$has_fields("URL")
)
# use together in gp():
# (note that you have to list the name of your custom check in
# the checks-argument as well....)
bad1 <- system.file("bad1", package = "goodpractice")
res <- gp(bad1, checks = c("url", "no_description_depends"),
          extra_preps = list("desc" = url_prep),
          extra_checks = list("url" = url_chk))
}
\seealso{
Other API:
\code{\link[=checks]{checks()}},
\code{\link[=failed_checks]{failed_checks()}},
\code{\link[=failed_positions]{failed_positions()}}
}
\concept{API}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{checks}
\alias{checks}
\title{List all checks performed}
\usage{
checks(gp)
}
\arguments{
\item{gp}{\code{\link{gp}} output.}
}
\value{
Character vector of check names.
}
\description{
List all checks performed
}
\examples{
path <- system.file("bad1", package = "goodpractice")
# Run a subset of all checks available
g <- gp(path, checks = all_checks()[9:16])
checks(g)
# Or run with named check groups
g <- gp(path, checks = checks_by_group("description", "namespace"))
checks(g)
}
\seealso{
Other API:
\code{\link{customization}},
\code{\link[=failed_checks]{failed_checks()}},
\code{\link[=failed_positions]{failed_positions()}}
}
\concept{API}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lists.R
\name{default_checks}
\alias{default_checks}
\title{List the names of default checks (excludes optional check sets)}
\usage{
default_checks()
}
\value{
Character vector of default check names
}
\description{
List the names of default checks (excludes optional check sets)
}
\examples{
default_checks()
}
\seealso{
Other check_groups:
\code{\link[=all_check_groups]{all_check_groups()}},
\code{\link[=all_checks]{all_checks()}},
\code{\link[=checks_by_group]{checks_by_group()}},
\code{\link[=describe_check]{describe_check()}},
\code{\link[=describe_check_groups]{describe_check_groups()}},
\code{\link[=tidyverse_checks]{tidyverse_checks()}}
}
\concept{check_groups}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lists.R
\name{describe_check}
\alias{describe_check}
\title{Describe one or more checks}
\usage{
describe_check(check_name = NULL)
}
\arguments{
\item{check_name}{Names of checks to be described.}
}
\value{
List of character descriptions for each \code{check_name}
}
\description{
Describe one or more checks
}
\examples{
describe_check("rcmdcheck_non_portable_makevars")
check_name <- c("no_description_depends",
                "lintr_assignment_linter",
                "no_import_package_as_a_whole",
                "rcmdcheck_missing_docs")
describe_check(check_name)
# Or to see all checks:
\dontrun{
  describe_check(all_checks())
}
}
\seealso{
Other check_groups:
\code{\link[=all_check_groups]{all_check_groups()}},
\code{\link[=all_checks]{all_checks()}},
\code{\link[=checks_by_group]{checks_by_group()}},
\code{\link[=default_checks]{default_checks()}},
\code{\link[=describe_check_groups]{describe_check_groups()}},
\code{\link[=tidyverse_checks]{tidyverse_checks()}}
}
\concept{check_groups}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{learn_skill_gp}
\alias{learn_skill_gp}
\title{Function for AI agents to run and learn how to use the 'goodpractice'
package.}
\usage{
learn_skill_gp()
}
\value{
The content of the file \code{system.file("skills",
"goodpractice4agents.md", package = "goodpractice"))}
}
\description{
You tell agents to directly use this function in order for them to learn the
skill. If you want to use the skill yourself to guide an agent, then use the
\link{use_skill_gp} function.
}
\examples{
\dontrun{
learn_gp_skill()
}
}
\seealso{
Other skills:
\code{\link[=use_skill_gp]{use_skill_gp()}}
}
\concept{skills}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lists.R
\name{checks_by_group}
\alias{checks_by_group}
\title{Select checks by check group}
\usage{
checks_by_group(...)
}
\arguments{
\item{...}{Group names as character strings. Use \code{\link[=all_check_groups]{all_check_groups()}} to
see available names.}
}
\value{
Character vector of check names
}
\description{
Returns the names of all checks that belong to the given group(s).
This makes it easy to run or inspect a specific category of checks
without knowing individual check names.
}
\examples{
# run only DESCRIPTION and namespace checks
checks_by_group("description", "namespace")

# see what the lintr group covers
checks_by_group("lintr")
# See all checks by group:
lapply(all_check_groups(), checks_by_group)
# use directly in gp()
\dontrun{
  gp(".", checks = checks_by_group("description", "lintr"))
}
}
\seealso{
Other check_groups:
\code{\link[=all_check_groups]{all_check_groups()}},
\code{\link[=all_checks]{all_checks()}},
\code{\link[=default_checks]{default_checks()}},
\code{\link[=describe_check]{describe_check()}},
\code{\link[=describe_check_groups]{describe_check_groups()}},
\code{\link[=tidyverse_checks]{tidyverse_checks()}}
}
\concept{check_groups}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lists.R
\name{describe_check_groups}
\alias{describe_check_groups}
\title{Describe available check groups}
\usage{
describe_check_groups()
}
\value{
A named list of each check group defined in \code{\link[=all_check_groups]{all_check_groups()}},
with text descriptions of each group.
}
\description{
Returns full descriptions of all registered check groups.
}
\examples{
# Names of all check groups:
all_check_groups()

# And corresponding descriptions:
describe_check_groups()
}
\seealso{
Other check_groups:
\code{\link[=all_check_groups]{all_check_groups()}},
\code{\link[=all_checks]{all_checks()}},
\code{\link[=checks_by_group]{checks_by_group()}},
\code{\link[=default_checks]{default_checks()}},
\code{\link[=describe_check]{describe_check()}},
\code{\link[=tidyverse_checks]{tidyverse_checks()}}
}
\concept{check_groups}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lists.R
\name{all_checks}
\alias{all_checks}
\title{List the names of all checks}
\usage{
all_checks()
}
\value{
Character vector of checks
}
\description{
List the names of all checks
}
\examples{
all_checks()
}
\seealso{
Other check_groups:
\code{\link[=all_check_groups]{all_check_groups()}},
\code{\link[=checks_by_group]{checks_by_group()}},
\code{\link[=default_checks]{default_checks()}},
\code{\link[=describe_check]{describe_check()}},
\code{\link[=describe_check_groups]{describe_check_groups()}},
\code{\link[=tidyverse_checks]{tidyverse_checks()}}
}
\concept{check_groups}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/api.R
\name{use_skill_gp}
\alias{use_skill_gp}
\title{Download a local copy of 'goodpractice4agents.md' to tell AI agents how to
use this package.}
\usage{
use_skill_gp(
  path = "goodpractice4agents.md",
  use_skills_subdir = FALSE,
  overwrite = FALSE
)
}
\arguments{
\item{path}{Local path where the file should be written. If \code{path} is
an existing directory, the file is written as \code{goodpractice4agents.md}
within it; otherwise \code{path} is treated as the full target file name.}

\item{use_skills_subdir}{If \code{TRUE}, the \code{path} will be edited to
write the file in to a \code{skills/} sub-directory.}

\item{overwrite}{If \code{FALSE} (default) and file specified by \code{path}
already exists, function will error and not overwrite the existing file.}
}
\value{
(Invisibly) the full path to the written file.
}
\description{
The 'goodpractice4agents.md' file contains sufficient instructions for most
AI agents to edit package code so that it passes most 'goodpractice'
checks. The file can be directly edited and tweaked for personal use cases.
An agent can be instructed to simply "run the file goodpractice4agents.md".
Alternatively, the file can be directly downloaded from GitHub at
\url{https://github.com/ropensci-review-tools/goodpractice/blob/main/inst/skills/goodpractice4agents.md}.
}
\details{
You can also instruct agents to call the companion function,
\link{learn_skill_gp}, directly in order to learn and use the skill itself.
}
\examples{
path <- file.path(tempdir(), "skill.md")
f <- use_skill_gp(path)
file.exists(f)
unlink(f)
}
\seealso{
Other skills:
\code{\link[=learn_skill_gp]{learn_skill_gp()}}
}
\concept{skills}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/lists.R
\name{tidyverse_checks}
\alias{tidyverse_checks}
\title{List the names of tidyverse style checks}
\usage{
tidyverse_checks()
}
\value{
Character vector of tidyverse check names
}
\description{
These checks are optional and not included in the default set.
They are powered by \code{\link[lintr]{lint_package}} using lintr's
default linter set and respect any \code{.lintr} configuration file
in the package root (e.g. to disable specific linters or add exclusions).
Add them via \code{checks = c(default_checks(), tidyverse_checks())}.
}
\examples{
tidyverse_checks()
}
\seealso{
Other check_groups:
\code{\link[=all_check_groups]{all_check_groups()}},
\code{\link[=all_checks]{all_checks()}},
\code{\link[=checks_by_group]{checks_by_group()}},
\code{\link[=default_checks]{default_checks()}},
\code{\link[=describe_check]{describe_check()}},
\code{\link[=describe_check_groups]{describe_check_groups()}}
}
\concept{check_groups}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/print.R
\name{print.goodPractice}
\alias{print.goodPractice}
\title{Print goodpractice results}
\usage{
\method{print}{goodPractice}(x, groups = NULL, positions_limit = 5, ...)
}
\arguments{
\item{x}{Object of class \code{goodPractice}, as returned by \code{\link[=gp]{gp()}}.}

\item{groups}{Name of check groups for which to print results, as vector of
one or more of \code{\link[=all_check_groups]{all_check_groups()}}.}

\item{positions_limit}{How many positions to print at most.}

\item{...}{Unused, for compatibility with \code{\link[base:print]{base::print()}} generic method.}
}
\description{
Print goodpractice results
}
\seealso{
Other output:
\code{\link[=export_json]{export_json()}},
\code{\link[=results]{results()}}
}
\concept{output}

\docType{import}
\name{reexports}
\alias{reexports}
\alias{reexported_fn}
\title{Objects exported from other packages}
\keyword{internal}
\description{
These objects are imported from other packages. Follow the links
below to see their documentation.

\describe{
  \item{otherpkg}{\code{\link[otherpkg]{reexported_fn}}}
}}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/package.R
\name{func1}
\alias{func1}
\title{Random function}
\usage{
func1()
}
\description{
Random function
}
\value{
Nothing useful.
}
\examples{
func1()
}

\name{my_func}
\alias{my_func}
\title{Calcualte the incremnet of a value}
\description{
This functon compuets the incremnet of the givn value.
}
\usage{
my_func(x)
}
\arguments{
\item{x}{A numbr to incremnet.}
}

\name{good_func}
\alias{good_func}
\title{Properly documented function}
\usage{
good_func(x)
}
\description{
Has everything: value, examples, runnable code.
}
\value{
A number.
}
\examples{
good_func(1)
}
